#include "icgenerator.h"
#include <algorithm>
#include <set>
#include <cassert>
#include "potential.h"
#include "massmodel.h"
#include "utils.h"
#include "managepotential.h"

namespace smile{

// sample points from spherical model 
void CICGeneratorDensitySpherical::createInitialConditions(size_t numPoints, smile::CPointMassSet<float> *points)
{
    double totalmass=model->getTotalMass(), centralmass=model->mass(0);
    float pointmass=static_cast<float>((totalmass-centralmass)/numPoints);
    points->data.reserve(numPoints);
    for(size_t ind=0; ind<numPoints; ind++)
    {
        double radius=0, encmass;
        if (quiet==1) 
            encmass=(ind+getRandomNumber())/numPoints;
        else if (quiet==2) 
            encmass = (ind+0.5)/numPoints;
        else
            encmass=getRandomNumber();
        encmass=(totalmass-centralmass)*encmass + centralmass;
        radius=model->rad(encmass);
        double X, Y, Z;
        getRandomUnitVector(&X, &Y, &Z);
        CPosVelPoint<float> pt(
            static_cast<float>(radius*X),
            static_cast<float>(radius*Y), 
            static_cast<float>(radius*Z), 0, 0, 0);
        points->add(pt, pointmass);
    }
}

// estimate the upper bound on energy as one half of the potential energy of the outermost particle
double computeEnergyCutoff(const CPotential* potential, const CPointMassSet<float> & points)
{
    double maxRad2=0;
    for(CPointMassSet<float>::Type::const_iterator p=points.data.begin(); p!=points.data.end(); p++)
        maxRad2=std::max<double>(maxRad2, pow_2(p->first.Pos[0])+pow_2(p->first.Pos[1])+pow_2(p->first.Pos[2]));
    double val=potential->Phi(sqrt(maxRad2),0,0);
    return val<0 ? val*0.5 : val+1.0;  // take a fail-safe value if val>=0 (shouldn't occur for a finite-mass model)
}

// initial conditions for velocity based on spherical Eddington distribution function
void CICGeneratorVelocityEddington::createInitialConditions(size_t /*numPoints*/, CPointMassSet<float> *points)
{
    const CMassModel* massModel = createMassModel(potential);
    if(massModel==NULL) {
        my_error(FUNCNAME, "Cannot initialize velocities");
        return;
    }
    const CMassModel* densityModel = 
        density==NULL || density==potential ? massModel :
        createMassModel(density, NUM_RADIAL_POINTS_SPHERICAL_MODEL, 0, 0, massModel);  // need to create a separate model for density
    if(densityModel==NULL) {
        my_error(FUNCNAME, "Cannot initialize velocities");
        return;
    }
    double Emax=computeEnergyCutoff(potential, *points);
    for(CPointMassSet<float>::Type::iterator point=points->data.begin(); point!=points->data.end(); point++)
    {
        // correction factor for escape velocity
        double radius=sqrt(pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1])+pow_2(point->first.Pos[2]));
        double poten=potential->Phi(point->first.Pos[0], point->first.Pos[1], point->first.Pos[2]);
        double velcorr = sqrt(poten / massModel->pot(radius));
        if(gsl_isnan(velcorr)) velcorr=1.0;
        double veloc, E;
        size_t ntrial=0;
        do{
            veloc=densityModel->samplev(radius) * velcorr;
            ntrial++;
            E=poten+veloc*veloc/2;
        } while(E >= Emax && ntrial<1e3);  // to avoid infinite loop in the case of weird potential
        if(E>=Emax) veloc=0;
        float prior=1;
        double X,Y,Z;
        if(weightSkewFactorL==0) {  // direct velocity vector isotropically
            getRandomUnitVector(&X, &Y, &Z);
            point->first.Vel[0]=static_cast<float>(veloc*X); 
            point->first.Vel[1]=static_cast<float>(veloc*Y); 
            point->first.Vel[2]=static_cast<float>(veloc*Z);
        } else {   // create more orbits with lower angular momentum
            double w=getRandomNumber()*2-1;
            double sign=w>0?1:-1; 
            w=fabs(w);
            double oneminuscosth=w*(2+(weightSkewFactorL-1)*w)/(1+weightSkewFactorL);  // 1-cos(theta) is more frequently sampled at low values
            prior=static_cast<float>(2*(1+(weightSkewFactorL-1)*w)/(1+weightSkewFactorL));   // this bias is accounted for in the choice of prior weight
            double Vpar=veloc*(1-oneminuscosth)*sign;   // component of velocity parallel to radius-vector
            double Vper=sqrt(pow_2(veloc)-pow_2(Vpar)); // ..and perpendicular
            getRandomUnitVector(&X, &Y, &Z);
            double WdotR=X*point->first.Pos[0]+Y*point->first.Pos[1]+Z*point->first.Pos[2];  // projection of random vector W onto the radius vector R
            double Rmag2=pow_2(point->first.Pos[0])+pow_2(point->first.Pos[1])+pow_2(point->first.Pos[2]);
            double Rmag=sqrt(Rmag2);
            X -= WdotR*point->first.Pos[0]/Rmag2;
            Y -= WdotR*point->first.Pos[1]/Rmag2;
            Z -= WdotR*point->first.Pos[2]/Rmag2; // subtract parallel component
            double Wmag=sqrt(pow_2(X)+pow_2(Y)+pow_2(Z));
            if(Wmag==0) Wmag=1;  // avoid singularity
            point->first.Vel[0] = static_cast<float>(point->first.Pos[0]/Rmag*Vpar + X/Wmag*Vper);
            point->first.Vel[1] = static_cast<float>(point->first.Pos[1]/Rmag*Vpar + Y/Wmag*Vper);
            point->first.Vel[2] = static_cast<float>(point->first.Pos[2]/Rmag*Vpar + Z/Wmag*Vper);
        }
        point->second*=prior;
    }
    if(densityModel!=massModel) delete densityModel;
    delete massModel;
};

}; // namespace
