/** \file   mkspherical.cpp
    \brief  Tool: Create spherical mass models from a text table, Nbody snapshot, or analytic density profile
    \author EV
    \date   2010-2015

This program uses spherical mass models for two different purposes and created from several possible sources:
either from a supplied table with r, m(r) values specifying enclosed mass as a function of radius,
or an N-body snapshot, in which case it first fits a penalized spline to compute M(r) from the snapshot,
or an analytic density profile, or potential expansion coefficients given in a text file.

The spherical model, in turn, may be used to compute a number of parameters as spline-interpolated functions
of radius (potential, radial/circular period, distribution function via Eddington inversion formula,
diffusion coefficients, etc.), or to generate an N-body snapshot, in which particles are distributed
according to the given density profile and isotropically in velocities.

In short, this is a generalization of tools such as `halogen` or `spherICs` for creating a spherical isotropic
model with a given arbitrary density profile, and at the same time a useful tool to study dynamical properties
of a given \Nbody system (or, rather, its spherically-symmetric isotropic counterpart).

*/

#include <vector>
#include <iostream>
#include <fstream>
#include "configfile.h"
#include "icgenerator.h"
#include "iosnapshot.h"
#include "managepotential.h"
#include "massmodel.h"
#include "potential.h"
#include "utils.h"

/// usage description
const char* description =
    "mkspherical: Create spherical isotropic models for a given mass profile using Eddington inversion.\n"
    "Input parameters must specify the model using one of the possible modes:\n"
    " 1) a text file with a table describing the cumulative mass profile: each line contains 'r M(r)' values\n"
    " 2) an N-body snapshot\n"
    " 3) a density profile from SMILE described by its name and some parameters\n"
    " 4) a potential coefficients file describing one of SMILE potential expansion types\n"
    " 5) an ini file with all potential parameters\n"
    "Output may consist of an extended table with many parameters as functions of radius and energy, and/or "
    "an N-body snapshot corresponding to the spherical isotropic model (or both a table and a snapshot)\n\n"
    "Command-line parameters and their default values:\n"
    "file=           input file (for modes 1,2,4,5); choice of mode is determined by file extension:"
    " .mass (1), .coef_*** (4), .ini (5), anything else (2), but mode 1 may be forced by density=Coefs argument\n"
    "density=        name of the density profile (for mode 3), in this case may provide some more potential parameters\n"
    "outtab=         output table with r,M,Phi,f(E) and other parameters\n"
    "rmin=0          output table inner nonzero radius [0 for auto-detect]\n"
    "rmax=0          output table outer radius [0 for auto-detect]\n"
    "npoints=100     output table number of grid points in radius [-1 to use input grid]\n"
    "mbh=0           additional point mass (black hole) at origin\n"
    "outsnap=        output snapshot file name (to create an N-body representation of the model)\n"
    "outformat=Text  output snapshot format [Text/Nemo/Gadget]\n"
    "nbody=0         number of points to create (must provide a value if outsnap is given)\n"
    "quiet=0         0=noisy 1=somewhat quiet 2=more quiet\n"
    "VERSION=2.5     Jan 2015 EV\n";

/// main program begins here
int main(int argc, char* argv[])
{
    if(argc<=1) {  // print help and exit
        std::cerr << description;
        return 0;
    }
    gsl_set_error_handler(&smile::gsl_verbose_error_handler);
    smile::CConfigPotential cfg=smile::defaultConfigPotential;
    cfg.Ncoefs_radial=0;  // default value for N-body snapshots -- will determine automatically based on N
    // parse command-line parameters as if they were provided in the [Potential] section of a hypothetical ini file
    smile::CConfigFile cfgFile("Potential", argc, argv);
    std::string outputtab   =cfgFile.getValueString("Potential/outtab");
    std::string outputsnap  =cfgFile.getValueString("Potential/outsnap");
    std::string outputformat=cfgFile.getValueString("Potential/outformat", "Text");
    double gridrmin=cfgFile.getValueDouble("Potential/rmin");
    double gridrmax=cfgFile.getValueDouble("Potential/rmax");
    int numNodes=cfgFile.getValueInt("Potential/npoints", 100);
    int nbody   =cfgFile.getValueInt("Potential/nbody");
    int quiet   =cfgFile.getValueInt("Potential/quiet");
    if((!outputsnap.empty() && nbody==0) || (outputsnap.empty() && nbody>0)) {
        std::cerr << "Must give both output snapshot filename (outsnap=...) and number of bodies (nbody=...)\n";
        return 0;
    }
    if(outputsnap.empty() && outputtab.empty()) {
        std::cerr << "Must give output snapshot filename (outsnap=...) or output table filename (outtab=...), or both\n";
        return 0;
    }
    smile::readConfigPotential(cfgFile, &cfg);  // parse the rest of parameters
    if(!cfg.NbodyFile.empty()) {  // file name was provided
        {   // open a block for temp variable
            std::ifstream strm(cfg.NbodyFile.c_str(), std::ios::in);
            if(!strm) {
                std::cerr << "Cannot read from file "<<cfg.NbodyFile<<"\n";
                return 0;
            }
        }
        if(smile::ends_with_str(cfg.NbodyFile, ".ini")) {  // name of ini file passed as file=... parameter
            smile::CConfigFile cfgFile1(cfg.NbodyFile);
            smile::readConfigPotential(cfgFile1, &cfg);
        } else {
            cfg.PotentialType=smile::coefFileType(cfg.NbodyFile);
            if(cfg.PotentialType==smile::CDensity::PT_UNKNOWN) { // none of the known file extensions -- assumed to be an N-body snapshot
                cfg.PotentialType=smile::CDensity::PT_SPHERICAL;
                cfg.DensityType=smile::CDensity::PT_NB;
            } else
                cfg.DensityType=smile::CDensity::PT_COEFS;
        }
    } else {
        if(cfg.Ncoefs_radial==0)  // assign a default value for analytic density profiles
            cfg.Ncoefs_radial=50;
    }
    if(cfg.PotentialType==smile::CDensity::PT_UNKNOWN)
        cfg.PotentialType=smile::CDensity::PT_SPHERICAL;  // default for this program
    if(cfg.PotentialType==smile::CDensity::PT_SPHERICAL && cfg.DensityType==smile::CDensity::PT_UNKNOWN) {
        std::cerr << "Must provide an input file=... or a valid name for density=... profile\n";
        return 0;
    }
    const smile::CPotential* potential=smile::createPotential(&cfg);  // NULL on error
    if(potential==NULL) {
        std::cerr << "Incorrect "<<smile::getPotentialNameByType(cfg.PotentialType)<<" potential parameters\n";
        return 0;
    }
    const smile::CMassModel* model = NULL;
    if(potential->PotentialType()==smile::CDensity::PT_SPHERICAL)
        model = static_cast<const smile::CPotentialSpherical*>(potential)->getModel();
    else
        model = smile::createMassModel(potential);
    if(model==NULL) {
        std::cerr << "Cannot create a finite-mass model\n";
        return 0;
    }
    std::vector<double> radii, inmass;  // r, M(r) table 
    model->getMassProfile(&radii, &inmass);
    if(cfg.Mbh!=0) {  // need to modify the mass model to account for a central black hole
        for(size_t i=0; i<inmass.size(); i++) 
            inmass[i]+=cfg.Mbh;
        model = new smile::CMassModel(radii, inmass);
        if(model->getStatus()!=smile::CMassModel::MS_OK && model->getStatus()!=smile::CMassModel::MS_DF_NEGATIVE) {
            std::cerr << "Cannot create a model with a black hole\n";  // shouldn't occur if a model without BH was ok.
            return 0;
        }
    }
    double rho0=0, drhodr=0, alpha=0;
    if(model->hasDensityCore(&rho0, &drhodr, &alpha))
        std::cerr << "Constant-density core determined: rho(r) = " << rho0 << "*(1-" << drhodr << "*r^" << alpha << ")\n";
    if(model->getStatus()==smile::CMassModel::MS_DF_NEGATIVE)
        std::cerr << "Warning, distribution function is negative\n";

    if(outputtab!="")   // output table with all computed quantities vs radius
    {
        std::ofstream out(outputtab.c_str(), std::ios_base::out);
        if(!out) {
            std::cerr << "Cannot write output table: " << outputtab << "\n";
            return 0;
        }
        std::vector<double> gridRadius;
        if(numNodes<0) {  // use input points, excluding r=0
            std::vector<double>::iterator beg=radii.begin();
            while(beg!=radii.end() && *beg<=0) beg++;
            gridRadius.assign(beg, radii.end());
            numNodes=static_cast<int>(gridRadius.size());
        } else {  // create a log-grid with reasonable inner and outer radii
            double zeroMass=model->mass(0), totalMass=model->getTotalMass()-zeroMass;
            if(gridrmin==0) {
                gridrmin=0.125;
                while(gridrmin>1e-6 && model->mass(gridrmin)-zeroMass>totalMass*1e-6) gridrmin/=2;  // a quick default algo
            }
            if(gridrmax==0) {
                gridrmax=8;
                while(gridrmax<1e6 && model->mass(gridrmax)-zeroMass<totalMass*(1-1e-3)) gridrmax*=2;
            }
            if(numNodes==0) numNodes=100;
            smile::createNonuniformGrid(numNodes, gridrmin, gridrmax, false, &gridRadius);
        }
        out << "#r\tM(r)\tPhi=E\trho\tf(E)\tM(E)\tt_epi(E)\tperiod(E)\tr_circ(E)\tL_circ(E)\tsigma\tsigma_los\tdens_proj\tD_EE\tD_E\tD_LL\tD_RR/R(0)\n";
        out << 0 << "\t" << model->mass(0) << "\t" << model->pot(0) << "\n";
        for(int i=0; i<numNodes; i++)
        {
            double rad = gridRadius[i];
            double pot = model->pot(rad);
            double DEE, DE, DLL, DRRoverR;
            model->DifCoefEL(pot, &DEE, &DE, &DLL, &DRRoverR, false);
            double distrf = model->distrf(pot);
            out << rad << "\t" << smile::convertToString(model->mass(rad),12) << "\t"
                << smile::convertToString(pot,10) << "\t" << smile::convertToString(model->dens(rad),10) << "\t"
                << distrf << "\t" << model->masse(pot) << "\t" << model->tepi(pot) << "\t"
                << model->trad(pot) << "\t" << model->rcirc(pot) << "\t" << model->lcirc(pot) << "\t"
                << model->veldisp(rad) << "\t" << model->losvd(rad) << "\t" << model->densproj(rad) << "\t"
                << DEE << "\t" << DE << "\t" << DLL << "\t" << DRRoverR << "\t";
            out << "\n";
        }
        out.close();
    }

    if(outputsnap!="" && nbody>0)
    {
        // generate initial conditions using derived d.f.
        smile::randomize();
        smile::CPointMassSet<float>* points_out=new smile::CPointMassSet<float>();
        smile::CICGeneratorDensitySpherical  icg_pos(model, quiet);
        smile::CICGeneratorVelocityEddington icg_vel(new smile::CPotentialSpherical(*model));
        icg_pos.createInitialConditions(nbody, points_out);  // assign positions
        icg_vel.createInitialConditions(nbody, points_out);  // assign velocities
        if(model->mass(0)!=0) {  // replace first point by a black hole
            points_out->at(0).first=smile::CPosVelPoint<float>();
            points_out->at(0).second=static_cast<float>(model->mass(0));
        }
        std::string cmdlinefull="mkspherical";
        char** argv0=argv+1;
        for(char *arg=*argv0; (arg = *argv0); argv0 ++ ){
            cmdlinefull += " " + std::string(arg);
        }
        smile::CBasicIOSnapshot* snap_out = smile::createIOSnapshotWrite(outputformat, outputsnap, cmdlinefull);
        if(snap_out!=NULL) {
            if(!snap_out->writeSnapshot(points_out)) 
                std::cerr << "Can't write output snapshot: " << outputsnap << "\n";
        } else {
            std::cerr << "Unknown output format: " << outputformat << "\n";
        }
        delete snap_out;
        delete points_out;
    }
    return 0;
}
