#include "orbitan.h"
#include <cassert>
#include <algorithm>
#include <cmath>
#include <set>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_multiroots.h>
#include "potential.h"
#include "utils.h"
#include "massmodel.h"
#include "orbitint.h"

namespace smile{ 

/* constructor of COrbit class for an existing orbit 
   (!!! defined here and not in orbitint.h because it uses a couple of runtime functions) -- bad design! */
template<typename NumT> COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const std::vector< CPosVelPoint<NumT> > &_traj, double _timeStep) :
  potential(_potential),
  configOrbit(_configOrbit), 
  initCond(_traj[0]), endCond(_traj.back())
{
    timeUnit = -2;
    state=OS_DONE;
    odeInt=NULL;
    timeEnd=timeMax=_timeStep*_traj.size();
    runtimeFncs.push_back( new COrbitRuntimeTrajectory(this, _traj, _timeStep));
    runtimeFncs.push_back( new COrbitRuntimePericenter(this, _traj, _timeStep));
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
        (*iter)->Finish();
};
// explicit instantiations of constructor for two variants of template argument
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const std::vector< CPosVelPoint<float> > &_traj, double _timeStep);
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const std::vector< CPosVelPoint<double> > &_traj, double _timeStep);

//----------------------------------------------------------------------------//

template<typename NumT> std::string CPericenterInformation<NumT>::toString() const
{
    return "Lmin\xB2="+convertToString(fitIntercept,4)+" + "+convertToString(fitSlope,3)+"*w [sct="+convertToString(fitScatter,3)+",sig="+convertToString(fitSignificance,5)+"]\nLcirc\xB2="+convertToString(Lcirc2,3)+"\n";
}
//template std::string CPericenterInformation<float >::toString() const;
//template std::string CPericenterInformation<double>::toString() const;

template<typename NumT> std::string COrbitInformation<NumT>::toString() const
{
    return Description 
        + "\nlfx=" + convertToString(lf[0],8)
        + "\nlfy=" + convertToString(lf[1],8)
        + "\nlfz=" + convertToString(lf[2],8)
        + "\nFreqDiff=" + convertToString(lfccdiff,3) 
        + (lambda>=0 ? "\nLyapunovExp=" + convertToString(lambda,3) : "")
        + "\nInertia=" + convertToString(inertia[0],3) + "," +  convertToString(inertia[1],3) + "," +  convertToString(inertia[2],3)
        + "\nLz=" + convertToString(Lavg[2],3)+"\xB1"+convertToString(Lvar[2],3) 
        + (Lavg[0]!=0 ? ", Lx=" + convertToString(Lavg[0],3)+"\xB1"+convertToString(Lvar[0],3) : "") 
        + "\nE=" + convertToString(Einit) + ", deltaE=" + convertToString(Ediff,3) + "\n";
}
//template std::string COrbitInformation<float >::toString() const;
//template std::string COrbitInformation<double>::toString() const;

template<typename NumT> std::string CTrajSampleInformation<NumT>::toString() const
{
    if(trajSample.empty()) return std::string(); 
    else return convertToString(trajSample.size())+" sampling points\n";
}
//template std::string CTrajSampleInformation<float >::toString() const;
//template std::string CTrajSampleInformation<double>::toString() const;

template<typename NumT> std::string CVelPerturberInformation<NumT>::toString() const
{
    return "deltaErelax="+convertToString(deltaErelax)+"\n";
}

// Runtime function for recording and analyzing trajectory by means of frequency analysis and Lyapunov exponent 

COrbitRuntimeTrajectory::COrbitRuntimeTrajectory(const smile::COrbit *_orbit) :
  CBasicOrbitRuntimeFnc(_orbit), 
  N_dim(orbit->getConfig()->N_dim) 
{
    Einit=totalEnergy(orbit->getPotential(), orbit->getInitCond(), orbit->getConfig()->Omega);
    Ediff=0;
    timeUnit=orbit->getTimeUnit();
    timeStep=timeUnit*orbit->getConfig()->samplingInterval;
    numSteps=0;
}

template<typename NumT> COrbitRuntimeTrajectory::COrbitRuntimeTrajectory( 
    const smile::COrbit *_orbit, const std::vector< CPosVelPoint<NumT> > &_traj, double _timeStep) :
  CBasicOrbitRuntimeFnc(_orbit), 
  N_dim(orbit->getConfig()->N_dim), 
  traj(_traj.begin(), _traj.end())
{
    Einit=totalEnergy(orbit->getPotential(), orbit->getInitCond(), orbit->getConfig()->Omega);
    Ediff=0;  // not quite true, but we won't bother calculating it..
    timeUnit=orbit->getTimeUnit();
    timeStep=_timeStep;
    numSteps=0;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeTrajectory::Timestep(const double told, const double t, CPosVelPoint<double> *vars)
{
    int i1=static_cast<int>(told/timeStep); 
    int i2=static_cast<int>(t/timeStep);

    // check energy conservation
    numSteps++;
    const unsigned int Ncheck=orbit->getPotential()->PotentialType()==CPotential::PT_NB ? 20 : 5;
    if(numSteps%Ncheck==0)  // do not check too frequently (potential evaluation may be costly)
    {
        double Ecur = totalEnergy(orbit->getPotential(), *vars, orbit->getConfig()->Omega, t);
        double Edifcur = fabs(Einit - Ecur);
        Ediff=std::max<double>(Edifcur, Ediff);
        energy.push_back(Ecur-Einit);
        enertime.push_back(t);
    }
    // store trajectory at regular intervals of time 
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            CPosVelPoint<double> pt;
            for(unsigned int c=0; c<N_dim; c++)
            {
                pt.Pos[c]=orbit->getPos(c, tt);
                pt.Vel[c]=orbit->getVel(c, tt);
            }
            traj.push_back(pt);
            if(orbit->getConfig()->calcLyapunovMethod>0)
                lnw.push_back(orbit->getLogDevVec(tt));
        }
    }
    return SR_TERMINATE;
}

template<typename NumT> COrbitInformation<NumT>* COrbitRuntimeTrajectory::analyzeOrbit(size_t start, size_t count, CSpectrum *spectrum, vectord *fa) const
{
    if(start>=traj.size())  // invalid call
        return new COrbitInformation<NumT>(traj.size()==0 ? "Empty" : "Invalid call: start>=size");
    if(count<=0 || count>traj.size()-start) count = traj.size() - start;   // passing zero for count means analysing entire orbit
    if(count<=1)  // empty orbit
        return new COrbitInformation<NumT>("?");
    CSpectrum tempSpectrum[N_DIM];
    CSpectrum* sl = spectrum!=NULL ? spectrum : tempSpectrum;   // if outputting spectrum is not necessary, use temp. array for storing the information
    double Lavg_d[N_DIM]={0,0,0}, Lvar_d[N_DIM]={0,0,0};
    std::string Description = performOrbitAnalysis(start, count, sl, fa, Lavg_d, Lvar_d);
    double lfccdiff = calcLFCCdiffusion(start, count);
    double lambda = calcLambda();
    double inertia_d[3];
    calcInertiaOverInterval(start, count, inertia_d);
    NumT leadfreq[N_DIM]={static_cast<NumT>(sl[0][0]), static_cast<NumT>(sl[1][0]), static_cast<NumT>((N_dim==3)?sl[2][0].freq:0)};
    NumT inertia[N_DIM] ={static_cast<NumT>(inertia_d[0]), static_cast<NumT>(inertia_d[1]), static_cast<NumT>(inertia_d[2])};
    NumT Lavg[N_DIM] ={static_cast<NumT>(Lavg_d[0]), static_cast<NumT>(Lavg_d[1]), static_cast<NumT>(Lavg_d[2])};
    NumT Lvar[N_DIM] ={static_cast<NumT>(Lvar_d[0]), static_cast<NumT>(Lvar_d[1]), static_cast<NumT>(Lvar_d[2])};
    // note that Ediff is returned for entire orbit integration time, not for the given interval
    return new COrbitInformation<NumT>(Description, static_cast<NumT>(Einit), static_cast<NumT>(Ediff), 
        static_cast<NumT>(timeUnit), leadfreq, static_cast<NumT>(lfccdiff), static_cast<NumT>(lambda), 
        inertia, Lavg, Lvar);
}
// explicit instantiations of a template function
template COrbitInformation<double>* COrbitRuntimeTrajectory::analyzeOrbit<double>(size_t start, size_t count, CSpectrum *spectrum, vectord *fa) const;

// perform FFT of orbit (output amplitudes to 'fa' array) and find 'maxnlines' lines, which will be placed into 'sl' array. 
// If fa=NULL, do not output frequency amplitudes
void COrbitRuntimeTrajectory::findLines(size_t start, size_t npoints, CSpectrum *sl, vectord *fa) const
{
    vectord trajSlice(npoints);
    const CConfigOrbit* configOrbit=orbit->getConfig();
    for(unsigned int c=0; c<N_dim; c++)
    {
        for(size_t i=0; i<npoints; i++)
            trajSlice[i]=traj[start+i].Pos[c];
        CSpectrumAnalyzerHunter spect(trajSlice, timeStep/timeUnit);
        spect.findSpectralLines(&(sl[c]), fa!=NULL? &(fa[c]) : NULL);
        int sw=0; CSpectralLine ftmp;
        switch(c)   // ensure that the first line in each coordinate has frequency >1 (for X), >X (for Y), and >Y (for Z); tolerate small deviations given in MIN_FREQ_*
        {
        case 0:
            if(configOrbit->freqStrictOrder) 
                while((sw<(int)sl[0].size()-1) && (sl[0][0].freq<configOrbit->minFreqX)) 
                { 
                    ftmp=sl[0][sw+1]; 
                    sl[0][sw+1]=sl[0][0]; 
                    sl[0][0]=ftmp; sw++; 
                }
            break;
        case 1:
            if(configOrbit->freqStrictOrder) 
                while((sw<(int)sl[1].size()-1) && (sl[1][0].freq<configOrbit->minFreqYZ*sl[0][0].freq)) 
                { 
                    ftmp=sl[1][sw+1]; 
                    sl[1][sw+1]=sl[1][0]; 
                    sl[1][0]=ftmp; sw++; 
                }
            break;
        case 2:
            if(configOrbit->freqStrictOrder && N_dim==3) 
                while((sw<(int)sl[2].size()-1) && (sl[2][0].freq<configOrbit->minFreqYZ*sl[1][0].freq)) 
                { 
                    ftmp=sl[2][sw+1]; 
                    sl[2][sw+1]=sl[2][0]; 
                    sl[2][0]=ftmp; sw++; 
                }
            break;
        }
    }
}

/// \cond INTERNAL_DOCS
void simplify(int &n1, int &n2)    // remove common factors
{
    int nmin=std::min<int>(n1, n2);
    int f=1;
    for(int k=1; k<nmin/2; k++)
        if((n1%k==0) && (n2%k==0)) f=k;
    n1/=f;
    n2/=f;
}

bool commensurable(double a, double b, double acc, int &na, int &nb)   // check that a:b ~=~ na:nb
{
    if(a*b==0) return(false);
    na=1; nb=0;
    while(na<=FREQ_RES_2)
    {
        int nb1=(int)floor(b/a*na+.5);
        if(nb1 && (fabs(nb1*a-na*b) < acc))
        {
            nb=nb1;
            return true;
        }
        else na++;
    }
    return false;
}

bool lineardepend(double f, double a, double b, double acc, int &ma, int &mb)   // check linear dependency: f ~=~ ma*a + mb*b
{
    ma=mb=0; 
    while(abs(ma)<=FREQ_RES_2)
    {
        mb=(int)floor((f-ma*a)/b+0.5);
        if(fabs(ma*a+mb*b-f) < acc) 
            return true;
        else 
            if(ma>0) ma=-ma; else ma=-ma+1;
    }
    return false;
}

bool lineardepend3(double f, double a, double b, double c, double acc, int &ma, int &mb, int &mc)   // check 3d linear dependency: f ~=~ ma*a + mb*b + mc*c
{
    ma=mb=mc=0; 
    while(abs(ma)<=FREQ_RES_2)
    {
      mb=0;
      while(abs(mb)<=FREQ_RES_2)
      {
        mc=(int)floor((f-ma*a-mb*b)/c+0.5);
        if(fabs(ma*a+mb*b+mc*c-f) < acc) 
            return true;
        else 
            if(mb>0) mb=-mb; else mb=-mb+1;
      }
      if(ma>0) ma=-ma; else ma=-ma+1;
    }
    return false;
}

bool commensurable3(double a, double b, double c, double acc, int &ma, int &mb, int &mc)
{
    for(int na=1; na<=FREQ_RES_3; na++)
        for(int nb=-FREQ_RES_3; nb<=FREQ_RES_3; nb++)
            for(int nc=-FREQ_RES_3; nc<=FREQ_RES_3; nc++)
                if(fabs(a*na+b*nb+c*nc) < acc)
                {
                    ma=na; mb=nb; mc=nc;
                    return true;
                }
    return false;
}
/// \endcond

std::string COrbitRuntimeTrajectory::performOrbitAnalysis(size_t start, size_t count, 
    CSpectrum out_lines[N_DIM], vectord *out_ampl, double Lavg[N_DIM], double Lvar[N_DIM]) const
{
    findLines(start, count, out_lines, out_ampl);
    std::string Description;
    double Labs[N_DIM];  // abs.value of ang.mom.components
    double Cavg[N_DIM];  // avg.value of C-th coordinate
    double Cmax[N_DIM];  // max.value of this coordinate
    int Lsign[N_DIM]={0,0,0};  // keep track of angular momentum sign for each coordinate: -1,0,+1,or 2 if it is not conserved
    for(unsigned int d=0; d<N_DIM; d++)
    {
        Lavg[d]=0;  // mean value of d-axis component of angular momentum
        Lvar[d]=0;  // variation in this component (r.m.s.)
        Labs[d]=0;  // mean absolute value of d-axis component of ang.mom.
        Cavg[d]=0;
        Cmax[d]=0;
    }
    double Omega=orbit->getConfig()->Omega;
    for(size_t i=start; i<start+count; i++)
    {
        for(unsigned int d=0; d<N_DIM; d++)
        {
            double L =
            d==0 ? traj[i].Pos[1]*traj[i].Vel[2] - traj[i].Pos[2]*traj[i].Vel[1] + Omega*traj[i].Pos[0]*traj[i].Pos[2] :
            d==1 ? traj[i].Pos[2]*traj[i].Vel[0] - traj[i].Pos[0]*traj[i].Vel[2] + Omega*traj[i].Pos[1]*traj[i].Pos[2] :
                   traj[i].Pos[0]*traj[i].Vel[1] - traj[i].Pos[1]*traj[i].Vel[0] - Omega*(pow_2(traj[i].Pos[0])+pow_2(traj[i].Pos[1]));
            Lavg[d] += L;
            Labs[d] += fabs(L);
            Lvar[d] += pow_2(L);
            if(Lsign[d]==0) // initialize ang.mom.sign (has not yet been identified)
                Lsign[d] = L>0 ? 1 : L<0 ? -1 : 0;
            else if((L>0 && Lsign[d]==-1) || (L<0 && Lsign[d]==1)) 
                Lsign[d]=2;  // sign is variable
            Cavg[d] += traj[i].Pos[d];
            Cmax[d] = std::max<double>(Cmax[d], fabs(traj[i].Pos[d]));
        }
    }
    for(unsigned int d=0; d<N_DIM; d++)
    {
        Lavg[d]/=count;
        Labs[d]/=count;
        Cavg[d]/=count;
        Lvar[d]=sqrt(std::max<double>(0, Lvar[d]/count-pow_2(Lavg[d])));
    }
    // determine orbit class
    /*
    bool tubez=(fabs(Lavg[2])>Labs[2]*OC_TUBE_AVGL);  // has definite sense of rotation about z axis (more robust indicator than 1:1 correspondence of leading freqencies)
    bool tubex=(N_dim==3) && (fabs(Lavg[0])>Labs[0]*OC_TUBE_AVGL);  // same for x axis
    bool tubey=(N_dim==3) && (fabs(Lavg[1])>Labs[1]*OC_TUBE_AVGL);  // same for y axis (although we don't expect to find y-axis tubes if axes are sorted properly)
    */
    bool tubex=Lsign[0]==1 || Lsign[0]==-1;
    bool tubey=Lsign[1]==1 || Lsign[1]==-1;
    bool tubez=Lsign[2]==1 || Lsign[2]==-1;
    double lfX=out_lines[0][0];
    double lfY=out_lines[1][0];
    double acc = timeUnit/(timeStep*count) * FREQ_ACCURACY;  // accuracy of comparison
    if(acc<FREQ_DIFF_REG_MIN) acc=FREQ_DIFF_REG_MIN;  // don't be too demanding.. (actually the accuracy limit is determined by energy conservation error, for simplicity ignored here)
    if(N_dim==2)
    {
        // try to find nx:ny correspondence of leading frequencies
        int nx=0, ny=0;
        if(tubez || commensurable(lfX, lfY, acc, nx, ny))    // boxlet family or tube (in case lfX==lfY)
        {
            double lf = (lfX/nx + lfY/ny)/2;  // mean value of common denominator 
            // find additional (libration) frequency
            double lfa = 0;
            bool regular=true;
            int rank=1;
            int ma=0, mb=0;
            for(unsigned int c=0; c<N_dim; c++)
            {
                for(size_t i=1; regular&&(i<out_lines[c].size()); i++)
                {
                    if(commensurable(out_lines[c][i], lf, acc, ma, mb)) 
                    {
                        if(((ma==1) || (abs(ma-mb)==1)) && (mb>1))
                            rank=mb;
                    }
                    else
                    {
                        double lfa1=std::min<double>(out_lines[c][i].freq, fabs(out_lines[c][0].freq-out_lines[c][i].freq));
                        if(lfa)
                        {
                            if(!lineardepend(out_lines[c][i], lf, lfa, acc, ma, mb)) 
                            {
                                regular=false;
                            }
                            else 
                                if(lfa1<lfa) 
                                {
                                    lfa=lfa1;
                                }
                        }
                        else
                        {
                            lfa=lfa1;
                        }
                    }
                }
            }
            if(tubez) {
                Description = "tube";
                if(Omega!=0) {
                    if(Omega*Lavg[2]>0) Description = "prograde "+Description;
                    else Description = "retrograde "+Description;
                }
            }
            else if((nx==1)&&(ny==2))
                Description = "banana";
            else if((nx==2)&&(ny==3))
                Description = "fish";
            else if((nx==3)&&(ny==4))
                Description = "pretzel";
            else 
                Description = convertToString(nx)+":"+convertToString(ny)+" resonant boxlet";
            Description += " rank "+convertToString(rank);
            if(!regular)
                Description = "chaotic "+Description;
        }
        else 
        {
            // for regular box orbit the remaining lines must be linear-dependent on two leading frequencies
            bool regular=true;
            int ma, mb;
            for(int c=0; c<2; c++)
                for(size_t i=1; regular&&(i<out_lines[c].size()); i++)
                    if(!lineardepend(out_lines[c][i], lfX, lfY, acc, ma, mb)) regular=false;
            if(regular)
                Description = "box";
            else
                Description = "chaotic box";
        }
    }
    else    // N_dim==3
    {
        // determine if there are resonances between each pair of coordinates
        bool res[3];
        int nres=0;          // number of resonances
        int na[3], nb[3];    // resonance quotients
        double lf[3]={0,0,0};// leading frequencies (1 for nres=3, 2 for nres=1, 3 for nres=0)
        int nlf=0;           // number of leading frequencies
        for(int d=0; d<3; d++)
        {
            res[d] = commensurable(out_lines[d][0], out_lines[(d+1)%3][0], acc, na[d], nb[d]);
            if(res[d]) nres++;
        }
 
        if(nres==2)  // check for consistency (there cannot be two resonances)
        {   // reconstruct missing resonance number dn
            int dn=!res[0] ? 0 : (!res[1] ? 1 : 2);
            na[dn]=nb[(dn+2)%3]*nb[(dn+1)%3];
            nb[dn]=na[(dn+1)%3]*na[(dn+2)%3];
            simplify(na[dn], nb[dn]);
            nres=3;
        }
        if(nres==0)
        {
            nlf=3;
            lf[0]=out_lines[0][0];
            lf[1]=out_lines[1][0];
            lf[2]=out_lines[2][0];
            int n1, n2, n3;
            if(commensurable3(lf[0],lf[1],lf[2],acc,n1,n2,n3))
            {
                Description = "("+convertToString(n1)+","+convertToString(n2)+","+convertToString(n3)+") thin orbit";
                nlf = 2;
            }
            else
            {
                // either box or pyramid (near black hole)
                if(fabs(Cavg[2])>Cmax[2]*OC_PYRAMID_AVGZ)
                    Description = "pyramid";
                else if(fabs(Cavg[0])>Cmax[0]*OC_PYRAMID_AVGZ)
                    Description = "x-pyramid";   // shouldn't exist normally
                else if(fabs(Cavg[1])>Cmax[1]*OC_PYRAMID_AVGZ)
                    Description = "y-pyramid";   // shouldn't exist normally
                else
                    Description = "box";
            }
        }
        if(nres==1)
        { 
            int de=res[0] ? 0 : (res[1] ? 1 : 2);
            nlf=2;
            lf[0] = (out_lines[de][0]/na[de] + out_lines[(de+1)%3][0]/nb[de])/2;
            lf[1] = out_lines[(de+2)%3][0];
            if(lf[1]==0)
                nlf=1;  // avoid degenerate cases
            switch(de)
            {
            case 0: Description = convertToString(na[0])+":"+convertToString(nb[0])+":*"; break;
            case 1: Description = "*:"+convertToString(na[1])+":"+convertToString(nb[1]); break;
            case 2: Description = convertToString(nb[2])+":*:"+convertToString(na[2]); break;
            }
            Description +=" resonance";
        }
        if(nres==3)
        {
            nlf=1;
            lf[0]=std::min<double>((out_lines[0][0].freq/na[0] + out_lines[1][0].freq/nb[0])/2, 
                std::min<double>  ((out_lines[1][0].freq/na[1] + out_lines[2][0].freq/nb[1])/2, 
                                   (out_lines[2][0].freq/na[2] + out_lines[0][0].freq/nb[2])/2));
            int nx=(int)floor(out_lines[0][0]/lf[0]+0.5); 
            int ny=(int)floor(out_lines[1][0]/lf[0]+0.5); 
            int nz=(int)floor(out_lines[2][0]/lf[0]+0.5); 
            Description = convertToString(nx)+":"+convertToString(ny)+":"+convertToString(nz)+" resonance";
        }
        if(tubex && !tubez && !tubey) 
        {
            Description = "LAT";
            // try to determine whether it's inner or outer LAT
            double radiusAtWaist=0, radiusAtEdge=0;
            for(size_t i=start; i<start+count; i++)
            {
                double R=sqrt(pow_2(traj[i].Pos[1])+pow_2(traj[i].Pos[2]));  // radius in y-z plane
                if(fabs(traj[i].Pos[0])<0.25*Cmax[0])   // near "waist" (X=0) 
                    radiusAtWaist = std::max<double>(radiusAtWaist, R);
                else
                    radiusAtEdge = std::max<double>(radiusAtEdge, R);
            }
            if(radiusAtWaist<radiusAtEdge)
                Description+="(inner)";
            else
                Description+="(outer)";
        }
        if(tubez && !tubey && !tubex) 
        {
            if(fabs(Cavg[2])>fabs(Cmax[2]*OC_PYRAMID_AVGZ))
                Description = "SAU";  // saucer orbit (a variant of short-axis tube)
            else
                Description = "SAT";
            if(Omega!=0) {
                if(Omega*Lavg[2]>0) Description = "prograde "+Description;
                else Description = "retrograde "+Description;
            }
        }
        if(tubey && !tubex && !tubez)
        {
            Description = "IAT";  // intermediate-axis tube, they simply shouldn't exist if the axes were ordered properly.
        }
        // sometimes an orbit appears to conserve the sign of more than one component of angular momentum
        if(tubex && tubey && !tubey)
            Description = "[L+I] tube";
        if(tubey && tubez && !tubex)
            Description = "[I+S] tube";
        if(tubex && tubez && !tubey)
            Description = "[L+S] tube";
        if(tubex && tubey && tubez)
            Description = "[L+I+S] tube";
        // check that every other freq is commensurable with base
        int ma, mb, mc;
        bool regular=true;
        for(int c=0; c<3; c++)
            for(size_t i=1; regular&&(i<out_lines[c].size()); i++)
            {
                switch(nlf)
                {
                case 1:
                    if(!commensurable(out_lines[c][i], lf[0], acc, ma, mb)) 
                    {
                        nlf++;
                        lf[1] = std::min<double>(out_lines[c][i].freq, fabs(lf[0]-out_lines[c][i].freq));
                    }
                    break;
                case 2:
                    if(!lineardepend(out_lines[c][i], lf[0], lf[1], acc, ma, mb))
                    {
                        nlf++;
                        lf[2] = std::min<double>(out_lines[c][i].freq, 
                            std::min<double>(fabs(lf[0]-out_lines[c][i].freq), fabs(lf[1]-out_lines[c][i].freq)));
                    }
                    break;
                case 3:
                    if(!lineardepend3(out_lines[c][i], lf[0], lf[1], lf[2], acc, ma, mb, mc))
                        regular=false;
                }
            }
        if(!regular)
            Description = "chaotic "+Description;
    }
    if(tubez) {  // compare radial and angular frequencies of a tube orbit
        vectord tim(count), rad(count), ang(count);
        double ang_add=0, prev_atan=0, curr_atan, rad_avg=0;
        for(size_t i=0; i<count; i++) {
            rad[i]=sqrt(pow_2(traj[start+i].Pos[0])+pow_2(traj[start+i].Pos[1]));
            rad_avg+=rad[i];
            curr_atan=atan2(traj[start+i].Pos[1], traj[start+i].Pos[0]);
            if(i>0 && curr_atan<-M_PI*0.5 && prev_atan>M_PI*0.5) {
                ang_add+=M_PI*2;
            } else if(i>0 && curr_atan>M_PI*0.5 && prev_atan<-M_PI*0.5) {
                ang_add-=M_PI*2;
            }
            tim[i]=i*timeStep;
            ang[i]=(curr_atan+ang_add);
            prev_atan=curr_atan;
        } 
        rad_avg/=count;
        for(size_t i=0; i<count; i++) rad[i]-=rad_avg;  // subtract the mean to facilitate detection of spectral lines
        double freq_kappa;  // epicyclic frequency (radial oscillations)
        double freq_omega;  // frequency of angular motion
        CSpectrum sl_rad;
        CSpectrumAnalyzerHunter spect(rad, timeStep/(2*M_PI));
        spect.findSpectralLines(&sl_rad, NULL);
        freq_kappa=sl_rad[0].freq;
        double fitSlope, fitIntercept, cov00, cov01, cov11, sumsq;
        // fit linear regression to ang(t) to estimate its slope
        gsl_fit_linear(&(tim.front()), 1, &(ang.front()), 1, count, &fitIntercept, &fitSlope, &cov00, &cov01, &cov11, &sumsq);
        freq_omega=fitSlope;
        int mk, mo;
        if(fabs(freq_omega)<M_PI/(timeStep*count)) Description+=" (CR)";  // corotation resonance
        else if(commensurable(freq_kappa, freq_omega, acc, mk, mo) && abs(mk)<=10 && abs(mo)<=10) {
            if(mk==2 && mo== 1) Description+=" (ILR)"; else  // inner Lindblad
            if(mk==2 && mo==-1) Description+=" (OLR)"; else  // outer Lindblad
            Description += " ("+convertToString(mk)+":"+convertToString(mo)+")";
        }
    }
    return Description;
}

void COrbitRuntimeTrajectory::calcLFCCoverInterval(size_t start, size_t count, double leadfreq[]) const
{
    CSpectrum tempSpectrum[N_DIM];
    findLines(start, count, tempSpectrum, NULL);
    for(unsigned int c=0; c<N_dim; c++)
        leadfreq[c]=(tempSpectrum[c].size()>0)? tempSpectrum[c][0].freq : -1;
}

// calculate rate of diffusion of leading frequencies in cartesian coordinates (LFCCs), which are related to the degree of chaos in the orbit
// determines the LFCCs for first and second halves of orbit, computes difference, returns delta f/f averaged over all coordinates
double COrbitRuntimeTrajectory::calcLFCCdiffusion(size_t start, size_t count) const
{
    if(count<=0) count=traj.size()-start;  // whole orbit by default
    size_t count2=count/2;
    double lf1[N_DIM], lf2[N_DIM];  // leading frequencies of 1st and 2nd half for each coordinate
    calcLFCCoverInterval(start, count2, lf1);
    calcLFCCoverInterval(start+count2, count2, lf2);
    double FreqDiff = 0;
#if FREQ_DIFF_METHOD == 0
    for(int c=N_dim-1; c>=0 && FreqDiff==0; c++)
      if(lf1[c] + lf2[c] > 0 )
        FreqDiff = fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]);
#else
    int nf=0;
    for(unsigned int c=0; c<N_dim; c++)
        if(lf1[c] + lf2[c] > 0 )
#if FREQ_DIFF_METHOD == 2
          if(fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]) < 1)
#endif
        {
            FreqDiff += fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]);
            nf++;
        }
    if(nf>0) FreqDiff /= nf;
#endif
#if 0
    double maxFreqDiff = pow_2(timeUnit/intTime);  // rough estimate of regular/chaotic divide
    if(maxFreqDiff<FREQ_DIFF_REG_MIN) maxFreqDiff=FREQ_DIFF_REG_MIN;

    if(FreqDiff<1*maxFreqDiff)
        DescriptionLFCCdiff = "regular";
    else 
    if(FreqDiff<2*maxFreqDiff)
        DescriptionLFCCdiff = "probably regular";
    else 
    if(FreqDiff<4*maxFreqDiff)
        DescriptionLFCCdiff = "probably chaotic";
    else 
        DescriptionLFCCdiff = "chaotic";
#endif
    return FreqDiff;
}

void COrbitRuntimeTrajectory::calcInertiaOverInterval(size_t start, size_t count, double inertia[]) const
{
    for(unsigned int c=0; c<N_DIM; c++)
        inertia[c] = 0;
    for(size_t i=0; i<count; i++)
    {
        for(unsigned int c=0; c<N_dim; c++)
            inertia[c] += pow_2(traj[i+start].Pos[c]);
    }
    for(unsigned int c=0; c<N_dim; c++)
        inertia[c] = sqrt(inertia[c]/count);
}

// calculates maximal Lyapunov characteristic exponent (if the deviation equation was integrated along with orbit)
double COrbitRuntimeTrajectory::calcLambda() const
{
    if(orbit->getConfig()->calcLyapunovMethod==0) 
        return -1;  // means that no Lyapunov exponent was computed
    double sumavg=0;     // to calculate average value of ln(|w|/t) over the linear-growth interval
    int cntavg=0;
    double sumlambda=0;  // to calculate average value of ln(|w|)/t over the exponential growth interval
    int cntlambda=0;
    int timeblock=static_cast<int>(timeUnit/timeStep*2);    // averaging interval (roughly orbital period * 2)
    std::vector<double> interval(timeblock);
    bool chaotic=false;
    for(int t=timeblock; t<static_cast<int>(traj.size()); t+=timeblock)
    {
        int cnt = std::min<int>(timeblock, static_cast<int>(traj.size())-t);
        if(!chaotic)
        {
            // compute median value over 'timeblock' points  to avoid outliers
            interval.resize(cnt);
            for(int i=0; i<cnt; i++)
                interval[i] = (lnw[i+t]-log((i+t)*timeStep));
            std::sort(interval.begin(), interval.end());
            double mid = interval[cnt/2];
            sumavg+=mid;
            cntavg++;
            if((cntavg>4) && (mid > sumavg/cntavg + 2))
                chaotic=true;     // from here on, assume that we got into t->infinity regime and start collecting data for estimating lambda
        }
        if(chaotic)
        {
            for(int i=0; i<cnt; i++)
            {
                sumlambda += lnw[i+t]/(i+t);
                cntlambda++;
            }
        }
    }
    if(chaotic)
        return (sumlambda / cntlambda * timeUnit / timeStep);
    else
        return 0;
}

// return maximal distance between adjacent points
double COrbitRuntimeTrajectory::getMaxDist(size_t start, size_t count) const
{
    if(start>=traj.size())  return -1; // invalid call
    if(count<=0 || count>traj.size()-start) count = traj.size() - start;   // passing zero for count means analysing entire orbit
    if(count<=1) return -1; // empty orbit
    double maxd=0;
    CPosPoint<double> prev=traj[start];
    for(size_t i=start+1; i<start+count; i++)
    {
        CPosPoint<double> cur=traj[i];
        maxd = std::max<double>(maxd, sqrt(pow_2(cur.Pos[0]-prev.Pos[0])+pow_2(cur.Pos[1]-prev.Pos[1])+pow_2(cur.Pos[2]-prev.Pos[2])));
        prev=cur;
    }
    return maxd;
}

//----------------------------------------------------------------------------//
// Timestep function for storing a number of sampling points from trajectory
COrbitRuntimeTrajSample::COrbitRuntimeTrajSample(const COrbit* _orbit, size_t _numSamplingPoints, double _timeStep) :
  CBasicOrbitRuntimeFnc(_orbit), 
  numSamplingPoints(_numSamplingPoints), 
  timeStep(_timeStep),
  N_dim(orbit->getConfig()->N_dim)
{}; 

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeTrajSample::Timestep(const double told, const double t, CPosVelPoint<double>*)
{
    int i1=static_cast<int>(told/timeStep); 
    int i2=static_cast<int>(t/timeStep);

    // store trajectory at regular intervals of time 
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            CPosVelPoint<float> pt;
            for(unsigned int c=0; c<N_dim; c++)
            {
                pt.Pos[c]=static_cast<float>(orbit->getPos(c, tt));
                pt.Vel[c]=static_cast<float>(orbit->getVel(c, tt));
            }
            trajEntire.push_back(pt);
        }
    }
    return SR_TERMINATE;
}

void COrbitRuntimeTrajSample::Finish()
{
    size_t numActual=trajEntire.size();
    if(numActual<1) return;
    numSamplingPoints = std::min<size_t>(numSamplingPoints, numActual-1);
    std::set<size_t> samplingIndexes;
    gsl_rng* randn=gsl_rng_alloc(gsl_rng_default);
    for(size_t i=0; i<numSamplingPoints; i++)
    {
        size_t np;
        do{
            np=static_cast<size_t>(floor(gsl_rng_uniform(randn)*(numActual-1)));
        }
        while(samplingIndexes.find(np)!=samplingIndexes.end());
        samplingIndexes.insert(np);   // try to avoid duplication
        trajSample.push_back(trajEntire[np]);
    }
    gsl_rng_free(randn);
}

COrbitRuntimeTrajSampleCreator::COrbitRuntimeTrajSampleCreator(unsigned int _numSamplingPointsDefault, double _intTimeMin) : 
  numSamplingPointsDefault(_numSamplingPointsDefault), intTimeMin(_intTimeMin)
{ 
    if(numSamplingPointsDefault==0) {
        my_error(FUNCNAME, "numSamplingPoints=0"); 
        return;
    }
};

void COrbitRuntimeTrajSampleCreator::addOrbitSpecial(const CPosVelPoint<float> &orbitIC, 
    size_t numSamplingPoints, double samplingTimestep)
{
    specialIC.push_back(orbitIC);
    specialNumSamplingPoints.push_back(numSamplingPoints);
    specialSamplingTimestep.push_back(samplingTimestep);
}

CBasicOrbitRuntimeFnc* COrbitRuntimeTrajSampleCreator::createRuntimeFnc(const COrbit* orbit) const
{ 
    // find if this orbit is listed in the 'special' list
    if(!specialIC.empty())
    {
        CPosVelPoint<double> initCond(orbit->getInitCond());
        for(size_t i=0; i<specialIC.size(); i++)
        {
            CPosVelPoint<double> specialCond(specialIC[i]);
            const double eps=1e-6;  // accuracy of comparison (float rather than double)
            if( gsl_fcmp(initCond.Pos[0], specialCond.Pos[0], eps)==0 &&
                gsl_fcmp(initCond.Pos[1], specialCond.Pos[1], eps)==0 &&
                gsl_fcmp(initCond.Pos[2], specialCond.Pos[2], eps)==0 &&
                gsl_fcmp(initCond.Vel[0], specialCond.Vel[0], eps)==0 &&
                gsl_fcmp(initCond.Vel[1], specialCond.Vel[1], eps)==0 &&
                gsl_fcmp(initCond.Vel[2], specialCond.Vel[2], eps)==0) {
                return new COrbitRuntimeTrajSample(orbit, specialNumSamplingPoints[i], specialSamplingTimestep[i]);
            }
        }
    }
    // compute timestep and create runtime function
    return new COrbitRuntimeTrajSample(orbit, numSamplingPointsDefault, 
        orbit->getTimeUnit()*intTimeMin/(numSamplingPointsDefault*10+1)); 
}

//----------------------------------------------------------------------------//
/// \cond INTERNAL_DOCS
// simple trajectory timestep function recording points in Poincare section
struct CPoincareParam {
    const COrbit* orbit;
    unsigned int psindex;
};

double findPS(double t, void* params)
{
    return static_cast<CPoincareParam*>(params)->orbit->getPos(static_cast<CPoincareParam*>(params)->psindex, t);
}
/// \endcond 

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimePoincare::Timestep(const double told, const double t, CPosVelPoint<double>*)
{
    // Poincare section
    double y2=orbit->getPos(PS2, t);
    double yprev=(told>0)?orbit->getPos(PS2, told):y2;
    if((ps.size()<MAX_POINCARE_POINTS) && (yprev<=0) && (y2>=0))  // put a point
    {
        double tcross=t;
        if(y2>yprev)
        {
            // interpolate coordinates at Poincare section crossing
            gsl_function F;
            CPoincareParam params;
            F.function=&findPS;
            F.params=&params;
            params.orbit=orbit;
            params.psindex=PS2;
            tcross = findRoot(&F, told, t, 1e-10);
        }
        double PS1x=orbit->getPos(PS1, tcross), PS1v=orbit->getVel(PS1, tcross);
        if((orbit->getPotential()->symmetry() & CPotential::ST_AXISYMMETRIC) == CPotential::ST_AXISYMMETRIC && 
            PS2==2 && PS1==1) {  // special treatment for axisymmetric systems - Y coordinate means R
            double X=orbit->getPos(0, tcross), Vx=orbit->getVel(0, tcross);
            double R=sqrt(pow_2(X)+pow_2(PS1x));
            PS1v=(X*Vx+PS1x*PS1v)/R;
            PS1x=R;
        }
        ps.push_back(paird(PS1x, PS1v));
        if(ps.size()==Ncross) {  // record last crossing in full glory
            lastCrossing.Pos[PS1]=PS1x;
            lastCrossing.Vel[PS1]=PS1v;
            lastCrossing.Pos[PS2]=0;
            lastCrossing.Vel[PS2]=orbit->getVel(PS2, tcross);
            int PS3=std::max<int>(std::min<int>(N_DIM-PS1-PS2,N_DIM-1),0);
            lastCrossing.Pos[PS3]=orbit->getPos(PS3, tcross);
            lastCrossing.Vel[PS3]=orbit->getVel(PS3, tcross); 
        }
    }
    return ps.size()>=Ncross ? SR_TERMINATE : SR_CONTINUE;
}

bool COrbitRuntimePoincare::getLastCrossing(CPosVelPoint<double>* cross) const
{
    if(ps.size()>=Ncross) {
        *cross=lastCrossing;
        return true;
    } else return false;
}

//----------------------------------------------------------------------------//
COrbitRuntimePericenter::COrbitRuntimePericenter(const COrbit* _orbit):
CBasicOrbitRuntimeFnc(_orbit)
{
    fitSlope=fitIntercept=fitScatter=fitSignificance=0;
    Lavg[0]=Lavg[1]=Lavg[2]=L2avg[0]=L2avg[1]=L2avg[2]=0;
    pericenterrmin=gsl_posinf();
    // init sign
    CPosVelPoint<double> ic = orbit->getInitCond();
    drdtsign= ic.Pos[0]*ic.Vel[0]+ic.Pos[1]*ic.Vel[1]+ic.Pos[2]*ic.Vel[2] >0 ? 1 : -1;
    // find approximate L2 for a circular orbit
    Lcirc2=findLcirc2(orbit->getPotential(), ic /*, orbit->getConfig()->Omega*/);
}

template<typename NumT> COrbitRuntimePericenter::COrbitRuntimePericenter(const COrbit* _orbit, 
    const std::vector< CPosVelPoint<NumT> > &traj, double timeStep):
CBasicOrbitRuntimeFnc(_orbit)
{
    fitSlope=fitIntercept=fitScatter=fitSignificance=0;
    pericenterrmin=gsl_posinf();
    Lavg[0]=Lavg[1]=Lavg[2]=L2avg[0]=L2avg[1]=L2avg[2]=0;
    drdtsign= traj[0].Pos[0]*traj[0].Vel[0]+traj[0].Pos[1]*traj[0].Vel[1]+traj[0].Pos[2]*traj[0].Vel[2] >0 ? 1 : -1;
    int N_dim=orbit->getConfig()->N_dim;  // shortcut
    for(size_t i=1; i<traj.size(); i++)
    {
        int prevsign=drdtsign;
        drdtsign= traj[i].Pos[0]*traj[i].Vel[0]+traj[i].Pos[1]*traj[i].Vel[1]+traj[i].Pos[2]*traj[i].Vel[2] >0 ? 1 : -1;
        if(drdtsign*prevsign<0 && drdtsign>0)
        {   // direction reversed - passed pericenter; do not try to refine exact time, just record current squared ang.mom.
            double Lx=(N_dim==3)? traj[i].Pos[1]*traj[i].Vel[2]-traj[i].Pos[2]*traj[i].Vel[1] : 0;
            double Ly=(N_dim==3)? traj[i].Pos[2]*traj[i].Vel[0]-traj[i].Pos[0]*traj[i].Vel[2] : 0;
            double Lz=traj[i].Pos[0]*traj[i].Vel[1]-traj[i].Pos[1]*traj[i].Vel[0];
            double L2=pow_2(Lx)+pow_2(Ly)+pow_2(Lz);
            double rperi=sqrt(pow_2(traj[i].Pos[0])+pow_2(traj[i].Pos[1])+pow_2(traj[i].Pos[2]));
            pericentert.push_back(i*timeStep);
            pericenterr.push_back(rperi);
            pericenterl2.push_back(L2);
            Lavg[0]+=Lx; L2avg[0]+=pow_2(Lx);
            Lavg[1]+=Ly; L2avg[1]+=pow_2(Ly);
            Lavg[2]+=Lz; L2avg[2]+=pow_2(Lz);
            if(rperi<pericenterrmin) pericenterrmin=rperi;
        }
    }
    // find approximate L2 for a circular orbit
    Lcirc2=findLcirc2(orbit->getPotential(), traj[0] /*, orbit->getConfig()->Omega*/);
}

/// \cond INTERNAL_DOCS
struct CPericenterParam {
    const COrbit* orbit;
};

double finddrdt(double t, void* params)
{
    const COrbit* o=static_cast<CPericenterParam*>(params)->orbit;
    return o->getPos(0,t)*o->getVel(0,t) + o->getPos(1,t)*o->getVel(1,t) + o->getPos(2,t)*o->getVel(2,t);
}
/// \endcond 

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimePericenter::Timestep(const double told, const double t, CPosVelPoint<double> *vars)
{
    if(t<=told) return SR_TERMINATE;  // shouldn't happen
    unsigned int N_dim=orbit->getConfig()->N_dim;  // shortcut
    int prevsign=drdtsign;
    drdtsign= vars->Pos[0]*vars->Vel[0] + vars->Pos[1]*vars->Vel[1] + vars->Pos[2]*vars->Vel[2] >0 ? 1 : -1;
    if(drdtsign*prevsign<0 && drdtsign>0)
    {
        // direction reversed -- passed pericenter; find exact passagee time
        gsl_function F;
        CPericenterParam params;
        params.orbit=orbit;
        F.function=&finddrdt;
        F.params=&params;
        double tperi=findRoot(&F, told, t, -1);
        if(tperi==-1)   // root finder failed - could happen if adding velocity perturbation has changed 
            tperi=told; // the sign of angular momentum at the beginning of this timestep to already positive
        pericentert.push_back(tperi);
        double yperi[N_DIM*2];
        for(unsigned int i=0; i<N_DIM; i++) {
            yperi[i]=orbit->getPos(i, tperi);
            yperi[i+N_DIM]=orbit->getVel(i, tperi);
        }
        double rperi=sqrt(pow_2(yperi[0])+pow_2(yperi[1])+pow_2(yperi[2]));
        pericenterr.push_back(rperi);
        if(rperi<pericenterrmin) pericenterrmin=rperi;
        double Lx=(N_dim==3)? yperi[1]*yperi[5]-yperi[2]*yperi[4] : 0;
        double Ly=(N_dim==3)? yperi[2]*yperi[3]-yperi[0]*yperi[5] : 0;
        double Lz=(N_dim==3)? yperi[0]*yperi[4]-yperi[1]*yperi[3] : yperi[0]*yperi[3]-yperi[1]*yperi[2];
        double L2=pow_2(Lx)+pow_2(Ly)+pow_2(Lz);
        pericenterl2.push_back(L2);
        Lavg[0]+=Lx; L2avg[0]+=pow_2(Lx);
        Lavg[1]+=Ly; L2avg[1]+=pow_2(Ly);
        Lavg[2]+=Lz; L2avg[2]+=pow_2(Lz);
    }
    return SR_TERMINATE;  // do not require anything special to be able to terminate integration
}

void COrbitRuntimePericenter::Finish()
{
    vectord lp(pericenterl2);  // !! here one may substitute either pericenter radius or pericenter ang.mom.sq., but the latter is more general
    std::sort(lp.begin(), lp.end());
    int maxn=static_cast<int>(lp.size());
    int nump=static_cast<int>(maxn*PERICENTER_FIT_FRACTION);
    if(nump<PERICENTER_FIT_MIN_POINTS) nump=PERICENTER_FIT_MIN_POINTS;
    if(nump>PERICENTER_FIT_MAX_POINTS) nump=PERICENTER_FIT_MAX_POINTS;
    if(nump>=maxn) nump=maxn-1;
    if(nump<3) { fitScatter=fitSignificance=fitSlope=fitIntercept=0; return; }
    // find out whether there are any conserved components of ang.mom.
    double L2min=0;  // minimum possible L^2 due to conserved quantities
    for(unsigned int d=0; d<N_DIM; d++)
    {
        Lavg[d]/=lp.size();
        L2avg[d]/=lp.size();
        const double eps=1e-5;  // relative accuracy of comparison (std.dev. in L[d] < eps*|L[d]| )
        if(gsl_fcmp(pow_2(Lavg[d]), L2avg[d], eps*eps)==0)
            L2min+=L2avg[d];
    }
    if(L2min>0)  // subtract it from the values for fitting
    {
        if(L2min>lp[0]) L2min=lp[0];  // to be sure that first point is not below that...
        for(int i=0; i<nump; i++)
            lp[i]-=L2min;
    }
    vectord np(nump)/*, weight(nump)*/;
    for(int i=0; i<nump; i++)
    {
        np[i]=(i+1.0)/maxn;
        /*weight[i] = (i<PERICENTER_FIT_MIN_POINTS ? 1.0 : (PERICENTER_FIT_MIN_POINTS+0.0)/i);*/  // not used -- may be improved in the future to give more weight to innermost points
    }
    double cov00, cov01, cov11, sumsq;
    // fit linear regression with and without constant term
    gsl_fit_linear(&(np.front()), 1, &(lp.front()), 1, nump, &fitIntercept, &fitSlope, &cov00, &cov01, &cov11, &sumsq);
    // compute expected dispersion assuming that chi^2/(N-2)=1
    double sigma_expected = fitSlope/maxn * sqrt(nump*1.0);
    double sigma_actual = sqrt(sumsq/(nump-2));
#if 0
    double sigma2_used = pow_2(std::max(sigma_expected, sigma_actual));
#else
    double sigma2_used = pow_2(sigma_actual);
#endif
    double chi2 = sumsq/sigma2_used;    // equals to nump-2 or less

    double fitSlopeM, cov11M, sumsqM;
    gsl_fit_mul(&(np.front()), 1, &(lp.front()), 1, nump, &fitSlopeM, &cov11M, &sumsqM);
    double deltachi2 = //(sumsqM/sumsq - 1)*(nump-2);
        sumsqM/sigma2_used - chi2;
    if(deltachi2 < 11.8 || fitIntercept<0)  // <3sigma
    {
        fitSlope=fitSlopeM;
        fitIntercept=0;
        sumsq=sumsqM;
    }
    // compute ratio of actual over expected dispersion
    fitScatter = sigma_actual/sigma_expected;
    // compute how significant (how many sigmas) is intercept!=0 with respect to zero intercept
    double stddevcount=0;
    if(deltachi2<50)  // less than ~7sigma - directly compute equivalent number of 'gaussian' sigmas from chi^2 tail
    {
        stddevcount = gsl_cdf_gaussian_Qinv(gsl_cdf_chisq_Q(deltachi2, 2)/2, 1);
    }
    else
    {
        double intrcept, stddev;
        gsl_fit_linear_est(0, fitIntercept, fitSlope, cov00, cov01, cov11, &intrcept, &stddev);
        stddevcount = intrcept/stddev;  // distance in std.dev.
    }
    fitSignificance=stddevcount;
    fitIntercept += L2min;  // add back conserved value
}

//----------------------------------------------------------------------------//
COrbitRuntimeVelPerturber::COrbitRuntimeVelPerturber(const COrbit* _orbit, const CMassModel* _model, double _mult) : 
  CBasicOrbitRuntimeFnc(_orbit), model(_model!=NULL ? new CMassModel(*_model) : NULL), mult(_mult), 
  intTime(0), deltaE(0), deltaE2(0), deltaEcumul(0), deltaL(0), deltaL2(0), deltaLcumul(0) 
{ 
    for(size_t i=0; i<15; i++) 
        avgCoefs[i]=0; 
    if(model==NULL || mult==0) return;
    // find energy below which the perturbations are so strong that orbit-averaged 
    // energy and momentum diffusion coefficients will be used instead of local velocity diffusion coefs
    Eorbitavg=model->getMinEnergyOrbitAvg(mult);
};

COrbitRuntimeVelPerturber::~COrbitRuntimeVelPerturber()
{
    delete model;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeVelPerturber::Timestep(const double told, const double t, CPosVelPoint<double> *vars)
{
    intTime+=t-told;
    if(model==NULL || mult==0) 
        return SR_TERMINATE;
    double timeinterval=t-told, timepassed=0, timestep;
    double rad=sqrt(pow_2(vars->Pos[0])+pow_2(vars->Pos[1])+pow_2(vars->Pos[2]));
    double pot=model->pot(rad);
    size_t numsteps=0;
    do{
        double dvpar, dv2par, dv2per;
        double vel=sqrt(pow_2(vars->Vel[0])+pow_2(vars->Vel[1])+pow_2(vars->Vel[2]));
        if(vel==0 || pot + vel*vel/2>0)
            return SR_TERMINATE;  // don't do anything for a non-moving particle (drag coef diverges), or for an unbound particle (let it go..)
        double mom=sqrt(   // current value of angular momentum (not used for computing instantaneous dif.coefs.)
            pow_2(vars->Pos[1]*vars->Vel[2]-vars->Pos[2]*vars->Vel[1]) +
            pow_2(vars->Pos[2]*vars->Vel[0]-vars->Pos[0]*vars->Vel[2]) +
            pow_2(vars->Pos[0]*vars->Vel[1]-vars->Pos[1]*vars->Vel[0]) );
        if(pot + vel*vel/2 < Eorbitavg) {  /// don't bother computing local coefs; just take orbit-averaged values and set new velocity
            double DEE, DE, deltaE;
            model->DifCoefEL(pot + vel*vel/2, &DEE, &DE);
            if(!gsl_finite(DEE+DE)) {
                my_error(FUNCNAME, 
                    "Error in averaged dif.coefs at E="+convertToString(pot+vel*vel/2)+", DEE="+
                    convertToString(DEE)+", DE="+convertToString(DEE));
                return SR_TERMINATE;
            }
            DEE*=mult;
            DE *=mult;
            timestep=timeinterval-timepassed;
            int ntrials=0;
            do{
                double rand1, rand2;
                getNormalRandomNumbers(&rand1, &rand2);
                deltaE = sqrt(DEE*timestep)*rand1 + DE*timestep;
                ntrials++;
            } while((!gsl_finite(deltaE) || deltaE<=-vel*vel/2) && ntrials<16);
            if(ntrials<16) {  // set new velocity with random direction and updated magnitude
                double vmag=sqrt(2*deltaE+vel*vel);
                getRandomUnitVector(&(vars->Vel[0]), &vars->Vel[1], &vars->Vel[2]);
                vars->Vel[0]*=vmag;
                vars->Vel[1]*=vmag;
                vars->Vel[2]*=vmag;
                //std::cerr << "+";
            }
            // add current values of drift and diffusion coefficients, for computing time-averaged values
            avgCoefs[0] += timestep*DEE;  // D_EE
            avgCoefs[6] += timestep*DE;   // D_E
            // skip accumulation of averaged values of D_EL, D_LL, D_L, as they are more costly to compute
            //avgCoefs[14] += timestep*mom;  // avg.ang.mom.
        } else { 
            // compute local velocity diffusion coefs
            model->DifCoefVel(rad, vel, &dvpar, &dv2par, &dv2per);
            dvpar *=mult;
            dv2par*=mult;
            dv2per*=mult;
            if(!gsl_finite(dvpar+dv2par+dv2per)) {
                my_error(FUNCNAME,
                    "Error in dif.coef. at r="+convertToString(rad)+",v="+convertToString(vel)+
                    ",E="+convertToString(model->pot(rad)+vel*vel/2)+": dvpar="+convertToString(dvpar)+
                    ",dv2par="+convertToString(dv2par)+",dv2per="+convertToString(dv2per));
                return SR_TERMINATE;  // some weird error, dump it
            }
            double maxdeltav=0.4*std::min<double>(vel, sqrt(std::max<double>(-2*pot-vel*vel,-0.001*pot)) );
            double maxtimestep=std::min<double>(pow_2(maxdeltav)/(dv2par+dv2per), maxdeltav/fabs(dvpar/vel));
            timestep = std::min<double>(maxtimestep, timeinterval-timepassed);   // may be necessary to perform several steps
            // add current values of drift and diffusion coefficients, for computing time-averaged values
            avgCoefs[0] += timestep*vel*vel*dv2par;  // D_EE
            avgCoefs[2] += timestep*(pow_2(mom/vel)*(dv2par-dv2per/2) + rad*rad/2*dv2per);  // D_LL
            avgCoefs[4] += timestep*mom*dv2par;  // D_EL
            avgCoefs[6] += timestep*(dvpar + dv2par/2 + dv2per/2);  // D_E
            avgCoefs[8] += timestep*(mom>0 ? mom/pow_2(vel)*dvpar + rad*rad/4/mom*dv2per : 0); // D_L
            //avgCoefs[14] += timestep*mom;  // avg.ang.mom.
            // sample actual value of velocity perturbation
            double newv[3];
            int ntrials=0;
            do {
                double rand1, rand2;
                getNormalRandomNumbers(&rand1, &rand2);
                double deltavpar = rand1*sqrt(dv2par*timestep) + dvpar/vel*timestep;
                double deltavper = rand2*sqrt(dv2per*timestep);
                if(!gsl_finite(deltavpar+deltavper)) {
                    my_error(FUNCNAME,
                        "Error in dif.coef. at r="+convertToString(rad)+",v="+convertToString(vel)+
                        ",E="+convertToString(model->pot(rad)+vel*vel/2)+": dvpar="+convertToString(dvpar)+
                        ",dv2par="+convertToString(dv2par)+",dv2per="+convertToString(dv2per)+
                        " results in deltavpar="+convertToString(deltavpar)+",deltavper="+convertToString(deltavper));
                    return SR_TERMINATE;  // some weird error, dump it
                }
                // assign parallel and perpendicular velocity components
                double rx, ry, rz;  // random vector, assuming 3d motion (diffusion coefs are not valid for 2d anyway)
                getRandomUnitVector(&rx, &ry, &rz);
                double rdotv=rx*vars->Vel[0]+ry*vars->Vel[1]+rz*vars->Vel[2];  // projection of random vector on the velocity
                rx -= rdotv*vars->Vel[0]/(vel*vel);
                ry -= rdotv*vars->Vel[1]/(vel*vel);
                rz -= rdotv*vars->Vel[2]/(vel*vel); // subtract parallel component
                double rmag=sqrt(pow_2(rx)+pow_2(ry)+pow_2(rz));
                if(rmag==0) rmag=1;  // avoid singularity
                // add perturbations to the velocity
                newv[0] = vars->Vel[0] + vars->Vel[0]/vel*deltavpar + rx/rmag*deltavper;
                newv[1] = vars->Vel[1] + vars->Vel[1]/vel*deltavpar + ry/rmag*deltavper;
                newv[2] = vars->Vel[2] + vars->Vel[2]/vel*deltavpar + rz/rmag*deltavper;
                ntrials++;
            }   // check that new energy is still below zero
            while( ntrials<16 && pot + (pow_2(newv[0])+pow_2(newv[1])+pow_2(newv[2]))/2 >= 0 );
            if(ntrials<16 && gsl_finite(newv[0]+newv[1]+newv[2])) {  // update velocity if it will not unbind the particle
                vars->Vel[0]=newv[0];
                vars->Vel[1]=newv[1];
                vars->Vel[2]=newv[2];
            }
        }
        // record actual changes in energy and momentum resulting from perturbations
        double mom_new=sqrt(
            pow_2(vars->Pos[1]*vars->Vel[2]-vars->Pos[2]*vars->Vel[1]) +
            pow_2(vars->Pos[2]*vars->Vel[0]-vars->Pos[0]*vars->Vel[2]) +
            pow_2(vars->Pos[0]*vars->Vel[1]-vars->Pos[1]*vars->Vel[0]) );
        //double deltaEstep= pow_2(deltavper)/2 + pow_2(deltavpar)/2 + vel*deltavpar;
        double deltaEstep=(pow_2(vars->Vel[0])+pow_2(vars->Vel[1])+pow_2(vars->Vel[2]) - vel*vel)*0.5;
        double deltaLstep= mom_new-mom;
        deltaEcumul+=deltaEstep; // current accumulated difference
        deltaE += deltaEcumul*timestep;
        deltaE2+= pow_2(deltaEcumul)*timestep;
        deltaLcumul+=deltaLstep; // current accumulated difference
        deltaL += deltaLcumul*timestep;
        deltaL2+= pow_2(deltaLcumul)*timestep;
        timepassed+=timestep;
        numsteps++;
        if(!gsl_finite(deltaE)) 
            my_error(FUNCNAME,
            "Energy is "+convertToString(deltaE)+" at time t="+convertToString(t));
    } while(timepassed<timeinterval && numsteps<100);   // last condition to avoid deadlock for v->0?
    return SR_REINIT;
}

void COrbitRuntimeVelPerturber::Finish()
{
    if(model==NULL || mult==0 || intTime==0) return;
    // compute actual mean and mean-square changes in energy and ang.mom. during the entire integration
    avgCoefs[10] = deltaE/intTime;
    avgCoefs[11] = sqrt(deltaE2/intTime);
    avgCoefs[12] = deltaL/intTime;
    avgCoefs[13] = sqrt(deltaL2/intTime);
    avgCoefs[14] = deltaEcumul;
#if 0  // unused
    for(size_t i=0; i<10; i+=2)
        avgCoefs[i]/=intTime;
    // compute expected orbit-averaged coefs from spherical mass model with the same mean E and L
    CPosVelPoint<double> pt=orbit->getInitCond();
    double E=model->pot(sqrt(pow_2(pt.Pos[0])+pow_2(pt.Pos[1])+pow_2(pt.Pos[2]))) +
        (pow_2(pt.Vel[0])+pow_2(pt.Vel[1])+pow_2(pt.Vel[2]))/2;
    if(!gsl_finite(E)) 
        my_error(FUNCNAME, "Error in initial energy E="+convertToString(E));
    E+=avgCoefs[10];  // = Einit + avg.deltaE
    if(!gsl_finite(E)) 
        my_error(FUNCNAME, "Error in averaged energy E="+convertToString(E));
    model->DifCoefEL(E, &(avgCoefs[1]), &(avgCoefs[7]), &(avgCoefs[3]));
    if(!gsl_finite(avgCoefs[1]+avgCoefs[3]+avgCoefs[7])) {
        my_error(FUNCNAME, "Error in computing averaged dif.coefs at E="+convertToString(E));
        avgCoefs[1]=avgCoefs[3]=avgCoefs[7]=0;  // dump it
    }
    for(size_t i=1; i<10; i+=2)
        avgCoefs[i]*=mult;
#endif
}

COrbitRuntimeVelPerturberCreator::COrbitRuntimeVelPerturberCreator(const CMassModel* _model, double _mult) :
  model(new CMassModel(*_model)), mult(_mult) {};

COrbitRuntimeVelPerturberCreator::~COrbitRuntimeVelPerturberCreator()
{ if(model!=NULL) delete model; }

/* ------- Frequency analysis routines -------- */
/** utility function to ensure that n does not have large prime divisors.
    if needed, decrease it (in the FFT primes are great evil). */
void tuneNonPrime(size_t *n)
{
    assert(n);
    size_t m=*n, f=2;
    while(f*f<=m)
        if(m%f==0) m/=f; else f++;
    if(m>=MAX_FFT_PRIME && *n>=MAX_FFT_PRIME)
    {
        (*n)--;
        tuneNonPrime(n);
    }
}

// perform FFT of orbit (output amplitudes to 'fa' array) and find 'maxnlines' lines, which will be placed into 'sl' array. 
// If fa=NULL, do not output frequency amplitudes
void CSpectrumAnalyzerHunter::findSpectralLines(smile::CSpectrum *spectralLines, smile::vectord *spectrumAmplitudes)
{
    size_t npoints=timeSeries.size();
    if(npoints==0 || spectralLines==NULL) return;
    tuneNonPrime(&npoints);
    gsl_fft_real_wavetable * wavetable = gsl_fft_real_wavetable_alloc (npoints);
    gsl_fft_real_workspace * workspace = gsl_fft_real_workspace_alloc (npoints);
    vectord in(timeSeries);   // copy of input array; FFT is performed in-place
    gsl_fft_real_transform (&(in.front()), 1, npoints, wavetable, workspace);
    gsl_fft_real_wavetable_free (wavetable);
    gsl_fft_real_workspace_free (workspace);
    std::vector<complexd> fc(npoints/2+1);  // working array of complex amplitudes
    fc[0]=complexd(in[0],0)  / (0.5*npoints);
    for(size_t i = 1; i < npoints/2; i++)
        fc[i]=complexd(in[2*i-1], in[2*i])/(0.5*npoints);
    if(npoints%2==0)
        fc[npoints/2]=complexd(in[npoints-1], 0)/(0.5*npoints);
    if(spectrumAmplitudes!=NULL)
    {  // output amplitudes only (for illustrative purposes) if necessary
        spectrumAmplitudes->resize(npoints/2+1, 0);
        for(size_t i=0; i<npoints/2+1; i++)
            spectrumAmplitudes->at(i)=abs(fc[i]);    // store amplitude of spectrum if needed - and then forget it; only the true complex spectrum is used hereafter
    }
    // perform frequency analysis
    spectralLines->clear();
    bool finish=false;
    double maxamp=0; double ma=0;
    while(!finish)
    {   // find next greatest line
        int m=0; 
        ma=0;
        for(int i=0; i<static_cast<int>(npoints/2)-2; i++) {
            if(abs(fc[i])>ma)
            {
                ma=abs(fc[i]);
                m=i;
            }
        }
        if(maxamp==0 && m>0) maxamp=ma;  // record max.amplitude if the frequency is non-zero
        if(ma>maxamp*MIN_SPECTRAL_LINE_AMPLITUDE)
        {   // add line
            // peak is somewhere near m. 
            if(m>0)
            {
                int m1=m;
                if(m>1 && abs(fc[m-1])>abs(fc[m+1]))  m1--;
                double s;   // fractional index of peak frequency 
                complexd Ac;// amplitude
                complexd f1 = (fc[m1] - (fc[m1-1]+fc[m1+1])/2.0);
                complexd f2 = (fc[m1+1] - (fc[m1]+fc[m1+2])/2.0);
                // C=f2/f1; check out that C<>0 and C<>infinity
                if(abs(f1) < abs(f2)*1e-4)
                { 
                    m1++;
                    f1=f2;
                    f2 = (fc[m1+1] - (fc[m1]+fc[m1+2])/2.0);
                }
                if((abs(f2) < abs(f1)*1e-4) && (m1>1))  // !!!what happens here???
                { 
                    m1--;
                    f2=f1;
                    f1 = (fc[m1] - (fc[m1-1]+fc[m1+1])/2.0);
                }
                complexd C = f2 / f1;
                double Ca=C.real();  // in ideal case, Im C=0
                double omega=2*M_PI/npoints;
                int method=1;  // Hunter's method
                if((Ca<-1.0/FREQ_PRECISE_THRESHOLD) && (Ca>-FREQ_PRECISE_THRESHOLD)/*fabs(C.imag())*0 < fabs(Ca)*/)
                {   // try Hunter's precise method
                    double sgn = (Ca<1)?+1:-1;
                    double t = (Ca*cos(omega)-1 + sgn * sqrt( pow_2(Ca*cos(omega)-1) - (1+2*Ca)*pow_2(sin(omega))))/sin(omega);
                    s = m1 + npoints/M_PI * atan(t);
                    complexd z = std::polar(1.0, 2*M_PI*(s-m1)/npoints); //complexd((1-pow_2(t))/(1+pow_2(t)), 2*t/(1+pow_2(t)));
                    complexd beta = std::polar(1.0, omega);
                    complexd P = (std::polar(1.0, 2*M_PI*s)-1.0) / (std::polar(1.0, 2*M_PI*(s-m1)/npoints)-1.0) / (1.0*npoints);
                    complexd g1 = P * (1.0 - 0.5*(z-1.0)*(1.0/(z*beta-1.0) + 1.0/(z*std::conj(beta)-1.0)));
                    Ac = f1/g1;
                }
                else
                {   // Follow the method by Carpintero&Aguilar(1998)
                    method=0;
                    double alpha = atan(sin(omega/2)/(cos(omega/2)+abs(fc[m1]/fc[m1+1])));
                    s = m1 + npoints*alpha/M_PI;
                    complexd P=(std::polar(1.0, 2*alpha*npoints)-1.0) / (std::polar(1.0, 2*alpha)-1.0) / (1.0*npoints);
                    //complexd Q=(std::polar(1.0, -2*M_PI*(s+m1))-1.0) / (std::polar(1.0, -2*M_PI*(s+m1)/npoints)-1.0) / (1.0*npoints);
                    Ac=fc[m1] / P;
                }
                spectralLines->push_back(CSpectralLine(s/(timeStep*npoints), abs(Ac), std::arg(Ac), method));
                if(spectralLines->size()>=MAX_SPECTRAL_LINES) finish=true;    // not to iterate forever
                // subtract line from spectrum
                if(!finish)
                {
                    if(s-floor(s) < 1e-5)    // frequency was very close to integer value, so subtract only this point
                    {
                        fc[(int)floor(s)]-=Ac;
                    } else
                        if(s-floor(s) > 1-1e-5)
                        {
                            fc[(int)floor(s)+1]-=Ac;
                        }
                        else   // subtract this line with its wings from the whole spectrum
                        {
                            complexd Amul = Ac / (1.0*npoints) * (std::polar(1.0, 2*M_PI*(s))-1.0);
                            complexd Acmul = std::conj(Ac) / (1.0*npoints) * (std::polar(1.0,-2*M_PI*(s))-1.0);
                            int jmin=std::max<int>(0, m-1000);
                            int jmax=std::min<int>(static_cast<int>(npoints/2)+1, m+1000);
                            for(int j=jmin; j<jmax; j++)
                            {
                                complexd linea = Amul / (std::polar(1.0, 2*M_PI*(s-j)/npoints)-1.0) +
                                    Acmul / (std::polar(1.0,-2*M_PI*(s+j)/npoints)-1.0);
                                fc[j] -= linea;
                            }
                        }
                }
            }
            else
            {   // zero frequency -- just nullify one array element (do not record zero frequencies)
                fc[m]=0;
            }
        }
        else finish=true;
    }
    std::sort(spectralLines->begin(), spectralLines->end());
    if(spectralLines->size()==0)
        spectralLines->push_back(CSpectralLine());  // add zero to ensure that array is non-empty
}

/** starting from the given initial conditions in a plane x[coord_section]=0, x'[coord_section]!=0, 
    integrate an orbit until it crosses the same plane for N-th time. 
    Return the final coordinates at this crossing. */
bool computeNperiods(const CConfigOrbit* configOrbit, const CPotential* potential, 
    const CPosVelPoint<double> &initCond, int coord_section, int Nperiods, CPosVelPoint<double> *endCond)
{
    vectorRuntimeFncCreators crts;
    int coord1= coord_section==1 ? 0 : 1; // doesn't matter, will record full phase-space point anyway
    crts.push_back(new COrbitRuntimePoincareCreator(coord1, coord_section, Nperiods));
    double minTime = computeTimeUnit(potential, initCond /*, configOrbit->Omega*/)/2.0;
    double maxTime = minTime*Nperiods*10.0;  // arbitrary!
    COrbit* orb=new COrbit(configOrbit, potential, initCond, &crts, minTime, maxTime);
    // Attention! if the poincare section is not crossed ever, the integration may continue for a very long time!
    orb->run();   // should perform Nperiods crossings through x[coord]=0 plane
    const COrbitRuntimePoincare* pi = static_cast<const COrbitRuntimePoincare*>(orb->getRuntimeFnc(0));
    bool success=pi->getLastCrossing(endCond);
    delete orb;
    delete crts[0];
    return success;
}

/// \cond INTERNAL_DOCS
struct CPeriodicOrbitFinderParam {
    const CConfigOrbit* configOrbit;
    const CPotential* potential;
    double E;
    unsigned int Nperiods;
    unsigned int coord_section;
    unsigned int coord_first;
    int nfnc;
};
/// \endcond
int gsl_wrapper_computeNperiods(const gsl_vector *initC, void* param, gsl_vector *values, gsl_matrix* jac)
{
    CPeriodicOrbitFinderParam* par=static_cast<CPeriodicOrbitFinderParam*>(param);
    unsigned int coord_second=N_DIM-par->coord_first-par->coord_section;
    if(coord_second==par->coord_first || coord_second==par->coord_section || coord_second>=N_DIM || gsl_isnan(gsl_vector_get(initC, 0)))
        return GSL_EBADFUNC;
    size_t numeval = jac==NULL ? 1 : 1+initC->size;
    double delta=1e-8;//LYAP_DEV_INIT;  //!!!? use an adaptive deviation magnitude
    CPosVelPoint<double> initCond, endCond, endCond0;
    for(size_t eval=0; eval<numeval; eval++) {
        initCond.Pos[par->coord_first]=gsl_vector_get(initC, 0) + delta*(eval==1);
        initCond.Vel[par->coord_first]=gsl_vector_get(initC, 1) + delta*(eval==2);
        if(initC->size>2) {
            initCond.Pos[coord_second]=gsl_vector_get(initC, 2) + delta*(eval==3);
            initCond.Vel[coord_second]=gsl_vector_get(initC, 3) + delta*(eval==4);
        } else {
            initCond.Pos[coord_second]=0;
            initCond.Vel[coord_second]=0;
        } 
        initCond.Pos[par->coord_section]=0;
        //std::cerr << "D="<<eval<<", IC=["<<initCond.Pos[0]<<","<<initCond.Pos[1]<<","<<initCond.Pos[2]<<";";
        double us2=2*(par->E-par->potential->Phi(initCond.Pos[0], initCond.Pos[1], initCond.Pos[2])) + 
            pow_2(par->configOrbit->Omega)*(pow_2(initCond.Pos[0])+pow_2(initCond.Pos[1])) - 
            (pow_2(initCond.Vel[par->coord_first]) + pow_2(initCond.Vel[coord_second]));
        if(us2<0) {
            //std::cerr << "no vel\n";
            return GSL_EBADFUNC;
        }
        double vs=sqrt(us2);
        if(par->coord_section==0) vs-=par->configOrbit->Omega*initCond.Pos[1]; else
        if(par->coord_section==1) vs+=par->configOrbit->Omega*initCond.Pos[0];
        initCond.Vel[par->coord_section]=vs;
        //std::cerr << initCond.Vel[0]<<","<<initCond.Vel[1]<<","<<initCond.Vel[2]<<"]";
        par->nfnc++;
        if(!computeNperiods(par->configOrbit, par->potential, initCond, par->coord_section, par->Nperiods, &endCond))
            return GSL_EBADFUNC;
        if(eval==0) {
            //std::cerr<<" EC-IC=["<<(endCond.Pos[0]-initCond.Pos[0])<<","<<(endCond.Pos[1]-initCond.Pos[1])<<","<<(endCond.Pos[2]-initCond.Pos[2])<<";"
            //    <<(endCond.Vel[0]-initCond.Vel[0])<<","<<(endCond.Vel[1]-initCond.Vel[1])<<","<<(endCond.Vel[2]-initCond.Vel[2])<<"]\n";
            endCond0=endCond;
            if(values!=NULL) {
                gsl_vector_set(values, 0, endCond.Pos[par->coord_first]-initCond.Pos[par->coord_first]);
                gsl_vector_set(values, 1, endCond.Vel[par->coord_first]-initCond.Vel[par->coord_first]);
                if(initC->size>2) {
                    gsl_vector_set(values, 2, endCond.Pos[coord_second]-initCond.Pos[coord_second]);
                    gsl_vector_set(values, 3, endCond.Vel[coord_second]-initCond.Vel[coord_second]);
                }
            }
        } else {  // set jacobian matrix
            //std::cerr<<" EC-EC0=["<<(endCond.Pos[0]-endCond0.Pos[0])<<","<<(endCond.Pos[1]-endCond0.Pos[1])<<","<<(endCond.Pos[2]-endCond0.Pos[2])<<";"
            //<<(endCond.Vel[0]-endCond0.Vel[0])<<","<<(endCond.Vel[1]-endCond0.Vel[1])<<","<<(endCond.Vel[2]-endCond0.Vel[2])<<"]\n";
            gsl_matrix_set(jac, 0, eval-1, (endCond.Pos[par->coord_first]-endCond0.Pos[par->coord_first]-delta*(eval==1))/delta);
            gsl_matrix_set(jac, 1, eval-1, (endCond.Vel[par->coord_first]-endCond0.Vel[par->coord_first]-delta*(eval==2))/delta);
            if(initC->size>2) {
                gsl_matrix_set(jac, 2, eval-1, (endCond.Pos[coord_second]-endCond0.Pos[coord_second])/delta);
                gsl_matrix_set(jac, 3, eval-1, (endCond.Vel[coord_second]-endCond0.Vel[coord_second])/delta);
            }
        }
    }
    return GSL_SUCCESS;
};

int gsl_wrapper_computeNperiods_f(const gsl_vector *initC, void* param, gsl_vector *values)
{
    return gsl_wrapper_computeNperiods(initC, param, values, NULL);
}

int gsl_wrapper_computeNperiods_df(const gsl_vector *initC, void* param, gsl_matrix* jac)
{
    return gsl_wrapper_computeNperiods(initC, param, NULL, jac);
}

bool findPeriodicOrbit(const CPotential* potential, double Omega, 
    unsigned int coord_section, unsigned int coord_first, bool threeD, unsigned int Nperiods, CPosVelPoint<double> *IC)
{
    size_t Nvars = (threeD ? 4 : 2);  // whether we are looking for a 2d or 3d system, which have 2d and 4d surfaces of section, correspondingly
    if(coord_section==coord_first || coord_section>=N_DIM || coord_first>=N_DIM)
        return false;  // invalid call
    unsigned int coord_second=N_DIM-coord_first-coord_section;
    CConfigOrbit configOrbit = {
        3,     // N_dim
        Omega, // Omega
        CBasicOdeIntegrator::SK_DOP853,  // integratorType
        1e-9,  // accuracyRel
        1e-9,  // accuracyAbs
        1e-4,  // accuracyIAS15
        0.4,   // accuracyTreecode
        false, // treecodeSymmetrizeTimestep
        0.005, // accuracyFixedTimestep
        0.0,   // samplingInterval
        0,     // calcLyapunovMethod
        0,     // numDevVectors
        0.0,   // relaxRate
        false, // freqStrictOrder
        0.0,   // minFreqX
        0.0    // minFreqYZ
    };
    //configOrbit.N_dim = !threeD && coord_first<2 && coord_section<2 ? 2 : 3;   // if we have z-coordinate involved anywhere, use integration in 3d
    gsl_vector *vec=gsl_vector_alloc(Nvars);
    gsl_multiroot_fdfsolver *solver=gsl_multiroot_fdfsolver_alloc(gsl_multiroot_fdfsolver_hybridsj, Nvars);
    gsl_multiroot_function_fdf F;
    CPeriodicOrbitFinderParam param;
    F.f=&gsl_wrapper_computeNperiods_f;
    F.df=&gsl_wrapper_computeNperiods_df;
    F.fdf=&gsl_wrapper_computeNperiods;
    F.n=Nvars;
    F.params=&param;
    param.configOrbit=&configOrbit;
    param.potential=potential;
    param.coord_section=coord_section;
    param.coord_first=coord_first;
    param.Nperiods=Nperiods;
    param.E=totalEnergy(potential, *IC, Omega);
    param.nfnc=0;
    CPosVelPoint<double> IC1;
    CPosVelPoint<double>* myIC=IC;
    if(IC->Pos[coord_section]!=0 || IC->Vel[coord_section]<0) 
    {  // we are not on the surface of section yet; compute the first intersection and start over from it
        if(!computeNperiods(&configOrbit, potential, *IC, coord_section, 1, &IC1))
            return false;
        param.nfnc++;
        myIC=&IC1;
    } 
    gsl_vector_set(vec, 0, myIC->Pos[coord_first]);
    gsl_vector_set(vec, 1, myIC->Vel[coord_first]);
    if(threeD) {
        gsl_vector_set(vec, 2, myIC->Pos[coord_second]);
        gsl_vector_set(vec, 3, myIC->Vel[coord_second]);
    }
    gsl_multiroot_fdfsolver_set(solver, &F, vec);
    int status, niter=0;
    double eps=1e-6;
    do {
        status = gsl_multiroot_fdfsolver_iterate(solver);
        if(status!=GSL_SUCCESS)
            break;
        status = gsl_multiroot_test_residual(solver->f, eps);
        niter++;
    } while(status==GSL_CONTINUE && niter<100);
    if(status==GSL_SUCCESS) {
        IC->Pos[coord_first]=gsl_vector_get(solver->x, 0);
        IC->Vel[coord_first]=gsl_vector_get(solver->x, 1);
        if(threeD) {
            IC->Pos[coord_second]=gsl_vector_get(solver->x, 2);
            IC->Vel[coord_second]=gsl_vector_get(solver->x, 3);
        } 
        IC->Pos[coord_section]=0;
        double vs=sqrt( 2*(param.E-potential->Phi(IC->Pos[0], IC->Pos[1], IC->Pos[2])) + 
            pow_2(Omega)*(pow_2(IC->Pos[0])+pow_2(IC->Pos[1])) - 
            (pow_2(IC->Vel[coord_first]) + pow_2(IC->Vel[coord_second])) );
        if(coord_section==0) vs-=Omega*IC->Pos[1]; else
        if(coord_section==1) vs+=Omega*IC->Pos[0];
        IC->Vel[coord_section]=vs;
    }
    //my_message(std::string(status==GSL_SUCCESS ? "Found" : "Not found")+" periodic orbit in "+
    //    convertToString(param.nfnc)+" integrations, "+convertToString(niter)+" iterations of solver");
    gsl_multiroot_fdfsolver_free(solver);
    gsl_vector_free(vec);
    return status==GSL_SUCCESS;
}

} // namespace
