/** \file    orbitan.h
    \brief   Orbit analysis classes
    \author  EV
    \date    2009-2014

    This file defines orbit analysis methods, implemented in various classes derived from 
    smile::CBasicOrbitRuntimeFnc, for instance, routines for spectrum analysis, 
    Poincare surface of section and other stuff.
    Each runtime function class has its own information object class derived from 
    smile::CBasicInformation, and a creator class derived from smile::CBasicRuntimeFncCreator.
*/

#pragma once
#include "common.h"
#include <complex>

namespace smile{

/* -------------- Spectrum analysis data and routines --------------- */

/** Definition of one spectral line in one-dimensional time series (nothing to do with spectroscopy!:) */
struct CSpectralLine
{
    double freq;  ///< frequency
    double ampl;  ///< amplitude
    double phase; ///< phase (complex amplitude = ampl * exp(I*phase)
    int method;   ///< 0: Carpintero&Aguilar, 1: Hunter (more precise); other variants are not implemented yet
    CSpectralLine(double _freq=0, double _ampl=0, double _phase=0, int _method=0) : 
      freq(_freq), ampl(_ampl), phase(_phase), method(_method) {}
    inline bool operator<(CSpectralLine S) const
    {   return (ampl>S.ampl); }    ///< sort in reverse order of amplitude
    inline operator double() const { return freq; }
};
/** An array of spectral lines in one-dimensional time series */
typedef std::vector<CSpectralLine> CSpectrum;
typedef std::complex<double> complexd;   ///< convenience shorthand for a complex number

/** Basic class for various methods of spectral analysis */
class CBasicSpectrumAnalyzer {
public:
    virtual ~CBasicSpectrumAnalyzer() {};
    /** Frequency analysis routine: perform Fourier transform and find up to Nmax lines in the spectrum.
        input is provided by the constructor in derived classes 
        (it may consist of one-dimensional real array of x(t) for the Hunter's method, 
        or a complex array of { x(t),v(t) } for the Laskar's method).
        \param[out]  spectralLines array (must be a pointer to an existing vector); 
        \param[out]  spectrumAmplitudes (absolute values of complex spectrum) if the pointer to this vector is not NULL
    */
    virtual void findSpectralLines(CSpectrum *spectralLines, vectord *spectrumAmplitudes=NULL) = 0;
};

/** Spectrum analysis based on Hunter's DFT method.
    It determines the location of spectral lines with a high accuracy, using Hanning filtering;
    the backup method of Carpintero&Aguilar without filtering is used in case of a poorly resolved line
*/
class CSpectrumAnalyzerHunter: public CBasicSpectrumAnalyzer {
public:
    /** Constructor takes the one-dimensional array of equally spaced x(t), and the timestep */
    CSpectrumAnalyzerHunter(const vectord &_timeSeries, const double _timeStep) : 
      timeSeries(_timeSeries), timeStep(_timeStep) {};
    virtual void findSpectralLines(CSpectrum *spectralLines, vectord *spectrumAmplitudes=NULL);
private:
    const vectord &timeSeries;   ///< reference to the source array of x(t)
    const double timeStep;       ///< spacing between elements of the array (used to scale the frequencies)
};

/* ------------- Definition of various orbit information classes ---------- */

/** information object containing orbit analysis results.
    It represents some general data about the orbit, like the textual description of 
    orbit class, frequencies and chaotic properties, etc. 
*/
template<typename NumT> class COrbitInformation: public CBasicInformation
{
public:
    explicit COrbitInformation(const std::string &_Description) :
      Description(_Description), Einit(0), Ediff(0), timeUnit(0), lfccdiff(0), lambda(0) 
    { 
        for(unsigned int d=0; d<N_DIM; d++) { 
            lf[d]=0; 
            inertia[d]=0; 
            Lavg[d]=0; 
            Lvar[d]=0; 
        } 
    };
    COrbitInformation(const std::string &_Description, const NumT _Einit, const NumT _Ediff, 
        const NumT _timeUnit, const NumT _lf[N_DIM], const NumT _lfccdiff, const NumT _lambda, 
        const NumT _inertia[N_DIM], const NumT _Lavg[N_DIM], const NumT _Lvar[N_DIM]) :
      Description(_Description), Einit(_Einit), Ediff(_Ediff), 
      timeUnit(_timeUnit), lfccdiff(_lfccdiff), lambda(_lambda) 
    { 
        for(unsigned int d=0; d<N_DIM; d++) { 
            lf[d]=_lf[d]; 
            inertia[d]=_inertia[d]; 
            Lavg[d]=_Lavg[d]; 
            Lvar[d]=_Lvar[d]; 
        } 
    };
    COrbitInformation(const COrbitInformation<float> &src) :
      Description(src.Description), Einit(static_cast<NumT>(src.Einit)), Ediff(static_cast<NumT>(src.Ediff)), 
      timeUnit(static_cast<NumT>(src.timeUnit)), lfccdiff(static_cast<NumT>(src.lfccdiff)), lambda(static_cast<NumT>(src.lambda)) 
    { 
        for(unsigned int d=0; d<N_DIM; d++) { 
            lf[d]=static_cast<NumT>(src.lf[d]); 
            inertia[d]=static_cast<NumT>(src.inertia[d]); 
            Lavg[d]=static_cast<NumT>(src.Lavg[d]); 
            Lvar[d]=static_cast<NumT>(src.Lvar[d]); 
        } 
    };
    COrbitInformation(const COrbitInformation<double> &src) :
      Description(src.Description), Einit(static_cast<NumT>(src.Einit)), Ediff(static_cast<NumT>(src.Ediff)), 
      timeUnit(static_cast<NumT>(src.timeUnit)), lfccdiff(static_cast<NumT>(src.lfccdiff)), lambda(static_cast<NumT>(src.lambda)) 
    { 
        for(unsigned int d=0; d<N_DIM; d++) { 
            lf[d]=static_cast<NumT>(src.lf[d]); 
            inertia[d]=static_cast<NumT>(src.inertia[d]); 
            Lavg[d]=static_cast<NumT>(src.Lavg[d]); 
            Lvar[d]=static_cast<NumT>(src.Lvar[d]); 
        } 
    };
    virtual INFOTYPE infoType() const { return sizeof(NumT)==sizeof(float) ? IT_TRAJ_ANALYSIS : IT_UNKNOWN; };  // streams support only float instantiations of this class
    virtual std::string toString() const;
    // get functions
    std::string getDescription() const { return Description; }
    NumT getEinit() const { return Einit; }
    NumT getEdiff() const { return Ediff; }
    NumT getTimeUnit() const { return timeUnit; }
    NumT getlf(unsigned int index) const { if(index<N_DIM) return lf[index]; else return -1; }
    NumT getlfccdiff() const { return lfccdiff; }
    NumT getlambda() const { return lambda; }
    NumT getinertia(unsigned int index) const { if(index<N_DIM) return inertia[index]; else return -1; }
    NumT getLavg(unsigned int index) const { if(index<N_DIM) return Lavg[index]; else return -1; }
    NumT getLvar(unsigned int index) const { if(index<N_DIM) return Lvar[index]; else return -1; }
private:
    std::string Description;     ///< description of orbit type
    NumT Einit, Ediff;           ///< initial energy and energy conservation error after integration
    NumT timeUnit;               ///< time unit for measuring frequencies
    NumT lf[N_DIM];              ///< leading frequencies
    NumT lfccdiff;               ///< rate of LFCC diffusion
    NumT lambda;                 ///< Lyapunov exponent
    NumT inertia[N_DIM];         ///< inertia tensor components (only diagonal in Cartesian coordinates)
    NumT Lavg[N_DIM];            ///< average components of angular momentum
    NumT Lvar[N_DIM];            ///< variation in angular momentum components
};

/// information object for pericenter data (distribution of angular momentum at pericenter passages)
template<typename NumT> class CPericenterInformation: public CBasicInformation
{
public:
    CPericenterInformation(const NumT _fitIntercept, const NumT _fitSlope, const NumT _fitScatter, const NumT _fitSignificance, const NumT _Lcirc2) : 
      fitIntercept(_fitIntercept), fitSlope(_fitSlope), fitScatter(_fitScatter), fitSignificance(_fitSignificance), Lcirc2(_Lcirc2) {};
    virtual INFOTYPE infoType() const { return sizeof(NumT)==sizeof(float) ? IT_PERICENTER : IT_UNKNOWN; };
    virtual std::string toString() const;
    /// get data (all output parameters must point to existing variables which are filled with the values of parameters)
    void getParams(NumT* _fitIntercept, NumT* _fitSlope, NumT* _fitScatter, NumT* _fitSignificance, NumT* _Lcirc2) const 
      { *_fitIntercept=fitIntercept; *_fitSlope=fitSlope; *_fitScatter=fitScatter; *_fitSignificance=fitSignificance; *_Lcirc2=Lcirc2; };
private:
    /// \name fitting coefficients for distribution of squared angular momentum at pericenter
    NumT fitIntercept;    ///< L_{min}^2 - estimated lower bound on angular momentum at pericenter passage (0 means a centrophilic orbit)
    NumT fitSlope;        ///< slope of the regression  L^2_{k} = L^2_{min} + slope*k/N_{peri}  for k=1..N_{peri}
    NumT fitScatter;      ///< scatter of points around the best-fit regression, normalized to the `typical' value (should generally be <0.5 for the regressionto be reliable)
    NumT fitSignificance; ///< difference of L_{min}^2 from zero, measured in `sigma's (if less than a few, probably the orbit is centrophilic)
    NumT Lcirc2;          ///< approximate squared angular momentum of a circular orbit
};

/// information object to store points from the Poincare surface of section
class CPoincareInformation: public CBasicInformation
{
public:
    CPoincareInformation(const std::vector< std::pair<double, double> > &_ps) : ps(_ps) {};
    virtual INFOTYPE infoType() const { return IT_POINCARE; };
    virtual std::string toString() const { return std::string(); }
    unsigned int size() const { return static_cast<unsigned int>(ps.size()); }
    const std::pair<double, double>& at(unsigned int index) const { return ps.at(index); }
private:
    std::vector< std::pair<double, double> > ps;  ///< Poincare section  (x - xdot plane)
};

/// information object keeping a trajectory sample (to generate an N-body model from a Schwarzschild model)
template<typename NumT> class CTrajSampleInformation: public CBasicInformation
{
public:
    explicit CTrajSampleInformation(const std::vector< CPosVelPoint<NumT> > &src) : trajSample(src) {};
    virtual INFOTYPE infoType() const { return sizeof(NumT)==sizeof(float) ? IT_TRAJ_SAMPLE : IT_UNKNOWN; };
    virtual std::string toString() const;
    /// get function: return the const reference to the entire trajectory
    const CPosVelDataFloat& getTraj() const { return trajSample; }
private:
    CPosVelDataFloat trajSample;   //< array of unsorted sampling points from the trajectory
};

/// information object for velocity perturber class 
/// (keep track of accumulated energy change due to relaxation)
template<typename NumT> class CVelPerturberInformation: public CBasicInformation
{
public:
    CVelPerturberInformation(const double _deltaErelax) : deltaErelax(static_cast<NumT>(_deltaErelax)) {};
    virtual INFOTYPE infoType() const { return IT_VEL_PERTURBER; };
    virtual std::string toString() const;
    /// get data
    const NumT getData() const { return deltaErelax; }
private:
    NumT deltaErelax;  ///< stores a single value -- the accumulated energy change
};

/* --------- several orbit runtime functions and their associated creator classes ------------- */

/** Class for recording orbit trajectory and doing various analysis (frequency, Lyapunov exponent, etc).
    It records the trajectory at equal intervals of time (specified by timeStep), 
    and upon completion performs frequency analysis and extracts other information, 
    like the orbit class, chaotic properties, etc. The trajectory itself 
    is used in computing these properties, but not returned in the COrbitInformation class.
*/
class COrbitRuntimeTrajectory: public CBasicOrbitRuntimeFnc
{
public:
    /** constructor for an empty orbit */
    COrbitRuntimeTrajectory(const COrbit* _orbit);
    /** constructor for an existing trajectory */
    template<typename NumT> COrbitRuntimeTrajectory(const COrbit* _orbit, 
        const std::vector< CPosVelPoint<NumT> > &_traj, double _timeStep);
    virtual FNCTYPE FncType() const {return FT_TRAJ_ANALYSIS; };
    virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
    virtual CBasicInformation* getData() const  { return analyzeOrbit<float>(0, 0); }

    /** extract frequencies and perform orbit classification.
        by default operate on the entire whole orbit, but may use part of orbit as well. */
    template<typename NumT> COrbitInformation<NumT>* analyzeOrbit(
        size_t start,                ///< [in]  the first point of the interval to be analyzed
        size_t count,                ///< [in]  duration of the interval to be analyzed; 0 means entire orbit
        CSpectrum *spec_lines=NULL,  ///< [out] list of spectral lines found for each coordinate; if NULL then no output is given
        vectord *spec_ampl=NULL      ///< [out] entire spectrum (amplitudes only) for each coordinate; if NULL then no output is given
    ) const;                         ///< \return a COrbitInformation object with all computed parameters
    // get functions
    /// returns the number of points in recorded trajectory
    size_t getTrajSize() const { return traj.size(); };
    /// retrieve a point from trajectory
    const CPosVelPoint<double>& getTraj(size_t index) const { return traj[index]; }
    /// return time interval between consecutive trajectory points
    double getTimeStep() const { return timeStep; }
    /// return the logarithm of magnitude of the deviation vector
    double getLnDevVec(size_t index) const { return lnw[index]; }
    /// return max distance between adjacent points in the given interval
    double getMaxDist(size_t start=0, size_t count=0) const;
    vectord energy, enertime;
private:
    const unsigned int N_dim;        ///< copied from orbit->potential->N_dim, for convenience
    double timeUnit, timeStep;       ///< copied from orbit->timeUnit, timeUnit*samplingInterval

    CPosVelDataDouble traj;          ///< stores entire trajectory
    vectord lnw;                     ///< stores ln(|w|) - norm of deviation vector, lambda - finite-time Lyapunov exponent
    double Einit, Ediff;             ///< initial energy and energy conservation error by the end of integration
    unsigned int numSteps;           ///< counts the number of steps taken, to check energy conservation every Ncheck-th step

    /// estimate Lyapunov exponent
    double calcLambda() const;

    /// perform orbit classification based on its spectrum and angular momentum
    std::string performOrbitAnalysis(
        size_t start,                ///< [in]  specifies the first point of the interval to be analyzed
        size_t count,                ///< [in]  duration of the interval to analyze
        CSpectrum spec_lines[N_DIM], ///< [out] list of spectral lines found for each coordinate
        vectord spec_ampl[N_DIM],    ///< [out] entire spectrum (amplitudes only) for each coordinate; if NULL then no output is given
        double Lavg[N_DIM],          ///< [out] average values of angular momentum for each coordinate
        double Lvar[N_DIM]           ///< [out] std.dev. of angular momentum in each coord
    ) const;                         ///< \return a string with the description of orbit class

    /** finds spectral lines in each coordinate.
        uses CBasicSpectrumAnalyzer to find all lines, then sort them in appropriate order 
        taking into account axis ordering and base frequency of a long-axis orbit. 
        \param[in] start is the first point to analyze,
        \param[in] count is the length of the interval to analyze;
        \param[out] spec_lines is the array of sorted spectral lines; 
        \param[out] spec_ampl contains the amplitudes of entire spectrum if not NULL. */
    void findLines(size_t start, size_t count, CSpectrum spec_lines[N_DIM], vectord spec_ampl[N_DIM]) const;

    /** calculate LFCCs over a part of orbit or the entire orbit.
        \param[in] start is the first point to analyze,
        \param[in] count is the length of the interval to analyze;
        \param[out] leadfreq is the array of leading frequencies in Cartesian coordinates. */
    void calcLFCCoverInterval(size_t start, size_t count, double leadfreq[N_DIM]) const;

    /** evaluate the rate of diffusion of leading frequencies over a given interval (entire orbit by default) */
    double calcLFCCdiffusion(size_t start, size_t count) const;

    /** compute the inertia tensor of the orbit or part of it.
        \param[in] start is the first point to analyze,
        \param[in] count is the length of the interval to analyze;
        \param[out] inertia is the array of mean-square values of each coordinate, 
        which are directly related to the components of inertia tensor. */
    void calcInertiaOverInterval(size_t start, size_t count, double inertia[N_DIM]) const;
};

/// corresponding creator class for Trajectory analysis timestep fnc
class COrbitRuntimeTrajectoryCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
    { return new COrbitRuntimeTrajectory(orbit); };
};

/** Trajectory sampler (record a given number of unsorted points from the trajectory, 
    to be used in creating an Nbody model from the Schwarzschild model).
*/
class COrbitRuntimeTrajSample: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimeTrajSample(const COrbit* _orbit, size_t _numSamplingPoints, double _timeStep);
    virtual FNCTYPE FncType() const {return FT_TRAJ_SAMPLE; };
    virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);  ///< record trajectory at regular intervals
    virtual void Finish();   ///< draw a necessary number of sampling points from the entire trajectory
    virtual CBasicInformation* getData() const { return new CTrajSampleInformation<float>(trajSample); };
private:
    size_t numSamplingPoints;      ///< required number of sampling points at the end of integration
    const double timeStep;         ///< time interval for recording the trajectory (must be less than duration of integration divided by numSamplingPoints)
    const unsigned int N_dim;      ///< dimensions of potential (stored for convenience)
    CPosVelDataFloat trajEntire;   ///< recorded trajectory
    CPosVelDataFloat trajSample;   ///< points sampled from the trajectory (unordered)
};

/** corresponding creator class for trajectory sampler.
    In the Schwarzschild model, orbits that have a high weight may need more sampling points 
    to represent them in the Nbody model. Consequently, if the number of sampling points 
    for an orbit was insufficient, it will be re-integrated before exporting to Nbody model, 
    with an increased number of sampling points. 

    This creator class takes care of these special orbits:
    it requests a default number of sampling points for all orbits except those ones 
    which initial conditions are listed in the numSamplingPointsSpecial array;
    these special points are requested to provide more samples.
    intTimeMin is the (minimal) length of integration time (in periods), 
    it is necessary to compute the interval between sampling points used by COrbitRuntimeTrajSample,
    since the runtime function does not know the total duration of the integration in advance.
*/
class COrbitRuntimeTrajSampleCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimeTrajSampleCreator(unsigned int _numSamplingPointsDefault, double _intTimeMin);
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const;
    /** Add the orbit with the given initial conditions to the list of "special" orbits,
        for which the number of sampling points should be larger than usual.
        This function is called from COrbitLibrary::exportNbody, when the number of points 
        sampled from a given orbit is less than required (i.e. its weight in the Schwarzschild 
        solution divided by the mass of one particle in N-body model).
        This orbit is added to the "special" list, and during reintegtation, it will be sampled 
        with the required number of points.
        \param[in] orbitIC specifies the initial conditions for the orbit
        \param[in] numSamplingPoints is the minimum required number of sampling points for this orbit
        \param[in] samplingTimestep additionally specifies the time interval between sampled points
        (because the total integration time for this orbit is not known by the creator class). **/
    void addOrbitSpecial(const CPosVelPoint<float> &orbitIC, size_t numSamplingPoints, double samplingTimestep);
private:
    const unsigned int numSamplingPointsDefault;  ///< standard number of sampling points for all orbits except for the special list
    const double intTimeMin;                      ///< (minimum) integration time in periods, needed to compute timestep
    std::vector<CPosVelPoint<float> > specialIC;  ///< list of special orbits, for which the number of sampling points needs to be larger
    std::vector<size_t> specialNumSamplingPoints; ///< list of required number of sampling points for special orbits
    std::vector<double> specialSamplingTimestep;  ///< list of required sampling timestep for special orbits
};

/** compute the Poincare surface of section for an orbit.
    When coord2=0 and coord2'>0, record a pair (coord1, coord1').
    The Poincare surface of section is used only for visualization in the GUI,
    it has no use in Schwarzschild modelling, although may be utilized to systematically 
    explore the phase space at a given energy (using orbit library in the Frequency Map mode).
    If Ncross>0, the orbit is integrated until it passes through coord2=0 plane at least Ncross times.
*/
class COrbitRuntimePoincare: public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimePoincare(const COrbit* _orbit, unsigned int _PS1, unsigned int _PS2, unsigned int _Ncross) : 
      CBasicOrbitRuntimeFnc(_orbit), PS1(_PS1), PS2(_PS2), Ncross(_Ncross) {};
    virtual FNCTYPE FncType() const {return FT_POINCARE; };
    virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
    virtual CBasicInformation* getData() const { return new CPoincareInformation(ps); };
    /// return full coordinates of the last point that the orbit crossed the surface of section, if it exists.
    bool getLastCrossing(CPosVelPoint<double>* cross) const;
private:
    const unsigned int PS1, PS2;                  ///< index of coordinates to use in Poincare section: PS1 is the coordinate to record, PS2 is the coordinate to pass through zero
    const unsigned int Ncross;                    ///< maximum number of points in the section, after which the orbit integration should terminate
    std::vector< std::pair<double, double> > ps;  ///< list of recorded points in the surface of section ({x, xdot}, where x is the coordinate specified by index PS1)
    CPosVelPoint<double> lastCrossing;            ///< keeps the coordinates of last point of crossing the surface of section
};

/// corresponding creator class for Poincare timestep fnc
class COrbitRuntimePoincareCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimePoincareCreator(unsigned int _PS1, unsigned int _PS2, unsigned int _Ncross=0) : 
      PS1(_PS1), PS2(_PS2), Ncross(_Ncross) {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
    { return new COrbitRuntimePoincare(orbit, PS1, PS2, Ncross); };
private:
    const unsigned int PS1, PS2;
    const unsigned int Ncross;
};
 
/// class for recording pericenter passages and analyzing their distribution 
/// to figure out if the orbit is centrophilic or not
class COrbitRuntimePericenter: public CBasicOrbitRuntimeFnc
{
public:
    /// create the runtime function for an orbit to be integrated later
    COrbitRuntimePericenter(const COrbit* _orbit);
    /// create the runtime function for an orbit with already existing trajectory
    template<typename NumT> COrbitRuntimePericenter(const COrbit* _orbit, 
        const std::vector< CPosVelPoint<NumT> > &_traj, double _timeStep);
    virtual FNCTYPE FncType() const {return FT_PERICENTER; };
    virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);  ///< records pericenter passages
    virtual void Finish();   ///< fit pericenter distribution and compute fitting coefficients
    virtual CBasicInformation* getData() const { return new CPericenterInformation<float>((float)fitIntercept, (float)fitSlope, (float)fitScatter, (float)fitSignificance, (float)Lcirc2); };
    size_t size() const { return pericentert.size(); };
    double gett(size_t index) const { return pericentert[index]; };
    double getr(size_t index) const { return pericenterr[index]; };
    double getl2(size_t index)const { return pericenterl2[index]; };
protected:
    int drdtsign;                      ///< whether approaching or receding from center on the previous timestep
    vectord pericentert, pericenterr;  ///< times and radii of consequent pericenter passages
    vectord pericenterl2;              ///< L^2 - squared magnitude of angular momentum
    double pericenterrmin;             ///< minimum pericenter radius reached so far
    double fitIntercept, fitSlope;     ///< fitting coefficients for L2 distribution
    double fitScatter;                 ///< scatter in linear fit (less=better, >0.5-poor fit) 
    double fitSignificance;            ///< significance of non-zero value of intercept (in sigmas)
    double Lcirc2;                     ///< approximate squared angular momentum of a circular orbit with this energy
    double Lavg[N_DIM], L2avg[N_DIM];  ///< accumulator for average value of i-th component of angular momentum and its square (to determine if there is a conserved component)
};

/// corresponding creator class for the pericenter analysis runtime function
class COrbitRuntimePericenterCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimePericenterCreator() {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
    { return new COrbitRuntimePericenter(orbit); };
};

/** class for applying random perturbations on velocity during orbit integration, 
    mimicking the effect of two-body relaxation.
    The diffusion coefficients come in two varieties: local (position-dependent) coefficients for velocity,
    and orbit-averaged coefficients for energy and angular momentum. 
    They are computed from a spherical mass model constructed as an approximation to the true (non-spherical) 
    potential used in orbit integration itself. There is a free parameter that determines the scaling of 
    these coefficients: its numerical value corresponds to \f$ (\ln\Lambda)/N \f$, where N is the number of 
    stars in the model galaxy, and log Lambda is the Coulomb logarithm (~log N).
    */
class COrbitRuntimeVelPerturber : public CBasicOrbitRuntimeFnc
{
public:
    COrbitRuntimeVelPerturber(const COrbit* _orbit, const CMassModel* _model, double _mult);
    virtual ~COrbitRuntimeVelPerturber();
    virtual FNCTYPE FncType() const { return FT_VEL_PERTURBER; };
    /** Modifies the velocity by applying a random perturbation to it
        and records accumulated changes in E and L resulting from the perturbation.
        \param[in,out] vars is the current position and velocity of the orbit; the velocity is modified.
        \param[in] told is the time at the beginning of timestep;
        \param[in] t is the time at the end of timestep. **/
    virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
    virtual void Finish();   ///< fit pericenter distribution and compute fitting coefficients
    virtual CBasicInformation* getData() const 
    { return new CVelPerturberInformation<float>(deltaEcumul); }
private:
    const CMassModel* model;   ///< spherical mass model used to compute diffusion coefficients
    const double mult;         ///< multiplier for the diffusion coefficients =(log Lambda)/N
    double Eorbitavg;          ///< energy below which we use orbit-averaged dif.coefs instead of local ones
    double intTime;            ///< total integration time 
    double deltaE, deltaE2, deltaEcumul; ///< accumulated drift and diffusion coefficients in energy and its total change
    double deltaL, deltaL2, deltaLcumul; ///< accumulated drift and diffusion coefficients in angular momentum and its total change
    double avgCoefs[15];       ///< time-averaged coefficients DEE, DLL, DEL, DE, DL and their counterparts in spherical model
};

/// corresponding creator class for the velocity perturber runtime function
class COrbitRuntimeVelPerturberCreator: public CBasicOrbitRuntimeFncCreator
{
public:
    COrbitRuntimeVelPerturberCreator(const CMassModel* _model, double _mult);
    ~COrbitRuntimeVelPerturberCreator();
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
    { return new COrbitRuntimeVelPerturber(orbit, model, mult); };
    const CMassModel* getModel() const { return model; }
private:
    const CMassModel* model;
    const double mult; 
};

/** Utility function to find a periodic orbit in the plane coord1:coord2
    which is closest to the provided initial conditions. */
bool findPeriodicOrbit(const CPotential* potential, double Omega, 
    unsigned int coord_section, unsigned int coord_first, bool threeD, unsigned int Nperiods, CPosVelPoint<double> *IC);

}  // namespace
