#include "orbitint.h"
#include <cassert>
#include <gsl/gsl_integration.h>
#include "potential.h"
#include "utils.h"
#include "odeint.h"

namespace smile{

/** default parameters for orbit integration and classification **/
const CConfigOrbit defaultConfigOrbit = {
    3,     ///< N_dim
    0.0,   ///< Omega
    CBasicOdeIntegrator::SK_DOP853,  ///< integratorType
    1e-9,  ///< accuracyRel
    1e-9,  ///< accuracyAbs
    1e-4,  ///< accuracyIAS15
    0.4,   ///< accuracyTreecode
    false, ///< treecodeSymmetrizeTimestep
    0.005, ///< accuracyFixedTimestep
    0.02,  ///< samplingInterval
    0,     ///< calcLyapunovMethod
    1,     ///< numDevVectors
    0.0,   ///< relaxRate
    true,  ///< freqStrictOrder
    0.98,  ///< minFreqX
    0.99   ///< minFreqYZ
};

/*-----------------   Below follows COrbit class  -------------*/

/* constructor for an empty orbit */
template<typename NumT> COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const CPosVelPoint<NumT> &_initCond, const vectorRuntimeFncCreators* creators,
    double _timeEnd, double _timeMax) :
  potential(_potential),
  configOrbit(_configOrbit),
  initCond(_initCond), endCond(initCond)
{
    timeUnit = -2;  // will be computed when necessary
    timeEnd=_timeEnd; 
    timeMax=_timeMax;
    if(timeMax<timeEnd) timeMax=timeEnd;
    state=OS_INITIALIZED;
    odeInt=NULL;
    if(creators!=NULL)  // construct timestep function objects using creator classes
        for(vectorRuntimeFncCreators::const_iterator iter=creators->begin(); iter!=creators->end(); iter++)
        {
            CBasicOrbitRuntimeFnc* fnc=(*iter)->createRuntimeFnc(this);
            if(fnc!=NULL) runtimeFncs.push_back(fnc);
        }
}
// explicit instantiations of constructor for two variants of template argument
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const CPosVelPoint<float> &_initCond, const vectorRuntimeFncCreators* creators,
    double _timeEnd, double _timeMax);
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const CPosVelPoint<double> &_initCond, const vectorRuntimeFncCreators* creators,
    double _timeEnd, double _timeMax);

COrbit::~COrbit()
{
    // if there were any timestep functions attached, destroy 'em all!
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
        delete (*iter);
}

std::string COrbit::toString(bool runtimeFncInfo) const 
{
    std::string strInfo;
    if(runtimeFncInfo) 
        for(size_t rf=0; rf<getInfoNum(); rf++)
        {
            const CBasicInformation* info=getNewInfo(rf);
            if(info!=NULL) {
                strInfo += info->toString();
                delete info;
            }
        }
    strInfo+="nStep="+convertToString(numSteps)+", nForceEval="+convertToString(numFncEval)+"\n";
    return strInfo;
}

double COrbit::getIntTime() const
{ 
    return state==OS_INITIALIZED ? 0 : odeInt!=NULL ? odeInt->getTime() : timeEnd; 
}

double COrbit::getTimeUnit() const
{
    if(timeUnit==-2)  // hasn't been computed yet
        // assignment to a mutable data member in const function! (sort of caching)
        timeUnit=computeTimeUnit(potential, initCond/*, configOrbit->Omega*/); 
    return timeUnit;
}

void COrbit::run()
{
    if(odeInt!=NULL) 
        return;   // can't run orbit integrator twice
    CBasicOdeIntegrator::STEPPERKIND stepperKind=configOrbit->integratorType;
    varEqMethod = configOrbit->calcLyapunovMethod;
    if(potential->PotentialType() == CPotential::PT_NB) 
    {
        if(stepperKind==CBasicOdeIntegrator::SK_DEFAULT)
            stepperKind=CBasicOdeIntegrator::SK_LEAPFROG_NB;
        varEqMethod=0;
    } else {
        if(stepperKind==CBasicOdeIntegrator::SK_LEAPFROG_NB) {
            my_error(FUNCNAME, "Can only use Leapfrog integrator with NBody potential");
            return;
        }
    }
    if(stepperKind==CBasicOdeIntegrator::SK_DEFAULT) 
        stepperKind=CBasicOdeIntegrator::SK_DOP853;
    needToTerminate=false;
    needrenorm=allowrenorm=false;
    drdtsign=0;
    numSteps=numFncEval=0;
    isCanonicalVel=true;
    Etoler=fabs(totalEnergy(potential, initCond)*10);  // used for energy conservation accuracy in determining timestep length
    if(Etoler==0) Etoler=(pow_2(initCond.Vel[0])+pow_2(initCond.Vel[1])+pow_2(initCond.Vel[2]))*5;  // somewhat arbitrary setting for non-finite potentials (which could have Phi>0)
    try{
        switch(stepperKind)
        {
        case CBasicOdeIntegrator::SK_LEAPFROG_NB:
            odeInt = new COdeIntegratorLeapfrog(this,
                configOrbit->accuracyTreecode, configOrbit->treecodeSymmetrizeTimestep);
            break;
        case CBasicOdeIntegrator::SK_DOP853:
            odeInt = new COdeIntegratorDOP853(this,
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_IAS15:
            odeInt = new COdeIntegratorIAS15(this, configOrbit->accuracyIAS15);
            if(configOrbit->Omega!=0)
                isCanonicalVel=false;   // IAS15 uses an ODE formulation in terms of {x, \dot x} instead of {x, v}
            break;
        case CBasicOdeIntegrator::SK_HERMITE:
            if(varEqMethod>1) {
                my_error(FUNCNAME, "Cannot use Hermite integrator with variational equation");
                return;
            }
            odeInt = new COdeIntegratorHermite(this, 
                10*pow(configOrbit->accuracyRel, 0.4),  // heuristic corresponding accuracy parameter
                configOrbit->Omega!=0);  // two corrector stages in case of rotating potential
            if(configOrbit->Omega!=0)
                isCanonicalVel=false;   // Hermite uses an ODE formulation in terms of {x, \dot x} instead of {x, v}
            break;
#ifdef HAVE_ODEINT
        case CBasicOdeIntegrator::SK_ODEINT_BS:
            odeInt = new COdeIntegratorOdeint<StepperBS>(this, 
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_DP5:
            odeInt = new COdeIntegratorOdeint<StepperDP5>(this,
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_CK5:
            odeInt = new COdeIntegratorOdeint<StepperCK5>(this,
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_BS3:
            odeInt = new COdeIntegratorOdeint<StepperBS3>(this,
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_RK4:
            odeInt = new COdeIntegratorOdeint<StepperRK4>(this,
                getTimeUnit()*configOrbit->accuracyFixedTimestep, 0);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_SYMPL4:
            if(configOrbit->Omega!=0) {
                my_error(FUNCNAME, "Symplectic Runge-Kutta does not work in rotating frame");
                return;
            }
            odeInt = new COdeIntegratorOdeint<StepperSympl4>(this,
                getTimeUnit()*configOrbit->accuracyFixedTimestep, 0);
            break;
#endif
        default:
            my_error(FUNCNAME, "Unknown orbit integrator type requested!");
            return;
        }
    }
    catch(std::exception& e) {
        my_error(FUNCNAME, std::string("An exception occured during creation of ODE integrator: ")+e.what());
        return;
    }
    state=OS_RUNNING;
    try{
        odeInt->integrateToTime(timeEnd);
    }
    catch(std::exception& e) {
        my_error(FUNCNAME, std::string("An exception occured during orbit integration: ")+e.what());
        return;
    }
    try{
        timeEnd=odeInt->getTime();
        const size_t N_dim=configOrbit->N_dim;
        const size_t numVars2 = N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
        endCond=CPosVelPoint<double>(
            odeInt->getInterpolatedSolution(0, timeEnd),
            odeInt->getInterpolatedSolution(1, timeEnd),
            N_dim==N_DIM ? odeInt->getInterpolatedSolution(2, timeEnd) : 0,
            odeInt->getInterpolatedSolution(numVars2+0, timeEnd),
            odeInt->getInterpolatedSolution(numVars2+1, timeEnd),
            N_dim==N_DIM ? odeInt->getInterpolatedSolution(numVars2+2, timeEnd) : 0);
        for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
            (*iter)->Finish();
    }
    catch(std::exception& e) {
        my_error(FUNCNAME, std::string("An exception occured during finalizing an orbit: ")+e.what());
        return;
    }
    state=OS_DONE;
    delete odeInt;
    odeInt=NULL;
}

void COrbit::initState(OdeStateType* state)
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
    state->assign(2*numVars2 + (potential->PotentialType() == CPotential::PT_NB), 0);
    for(unsigned int d=0; d<N_dim; d++)
    {
        (*state)[d]         = initCond.Pos[d];
        (*state)[d+numVars2]= initCond.Vel[d];
    }
    if(!isCanonicalVel) {
        (*state)[numVars2+0] += configOrbit->Omega*initCond.Pos[1];
        (*state)[numVars2+1] -= configOrbit->Omega*initCond.Pos[0];
    }
    if(varEqMethod>0)
    {
        // initialize deviation vector(s)
        for(unsigned int dv=1; dv<=configOrbit->numDevVectors; dv++) {
            CPosVelPoint<double> devVector;
            double w2=0;
            for(unsigned int k=0; k<N_dim; k++) 
            { 
                devVector.Pos[k]=getRandomNumber(); 
                devVector.Vel[k]=getRandomNumber(); 
                w2+=pow_2(devVector.Pos[k]) + pow_2(devVector.Vel[k]); 
            }
            double wmag=sqrt(w2);
            if(varEqMethod==2) {  // variational equation
                lnwadd=0;
                for(unsigned int k=0; k<N_dim; k++) 
                {
                    devVector.Pos[k]/=wmag;
                    devVector.Vel[k]/=wmag;
                }
            } else if(varEqMethod==1) {  // nearby orbit
                lnwadd = -log(LYAP_DEV_INIT);
                for(unsigned int k=0; k<N_dim; k++) 
                {
                    devVector.Pos[k]=initCond.Pos[k] + devVector.Pos[k]/wmag * LYAP_DEV_INIT;
                    devVector.Vel[k]=initCond.Vel[k] + devVector.Vel[k]/wmag * LYAP_DEV_INIT;
                }
            } else 
                my_error(FUNCNAME, "Unknown method for computing deviation vector!");
            for(unsigned int d=0; d<N_dim; d++)
            {
                (*state)[N_dim*dv+d]=devVector.Pos[d];
                (*state)[N_dim*dv+d+numVars2]=devVector.Vel[d];
            }
        }
    }
}

void COrbit::odeFnc(double t, const OdeStateType& x, OdeStateType* dx)
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
    double Omega=configOrbit->Omega;
    double force[N_DIM];
    double forceDeriv[N_DIM*(N_DIM+1)/2];
    double coord[N_DIM] = {x[0], x[1], N_dim==N_DIM?x[2]:0};
    bool hermite = configOrbit->integratorType == CBasicOdeIntegrator::SK_HERMITE;
    bool treecode= potential->PotentialType() == CPotential::PT_NB;
    double* fDer = hermite||treecode ? forceDeriv : NULL;
    switch(varEqMethod) {
    case 0:   // compute only the original orbit
        potential->Force(coord, t, force, fDer);
        break;
    case 1: { // calculation of a nearby orbit
        for(unsigned int dv=1; dv<=configOrbit->numDevVectors; dv++) {
            double coord2[N_DIM] = {x[N_dim*dv], x[N_dim*dv+1], N_dim==N_DIM?x[N_dim*dv+2]:0};
            potential->Force(coord2, t, force, fDer);
            numFncEval++;
            dx->at(numVars2+N_dim*dv+0)=force[0];
            dx->at(numVars2+N_dim*dv+1)=force[1];
            if(N_dim==N_DIM) 
                dx->at(numVars2+N_dim*dv+2)=force[2];
            if(hermite) {
                double Vx=x[numVars2+N_dim*dv+0];
                double Vy=x[numVars2+N_dim*dv+1];
                dx->at(N_dim*dv+0) = fDer[0]*Vx + fDer[3]*Vy - Omega*force[1];
                dx->at(N_dim*dv+1) = fDer[1]*Vy + fDer[3]*Vx + Omega*force[0];
                if(N_dim==N_DIM) {
                    double Vz=x[numVars2+N_dim*dv+2];
                    dx->at(N_dim*dv+0)+= fDer[5]*Vz;
                    dx->at(N_dim*dv+1)+= fDer[4]*Vz;
                    dx->at(N_dim*dv+2) = fDer[2]*Vz + fDer[5]*Vx + fDer[4]*Vy;
                }
            }
        }
        potential->Force(coord, t, force, fDer);  // compute the original orbit
        break;
    }
    case 2: { // compute the forces and force derivatives for the variational equation
        potential->Force(coord, t, force, forceDeriv);
        for(unsigned int dv=1; dv<=configOrbit->numDevVectors; dv++) {
            double wx=x[N_dim*dv+0];
            double wy=x[N_dim*dv+1];
            double wz=(N_dim==N_DIM?x[N_dim*dv+2]:0);
            dx->at(numVars2+N_dim*dv+0)=wx*forceDeriv[0] + wy*forceDeriv[3] + wz*forceDeriv[5];
            dx->at(numVars2+N_dim*dv+1)=wx*forceDeriv[3] + wy*forceDeriv[1] + wz*forceDeriv[4];
            if(N_dim==N_DIM) 
                dx->at(numVars2+N_dim*dv+2)=wx*forceDeriv[5] + wy*forceDeriv[4] + wz*forceDeriv[2];
        }
        break;
    }
    default: 
        dx->assign(x.size(),0);
    }
    numFncEval++;
    if(hermite) {  // store jerk instead of velocities in the first half of "dx" array
        double Vx=x[numVars2+0];
        double Vy=x[numVars2+1];
        dx->at(0) = fDer[0]*Vx + fDer[3]*Vy - Omega*force[1];
        dx->at(1) = fDer[1]*Vy + fDer[3]*Vx + Omega*force[0];
        if(N_dim==N_DIM) {
            double Vz=x[numVars2+2];
            dx->at(0)+= fDer[5]*Vz;
            dx->at(1)+= fDer[4]*Vz;
            dx->at(2) = fDer[2]*Vz + fDer[5]*Vx + fDer[4]*Vy;
        }
    } else {  // position derivatives are velocities
        for(size_t i=0; i<numVars2; i++)
            dx->at(i)=x[i+numVars2];  // d x_i/dt = v_i
    }
    dx->at(numVars2+0)=force[0];
    dx->at(numVars2+1)=force[1];
    if(N_dim==N_DIM) dx->at(numVars2+2)=force[2];
    if(Omega!=0) {
        if(isCanonicalVel) {
            dx->at(0) += Omega*x[1];
            dx->at(1) -= Omega*x[0];
            dx->at(numVars2+0) += Omega*x[numVars2+1];
            dx->at(numVars2+1) -= Omega*x[numVars2+0];
            if(varEqMethod>0) 
                for(unsigned int dv=1; dv<=configOrbit->numDevVectors; dv++) {
                    dx->at(N_dim*dv+0) += Omega*x[N_dim*dv+1];
                    dx->at(N_dim*dv+1) -= Omega*x[N_dim*dv+0];
                    dx->at(numVars2+N_dim*dv+0) += Omega*x[numVars2+N_dim*dv+1];
                    dx->at(numVars2+N_dim*dv+1) -= Omega*x[numVars2+N_dim*dv+0];
                }
        } else {
            dx->at(numVars2+0) += Omega*( 2*x[numVars2+1]+Omega*x[0]);
            dx->at(numVars2+1) += Omega*(-2*x[numVars2+0]+Omega*x[1]);
            dx->at(0) += Omega*( 3*dx->at(numVars2+1) + Omega*( 3*x[numVars2+0] - Omega*x[1] ) );
            dx->at(1) += Omega*(-3*dx->at(numVars2+0) + Omega*( 3*x[numVars2+1] + Omega*x[0] ) );
            if(varEqMethod==1) 
                for(unsigned int dv=1; dv<=configOrbit->numDevVectors; dv++) {
                    dx->at(numVars2+N_dim*dv+0) += Omega*( 2*x[numVars2+N_dim*dv+1]+Omega*x[N_dim*dv+0]);
                    dx->at(numVars2+N_dim*dv+1) += Omega*(-2*x[numVars2+N_dim*dv+0]+Omega*x[N_dim*dv+1]);
                    dx->at(N_dim*dv+0) += Omega*( 3*dx->at(numVars2+N_dim*dv+1) + 
                        Omega*( 3*x[numVars2+N_dim*dv+0] - Omega*x[N_dim*dv+1] ) );
                    dx->at(N_dim*dv+1) += Omega*(-3*dx->at(numVars2+N_dim*dv+0) + 
                        Omega*( 3*x[numVars2+N_dim*dv+1] + Omega*x[N_dim*dv+0] ) );
                }
        }
    }
    if(treecode && dx->size()==x.size()+1)  // special trick for leapfrog
        dx->back()=forceDeriv[0];  // CPotentialNB returns distance to closest neighbour in this variable
}

double COrbit::getRelErrorToleranceFactor(const OdeStateType& state) const
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
    double vel2=0;
    for(size_t i=numVars2; i<numVars2+N_dim; i++)
        vel2+=pow_2(state[i]);
    return std::min<double>(1.0, Etoler/vel2);
}

COrbit::STEPRESULT COrbit::processStep(double timePrev, double timeCurr, OdeStateType* state)
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
    numSteps++;
    CPosVelPoint<double> pt(
        state->at(0), state->at(1), N_dim==N_DIM ? state->at(2) : 0,
        state->at(numVars2+0), state->at(numVars2+1), N_dim==N_DIM ? state->at(numVars2+2) : 0);
    if(!isCanonicalVel) {
        pt.Vel[0] -= configOrbit->Omega*pt.Pos[1];
        pt.Vel[1] += configOrbit->Omega*pt.Pos[0];
    }
    bool needToContinue=false;
    bool needToReinit=false;
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
    { 
        // note that if runtime funtions modify pt, their order is important! 
        // in fact, it doesn't make sense to have more than one function modifying the current state,
        // and this only one should be the last in sequence.
        CBasicOrbitRuntimeFnc::STEPRESULT sr = (*iter)->Timestep(timePrev, timeCurr, &pt);
        if(sr==CBasicOrbitRuntimeFnc::SR_CONTINUE) needToContinue=true;
        if(sr==CBasicOrbitRuntimeFnc::SR_REINIT) needToReinit=true;
    }
    if(needToReinit) 
    {   // store modified phase-space variables into ODE state vector
        state->at(0)=pt.Pos[0];
        state->at(1)=pt.Pos[1];
        state->at(numVars2)=pt.Vel[0];
        state->at(numVars2+1)=pt.Vel[1];
        if(N_dim==N_DIM) {
            state->at(2)=pt.Pos[2];
            state->at(numVars2+2)=pt.Vel[2];
        }
        if(!isCanonicalVel) {
            state->at(numVars2+0) += configOrbit->Omega*pt.Pos[1];
            state->at(numVars2+1) -= configOrbit->Omega*pt.Pos[0];
        }
    }
    // check for escaping orbits (very crude!!)
    if(pow_2(pt.Pos[0])+pow_2(pt.Pos[1])+pow_2(pt.Pos[2]) > 
       (pow_2(initCond.Pos[0])+pow_2(initCond.Pos[1])+pow_2(initCond.Pos[2]))*1e6 )  // could be escaping, check total energy
        if(totalEnergy(potential, pt)>0)
            needToTerminate=true;
    if(needToTerminate || (timeCurr>=timeEnd && !needToContinue) || timeCurr>=timeMax)
        return SR_TERMINATE;

    // watch for peri/apocenter passage (only for Lyapunov deviation vector renormalization,
    // which should be made at apocenter to avoid unacceptably large error growth)
    int prevsign=drdtsign;
    drdtsign= pt.Pos[0]*pt.Vel[0]+pt.Pos[1]*pt.Vel[1]+pt.Pos[2]*pt.Vel[2] >0 ? 1 : -1;
    if(varEqMethod>0 && prevsign>0 && drdtsign<=0 && needrenorm) 
        allowrenorm=true; // passed apocenter
 
    // check if deviation vector requires renormalization
    if(varEqMethod == 2) {   // variational equation
        double wnorm=0;
        for(unsigned int indx=0; indx<N_dim; indx++) {
            wnorm+=pow_2((*state)[indx+N_dim]);
            wnorm+=pow_2((*state)[indx+N_dim+numVars2]);
        }
        wnorm=sqrt(wnorm);
        if(wnorm>LYAP_VAREQ_DEV_MAX)
        {
            lnwadd+=log(wnorm);
            for(unsigned int indx=0; indx<N_dim; indx++) {
                (*state)[indx+N_dim]/=wnorm;
                (*state)[indx+N_dim+numVars2]/=wnorm;
            }
#ifdef DEBUGPRINT
            my_message(FUNCNAME, 
                "renormalizing deviation vector at t="+convertToString(timeCurr)+" by factor "+convertToString(wnorm));
#endif
            return SR_REINIT;
        }
    } else if(varEqMethod == 1) {   // nearby trajectory
        double wnorm=0;
        for(unsigned int indx=0; indx<N_dim; indx++) {
            wnorm+=pow_2(state->at(indx+N_dim)-state->at(indx));
            wnorm+=pow_2(state->at(indx+N_dim+numVars2)-state->at(indx+numVars2));
        }
        wnorm=sqrt(wnorm)/LYAP_DEV_INIT;
        if(wnorm>LYAP_DEV_MAX/LYAP_DEV_INIT)
        {
            needrenorm=true;   // it will be performed later, when allowed
        }
        if(needrenorm && allowrenorm)
        {
            // ensure that we do not decrease normalization
            if(wnorm<2) wnorm=2;
            lnwadd+=log(wnorm);
            for(unsigned int indx=0; indx<N_dim; indx++) {
                state->at(indx+N_dim) = 
                    state->at(indx) + (state->at(indx+N_dim)-state->at(indx))/wnorm;
                state->at(indx+N_dim+numVars2) = 
                    state->at(indx+numVars2) + (state->at(indx+N_dim+numVars2)-state->at(indx+numVars2))/wnorm;
            }
            needrenorm = allowrenorm = false;
#ifdef DEBUGPRINT
            my_message(FUNCNAME, 
                "renormalizing nearby trajectory at t="+convertToString(timeCurr)+" by factor "+convertToString(wnorm));
#endif
            return SR_REINIT;
        }
    }
    return needToReinit ? SR_REINIT : SR_CONTINUE;
}

double COrbit::getPos(unsigned int c, double t) const 
{
    if(odeInt==NULL || c>=configOrbit->N_dim) return 0;
    return odeInt->getInterpolatedSolution(c, t);
}

double COrbit::getVel(unsigned int c, double t) const 
{
    if(odeInt==NULL || c>=configOrbit->N_dim) return 0;
    size_t numVars2 = configOrbit->N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
    double result=odeInt->getInterpolatedSolution(c+numVars2, t);
    if(!isCanonicalVel) {
        if(c==0) result -= configOrbit->Omega*odeInt->getInterpolatedSolution(1, t);
        if(c==1) result += configOrbit->Omega*odeInt->getInterpolatedSolution(0, t);
    }
    return result;
}

double COrbit::getLogDevVec(double t) const 
{
    const size_t N_dim=configOrbit->N_dim;
    if(odeInt==NULL || varEqMethod==0) return 0;
    size_t numVars2 = N_dim * (varEqMethod>0 ? (1+configOrbit->numDevVectors) : 1);
    unsigned int dv=1;
        double wx = odeInt->getInterpolatedSolution(N_dim*dv,t);
        double wy = odeInt->getInterpolatedSolution(N_dim*dv+1,t);
        double wz = N_dim==N_DIM ? odeInt->getInterpolatedSolution(N_dim*dv+2,t) : 0;
        double wvx= odeInt->getInterpolatedSolution(numVars2+N_dim*dv,t);
        double wvy= odeInt->getInterpolatedSolution(numVars2+N_dim*dv+1,t);
        double wvz= N_dim==N_DIM ? odeInt->getInterpolatedSolution(numVars2+N_dim*dv+2,t) : 0;
        if(varEqMethod==1) {   // deviation vector is computed as the difference between two nearby orbits
            wx -= odeInt->getInterpolatedSolution(0, t);
            wy -= odeInt->getInterpolatedSolution(1, t);
            wvx-= odeInt->getInterpolatedSolution(numVars2, t);
            wvy-= odeInt->getInterpolatedSolution(numVars2+1, t);
            if(N_dim==N_DIM) {
                wz -= odeInt->getInterpolatedSolution(2, t);
                wvz-= odeInt->getInterpolatedSolution(numVars2+2, t);
            }
        }
        return (log(pow_2(wx) + pow_2(wy) + pow_2(wz) + pow_2(wvx) + pow_2(wvy) + pow_2(wvz))/2 + lnwadd);
}

/// \cond INTERNAL_DOCS
struct CRcircParam{
    const CPotential* P;
    double E;
    unsigned int direction;
};

double findRcirc(double r, void* params)
{
    double X=0, Y=0, Z=0;
    switch(static_cast<CRcircParam*>(params)->direction)
    {
    case 0: X=r; break;
    case 1: Y=r; break;
    case 2: Z=r; break;
    }
    const CPotential* P=static_cast<CRcircParam*>(params)->P;
    double V2 = 2*( static_cast<CRcircParam*>(params)->E - P->Phi(X, Y, Z) );
    double coord[N_DIM]={X,Y,Z};
    double force[N_DIM+1]={0,0,0,0};
    P->Force(coord, 0, force, NULL);
    return V2/r + force[static_cast<CRcircParam*>(params)->direction];
}

double integrandTorb(double x, void* params)
{
    if(x==0) return 0;
    double E=((CRcircParam*)params)->E;
    double V2=2*(E-((CRcircParam*)params)->P->Phi(x,0,0));
    if(V2<=0) return 0; else return 1/sqrt(V2);
}

/// Utility function to find approximate L^2 of a circular orbit with given energy
double findLcirc2(const CPotential* potential, const CPosVelPoint<double> &initCond)
{
    double E=totalEnergy(potential, initCond);
    gsl_function F;
    CRcircParam params;
    params.E=E;
    params.P=potential;
    F.function=&findRcirc;
    F.params=&params;
    // average over 3 coordinates
    double Lcirc2=0;
    unsigned int numAvg=0;
    for(unsigned int d=0; d<N_DIM; d++)
    {
        // find r such that v^2/r = F(r) and v^2/2+Phi(r)=E
        params.direction=d;
        double rupp=findIntersection(potential, E, d==0?1.:0, d==1?1.:0, d==2?1.:0);
        if(rupp<=0) return 0;  // invalid arguments (i.e. energy is above escape)
        double rlow=rupp/4;
        double rcirc=findRoot(&F, rlow, rupp, -1);
        if(rcirc>0) {
            Lcirc2 += 2*(E - potential->Phi( d==0?rcirc:0, d==1?rcirc:0, d==2?rcirc:0))*pow_2(rcirc);
            numAvg++;
        }
    }
    if(numAvg) Lcirc2/=numAvg;
    return Lcirc2;
}
/// \endcond

template<typename NumT>
double computeTimeUnit(const CPotential* potential, const CPosVelPoint<NumT> &initCond)
{
    // calculate T_orb (for x-axis orbit)
    // NB: computed value does not take into account figure rotation!
    double E=totalEnergy(potential, initCond);
    double Xmax=longAxisRadius(potential, E, 0);
    if(Xmax<0) return -1;  // error
    double result, error;
    gsl_function F;
    CRcircParam PP;
    PP.E=E;
    PP.P=potential;
    F.function=&integrandTorb;
    F.params=&PP;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    gsl_integration_qags (&F, 0, Xmax, 0, EPSREL_POTENTIAL_INT, 1000, w, &result, &error); 
    gsl_integration_workspace_free (w);
    return 4*result;
}
template double computeTimeUnit(const CPotential* potential, const CPosVelPoint<float> &initCond);
template double computeTimeUnit(const CPotential* potential, const CPosVelPoint<double> &initCond);

} // namespace
