#include "raga_core.h"
#include <iostream>
#include <fstream>
#include <cassert>
#include <ctime>
#ifndef WIN32
#include <unistd.h>
#endif
#include <omp.h>
#include "configfile.h"
#include "managepotential.h"
#include "iosnapshot.h"
#include "massmodel.h"
#include "utils.h"
#include "raga_fp.h"
#include "raga_tasks.h"

namespace smile {

template<typename NumT> double totalEnergySph(const CMassModel* model, const CPosVelPoint<NumT> &point)
{
    return model->pot(sqrt(pow_2(point.Pos[0])+pow_2(point.Pos[1])+pow_2(point.Pos[2])))
                    + 0.5*(pow_2(point.Vel[0])+pow_2(point.Vel[1])+pow_2(point.Vel[2]));
}

void computeTotalEnergyModel(const CPotential* potential, const CConfigPotential& configPotential, 
    const CPointMassSet<double>* particles, double time, double* Etot, double* Esum)
{
    assert(particles);
    assert(potential);
    assert(Etot);
    assert(Esum);
    *Etot=0; *Esum=0;
    if(potential->PotentialType()==CPotential::PT_COMPOSITE) {
        double stellarPotentialCenter=static_cast<const CPotentialComposite*>(potential)->component(0)->Phi(0,0,0,time);  // potential at origin excluding BH
        *Etot += configPotential.Mbh*stellarPotentialCenter*0.5;  // add the energy of BH in the stellar potential 
        *Esum += configPotential.Mbh*stellarPotentialCenter;
        if(configPotential.binary_sma>0) {  // add internal energy of binary BH
            *Etot -= pow_2(configPotential.Mbh)/(2*configPotential.binary_sma) * 
                configPotential.binary_q/pow_2(1+configPotential.binary_q);
            *Esum -= pow_2(configPotential.Mbh)/(2*configPotential.binary_sma)*3 * 
                configPotential.binary_q/pow_2(1+configPotential.binary_q);
        }
    }
    for(size_t ip=0; ip<particles->size(); ip++) {
        const CPosVelPoint<double>& point=particles->at(ip).first;
        double Epot=potential->Phi(point.Pos[0], point.Pos[1], point.Pos[2], time);
        double Ekin=(pow_2(point.Vel[0])+pow_2(point.Vel[1])+pow_2(point.Vel[2]))*0.5;
        if(gsl_isnan(Epot+Ekin)) {
            my_error(FUNCNAME, "Error in particle #"+convertToString(ip)+"'s energy");
            *Etot=*Esum=gsl_nan();
            return;
        } else {
            *Etot+=particles->at(ip).second * (Ekin+Epot*0.5);
            *Esum+=particles->at(ip).second * (Ekin+Epot);
        }
    }
}

CRagaCore::~CRagaCore()
{
    if(particles!=NULL) delete particles;
    if(massmodel!=NULL) delete massmodel;
    if(potential!=NULL) delete potential;
    for(size_t i=0; i<ragaTasks.size(); i++)
        delete ragaTasks[i];
}

bool CRagaCore::init(std::string configFile)
{
    my_message("", "Raga (SMILE version "SMILEVERSION")  build "__DATE__);
    gsl_set_error_handler(&gsl_verbose_error_handler);
    // load specific RAGA settings and check the validity of parameters
    if(!loadSettings(configFile))
        return false;

    // read input snapshot
    CBasicIOSnapshot* insnap=createIOSnapshotRead(configRaga.fileInput);
    if((insnap==NULL || !insnap->readSnapshot(particles) || particles->size()==0) && configRaga.mode!=CConfigRaga::RM_FP) {
        my_error(FUNCNAME, "Error reading initial snapshot "+configRaga.fileInput);
        return false;
    }   // for Fokker-Planck the input snapshot not required but may be provided; for other modes it is required
    delete insnap;

    // initialize potential and spherical mass model (if necessary)
    potential=createPotential(&configPotential);
    if(potential==NULL) {
        my_error(FUNCNAME, "Error creating potential");
        return false;
    }
    if(configPotential.Mbh>0) {
        std::vector<const CPotential*> components(1, potential);
        potential = new CPotentialComposite(components, configPotential.Mbh, 
            configPotential.binary_q, configPotential.binary_sma, 
            configPotential.binary_ecc, configPotential.binary_phase);
    }
    if(configOrbit.relaxRate>0 || configRaga.mode!=CConfigRaga::RM_RAGA) {
        massmodel=createMassModel(potential);
        if(massmodel==NULL) {
            my_error(FUNCNAME, "Error creating spherical mass model");
            return false;
        }
    }
    if(configRaga.mode!=CConfigRaga::RM_RAGA && potential->PotentialType()!=CPotential::PT_SPHERICAL) { 
        // replace the original potential with the clone of the spherical model (!!! is it required at all???)
        delete potential;
        potential=new CPotentialSpherical(*massmodel);
    }
    if(configRaga.mode==CConfigRaga::RM_FP) {
        runFokkerPlanck();
        return false;
    }

    // initialize various tasks, depending on the config file settings. Order *IS* important!
    if(configRaga.timestepOutput>0 && !configRaga.fileOutputParticles.empty()) 
    {   // output trajectories of particles
        ragaTasks.push_back(new CRagaTaskTrajectory(particles, &potential, 
            configRaga.fileOutputParticles, configRaga.fileOutputFormat, 
            "Raga "+configFile, configRaga.timestepOutput, configRaga.timeInit));
    }
    if(configOrbit.relaxRate>0) 
    {   // relaxation (velocity perturbation)
        ragaTasks.push_back(new CRagaTaskRelaxation(particles, &massmodel,
            &configRaga.episodeLength, configOrbit.relaxRate, configRaga.applyEnergyCorrection));
    }
    if(configRaga.captureRadius>0 && configPotential.Mbh>0) 
    {   // capture of stars by a central black hole
        ragaTasks.push_back(new CRagaTaskLosscone(particles, &(configPotential.Mbh),
            configRaga.captureRadius, configRaga.fileOutputLosscone));
    }
    if(configPotential.binary_sma>0 && configPotential.binary_q>0 && configPotential.Mbh>0) 
    {   // binary black hole evolution
        ragaTasks.push_back(new CRagaTaskBinaryBH(particles, &potential, &configPotential, 
            configRaga.speedOfLight, configRaga.fileOutputBinary, configRaga.timeInit));
    }
    if(configRaga.updatePotential || configPotential.binary_sma>0 || massmodel!=NULL) 
    {   // update potential and distr.fnc.
        ragaTasks.push_back(new CRagaTaskPotential(particles, &potential, &massmodel, 
            &configRaga.episodeLength, configRaga.numSamplesPerEpisode, &configPotential, 
            configRaga.fileOutputPotential, configRaga.timestepOutput, configRaga.updatePotential, 
            configRaga.applyEnergyCorrection, configRaga.timeInit));
    }
    return true;  // all ok
}

void CRagaCore::run()
{
    bool done=false;
    int numorb=particles->size();
    size_t numtasks=ragaTasks.size();
    timeSim=configRaga.timeInit;
    vectorRuntimeFncCreators creators(numtasks);
    for(size_t it=0; it<numtasks; it++)
        creators[it]=ragaTasks[it];  // simple pointer type conversion
    std::time_t timeStartEpisode, timeCurr, timePrev;
    // main loop
    std::ofstream strmLog;
    if(!configRaga.fileLog.empty()) {
        strmLog.open(configRaga.fileLog.c_str());
        double Etot, Esum;
        computeTotalEnergyModel(potential, configPotential, particles, 0, &Etot, &Esum);
        strmLog << timeSim << "\t: " << Etot <<"\t" << Esum<<"\n";
    }
    while(!done) {
        my_message(FUNCNAME, "Starting episode at time "+convertToString(timeSim));
        double episodeLength=configRaga.episodeLength;  // duration of this episode

        timeStartEpisode=std::time(NULL);
        timePrev=timeStartEpisode-1;
        size_t nComplete=0;
        // loop over the particles
#pragma omp parallel for schedule(dynamic, 16)
        for(int norb=0; norb<numorb; norb++) {
            timeCurr=std::time(NULL);
            if(std::difftime(timeCurr, timePrev)>1) {
                std::cerr << (convertToString(nComplete)+" orbits done\r");
                timePrev=timeCurr;
            }
            double particleIntTime=episodeLength;
            if(configRaga.captureRadius>0 && massmodel!=NULL) {
                // if losscone runtime function is active, it will keep integraing until maxIntTime limit is reached, 
                // or until the particle is captured. Thus the lower limit on integration time is set to just one radial period.
                particleIntTime=massmodel->trad(totalEnergySph(massmodel, particles->at(norb).first));
                if(!gsl_finite(particleIntTime) || particleIntTime<=0 || particleIntTime>episodeLength) 
                    particleIntTime=episodeLength;
            }
            if(particles->at(norb).second>0) {   // run only non-zero-mass particles
                COrbit* orbit=new COrbit(&configOrbit, potential, particles->at(norb).first, 
                    &creators, particleIntTime, episodeLength);
                orbit->run();
                for(size_t ninf=0; ninf<numtasks; ninf++)
                {
                    ragaTasks[ninf]->setInfoOrbit(norb, orbit->getNewInfo(ninf));  // may be NULL!
                }
                particles->at(norb).first=orbit->getEndCond();
                delete(orbit);
            } else {
                for(size_t ninf=0; ninf<numtasks; ninf++)
                {
                    ragaTasks[ninf]->setInfoOrbit(norb, NULL);
                }
            }
//#pragma omp atomic
            ++nComplete;
        }   // end parallel for
        timeCurr=std::time(NULL);
        my_message(FUNCNAME, convertToString(numorb)+" orbits total, "+
            convertToString(numorb/std::max<double>(1, std::difftime(timeCurr, timeStartEpisode)))+" orbits/s");
        double Etot, Esum;
        computeTotalEnergyModel(potential, configPotential, particles, episodeLength, &Etot, &Esum);
        my_message(FUNCNAME, "Episode done at t="+convertToString(timeSim+episodeLength)+", total energy="+
            convertToString(Etot)+", sumE="+convertToString(Esum));
        strmLog << (timeSim+episodeLength) << "\t: " << Etot << "\t" << Esum << "\n";
        // finish episode by calling corresponding function for each task
        for(size_t n=0; n<numtasks; n++) {
            done |= ragaTasks[n]->episodeDone(timeSim, episodeLength);
            computeTotalEnergyModel(potential, configPotential, particles, n==numtasks-1?0:episodeLength, &Etot, &Esum);  // last task is potential update and so the time parameter must be 0, otherwise equal to episode length
            my_message(FUNCNAME, "Task #"+convertToString(n)+" done at t="+convertToString(timeSim+episodeLength)+
                ", total energy="+convertToString(Etot)+", sumE="+convertToString(Esum));
            strmLog << (timeSim+episodeLength) << "\t"<<n<<" " << Etot << "\t" << Esum;
            if(n==numtasks-1) {
                const CPotential* p=potential->PotentialType()==CPotential::PT_COMPOSITE ?
                static_cast<const CPotentialComposite*>(potential)->component(0) : potential;
                if(configPotential.binary_sma>0)
                    strmLog << "\t" << -pow_2(configPotential.Mbh)/(2*configPotential.binary_sma) * 
                        configPotential.binary_q/pow_2(1+configPotential.binary_q);
                double potcenter=p->Phi(0,0,0,timeSim+episodeLength);
                if(configPotential.Mbh>0)
                    strmLog << "\t" << configPotential.Mbh*potcenter*0.5;
                strmLog << "\t" << potcenter;
            }
            strmLog <<"\n";
        }
        strmLog << std::flush;
        timeSim+=episodeLength;
        if(timeSim>=configRaga.timeTotal) done=true;
    }
}

bool CRagaCore::loadSettings(const std::string& defaultConfigFileName)
{
    std::string configFileName(defaultConfigFileName.empty() ? "raga.ini" : defaultConfigFileName);
    if(!fileExists(configFileName)) {
        my_error(FUNCNAME, "Cannot read configuration from "+configFileName);
        return false;
    }
    CConfigFile settings(configFileName);
    readConfigPotential(settings, &configPotential);
    readConfigOrbit(settings, &configOrbit);
    if(configOrbit.N_dim!=3) {
        my_error(FUNCNAME, "Run is only possible in three dimensions");
        return false;
    }
    std::string modeName = settings.getValueString("Raga/mode");
    if(modeName=="FokkerPlanck") configRaga.mode=CConfigRaga::RM_FP; else
    if(modeName=="Isotropic")    configRaga.mode=CConfigRaga::RM_SPH_ISO; else
    if(modeName=="Anisotropic")  configRaga.mode=CConfigRaga::RM_SPH_ANISO; else
    configRaga.mode=CConfigRaga::RM_RAGA;

    configRaga.fileInput = settings.getValueString("Raga/fileInput");
    if((configRaga.fileInput=="" || !fileExists(configRaga.fileInput)) && configRaga.mode!=CConfigRaga::RM_FP) {
        my_error(FUNCNAME, "Input file "+configRaga.fileInput+" does not exist ([Raga]/fileInput)");
        return false;
    }
    // check if input snapshot should also be used for the initial potential
    if((configPotential.PotentialType==CPotential::PT_BSE || 
        configPotential.PotentialType==CPotential::PT_SPLINE || 
        configPotential.PotentialType==CPotential::PT_SPHERICAL ) && 
        configPotential.DensityType==CDensity::PT_NB && configPotential.NbodyFile.empty())  // use nbody potential, but the filename was not provided
        configPotential.NbodyFile=configRaga.fileInput;
    configRaga.fileOutputParticles = settings.getValueString("Raga/fileOutput");
    configRaga.fileOutputPotential = settings.getValueString("Raga/fileOutputPotential");
    configRaga.fileOutputLosscone = settings.getValueString("Raga/fileOutputLosscone");
    configRaga.fileOutputBinary = settings.getValueString("Raga/fileOutputBinary");
    configRaga.fileLog = settings.getValueString("Raga/fileLog", configRaga.fileInput+".log");
    configRaga.timeTotal = settings.getValueDouble("Raga/timeTotal");
    configRaga.timeInit  = settings.getValueDouble("Raga/timeInit");
    if(configRaga.timeTotal<=configRaga.timeInit) {
        my_error(FUNCNAME, "Total simulation time should be positive ([Raga]/timeTotal)");
        return false;
    }
    configRaga.timestepOutput = settings.getValueDouble("Raga/timestepOutput");
    configRaga.fileOutputFormat = settings.getValueString("Raga/fileOutputFormat", "t");
    if(configRaga.timestepOutput<=0 || configRaga.fileOutputParticles.empty()) {  // not a critical error
        my_message(FUNCNAME, "Warning, output snapshots disabled ([Raga]/timestepOutput, [Raga]/fileOutput)");
    } else {
        if(fileExists(configRaga.fileOutputParticles)) {  // need to erase it beforehand, or NEMO output routines will crash
            my_message(FUNCNAME, "Output file "+configRaga.fileOutputParticles+" exists! Overwrite? [Y/N]");
            std::string str;
            std::cin >> str;
            if(str[0]=='Y' || str[0]=='y') {
#ifdef WIN32
                _unlink(configRaga.fileOutputParticles.c_str());
#else
                unlink(configRaga.fileOutputParticles.c_str());
#endif
            } else {
                return false;
            }
        }
        CBasicIOSnapshot* snap=createIOSnapshotWrite(configRaga.fileOutputFormat, configRaga.fileOutputParticles);
        if(snap) delete snap;
        else {
            my_error(FUNCNAME, "Unrecognized output file format");
            return false;
        }
    }
    configRaga.episodeLength  = settings.getValueDouble("Raga/episodeLength", 0);
    configRaga.updatePotential= settings.getValueBool  ("Raga/updatePotential", true);
    if(configRaga.episodeLength<=0 || !configRaga.updatePotential) {  // not a error at all
        my_message(FUNCNAME, "Dynamic potential update is disabled ([Raga]/episodeLength, [Raga]/updatePotential)");
        if(configRaga.episodeLength<=0) configRaga.episodeLength=configRaga.timeTotal;
    }
    if(configRaga.updatePotential && !(  // check potential type
        configPotential.PotentialType==CPotential::PT_BSE || 
        configPotential.PotentialType==CPotential::PT_SPLINE || 
        configPotential.PotentialType==CPotential::PT_SPHERICAL ) ) {
        my_error(FUNCNAME, "Dynamic potential update is only possible for BSE, Spline, or Spherical potential ([Potential]/PotentialType)");
        return false;
    }
    configRaga.factorTimestep = settings.getValueDouble("Raga/factorTimestep", 0.2);
    configRaga.numSamplesPerEpisode = std::max<int>(1, settings.getValueInt("Raga/numSamplesPerEpisode", 1));
    configRaga.gridSize = settings.getValueInt("Raga/gridSize", 100);
    configRaga.captureRadius = settings.getValueDouble("Raga/captureRadius", 0);
    configRaga.speedOfLight = settings.getValueDouble("Raga/speedOfLight", 0);
    configRaga.applyEnergyCorrection = settings.getValueBool("Raga/applyEnergyCorrection", true);
    if(configPotential.binary_sma>0 || configRaga.captureRadius>0) {
        if(configPotential.Mbh>0)
            configRaga.applyEnergyCorrection=false;  // don't use it for loss-cone problems, since it changes angular momenta as well
        else
            my_message(FUNCNAME, "No central black hole!");
    }
    return true;
}

void CRagaCore::runFokkerPlanck()
{
    assert(massmodel);
    size_t gridsize=configRaga.gridSize;
    double timestepSample=configRaga.episodeLength/configRaga.numSamplesPerEpisode;
    CSolverFP* solver=NULL;
    if(particles!=NULL && particles->size()>0) {
        std::vector<double> particleEnergies(particles->size()), particleMasses(particles->size());
        for(size_t ip=0; ip<particles->size(); ip++)
        {
            particleEnergies[ip] = totalEnergySph(massmodel, particles->at(ip).first);
            particleMasses[ip]=particles->at(ip).second;
        }
        solver=new CSolverFP(massmodel, gridsize, timestepSample, configOrbit.relaxRate, &particleMasses, &particleEnergies);
    } else
        solver=new CSolverFP(massmodel, gridsize, timestepSample, configOrbit.relaxRate);
    //my_message(FUNCNAME, "max.diag.coef.="+convertToString(solver->getmaxdiag()));
    std::ofstream strmLog(configRaga.fileLog.c_str());
    strmLog << std::setprecision(12);
    double nextOutputTime=0;
    double nextUpdateTime=configRaga.episodeLength;
    do{
        if(timeSim>=nextOutputTime) {
            solver->exportGrid((configRaga.fileOutputPotential+convertToString(timeSim)).c_str());
            exportModel(*massmodel, (configRaga.fileOutputPotential+convertToString(timeSim)+".tab").c_str());
            nextOutputTime=timeSim+configRaga.timestepOutput;
        }
        strmLog << timeSim << "\t" << solver->etot() << "\n";
        if(timeSim>=nextUpdateTime) {
            solver->reinitDF(configRaga.updatePotential);
            double veldisp=massmodel->veldisp(0);
            double trelc=0.34*pow(veldisp,3)/massmodel->dens(0) / configOrbit.relaxRate;
            double rcore=3*veldisp/sqrt(4*M_PI*massmodel->dens(0));
            strmLog << timeSim << "\t" << solver->etot() << "\t" 
                << massmodel->getPotentialAtCenter() << "\t" //<< massmodel->epot() << "\t" << massmodel->epotf() << "\t" 
                //<< massmodel->ekin() << "\t" << solver->ekin() << "\t" << massmodel->etot() << "\t" << massmodel->dens(0) << "\t"
                << trelc << "\t" << rcore << "\n" << std::flush;
            if(!gsl_isinf(trelc) && !gsl_isnan(trelc) && trelc<configRaga.episodeLength*5) {
                my_message(FUNCNAME, "At time "+convertToString(timeSim)+" cut the hair in two and reinit solver");
                configRaga.episodeLength/=2;
                configRaga.timestepOutput/=2;
                timestepSample/=2;
                /*solver->tau/=2;
                solver->reinitDC();*/
                delete solver;
                solver=new CSolverFP(massmodel, gridsize, timestepSample, configOrbit.relaxRate);
            }
            nextUpdateTime=timeSim+configRaga.episodeLength;
        }
        if(timeSim<configRaga.timeTotal) 
            solver->step();
        timeSim+=timestepSample;
    } while(timeSim<=configRaga.timeTotal);
    delete solver;
}

bool fileExists(const std::string& fileName)
{
    std::ifstream infile(fileName.c_str());
    return infile.good();
}

}  // namespace
