/**
    This program performs scattering experiments (stars from a homogeneous isotropic 
    Maxwellian background interacting with a binary supermassive black hole).
    It replaces the classes defined in raga_core.cpp with only the required subset 
    of functionality.
*/
#include "raga_core.h"
#include <iostream>
#include <fstream>
#include <cassert>
#include <omp.h>
#include "configfile.h"
#include "utils.h"
#include "raga_tasks.h"

namespace smile {

    double sigma=1.0;
    double bmax;
    double mpart;
    double rinit;
    double maxrmin;
    int nparticles;
    double totalmass;

class COrbitRuntimeBinaryScatter: public COrbitRuntimeBinaryBH
{
public:
    COrbitRuntimeBinaryScatter(const COrbit* _orbit) :
      COrbitRuntimeBinaryBH(_orbit, rinit) {};
    virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars) {
        COrbitRuntimeBinaryBH::Timestep(told, t, vars);
        double r2=pow_2(vars->Pos[0])+pow_2(vars->Pos[1])+pow_2(vars->Pos[2]);
        double v2=pow_2(vars->Vel[0])+pow_2(vars->Vel[1])+pow_2(vars->Vel[2]);
        if (r2 > rinit*rinit && 
            vars->Pos[0]*vars->Vel[0]+vars->Pos[1]*vars->Vel[1]+vars->Pos[2]*vars->Vel[2] > 0 &&
            orbit->getPotential()->Phi(vars->Pos[0], vars->Pos[1], vars->Pos[2], t) + v2/2 > 0) 
            return SR_TERMINATE;
        return SR_CONTINUE;
    }
};

class CRagaTaskBinaryScatter: public CRagaTaskBinaryBH
{
public:
    CRagaTaskBinaryScatter(
        CPointMassSet<double>* const _particles, 
        const CPotential* const* _ptrPotential, 
        CConfigPotential* _ptrConfigPotential) :
      CRagaTaskBinaryBH(_particles, _ptrPotential, _ptrConfigPotential, 0, "", 0) {};
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
    { return new COrbitRuntimeBinaryScatter(orbit); };
};

CRagaCore::~CRagaCore()
{
    if(particles!=NULL) delete particles;
    for(size_t i=0; i<ragaTasks.size(); i++)
        delete ragaTasks[i];
}

// analytically calculate the Keplerian hyperbolic orbit for a star coming from infinity
void kepler_hyper_orbit(double Mbh, double vinf, double b, double r0, double* x, double* y, double* vx, double* vy, double* rmin)
{
    double p = pow_2(vinf*b)/Mbh;
    double ecc = sqrt(1 + pow_2(b*vinf*vinf/Mbh));
    double beta0 = acos((1-p/r0)/ecc)-acos(1/ecc);
    *x = -r0*cos(beta0);
    *y =  r0*sin(beta0);
    double tanb0 = -(*y)/(*x);
    *vx = vinf/(1+tanb0*tanb0) * (sqrt((1+2*Mbh/(vinf*vinf*r0))*(1+tanb0*tanb0)-pow_2(b/(*x))) - tanb0*b/(*x));
    *vy = ((*y)*(*vx)-b*vinf)/(*x);
    *rmin = p/(1+ecc);  // distance of closest approach
}

bool CRagaCore::init(std::string configFile)
{
    my_message("", "Raga-scatter  build "__DATE__);
    gsl_set_error_handler(&gsl_verbose_error_handler);
    if(!loadSettings(configFile))
        return false;
    std::vector<const CPotential*> components(1, new CPotentialHarmonic(0, 1, 1));
    potential = new CPotentialComposite(components, configPotential.Mbh, 
        configPotential.binary_q, configPotential.binary_sma, configPotential.binary_ecc, configPotential.binary_phase);
    particles=new CPointMassSet<double>();
    particles->data.resize(nparticles);
    randomize();
    bmax   =configPotential.binary_sma*100;
    maxrmin=configPotential.binary_sma*3;
    rinit  =configPotential.binary_sma*100;
    mpart  =0.01*configPotential.Mbh/nparticles;
    int nparttotal=0;
    // seed a large number of initial conditions and choose only the ones that come close to the binary
    for(int i=0; i<nparticles; i++) {
        double rmin,vinf, vx, vy, vz, px, py, pz=0;
        do{
            getNormalRandomNumbers(&vx, &vy);
            getNormalRandomNumbers(&vx, &vz);
            vinf=sqrt(vx*vx+vy*vy+vz*vz)*sigma;
            double b = bmax*sqrt(getRandomNumber());
            kepler_hyper_orbit(configPotential.Mbh, vinf, b, rinit, &px, &py, &vx, &vy, &rmin);
            nparttotal++;
        } while(rmin>maxrmin);
        vz=0;
        double matr[9];
        getRandomRotationMatrix(matr);
        particles->data[i].first=CPosVelPoint<double>(
            px*matr[0]+py*matr[1]+pz*matr[2], px*matr[3]+py*matr[4]+pz*matr[5], px*matr[6]+py*matr[7]+pz*matr[8], 
            vx*matr[0]+vy*matr[1]+vz*matr[2], vx*matr[3]+vy*matr[4]+vz*matr[5], vx*matr[6]+vy*matr[7]+vz*matr[8]);
        particles->data[i].second=mpart*vinf/sigma;
    }
    totalmass=mpart*nparttotal;
    double ah=configPotential.binary_q/(1+configPotential.binary_q)*configPotential.Mbh/4/pow_2(sigma);
    std::cerr << "a="<<configPotential.binary_sma<<", a_h="<<ah<<", ecc="<<configPotential.binary_ecc<<
        ", q="<<configPotential.binary_q<<", total Np="<<nparttotal<<"\n";
    ragaTasks.push_back(new CRagaTaskBinaryScatter(particles, &potential, &configPotential));
    return true;  // all ok
}

void CRagaCore::run()
{
    vectorRuntimeFncCreators creators(1, ragaTasks[0]);
    double inttime=0;
    double mininttime=rinit/sigma, maxinttime=0;
    // loop over the particles
#pragma omp parallel for schedule(dynamic, 256)
    for(int norb=0; norb<nparticles; norb++) {
        COrbit orbit(&configOrbit, potential, particles->at(norb).first, 
            &creators, mininttime, mininttime*200);
        orbit.run();
        const CBinaryBHInformation* info=static_cast<const CBinaryBHInformation*>(orbit.getNewInfo(0));
        ragaTasks[0]->setInfoOrbit(norb, info);
        inttime+=orbit.getIntTime();
        maxinttime=std::max<double>(maxinttime, orbit.getIntTime());
        if(norb%256==0) std::cerr << norb<<"/"<<nparticles<<"     \r";
    }   // end parallel for
    //std::cerr << "avg.inttime="<<(inttime/nparticles)<<", min/max.inttime="<<mininttime<<"/"<<maxinttime<<"\n";
    double oneovera=1/configPotential.binary_sma;
    double ecc=configPotential.binary_ecc;
    ragaTasks[0]->episodeDone(0, 1);
    double deltaoneovera=1/configPotential.binary_sma-oneovera;
    double H=(sigma*sigma*M_PI*bmax*bmax/totalmass*deltaoneovera);
    double K=((configPotential.binary_ecc-ecc)/configPotential.binary_sma/deltaoneovera);
    std::cerr << "H=" << H << ", K=" << K << "\n";
}

int argc;
char** argv;
    
bool CRagaCore::loadSettings(const std::string& )
{
    CConfigFile settings("Potential", argc, argv);
    configPotential=defaultConfigPotential;
    configOrbit=defaultConfigOrbit;
    configOrbit.accuracyRel=1e-10;
    configOrbit.accuracyAbs=1e-14;
    nparticles = settings.getValueInt("Potential/n", 10000);                // number of particles in the scattering experiment
    configPotential.Mbh = settings.getValueDouble("Potential/m", 1);        // total mass of the binary black hole
    configPotential.binary_sma = settings.getValueDouble("Potential/a", 0); // binary semimajor axis
    configPotential.binary_ecc = settings.getValueDouble("Potential/e", 0); // binary eccentricity
    configPotential.binary_q = settings.getValueDouble("Potential/q", 1);   // binary mass ratio
    sigma = settings.getValueDouble("Potential/s", sigma);                  // velocity dispersion of stars
    return configPotential.binary_sma>0;
}

// placeholders for unused functions
void CRagaCore::runFokkerPlanck() {}
void computeTotalEnergyModel(const CPotential* , const CConfigPotential& , const CPointMassSet<double>* , double , double* , double* ) {};
bool fileExists(const std::string& ) { return false; }
void exportModel(smile::CMassModel const&, char const*) {}

}  // namespace

int main(int argc, char *argv[])
{
    if(argc<=1) {
        smile::my_error("scatter", "Required command-line parameters: a, [m=1], [q=1], [e=0], [n=10000], [s=1]");
        return 0;
    }
    smile::argc=argc;
    smile::argv=argv;
    smile::CRagaCore c;
    if(!c.init(""))
        smile::my_error("scatter", "Exiting..");
    else
        c.run();
}

