#include "raga_tasks.h"
#include <cassert>
#include <fstream>
#include <algorithm>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_math.h>
#include "configfile.h"
#include "iosnapshot.h"
#include "massmodel.h"
#include "managepotential.h"
#include "orbitint.h"
#include "utils.h"
#include "raga_fp.h"

namespace smile {

template<typename NumT> double totalEnergySph(const CMassModel* model, const CPosVelPoint<NumT> &point)
{
    return model->pot(sqrt(pow_2(point.Pos[0])+pow_2(point.Pos[1])+pow_2(point.Pos[2])))
                    + 0.5*(pow_2(point.Vel[0])+pow_2(point.Vel[1])+pow_2(point.Vel[2]));
}

/*------ runtime functions ------*/

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeTrajEpisode::Timestep(const double told, const double t, CPosVelPoint<double>* /*vars*/)
{
    int i1=static_cast<int>(told/timeStep); 
    int i2=static_cast<int>(t/timeStep);
    // store trajectory at regular intervals of time 
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if (((tt>told) && (tt<=t)) || ((i==0) && (told==0)))  ///!!! added 0th point explicitly
        {
            CPosVelPoint<float> pt;
            for(unsigned int c=0; c<N_DIM; c++)
            {
                pt.Pos[c]=static_cast<float>(orbit->getPos(c, tt));
                pt.Vel[c]=static_cast<float>(orbit->getVel(c, tt));
            }
            traj.add(pt, static_cast<float>(timeStep));
        }
    }
    return SR_TERMINATE;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeLosscone::Timestep(const double told, const double t, CPosVelPoint<double>* vars)
{
    if(tcapt>0)  // already swallowed
        return SR_TERMINATE;
    size_t nperi=size();
    COrbitRuntimePericenter::Timestep(told, t, vars);
    if(size()>nperi) {   // array has grown -- just passed a pericenter
        double thisPericenterRadius=getr(nperi);
        double thisL2=getl2(nperi);
        if(thisPericenterRadius<=captureRadius)  {  // just had been captured by black hole
            tcapt=gett(nperi);
            L2=thisL2;
            deltaL2=prevL2>0 ? prevL2-thisL2 : -1;
            Ecapt=totalEnergy(orbit->getPotential(), *vars);
            pericenterE.push_back(Ecapt);
            return SR_TERMINATE;
        }
        prevL2=thisL2;
        if(thisPericenterRadius<=maxPericenterTrackingRadius)
            pericenterE.push_back(totalEnergy(orbit->getPotential(), *vars));
        else {  // don't store this pericenter passage (to save memory)
            pericentert.pop_back();
            pericenterr.pop_back();
            pericenterl2.pop_back();
        }
        maxradius=0;  // reset value after each pericenter passage
    }
    double radius=sqrt(pow_2(vars->Pos[0])+pow_2(vars->Pos[1])+pow_2(vars->Pos[2]));
    // record the position at some point before the next pericenter passage (which may turn out to be the last one..)
    // this may be later used to determine what was the type of orbit just before capture (e.g. tube/saucer/pyramid/...)
    if(radius>maxradius) {
        maxradius=radius;
        capturePoint=*vars;  // store the apocenter point...
    }
    else if(radius>sqrt(captureRadius*maxradius))    // ...or some point in between apocenter and pericenter, but not too close to capture radius (to avoid roundoff errors)
        capturePoint=*vars;
    return SR_CONTINUE;
}

/// \cond INTERNAL_DOCS
struct CBinaryBHParam {
    const COrbit* orbit;
    double rcrit2;
};

double findtrcrit(double t, void* params)
{
    const COrbit* o=static_cast<CBinaryBHParam*>(params)->orbit;
    double cmx=0, cmy=0, cmz=0;
    double r2=pow_2(o->getPos(0,t)-cmx) + pow_2(o->getPos(1,t)-cmy) + pow_2(o->getPos(2,t)-cmz);
    return r2-static_cast<CBinaryBHParam*>(params)->rcrit2;
}
/// \endcond 

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeBinaryBH::Timestep(const double told, const double t, CPosVelPoint<double>* vars)
{
    CPosVelPoint<double> pt1(orbit->getPos(0,told), orbit->getPos(1,told), orbit->getPos(2,told), 0, 0, 0);
    double rprev2=pow_2(pt1.Pos[0]) + pow_2(pt1.Pos[1]) + pow_2(pt1.Pos[2]);
    double rcurr2=pow_2(vars->Pos[0])+pow_2(vars->Pos[1])+pow_2(vars->Pos[2]);
    if(rcurr2>rcrit2 && rprev2>=rcrit2) {  // both ends of timestep are outside rcrit
        return SR_TERMINATE;  // no further action required
    }
    double t1=told, t2=t;
    bool startencounter=(rcurr2<rcrit2 && rprev2>=rcrit2) || (told==0 && (rcurr2<rcrit2 || rprev2<rcrit2));
    bool endencounter  =rcurr2>=rcrit2 && rprev2<rcrit2;
    if(startencounter || endencounter) {
        // moved into or out of rcrit; need to find the exact moment of entering/leaving the critical sphere
        gsl_function F;
        CBinaryBHParam params;
        params.orbit=orbit;
        params.rcrit2=rcrit2;
        F.function=&findtrcrit;
        F.params=&params;
        double trcrit=findRoot(&F, told, t, -1);
        if(trcrit<0) 
            trcrit=told;
        if(endencounter) {  // moved out of rcrit
            t2=trcrit;
        } else {  // moved into rcrit
            t1=trcrit;
        }
    }
    if(t1!=told) {
        pt1.Pos[0]=orbit->getPos(0,t1);
        pt1.Pos[1]=orbit->getPos(1,t1);
        pt1.Pos[2]=orbit->getPos(2,t1);
    }  // otherwise pt1.Pos is already assigned
    pt1.Vel[0]=orbit->getVel(0,t1);
    pt1.Vel[1]=orbit->getVel(1,t1);
    pt1.Vel[2]=orbit->getVel(2,t1);
    CPosVelPoint<double> pt2(
        orbit->getPos(0,t2), orbit->getPos(1,t2), orbit->getPos(2,t2),
        orbit->getVel(0,t2), orbit->getVel(1,t2), orbit->getVel(2,t2));
    double E1=totalEnergy(orbit->getPotential(), pt1, 0, t1);
    double L1=pt1.Pos[0]*pt1.Vel[1]-pt1.Pos[1]*pt1.Vel[0];
    double E2=totalEnergy(orbit->getPotential(), pt2, 0, t2);
    double L2=pt2.Pos[0]*pt2.Vel[1]-pt2.Pos[1]*pt2.Vel[0];
    if(startencounter) {
        double L=sqrt(pow_2(pt1.Pos[1]*pt1.Vel[2]-pt1.Pos[2]*pt1.Vel[1])+
                      pow_2(pt1.Pos[2]*pt1.Vel[0]-pt1.Pos[0]*pt1.Vel[2])+pow_2(L1));
        encounters.push_back(CBinaryEncounter(E1, L, L1, t1));
    }
    double deltaLtorque=0;  // change in L_z due to torque from non-spherical stellar potential
    const CPotential* poten=static_cast<const CPotentialComposite*>(orbit->getPotential())->component(0);
    if((poten->symmetry() & CPotential::ST_ZROTSYM) == 0) { // potential has z-torque
        // integrate the torque over the current timestep using 2-point Gauss rule
        const double interior_point=0.211324865; // (1-sqrt(1./3))/2
        double force[N_DIM];
        double tmed=(t1*interior_point+t2*(1-interior_point));
        double xyz[N_DIM] = {orbit->getPos(0,tmed), orbit->getPos(1,tmed), orbit->getPos(2,tmed)};
        poten->Force(xyz, tmed, force, NULL);
        if(gsl_finite(force[0]+force[1]+force[2]))
            deltaLtorque = xyz[0]*force[1]-xyz[1]*force[0];
        else
            my_error(FUNCNAME, "Error in computing force at t="+convertToString(tmed)+
                ", x="+convertToString(xyz[0])+", y="+convertToString(xyz[1])+", z="+convertToString(xyz[0]));
        tmed=(t2*interior_point+t1*(1-interior_point));
        xyz[0]=orbit->getPos(0,tmed); xyz[1]=orbit->getPos(1,tmed); xyz[2]=orbit->getPos(2,tmed);
        poten->Force(xyz, tmed, force, NULL);
        if(gsl_finite(force[0]+force[1]+force[2]))
            deltaLtorque += xyz[0]*force[1]-xyz[1]*force[0];
        else
            my_error(FUNCNAME, "Error in computing force at t="+convertToString(tmed)+
                ", x="+convertToString(xyz[0])+", y="+convertToString(xyz[1])+", z="+convertToString(xyz[0]));
        deltaLtorque *= 0.5*(t2-t1);
    }
    if(encounters.size()==0) {  // shouldn't occur!
        encounters.push_back(CBinaryEncounter(E1, 0, L1, t1));
        my_error(FUNCNAME, "Undetected start of encounter");
    }
    encounters.back().deltaE += E2-E1;
    encounters.back().deltaL += L2-L1-deltaLtorque;
    encounters.back().deltaLtorq += deltaLtorque;
    encounters.back().tend = t2;
    return SR_TERMINATE;
}

/*------ RAGA tasks -------*/

CRagaTaskTrajectory::CRagaTaskTrajectory(
    CPointMassSet<double>* _particles, 
    const CPotential* const * _ptrPotential, 
    const std::string& _outputFileName, 
    const std::string& _outputFileFormat, 
    const std::string& _outputFileHeader, 
    double _timestepOutput, 
    double timeInit
  ) :
  CBasicRagaTask(_particles), 
  ptrPotential(_ptrPotential), 
  outputFileName(_outputFileName), 
  outputFileFormat(_outputFileFormat), 
  outputFileHeader(_outputFileHeader), 
  timestepOutput(_timestepOutput), 
  prevOutputTime(timeInit)
{
    writeSnapshot(static_cast<const CPointMassSet<float> >(*particles), timeInit);  // write output snapshot at time 0
}

template<typename NumT> bool CRagaTaskTrajectory::writeSnapshot(const CPointMassSet<NumT> &points, double dtime)
{
    CBasicIOSnapshot* snap=createIOSnapshotWrite(outputFileFormat, outputFileName, outputFileHeader, dtime, true);
    if(!snap) return false;
    bool result=snap->writeSnapshot(&points);
    delete snap;
    return result;
}

bool CRagaTaskTrajectory::episodeDone(double episodeStart, double episodeLength)
{
#if 0
    /// unimplemented: if timestepOutput<episodeLength, should store intermediate points as well...
    for(size_t step=0; step<configRaga.numOutputsPerStep; step++)
    {
        points.data.clear();
        double timeOut=timeSim+(step+1)*configRaga.timestepOutput;
        for(size_t ind=0; ind<initLib->size(); ind++) {
            const CBasicInformation* info=initLib->getOrbitDesc(ind)->getInfo(rfOutputTraj);
            if(info!=NULL && info->infoType()==CBasicInformation::IT_TRAJ_SAMPLE) {
                const CPointMassSet<float>::Type& data=static_cast<const CTrajEpisodeInformation*>(info)->data.data;
                if(step<data.size())
                    points.add(data[step].first, particleMasses[ind]);
            }
        }
        writeSnapshot(points, timeOut);
    }
#endif
    if(episodeStart+episodeLength>=prevOutputTime+timestepOutput) {
        writeSnapshot(static_cast<const CPointMassSet<float> >(*particles), episodeStart+episodeLength); ///!!! need to write all intermediate snapshots as well
        prevOutputTime=episodeStart+episodeLength;
    }
    return false;
}

CRagaTaskPotential::CRagaTaskPotential(
    CPointMassSet<double>* const _particles,
    const CPotential** const _ptrPotential,
    CMassModel** const _ptrMassModel,
    const double* _ptrEpisodeLength,
    size_t _numSamplesPerEpisode,
    const CConfigPotential* _ptrConfigPotential,
    const std::string& _outputPotentialFileName,
    double _timestepOutput,
    bool _updatePotential,
    bool _applyEnergyCorrection,
    double timeInit
) : 
  CBasicRagaTask(_particles),
  ptrPotential(_ptrPotential),
  ptrMassModel(_ptrMassModel),
  ptrEpisodeLength(_ptrEpisodeLength),
  numSamplesPerEpisode(_numSamplesPerEpisode),
  ptrConfigPotential(_ptrConfigPotential),
  outputPotentialFileName(_outputPotentialFileName),
  updatePotential(_updatePotential),
  applyEnergyCorrection(_applyEnergyCorrection),
  timestepOutput(_timestepOutput),
  prevOutputTime(timeInit)
{
    exportPotential(timeInit);
};

bool CRagaTaskPotential::exportPotential(double time)
{
    if(outputPotentialFileName.empty()) return false;
    if(*ptrMassModel != NULL) {  // export spherical mass model statistics
        exportModel(**ptrMassModel, (outputPotentialFileName+convertToString(time)+".tab").c_str());
    }
    const CPotential* thePotential=(*ptrPotential)->PotentialType()==CPotential::PT_COMPOSITE ?
        static_cast<const CPotentialComposite*>(*ptrPotential)->component(0) :
        *ptrPotential;  // obtain the stellar component of potential (without BH)
    return writePotential(outputPotentialFileName+convertToString(time) +
        coefFileExtension(thePotential->PotentialType()), thePotential);
}

bool CRagaTaskPotential::episodeDone(double episodeStart, double episodeLength)
{
    // obtain all sampling points stored for each orbit
    assert(infoOrbit.size()==particles->size());
    CPointMassSet<float> points;
    points.data.reserve(infoOrbit.size()*numSamplesPerEpisode);
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_TRAJ_SAMPLE) {
            const CPointMassSet<float>::Type& data=static_cast<const CTrajEpisodeInformation*>(infoOrbit[ip])->data.data;
            for(size_t i=1; i<data.size(); i++) { ///!!! starting from 1th point, since 0th is the initial condition
                points.add(data[i].first, static_cast<float>(mass*data[i].second/episodeLength));
            }
        }
    }
    // procedure for updating the potential is specific to each PotentialType
    const CPotential* newPotential=NULL;
    const CPotential* oldPotential=(*ptrPotential)->PotentialType()==CPotential::PT_COMPOSITE ? 
        static_cast<const CPotentialComposite*>(*ptrPotential)->component(0) : *ptrPotential;  // obtain the stellar component of potential (without BH)
    if(updatePotential) {  // recompute potential at the end of episode
        if(oldPotential->PotentialType()==CPotential::PT_BSE) {
            newPotential = new CPotentialBSE(
                static_cast<const CPotentialBSE*>(oldPotential)->getAlpha(),   // keep the same shape parameter as before (don't use configPotential.Alpha as it may be 0-autodetect)
                ptrConfigPotential->Ncoefs_radial, ptrConfigPotential->Ncoefs_angular,
                points, ptrConfigPotential->SymmetryType, true);
        } else if(oldPotential->PotentialType()==CPotential::PT_SPLINE) {
            std::vector< vectord > oldCoefs;
            vectord gridRadii;
            static_cast<const CPotentialSpline*>(oldPotential)->getCoefs(&gridRadii, &oldCoefs, true);  // use the same radial grid as before
            newPotential = new CPotentialSpline(
                ptrConfigPotential->Ncoefs_radial, ptrConfigPotential->Ncoefs_angular,
                points, ptrConfigPotential->SymmetryType, 0, &gridRadii);
        } else if(oldPotential->PotentialType()==CPotential::PT_SPHERICAL) {
            vectord gridRadii, oldMass;
            static_cast<const CPotentialSpherical*>(oldPotential)->getModel()->getMassProfile(&gridRadii, &oldMass);  // use the same radial grid
            newPotential = new CPotentialSpherical(points, ptrConfigPotential->splineSmoothFactor, &gridRadii);
            CMassModel::MODELSTATUS status=static_cast<const CPotentialSpherical*>(newPotential)->getModel()->getStatus();
            switch(status)
            {
            case CMassModel::MS_DF_NEGATIVE:
            case CMassModel::MS_DF_UNDEFINED:
                my_message(FUNCNAME, "Warning, isotropic DF is poorly defined");  // it is not used anyway
                break;
            case CMassModel::MS_OK:
                break;
            case CMassModel::MS_MASS_NOT_ASC:
                my_message(FUNCNAME, "Mass not increasing with radius; retained old potential");
                delete newPotential;
                return true;
            default:  // shouldn't occur...
                my_error(FUNCNAME, "Error in creating new CPotentialSpherical");
                delete newPotential;
                return true;
            }
        } else {  // shouldn't happen - we have checked potential type at start
            my_error(FUNCNAME, "Unknown potential type is set to be updated");
            return true;  // signal of error
        }
    } else {  // updatePotential==false means that we don't change the stellar potential, but still should reinit the total potential (with black hole)
        newPotential=oldPotential->clone();
    }
    if(ptrConfigPotential->Mbh>0) {  // create a composite potential with black hole(s)
        std::vector<const CPotential*> components(1, newPotential);
        newPotential = new CPotentialComposite(components, ptrConfigPotential->Mbh,
            ptrConfigPotential->binary_q, ptrConfigPotential->binary_sma, 
            ptrConfigPotential->binary_ecc, ptrConfigPotential->binary_phase);
    }
    if(applyEnergyCorrection) {
        // apply correction for particle energy changes due to time-dependent potential
        double totalEcorr=0, totalMass=0;
        for(size_t ip=0; ip<infoOrbit.size(); ip++) {
            double pmass=particles->at(ip).second;
            if(pmass==0 || infoOrbit[ip]==NULL) 
                continue;
            const CPosVelPoint<float>& pOld = static_cast<const CTrajEpisodeInformation*>(infoOrbit[ip])->data.data[0].first;  // 0th point is the initial condition
            const CPosVelPoint<double>&pNew = particles->at(ip).first;
            double oldPotOldPos=(*ptrPotential)->Phi(pOld.Pos[0], pOld.Pos[1], pOld.Pos[2], 0);
            double oldPotNewPos=(*ptrPotential)->Phi(pNew.Pos[0], pNew.Pos[1], pNew.Pos[2], episodeLength);
            double newPotOldPos=newPotential->Phi(pOld.Pos[0], pOld.Pos[1], pOld.Pos[2], -episodeLength);
            double newPotNewPos=newPotential->Phi(pNew.Pos[0], pNew.Pos[1], pNew.Pos[2], 0);
            totalEcorr+=pmass*(newPotOldPos-oldPotNewPos)*0.5;
            double deltaEpart =   // this is how much we need to correct energy of this particle
                0.5*(newPotOldPos - oldPotOldPos)
               -0.5*(newPotNewPos - oldPotNewPos);
            double Vnew2 = pow_2(pNew.Vel[0])+pow_2(pNew.Vel[1])+pow_2(pNew.Vel[2]);
            double Vupd2 = Vnew2 + deltaEpart*2;
            if(Vupd2<Vnew2*0.25) {   // correction is too large
                //my_error("At time "+convertToString(episodeStart+episodeLength)+" couldn't correct #"+convertToString(ip)+"'s Energy "+
                //         convertToString(totalEnergy(*ptrPotential, particles->at(ip).first))+" by "+convertToString(deltaEpart));
                Vupd2=Vnew2*0.25;  // reduce correction
                totalEcorr+=pmass*(Vupd2/2-Vnew2/2-deltaEpart); // and append undercorrected energy to the total correction budget
            }
            totalMass +=pmass;
            double Vmul=sqrt(Vupd2/Vnew2);
            for(size_t c=0; c<3; c++) particles->at(ip).first.Vel[c]*=Vmul;
        }
#ifdef DEBUGPRINT
        my_message(FUNCNAME, 
            "Need additional correction -E="+convertToString(totalEcorr));
#endif
        // another round of corrections, to cancel any residual imbalance of energy
        for(size_t ip=0; ip<infoOrbit.size(); ip++) {
            double deltaEpart = -totalEcorr/totalMass;
            double Vold2 = pow_2(particles->at(ip).first.Vel[0])+pow_2(particles->at(ip).first.Vel[1])+pow_2(particles->at(ip).first.Vel[2]);
            double Vnew2 = Vold2 + deltaEpart*2;
            if(Vnew2<Vold2*0.25) {   // correction is too large
                Vnew2=Vold2*0.25;
            }
            double Vmul=(Vold2==0?1:sqrt(Vnew2/Vold2));
            for(size_t c=0; c<3; c++) particles->at(ip).first.Vel[c]*=Vmul;
        }
    }
    delete (*ptrPotential);
    *ptrPotential=newPotential;

    // compute new distribution function in energy used to calculate diffusion coefficients
    if(*ptrMassModel != NULL) {   // massmodel is used
        if(updatePotential) {     // potential has changed after episode
            CMassModel* newModel = createMassModel(*ptrPotential);  // create a new model with the new potential
            if(newModel!=NULL) {
                delete *ptrMassModel; // delete spherical model associated with the old potential
                *ptrMassModel=newModel;
            }
            else
                my_error(FUNCNAME, "Cannot update spherical mass model with the new potential; retained the old one");
        }
        vectord pointMasses(points.size()), pointEnergies(points.size());
        for(size_t i=0; i<points.size(); i++) {
            pointMasses[i]=points[i].second;
            pointEnergies[i]=totalEnergySph(*ptrMassModel, points[i].first);
        }
        vectord energyBins, distrFnc;
        //computeDistrFunc(model, sampleMasses, sampleEnergies, configRaga.gridSize, &energyBins, &distrFnc);
        if(!fitMassModel(pointEnergies, pointMasses, &energyBins, &distrFnc, ptrConfigPotential->splineSmoothFactor, 
            //configRaga.gridSize, configRaga.numSamplesPerStep*5, 
            0, 0, CMassModel::FT_DF_ENERGY, *ptrMassModel)) {
            my_error(FUNCNAME, "Cannot compute distribution function");
        } else {
            /*std::string fn=(configRaga.fileOutputPotential+convertToString(timeSim+configRaga.episodeLength)+".fite");
            remove(fn.c_str());
            rename("temp.fit", fn.c_str()); */
            if(!(*ptrMassModel)->initDistrFuncEnergy(energyBins, distrFnc)) {
                my_error(FUNCNAME, "Could not re-initialize distribution function");
                std::ofstream strm(("baddf"+convertToString(episodeStart+episodeLength)).c_str());
                for(size_t b=0; b<energyBins.size(); b++)
                    strm << energyBins[b] << "\t" << distrFnc[b] << "\n";
            }
        }
        // recompute diffusion coefficients for the new potential and distribution function
        (*ptrMassModel)->initSplineDifCoefE();
        (*ptrMassModel)->initSplineDifCoefV();
    }
    if(episodeStart+episodeLength>=prevOutputTime+timestepOutput) {
        exportPotential(episodeStart+episodeLength);
        prevOutputTime=episodeStart+episodeLength;
    }
    return false;
}

CRagaTaskRelaxation::CRagaTaskRelaxation(
    CPointMassSet<double>* const _particles, 
    const CMassModel* const* _ptrMassModel, 
    double* _ptrEpisodeLength, 
    double _relaxRate, 
    bool _applyEnergyCorrection
) :
  CBasicRagaTask(_particles), 
  ptrMassModel(_ptrMassModel), 
  ptrEpisodeLength(_ptrEpisodeLength), 
  relaxRate(_relaxRate),
  applyEnergyCorrection(_applyEnergyCorrection)
{}

bool CRagaTaskRelaxation::episodeDone(double episodeStart, double episodeLength)
{
    if(!applyEnergyCorrection)
        return false;
    // task: correct particle energies for fluctuations due to relaxation
    assert(infoOrbit.size()==particles->size());
    vectord accumProportionParticle(infoOrbit.size());
    double accumProportion=0, deltaE=0;
    // compute energy change of each particle and its average energy diffusion coefficient
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_VEL_PERTURBER) {
            float vpdata = static_cast<const CVelPerturberInformation<float>*>(infoOrbit[ip])->getData();
            // accumulated energy change by the end of interval resulting from relaxation only 
            // (not accounting for systematic change, e.g. due to time-dependent potential)
            deltaE += vpdata*mass;
            // FIXME: in the previous version, the energy disbalance was split in proportion to 
            // the average value of energy diffusion coefficient
            // now it is temporarily disabled and delta-e is uniformly distributed between particles
            accumProportionParticle[ip]=1;//sqrt(vpdata);
            accumProportion+=accumProportionParticle[ip]*mass;
        }
    }
    if(accumProportion==0 || !gsl_finite(accumProportion)) accumProportion=1;  // shouldn't probably happen, but will prevent division by 0

    // apply correction for energy fluctuations by changing particle velocity magnitudes
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double deltaEpart=accumProportionParticle[ip]/accumProportion * deltaE;
        double Vold2 = pow_2(particles->at(ip).first.Vel[0])+pow_2(particles->at(ip).first.Vel[1])+pow_2(particles->at(ip).first.Vel[2]);
        double Vnew2 = Vold2 - deltaEpart*2;
        if(Vnew2<Vold2*0.25) {   // correction is too large
            Vnew2=Vold2*0.25;
        }
        double Vmul=sqrt(Vnew2/Vold2);
        for(size_t c=0; c<3; c++) particles->at(ip).first.Vel[c]*=Vmul;
    }
    return false;
}

CRagaTaskLosscone::CRagaTaskLosscone(
    CPointMassSet<double>* const _particles, 
    double* _ptrMbh,
    double _captureRadius, 
    const std::string& _outputLossconeFileName
) :
  CBasicRagaTask(_particles), 
  ptrMbh(_ptrMbh), 
  captureRadius(_captureRadius), 
  outputLossconeFileName(_outputLossconeFileName) 
{
    particleMassInit.resize(particles->size());
    for(size_t ip=0; ip<particles->size(); ip++)
        particleMassInit[ip]=particles->at(ip).second;
    particlePericenterMinDist.assign(particles->size(), -1);  // default value means no information

    std::ofstream strmCaptD((outputLossconeFileName+".captdata").c_str());
    strmCaptD << "tcapt\tL2\tdeltaL2\tE\tmass\tindex\n";
    std::ofstream strmCaptO((outputLossconeFileName+".captorb" ).c_str());
    strmCaptO << "x\ty\tz\tvx\tvy\tvz\tweight\n";
}

bool CRagaTaskLosscone::episodeDone(double episodeStart, double episodeLength)
{
    // task: track loss-cone captures and optionally record pericenter distances
    assert(infoOrbit.size()==particles->size());
    double capturedMass=0;
    std::vector< std::pair<double, size_t> > captures;
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_PERICENTER) {
            const CLossconeInformation* info = static_cast<const CLossconeInformation*>(infoOrbit[ip]);
            if(info->L2>0) {
                capturedMass += mass;
                captures.push_back(std::pair<double, size_t>(info->tcapt, ip));
            }
#if 0
            if(maxPericenterTrackingRadius>0 && info->pericentert.size()>0) {  // store pericenter passages in text files
                std::ofstream strmPeri((convertToString(ip)+".peri").c_str(), std::ios::out | std::ios_base::app);
                for(size_t k=0; k<info->pericentert.size(); k++) {
                    strmPeri << (convertToString(episodeStart+info->pericentert[k],10) + "\t" +
                       convertToString(info->pericenterr[k]) + "\t" + convertToString(info->pericenterE[k]) + "\n");
                    if(particlePericenterMinDist[ip]<0 || info->pericenterr[k]<particlePericenterMinDist[ip])
                        particlePericenterMinDist[ip]=info->pericenterr[k];   // update the minimum distance reached by this particle
                }
            }
#endif
        }
    }
    if(captures.size()>0) {
        std::sort(captures.begin(), captures.end(), comparepair<double, size_t>);
        std::ofstream strmCaptD((outputLossconeFileName+".captdata").c_str(), std::ios_base::app);
        std::ofstream strmCaptO((outputLossconeFileName+".captorb" ).c_str(), std::ios_base::app);
        strmCaptO << std::setprecision(12);
        for(size_t c=0; c<captures.size(); c++) {
            size_t ip=captures[c].second;  // index of captured particle
            const CLossconeInformation* info=static_cast<const CLossconeInformation*>(infoOrbit[ip]);
            strmCaptD << (episodeStart+info->tcapt) << "\t" << info->L2 << "\t" 
                << info->deltaL2 << "\t" << info->Ecapt << "\t" << particles->at(ip).second << "\t" << ip << "\n";
            strmCaptO 
                << info->capturePoint.Pos[0] << "\t" << info->capturePoint.Pos[1] << "\t" << info->capturePoint.Pos[2] << "\t"
                << info->capturePoint.Vel[0] << "\t" << info->capturePoint.Vel[1] << "\t" << info->capturePoint.Vel[2] << "\t"
                << particles->at(ip).second << "\n";
            particles->at(ip).second=0;  // eliminate this particle
        }
        my_message(FUNCNAME,
            "By time "+convertToString(episodeStart+episodeLength)+" captured "+convertToString(captures.size())+
            " particles, total mass="+convertToString(capturedMass));
        if(ptrMbh!=NULL)
            *ptrMbh += capturedMass;
    }
#if 0
    // update the list of min.reached radii
    std::ofstream strmList((outputLossconeFileName+".listperi").c_str());
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        if(particleMassInit[ip]>0 && particlePericenterMinDist[ip]>0)
            strmList << ip << "\t" << particleMassInit[ip] << "\t" << particlePericenterMinDist[ip] << "\n";
    }
#endif
    return false;
}

CRagaTaskBinaryBH::CRagaTaskBinaryBH(
    CPointMassSet<double>* const _particles, 
    const CPotential* const* _ptrPotential, 
    CConfigPotential* _ptrConfigPotential,
    double speedOfLight,
    const std::string& _outputBinaryFileName,
    double timeInit
) : 
  CBasicRagaTask(_particles), 
  ptrPotential(_ptrPotential), 
  ptrConfigPotential(_ptrConfigPotential),
  cminus5(speedOfLight>0 ? pow(speedOfLight, -5.) : 0),
  outputBinaryFileName(_outputBinaryFileName) 
{
    if(!outputBinaryFileName.empty()) {
        std::ofstream strmBinary((outputBinaryFileName).c_str());
        strmBinary << "t\tnumOrb\t1/a\tecc\n";
        strmBinary << timeInit << "\t0\t" << convertToString(ptrConfigPotential->binary_sma) << "\t" <<
            convertToString(ptrConfigPotential->binary_ecc) << "\n";
        strmBinary.close();
#if 0
        strmBinary.open((outputBinaryFileName+"_enc").c_str());
        strmBinary << "t\tm\tE\tL\tLz\tdE\tdLz\td_ecc\n";
#endif
    }
}

bool CRagaTaskBinaryBH::episodeDone(double episodeStart, double episodeLength)
{
    // task: track particle energy and angular momentum changes due to the binary black hole, 
    // and adjust the BBH orbit using the conservation laws
    if(ptrConfigPotential->binary_sma<=0 || ptrConfigPotential->Mbh<=0) 
        return false;
    assert(infoOrbit.size()==particles->size());
    double Mbh=ptrConfigPotential->Mbh;
    double sma=ptrConfigPotential->binary_sma;
    double q=ptrConfigPotential->binary_q;
    double ecc2=pow_2(ptrConfigPotential->binary_ecc);
    double deltaE=0;
    double deltaL=0;
#if 0
    double Lnorm=sqrt(2*Mbh*sma);
    std::ofstream strmBinaryEnc((outputBinaryFileName+"_enc").c_str(), std::ios_base::app);
#endif
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_POINCARE) {  // poincare is a codename for BinaryBH runtime function...
            const CBinaryBHInformation* info = static_cast<const CBinaryBHInformation*>(infoOrbit[ip]);
            for(size_t ne=0; ne<info->encounters.size(); ne++) {
                deltaE+=mass*info->encounters[ne].deltaE;
                deltaL+=mass*info->encounters[ne].deltaL;
#if 0
                if(info->encounters[ne].Lstart/Lnorm<2.0)  strmBinaryEnc << 
                    (episodeStart+info->encounters[ne].tbegin) << "\t" << 
                    ip << "\t" << info->encounters[ne].Estart << "\t" << 
                    (info->encounters[ne].Lstart/Lnorm) << "\t" << 
                    (info->encounters[ne].Lzstart/Lnorm) << "\t" << 
                    info->encounters[ne].deltaE << "\t" << info->encounters[ne].deltaL << "\t" << 
                    info->encounters[ne].deltaLtorq << "\t" << 
                    (episodeStart+info->encounters[ne].tend) << "\n";
#endif
            }
        }
    }
#if 0
    strmBinaryEnc.close();
#endif
    double dphase=sqrt(Mbh/(sma*sma*sma))*episodeLength;
    //double Ebin=-Mbh*Mbh*q/(pow_2(q+1)*2*sma);
    //double Lbin=sqrt((1-ecc2)*Mbh*sma)*Mbh*q/pow_2(q+1);

    // apply changes due to interactions with stars and GW
    double S_star=2*pow_2(q+1)/(q*Mbh*Mbh) * deltaE/episodeLength;       // hardening rate due to stars: d(1/a)/dt
    double E_star=(1-ecc2)*(2*pow_2(q+1)/(q*Mbh*sqrt((1-ecc2)*Mbh*sma)) * deltaL/episodeLength - sma*S_star); // d(e^2)/dt due to stars
    double S_gw_avg=0, E_gw_avg=0;
    double timepassed=0;
    do{   // evolve sma and ecc, in several sub-steps if necessary
        double S_gw = cminus5 * pow(Mbh, 3.0) * pow(sma, -5.0) * q/pow_2(q+1) * 
            pow(1-ecc2, -3.5) * (192+584*ecc2+74*ecc2*ecc2)/15;
        double E_gw =-cminus5 * pow(Mbh, 3.0) * pow(sma, -4.0) * q/pow_2(q+1) * 
            pow(1-ecc2, -2.5) * (304+121*ecc2)*2/15 * ecc2;
        double tstep = 0.01*std::min<double>( std::min<double>(1-ecc2, ecc2+1e-3)/fabs(E_star+E_gw), 1/sma/fabs(S_star+S_gw) );
        if(!gsl_finite(tstep) || tstep>episodeLength-timepassed) 
            tstep=episodeLength-timepassed;
        if(1/sma/S_gw/4<episodeLength-timepassed && ecc2<0.01) {  // imminent coalescence due to GW emission
            timepassed+=1/sma/S_gw/4;
            episodeLength=timepassed;  // will exit the loop and record the exact coalescence time in file
            sma=ecc2=S_gw_avg=0;
        } else {
            sma=1/(1/sma+tstep*(S_star+S_gw));
            ecc2+=tstep*(E_star+E_gw);
            ecc2=std::min<double>(0.999*0.999, std::max<double>(0, ecc2));
            S_gw_avg+=S_gw*tstep;
            E_gw_avg+=E_gw*tstep;
            timepassed+=tstep;
        }
    } while(timepassed<episodeLength*(1-1e-10));
    S_gw_avg/=episodeLength;
    ptrConfigPotential->binary_phase+=dphase;
    ptrConfigPotential->binary_sma=sma;
    ptrConfigPotential->binary_ecc=sqrt(ecc2);
    my_message(FUNCNAME, 
        "Hardening rate: "+convertToString(S_star)+(cminus5?"+"+convertToString(S_gw_avg):"")+
        ", new a="+convertToString(ptrConfigPotential->binary_sma)+
        ", ecc="+convertToString(ptrConfigPotential->binary_ecc)+
        ", Norbits="+convertToString(dphase/(2*M_PI)));
    ptrConfigPotential->binary_phase -= 2*M_PI*floor(ptrConfigPotential->binary_phase/(2*M_PI));
    if(!outputBinaryFileName.empty()) {
        std::ofstream strmBinary((outputBinaryFileName).c_str(), std::ios_base::app);
        strmBinary << (episodeStart+episodeLength) << "\t" << (dphase/(2*M_PI)) << "\t" <<
            convertToString(ptrConfigPotential->binary_sma) << "\t" <<
            convertToString(ptrConfigPotential->binary_ecc) << "\t" << 
#if 0
            (convertToString(S_star) + "\t" + convertToString(E_star) + 
            (cminus5>0?"\t"+convertToString(S_gw_avg)+"\t"+convertToString(E_gw_avg):"")) << 
#endif
            "\n";
    }
    return false;
}

};  // namespace
