/** \file    raga_tasks.h
    \brief   Various tasks that can be performed by Raga code
    \author  EV
    \date    2013-2015


This module implements the driver routine for self-consistent secular evolution code RAGA.
The input parameters to the code are:
  - input orbit library file (i.e. nbody snapshot serving as initial conditions
  - initial potential (may also be computed from initial snapshot)
  - total integration time (T_final)
  - interval for outputting snapshots of current locations of particles (T_output)
  - interval for updating potential (one step) (T_update)
  - number of potential update steps to compute the time-averaged potential (N_update)
  - number of sampling points in trajectory of each particle per update step (N_sample)
  - prescribed two-body relaxation rate
  - prescribed symmetry of potential?

  If N_update>>1, the potential evolution is almost adiabatic; if N_update=1 then it is more step-like.
  If T_update>=T_output, the run is essentially in a fixed potential 
  (N_update=0 also disables potential update entirely).
  If N_sample=1 and T_update=one integration timestep, this is equivalent to standard SCF approach.
*/
#pragma once
#include "potential.h"
#include "orbitan.h"

namespace smile{

    /** Trajectory stored at regular intervals during one episode. 
        The difference from CTrajSampleInformation is that here point masses are also stored.
        This information object is used for writing down the output file with trajectories, 
        and for sampling the trajectories for updating the potential and distribution function.
    */
    class CTrajEpisodeInformation: public CBasicInformation
    {
    public:
        CTrajEpisodeInformation(const CPointMassSet<float>& src) : 
          data(src) {};
        virtual INFOTYPE infoType() const { return IT_TRAJ_SAMPLE; };
        virtual std::string toString() const { return ""; }
        CPointMassSet<float> data;
    };

    class COrbitRuntimeTrajEpisode: public CBasicOrbitRuntimeFnc
    {
    public:
        COrbitRuntimeTrajEpisode(const COrbit* _orbit, double _timeStep) : 
          CBasicOrbitRuntimeFnc(_orbit), timeStep(_timeStep), lastTime(0)
        {};
        virtual FNCTYPE FncType() const {return FT_TRAJ_SAMPLE; };
        virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
        virtual CBasicInformation* getData() const { return new CTrajEpisodeInformation(traj); };
    private:
        const double timeStep;        ///< time interval for recording the trajectory (must be less than duration of integration divided by numSamplingPoints)
        CPointMassSet<float> traj;    ///< recorded trajectory
        double lastTime;
    };

    class CLossconeInformation: public CBasicInformation
    {
    public:
        CLossconeInformation(double _L2, double _deltaL2, double _tcapt, double _Ecapt, const CPosVelPoint<double> &_capturePoint,
            const vectord& _pericentert, const vectord& _pericenterr, const vectord& _pericenterE) : 
          L2(_L2), deltaL2(_deltaL2), tcapt(_tcapt), Ecapt(_Ecapt), capturePoint(_capturePoint),
          pericentert(_pericentert), pericenterr(_pericenterr), pericenterE(_pericenterE) {};
        virtual INFOTYPE infoType() const { return IT_PERICENTER; };
        virtual std::string toString() const { return std::string(); }
        double L2, deltaL2, tcapt, Ecapt;
        size_t ind;
        CPosVelPoint<double> capturePoint;
        vectord pericentert, pericenterr, pericenterE;
    };

    /** Keep track of close approach to the supermassive black hole */
    class COrbitRuntimeLosscone: public COrbitRuntimePericenter
    {
    public:
        COrbitRuntimeLosscone(const COrbit* _orbit, double _captureRadius, double _maxPericenterTrackingRadius) :
          COrbitRuntimePericenter(_orbit), captureRadius(_captureRadius), maxPericenterTrackingRadius(_maxPericenterTrackingRadius),
          tcapt(0), Ecapt(0), L2(0), prevL2(0), deltaL2(0) {};
        virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
        virtual void Finish() {};
        virtual CBasicInformation* getData() const 
        { return new CLossconeInformation(L2, deltaL2, tcapt, Ecapt, capturePoint, pericentert, pericenterr, pericenterE); };
    private:
        const double captureRadius;
        const double maxPericenterTrackingRadius;
        double tcapt, Ecapt, L2, prevL2, deltaL2, maxradius;
        vectord pericenterE;
        CPosVelPoint<double> capturePoint;
    };

    struct CBinaryEncounter {
        double Estart, Lstart, Lzstart; // energy and z-component of angular momentum at the beginning of encounter
        double deltaE, deltaL, deltaLtorq; // accumulated change in energy and angular momentum during encounter
        double tbegin, tend;
        CBinaryEncounter(double E, double L, double Lz, double t) : 
          Estart(E), Lstart(L), Lzstart(Lz), deltaE(0), deltaL(0), deltaLtorq(0), tbegin(t), tend(t) {};
    };

    class CBinaryBHInformation: public CBasicInformation
    {
    public:
        CBinaryBHInformation(const std::vector<CBinaryEncounter> &_encounters) : encounters(_encounters) {};
        virtual INFOTYPE infoType() const { return IT_POINCARE; };
        virtual std::string toString() const { return ""; }
        std::vector<CBinaryEncounter> encounters;
    };

    class COrbitRuntimeBinaryBH: public CBasicOrbitRuntimeFnc
    {
    public:
        COrbitRuntimeBinaryBH(const COrbit* _orbit, double _rcrit) :
          CBasicOrbitRuntimeFnc(_orbit), rcrit2(_rcrit*_rcrit) {};
        virtual FNCTYPE FncType() const {return FT_POINCARE; }; // should use one of pre-defined enums, no matter which..
        virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
        virtual CBasicInformation* getData() const { return new CBinaryBHInformation(encounters); };
    private:
        const double rcrit2;      // radius of critical sphere (few times binary separation), squared
        std::vector<CBinaryEncounter> encounters;  // list of close encounters with the binary during the episode
    };

    /** Base class for tasks that are performed on each orbit (as RuntimeFunctions) 
        and before/after each episode.
        These tasks are descendants of RuntimeFncCreator class, which means that they are used 
        to create associated runtime functions for each orbit.
        They receive the pointer to the entire orbit library, so that at the end of each episode 
        they may walk through each COrbitDesc and collect necessary data from its information objects.
        The index of the relevant information object is passed as a parameter in the constructor
        (it is also the index of the given task in the array of RuntimeFncCreators).
    */
    class CBasicRagaTask: public CBasicOrbitRuntimeFncCreator
    {
    public:
        explicit CBasicRagaTask(CPointMassSet<double> * const _particles): 
          particles(_particles) 
        { infoOrbit.assign(particles->size(), NULL); };
        /// perform tasks at the end of each episode; return true if the entire simulation needs to be terminated
        virtual bool episodeDone(double episodeStart, double episodeLength)=0;
        void setInfoOrbit(size_t ind, const CBasicInformation* info) 
        { if(ind<infoOrbit.size()) { delete infoOrbit[ind]; infoOrbit[ind]=info; } }
    protected:
        CPointMassSet<double>* particles;
        std::vector<const CBasicInformation*> infoOrbit;
    };

    class CRagaTaskTrajectory: public CBasicRagaTask
    {
    public:
        CRagaTaskTrajectory(
            CPointMassSet<double>* const _particles, 
            const CPotential* const * _ptrPotential, 
            const std::string& _outputFileName, 
            const std::string& _outputFileFormat, 
            const std::string& _outputFileHeader, 
            double _timestepOutput,
            double timeInit
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeTrajEpisode(orbit, timestepOutput); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CPotential* const* ptrPotential;
        const std::string outputFileName;
        const std::string outputFileFormat;
        const std::string outputFileHeader;
        double timestepOutput;
        double prevOutputTime;
        template <typename NumT> bool writeSnapshot(const CPointMassSet<NumT> &points, double time);
    };

    class CRagaTaskPotential: public CBasicRagaTask
    {
    public:
        CRagaTaskPotential(
            CPointMassSet<double>* const _particles, 
            const CPotential** const _ptrPotential, 
            CMassModel** const _ptrMassModel, 
            const double* _ptrEpisodeLength, 
            size_t _numSamplesPerEpisode, 
            const CConfigPotential* _ptrConfigPotential, 
            const std::string& _outputPotentialFileName, 
            double _timestepOutput, 
            bool _updatePotential,
            bool _applyEnergyCorrection,
            double timeInit
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeTrajEpisode(orbit, (*ptrEpisodeLength)/numSamplesPerEpisode); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CPotential** ptrPotential;
        CMassModel** ptrMassModel;
        const double* ptrEpisodeLength;
        const size_t numSamplesPerEpisode;
        const CConfigPotential* ptrConfigPotential;
        const std::string outputPotentialFileName;
        const bool updatePotential;
        const bool applyEnergyCorrection;
        double timestepOutput;
        double prevOutputTime;
        bool exportPotential(double time);
    };

    class CRagaTaskRelaxation: public CBasicRagaTask
    {
    public:
        CRagaTaskRelaxation(
            CPointMassSet<double>* const _particles, 
            const CMassModel* const* _ptrMassModel, 
            double* _ptrEpisodeLength, 
            double _relaxRate,
            bool _applyEnergyCorrection
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeVelPerturber(orbit, *ptrMassModel, relaxRate); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CMassModel* const* ptrMassModel;
        double* ptrEpisodeLength;
        const double relaxRate;
        const bool applyEnergyCorrection;
        double avgtrelc;
    };

    class CRagaTaskLosscone: public CBasicRagaTask
    {
    public:
        CRagaTaskLosscone(
            CPointMassSet<double>* const _particles, 
            double* _ptrMbh,
            double _captureRadius,
            const std::string& _outputLossconeFileName
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeLosscone(orbit, captureRadius, 0); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        double* ptrMbh;  // pointer to configPotential.Mbh, to be updated when particles are captured
        const double captureRadius;
        const std::string outputLossconeFileName;
        double prevTimeSim;
        vectord particleMassInit; 
        vectord particlePericenterMinDist;
    };

    class CRagaTaskBinaryBH: public CBasicRagaTask
    {
    public:
        CRagaTaskBinaryBH(
            CPointMassSet<double>* const _particles, 
            const CPotential* const* _ptrPotential, 
            CConfigPotential* _ptrConfigPotential,
            double speedOfLight,
            const std::string& _outputBinaryFileName,
            double timeInit
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeBinaryBH(orbit, ptrConfigPotential->binary_sma*10); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CPotential* const* ptrPotential;
        CConfigPotential* ptrConfigPotential;
        const double cminus5;
        const std::string outputBinaryFileName;
    };

    /// utility function to compute total energy of a particle in a spherical mass model
    template<typename NumT> double totalEnergySph(const CMassModel* model, const CPosVelPoint<NumT> &point);

}; // namespace
