/**
    This program numerically integrates the coupled equations for the evolution of
    semimajor axis and eccentricity of the binary black hole under two effects:
    stellar-dynamical hardening and gravitational-wave emission.
    It computes the coalescence time and optionally a detailed evolution track.
*/
#include <iostream>
#include <cmath>
#include <gsl/gsl_sf_hyperg.h>
#define LIBSMILEPOT
#include "configfile.h"
#include "utils.h"

// extension for 2F1 into the range x<-1 from Heiko Bauke <heiko.bauke@snafu.de>
double hyperg_2F1(double a, double b, double c, double x)
{
  if (-1.<=x and x<1.)
    return gsl_sf_hyperg_2F1(a, b, c, x);
  if (x<-1.) {
    // gsl_sf_hyperg_2F1 may have problems with negative arguments
    if (c-a<0)
      return pow(1.-x, -a)*
        gsl_sf_hyperg_2F1(a, c-b, c, x/(x-1.));
    if (c-b<0)
      return pow(1.-x, -b)*
        gsl_sf_hyperg_2F1(c-a, c, c, x/(x-1.));
    // choose one of two equivalent formulas which is expected to be more accurate
    if (a*(c-b)<(c-a)*b)
      return pow(1.-x, -a)*
        gsl_sf_hyperg_2F1(a, c-b, c, x/(x-1.));
    else
      return pow(1.-x, -b)*
        gsl_sf_hyperg_2F1(c-a, b, c, x/(x-1.));
  }
  // insert some error handling for x>=1
  return gsl_nan();
}

double fe2(double e2) { return pow(1-e2, 3.5)/(1+e2*(73./24+37./96*e2)); }

int main(int argc, char *argv[])
{
    if(argc<=1) {
        std::cout << "Compute evolution of semimajor axis and eccentricity of black hole binary, and its coalescence time.\n"
        "Usage: tcoal param=value param2=value2 ... , where parameters [default values] are: \n"
        " a [???] - initial semimajor axis of hard binary in N-body units\n"
        " c [???] - speed of light in N-body units\n"
        " m [1]   - combined mass of two black holes in N-body units\n"
        " q [1]   - mass ratio between lighter and heavier components of the binary\n"
        " e [0]   - initial eccentricity\n"
        " s [???] - initial stellar-dynamical hardening rate in N-body units\n"
        " b [0]   - exponent in the expression describing the decrease of the hardening rate with time (b=0 means no decrease)\n"
        " o [f]   - full evolution output [t] or coalescence time only [f]\n"
        " ea [0.3], eb [0.6], ec[0.2] - parameters in the expression describing the stellar-dynamical eccentricity increase (ea=0 means no increase)\n";
        return 0;
    }
    smile::CConfigFile settings("cfg", argc, argv);
    double Mbh = settings.getValueDouble("cfg/m", 1);  // total mass of two black holes
    double a_h = settings.getValueDouble("cfg/a", 0);  // initial value of semimajor axis (a_h)
    double e_h = settings.getValueDouble("cfg/e", 0);  // binary eccentricity
    double q   = settings.getValueDouble("cfg/q", 1);  // binary mass ratio
    double c   = settings.getValueDouble("cfg/c", 0);  // speed of light
    double s   = settings.getValueDouble("cfg/s", 0);  // stellar-dynamical hardening rate S=d(1/a)/dt at a=a_h
    double b   = settings.getValueDouble("cfg/b", 0);  // S(a)=S(a_h)*(a/a_h)^b
    double ea  = settings.getValueDouble("cfg/ea", 0.3);
    double eb  = settings.getValueDouble("cfg/eb", 0.6);
    double ec  = settings.getValueDouble("cfg/ec", 0.2);  // K = d e / d ln(1/a) = ea*e*(1-e^2)^eb/(1+a/(ec*a_h))
    bool fulloutput = settings.getValueBool("cfg/o", false);  // whether to print entire evolution
    if(a_h==0 || c==0 || s==0) { std::cout << "insufficient parameters, exiting.\n"; return 0; }
    double mul = 64./5*pow(Mbh,3)/pow(c*a_h,5)*q/(1+q*(q+2));
    double epsh= mul/fe2(e_h*e_h)/s;
    double hg_end=hyperg_2F1(1, 4/(5+b), (9+b)/(5+b), -1/epsh);
    double tcoal_est=hg_end / (4*epsh*s*a_h);
    double chi=1;  // a_h/a
    double e2=e_h*e_h; // eccentricity^2
    double t=0;
    double a_gw=0, t_gw=0;
    double cumul_S_s=0;
    double maxecc=0;
    const double tau=0.002;  // factor in timestep assignment
    if(fulloutput) std::cout << "time    \tt_analytic\tsemimajor_axis\tecc\tS_star\tS_gw\n";
    while(42) {
        //double S_s=s*pow(chi, -b);
        double S_s=s*pow(1+t*s*a_h*(b+1), -b/(b+1));
        double S_g=mul*pow(chi, 5)/fe2(e2);
        double de2dt_s=ea*2*e2*pow(1-e2,eb)/(1+1/ec/chi) * S_s*a_h/chi;
        double de2dt_g=-mul*a_h*pow(chi, 4)*e2*(304+121*e2)/96/pow(1-e2,2.5);
        double dt=tau * std::min<double>( fabs((e2-1)/(de2dt_g+de2dt_s)), chi/(S_s+S_g)/a_h);
        double t_analytic=(hg_end - hyperg_2F1(1, 4/(5+b), (9+b)/(5+b), -pow(chi,-5-b)/epsh)*pow(chi,-4))/(4*epsh*s*a_h);
        if(fulloutput) std::cout << (
            smile::pp(t,14)+"\t"+smile::pp(t_analytic,14)+"\t"+
            smile::pp(a_h/chi,14)+"\t"+smile::pp(sqrt(e2),7)+"\t"+
            smile::pp(S_s,7)+"\t"+//smile::pp(de2dt_s,10)+"\t"+
            smile::pp(S_g,7)+"\t"+/*smile::pp(de2dt_g,10)+"\t"+smile::pp(cumul_S_s,7)+*/"\n");
        if(a_gw==0 && S_g>=S_s) { a_gw=a_h/chi; t_gw=t; }
        if(S_s<S_g*1e-4 && e2<1e-3) { // end of evolution
            t+=chi/S_g/a_h/4;
            if(fulloutput)
                std::cout << (smile::pp(t,14)+"\t"+smile::pp(tcoal_est,14)+"\t0\t0\tgameover\teps:\t"+smile::pp(epsh,8)+"\n");
            //double chi_gw_est=pow(epsh, -1/(5+b));
            //double t_gw_est=(hg_end - hyperg_2F1(1, 4/(5+b), (9+b)/(5+b), -1)*pow(chi_gw_est,-4))/(4*epsh*s*a_h);
            //double t_approx=1/(s*a_h*pow(epsh, (1+b)/(5+b)));
            double t_approx=pow(epsh, -0.2)/(a_h*s);
            std::cout << ("T_coal: "+smile::pp(t,10)+// "\tT_coal,est: "+smile::pp(tcoal_est,10)+
                "\ta_GW: "+smile::pp(a_gw,10)+// "\ta_GW,est: "+smile::pp(a_h/chi_gw_est,10)+
                "\tt_GW: "+smile::pp(t_gw,10)+// "\tt_GW,est: "+smile::pp(t_gw_est,10)+
                "\tt_approx: "+smile::pp(t_approx,10)+
                "\te_max: "+smile::pp(sqrt(maxecc),7)+"\n");
            return 0;
        }
        chi += a_h*(S_s+S_g)*dt;
        e2  += (de2dt_s+de2dt_g)*dt;
        t   += dt;
        cumul_S_s += S_s*dt;
        maxecc=std::max<double>(maxecc,e2);
    }
}
