#include "potential.h"
#include <cmath>
#include <algorithm>
#include <cassert>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_legendre.h>
#include <gsl/gsl_sf_gegenbauer.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_hyperg.h>
#include <gsl/gsl_sf_bessel.h>
#include <gsl/gsl_sf_trig.h>
#include <gsl/gsl_sf_ellint.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_multifit.h>
#include "utils.h"
#include "massmodel.h"

namespace smile{

/** default parameters for all possible potentials **/
const CConfigPotential defaultConfigPotential = {
    1.0, ///< mass
    1.0, ///< scalerad
    1.0, ///< scalerad2
    1.0, ///< q (y/z)
    1.0, ///< p (z/x)
    1.0, ///< Gamma
    4.0, ///< sersicIndex
    20,  ///< Ncoefs_radial
    6,   ///< Ncoefs_angular
    20,  ///< Ncoefs_vertical
    0.0, ///< Alpha
    1.0, ///< Rmax
    -2.0,///< treecodeEps
    0.5, ///< treecodeTheta
    CDensity::PT_UNKNOWN,  ///< PotentialType
    CDensity::PT_PLUMMER,  ///< DensityType
    CDensity::ST_TRIAXIAL, ///< SymmetryType
    1.0, ///< splineSmoothFactor
    0.0, ///< splineRMin
    0.0, ///< splineRMax
    0.0, ///< splineZMin
    0.0, ///< splineZMax
    "",  ///< NbodyFile
    0.0, ///< Mbh
    0.0, ///< binary_q
    0.0, ///< binary_sma
    0.0, ///< binary_ecc
    0.0  ///< binary_phase
};

//----------------------------------------------------------------------------//

/// \cond INTERNAL_DOCS
// helper functions for integrating density over volume
struct CDensityParam{
    const CDensity* P;
    double r, rcostheta, rsintheta, mumble;
};
double getDensityPhi(double phi, void* params)
{
    double X=((CDensityParam*)params)->rsintheta*cos(phi);
    double Y=((CDensityParam*)params)->rsintheta*sin(phi);
    double Z=((CDensityParam*)params)->rcostheta;
    double val=((CDensityParam*)params)->P->Rho(X, Y, Z);
    if((((CDensityParam*)params)->P->symmetry() & CDensity::ST_PLANESYM)==CDensity::ST_PLANESYM)
        return val;
    // otherwise need to add contributions from other octants
    val +=((CDensityParam*)params)->P->Rho(X, Y,-Z)
        + ((CDensityParam*)params)->P->Rho(X,-Y, Z)
        + ((CDensityParam*)params)->P->Rho(X,-Y,-Z)
        + ((CDensityParam*)params)->P->Rho(-X, Y, Z)
        + ((CDensityParam*)params)->P->Rho(-X, Y,-Z)
        + ((CDensityParam*)params)->P->Rho(-X,-Y, Z)
        + ((CDensityParam*)params)->P->Rho(-X,-Y,-Z);
    return val/8;
}
double getDensityTheta(double costheta, void* params)
{
    gsl_function F;
    F.function=&getDensityPhi;
    F.params=params;
    ((CDensityParam*)params)->rcostheta=((CDensityParam*)params)->r*costheta;
    ((CDensityParam*)params)->rsintheta=((CDensityParam*)params)->r*sqrt(1-pow_2(costheta));
    double result, error;
    size_t neval;
    if((((CDensityParam*)params)->P->symmetry() & CDensity::ST_AXISYMMETRIC)==CDensity::ST_AXISYMMETRIC)
        result=M_PI_2*((CDensityParam*)params)->P->Rho(
            ((CDensityParam*)params)->rsintheta, 0, ((CDensityParam*)params)->rcostheta);
    else
        gsl_integration_qng(&F, 0, M_PI_2, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result;
}
double getDensityR(double rscaled, void* params)
{
    gsl_function F;
    F.function=&getDensityTheta;
    F.params=params;
    double r=rscaled/(1-rscaled);
    ((CDensityParam*)params)->r=r;
    double result, error;
    size_t neval;
    if((((CDensityParam*)params)->P->symmetry() & CDensity::ST_SPHERICAL)==CDensity::ST_SPHERICAL)
        result=M_PI_2*((CDensityParam*)params)->P->Rho(r, 0, 0);
    else
        gsl_integration_qng(&F, 0, 1, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result*8*r*r/pow_2(1-rscaled);
}
// another implementation based on cylindrical, not spherical coordinates
double getDensityZ(double Zscaled, void* params)
{
    gsl_function F;
    F.function=&getDensityPhi;
    F.params=params;
    ((CDensityParam*)params)->rcostheta=Zscaled/(1-Zscaled);
    double result, error;
    size_t neval;
    if((((CDensityParam*)params)->P->symmetry() & CDensity::ST_AXISYMMETRIC)==CDensity::ST_AXISYMMETRIC)
        result=M_PI_2*((CDensityParam*)params)->P->Rho(
            ((CDensityParam*)params)->rsintheta, 0, ((CDensityParam*)params)->rcostheta);
    else
        gsl_integration_qng(&F, 0, M_PI_2, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result/pow_2(1-Zscaled);
}    
double getDensityRcyl(double Rscaled, void* params)
{
    gsl_function F;
    F.function=&getDensityZ;
    F.params=params;
    double R=Rscaled/(1-Rscaled);
    ((CDensityParam*)params)->rsintheta=R;
    // naming hell: "mumble" is the maximum value of z beyond which density is assumed to be zero,
    // and "r" refers to upper limit of integration in spherical radius
    double maxZ = std::min<double>(((CDensityParam*)params)->mumble, sqrt(pow_2(((CDensityParam*)params)->r)-R*R));
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, maxZ/(1+maxZ), 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result*8*R/pow_2(1-Rscaled);
}
/// \endcond

// base class for density-only model, without corresponding potential: CDensity

double CDensity::Mass(const double r) const
{
    if(r<=0) return 0;
    // by default, integrate over density inside given radius; may be replaced by cheaper and more approximate evaluation for child classes
    gsl_function F;
    CDensityParam params;
    F.function=&getDensityR;
    F.params=&params;
    params.P=this;
    double result, error;
    /*gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    gsl_integration_qag(&F, 0, r/(1+r), 0, EPSREL_DENSITY_INT, 1000, GSL_INTEG_GAUSS51, w, &result, &error);
    gsl_integration_workspace_free(w);*/
    size_t neval;
    gsl_integration_qng(&F, 0, r/(1+r), 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result;
}

double CDensity::totalMass() const
{
    double rad=32;
    double mass1, mass2=Mass(rad), mass3=Mass(rad*2), massest=0, massestprev;
    int num_neg=0;
    const int max_num_neg=4;
    do{
        rad*=2;
        mass1=mass2; mass2=mass3; mass3=Mass(rad*2);
        if(mass2==mass3) return mass3;  // mass doesn't seem to grow with raduis anymore
        massestprev=massest>0 ? massest : mass3;
        massest=(mass2*mass2-mass1*mass3)/(2*mass2-mass1-mass3);
        if(gsl_isnan(massest) || massest<0) num_neg++;  // increase counter of 'bad' attempts (negative means that mass is growing at least logarithmically with radius)
    } while(rad<MAX_RADIUS_CUTOFF && mass2!=mass3 && 
            (massest<0 || fabs(massestprev-massest)/massest>EPSREL_DENSITY_INT) && num_neg<max_num_neg);
    if(fabs((massestprev-massest)/massest)>EPSREL_DENSITY_INT) return -1;   // total mass seems to be infinite
    else return massest;
}

/// \cond INTERNAL_DOCS
double findRadiusByMass(double rscaled, void* params)
{
    double r=std::min<double>(rscaled/(1-rscaled), MAX_RADIUS_CUTOFF);   // map [0:1] to [0:infinity)
    return (r<MAX_RADIUS_CUTOFF ? ((CDensityParam*)params)->P->Mass(r) : ((CDensityParam*)params)->mumble)
        - ((CDensityParam*)params)->r;  // params->r holds mass fraction, params->mumble is the totalMass
}
/// \endcond
double CDensity::getRadiusByMass(const double m) const
{
    double totMass=totalMass();
    if(totMass<0 || m>=totMass) return -1;
    gsl_function F;
    CDensityParam PP;
    PP.r=m;
    PP.mumble=totMass;
    PP.P=this;
    F.function=&findRadiusByMass;
    F.params=&PP;
    double rscaled=findRoot(&F, 0, 1, -1);
    return rscaled/(1-rscaled);
}

void CDensity::getRadiiByMass(const vectord &masses, vectord *radii) const
{
    assert(radii!=NULL);
    radii->assign(masses.size(), -1);
    double totMass=totalMass();
    if(totMass<0) return;
    gsl_function F;
    CDensityParam PP;
    PP.P=this;
    PP.mumble=totMass;
    F.function=&findRadiusByMass;
    F.params=&PP;
    double rscaled_low=0;
    for(size_t i=0; i<masses.size(); i++)
    {
        if(i>0 && masses[i]<masses[i-1]) rscaled_low=0;  // masses should be increasing in the input array, but if not, reset the lower search bound to 0
        PP.r=masses[i];
        double root=findRoot(&F, rscaled_low, 1, -1);
        if(root<0) return;
        radii->at(i)=root/(1-root);
        rscaled_low=root;
    }
}

bool CDensity::checkMassMonotonic() const
{
    double massprev=Mass(0);
    if(gsl_isnan(massprev) || massprev!=0) return false;  // mass should be zero at origin (not including Mbh)
    for(double rad=0.01; rad<std::min(MAX_RADIUS_CUTOFF,10000.0); rad*=1.25)
    {
        double mass=Mass(rad);
        if(mass<massprev*(1-1e-10)) return false;
        massprev=mass;
    }
    return true;  // no apparent evidence for weird behaviour..
}

bool CDensity::checkDensityNonzero() const
{
    for(double rad=1e-4; rad<std::min(MAX_RADIUS_CUTOFF,10000.0); rad*=1.5)
    {
        if(Rho(rad, 0, 0) <=0 || Rho(0, rad, 0) <=0 || Rho(0, 0, rad) <=0)
            return false;
    }
    return true;  // no apparent evidence for weird behaviour..
}

double CDensity::getGamma() const
{
    double mass1, mass2, mass3;
    double rad=1./1024;
    do{
        mass2=Mass(rad);
        if(mass2<=0) rad*=2;
    }while(rad<1 && mass2==0);
    mass3=Mass(rad*2);
    if(mass2<=0 || mass3<=0) return -1; // apparent error
    double alpha1, alpha2=log(mass3/mass2)/log(2.), gamma1=-1, gamma2=3-alpha2;
    do{
        rad/=2;
        mass1=Mass(rad);
        if(mass1<=0) return gamma2;
        alpha1=log(mass2/mass1)/log(2.);
        gamma2=gamma1<0 ? 3-alpha1 : gamma1;  // rough estimate
        gamma1=3- (2*alpha1-alpha2); // extrapolated estimate
        alpha2=alpha1;
        mass3=mass2; mass2=mass1;
    } while(fabs(gamma1-gamma2)>0.001);
    if(fabs(gamma1)<1e-3) gamma1=0;
    return gamma1;
}
//----------------------------------------------------------------------------//

/// \cond INTERNAL_DOCS
// helper functions for long axis radius/period evaluation
struct CPotentialParam{
    const CPotential* P;
    double E, X, Y, Z, R, Omega;
    int m;
};

double k_intersect(double k, void* params)
{
    double E= ((CPotentialParam*)params)->E;
    double X= ((CPotentialParam*)params)->X;
    double Y= ((CPotentialParam*)params)->Y;
    double Z= ((CPotentialParam*)params)->Z;
    return E - ((CPotentialParam*)params)->P->Phi(X*k,Y*k,Z*k) + 0.5*pow_2(((CPotentialParam*)params)->Omega*k)*(X*X+Y*Y);
}
/// \endcond

double longAxisRadius(const CPotential* potential, double E, double Omega)
{
    return findIntersection(potential, E, 1, 0, 0, Omega);
}

double findIntersection(const CPotential* potential, double E, double X, double Y, double Z, double Omega)
{
    gsl_function F;
    CPotentialParam PP;
    PP.E=E;
    PP.P=potential;
    PP.X=X;
    PP.Y=Y;
    PP.Z=Z;
    PP.Omega=Omega;
    F.function=&k_intersect;
    F.params=&PP;
    // first bracket the root from below and above
    double klow=1, kupp=1, Elow, Eupp, Eupprev;
    do {
        klow/=2;
        Elow=potential->Phi(klow*X,klow*Y,klow*Z) - 0.5*pow_2(Omega*klow)*(X*X+Y*Y);
    } while ((Elow>E) && (klow>MIN_RADIUS_CUTOFF));
    if(klow<MIN_RADIUS_CUTOFF) 
        return -1;  // signal of error (E lower than minimum of the potential)
    Eupprev=Eupp=Elow;
    kupp=klow;
    do {
        kupp*=2;
        Eupp=potential->Phi(kupp*X,kupp*Y,kupp*Z) - 0.5*pow_2(Omega*kupp)*(X*X+Y*Y);
        if(Omega!=0 && Eupp<Eupprev)
        {   // special case: a rotating frame and we appear to catch the maximim of effective potential
            double kmax=findMinGuess(&F, kupp/4, kupp);  // lower limit is two steps back
            Eupp=potential->Phi(kmax*X,kmax*Y,kmax*Z) - 0.5*pow_2(Omega*kmax)*(X*X+Y*Y);
            if(kmax==kupp/2 || E-Eupp>0)
                return -1;   // energy is above the maximum of effective potential, or the maximum cannot be found
            else 
                kupp=kmax;   // will find the intersection inside corotation
        }
        Eupprev=Eupp;  // store value at previous iteration
    } while ((Eupp<E) && (kupp<MAX_RADIUS_CUTOFF));
    return findRoot(&F, klow, kupp);
}

void findLagrangianPoints(const CPotential* potential, double Omega, double* L1X, double *L4Y)
{
    assert(L1X!=NULL && L4Y!=NULL);
    *L1X=gsl_posinf(); *L4Y=gsl_posinf(); 
    if(Omega==0) return; 
    gsl_function F;
    CPotentialParam PP;
    PP.E=0;
    PP.P=potential;
    PP.Z=0;
    PP.Omega=Omega;
    F.function=&k_intersect;
    F.params=&PP;
    double Xlow, Xupp, Phi2, PhiL;
    bool found;
    for(int XorY=0; XorY<2; XorY++) {
        Xlow=1;
        found=false;
        PP.X=1-XorY;
        PP.Y=XorY;
        PhiL=potential->Phi(PP.X*Xlow,PP.Y*Xlow,0) - 0.5*pow_2(Omega*Xlow);
        do{
            Phi2=potential->Phi(PP.X*Xlow/2,PP.Y*Xlow/2,0) - 0.5*pow_2(Omega*Xlow/2);
            if(Phi2>=PhiL) { Xlow/=2; PhiL=Phi2; } 
            else found=true;
        }
        while(!found && Xlow>MIN_RADIUS_CUTOFF);
        found=false;
        if(Xlow<MIN_RADIUS_CUTOFF) return;  // signal of error
        Xupp=Xlow*2;
        do{
            Phi2=potential->Phi(PP.X*Xupp,PP.Y*Xupp,0) - 0.5*pow_2(Omega*Xupp);
            if(Phi2>=PhiL) Xupp*=2; 
            else found=true;
        }
        while(!found && Xupp<MAX_RADIUS_CUTOFF);
        if(Xupp>MAX_RADIUS_CUTOFF) return; 
        if(XorY==0) *L1X = findMinGuess(&F, Xlow, Xupp);
        else *L4Y = findMinGuess(&F, Xlow, Xupp);
    }
}

void kepler_orbit(double mean_anomaly, double ecc, double *x, double *y)
{
    double sinE=0,cosE=1;
    if(ecc>0) {
#if 0
        double num_revol=floor(mean_anomaly/(2*M_PI));
        mean_anomaly -= 2*M_PI*num_revol;
#else   // a more precise algo from GSL
        const double P1 = 4 * 7.85398125648498535156e-01;
        const double P2 = 4 * 3.77489470793079817668e-08;
        const double P3 = 4 * 2.69515142907905952645e-15;
        const double TwoPi = 2*(P1 + P2 + P3);
        const double w = 2*floor(mean_anomaly/TwoPi);
        mean_anomaly = ((mean_anomaly - w*P1) - w*P2) - w*P3;
#endif
        bool signm=mean_anomaly>M_PI;
        double E=M_PI_2+(M_PI/8)/ecc*(
            sqrt(M_PI*M_PI-8*M_PI*ecc+16*ecc*(ecc+(signm?2*M_PI-mean_anomaly:mean_anomaly)))-M_PI);
        if(signm) E=2*M_PI-E;
        double deltaE=0;
        int niter=0;
        do {  // Newton's method for solving the Kepler equation (f=0)
            sinE=sin(E);
            cosE=cos(E);
            double f=E-ecc*sinE-mean_anomaly;
            double fp=1-ecc*cosE;
            deltaE=-f/fp;
            deltaE=-f/(fp+0.5*deltaE*ecc*sinE);  // refinement using second derivative (thanks to A.Gurkan)
            E+=deltaE;
            niter++;
        } while(fabs(deltaE)>1e-12 && niter<42);
        if(niter==42) my_message(FUNCNAME, "Max #iterations exceeded at m="+convertToString(mean_anomaly)+", ecc="+convertToString(ecc));
    } else {
        sinE=sin(mean_anomaly);
        cosE=cos(mean_anomaly);
    }
    *x=cosE-ecc;
    *y=sqrt(1-ecc*ecc)*sinE;
}

// class for multicomponent potential/density model: CPotentialComposite
CPotentialComposite::CPotentialComposite(const std::vector<const CPotential*> &_components, double _Mbh, 
                                         double _q, double _sma, double _ecc, double _phase) :
  CPotential(), Mbh(_Mbh), q(_q), sma(_sma), ecc(_ecc), phase(_phase)
{
    components.resize(_components.size());
    for(size_t index=0; index<_components.size(); index++) {
        if(_components[index]==NULL) {
            my_error(FUNCNAME, "Empty pointer provided as component "+convertToString(index));
            components.resize(index);
            return;
        }
        components[index]=_components[index]->clone();
    }
};

CPotentialComposite::~CPotentialComposite()
{
    for(size_t index=0; index<components.size(); index++)
        delete components[index];
};

CPotential::SYMMETRYTYPE CPotentialComposite::symmetry() const
{
    int sym=static_cast<int>(ST_SPHERICAL);
    for(size_t index=0; index<components.size(); index++)
        sym &= static_cast<int>(components[index]->symmetry());
    return static_cast<SYMMETRYTYPE>(sym);
};

double CPotentialComposite::Mass(const double r) const
{
    double result=Mbh;
    for(size_t index=0; index<components.size(); index++)
        result += components[index]->Mass(r);
    return result;
};

double CPotentialComposite::Rho(double X, double Y, double Z, double t) const
{
    double result=0;
    for(size_t index=0; index<components.size(); index++)
        result += components[index]->Rho(X, Y, Z, t);
    return result;
};

double CPotentialComposite::Phi(double X, double Y, double Z, double t) const
{
    double result=0;
    for(size_t index=0; index<components.size(); index++)
        result += components[index]->Phi(X, Y, Z, t);
    if(Mbh!=0) {
        if(sma>0) {
            double omegabh=sma>0?sqrt(Mbh/(sma*sma*sma)):0;
            double bhX=0, bhY=0;
            kepler_orbit(omegabh*t+phase, ecc, &bhX, &bhY);
            bhX*=sma/(1+q);
            bhY*=sma/(1+q);
            result += -Mbh/(1+q)/sqrt(pow_2(X-bhX*q) + pow_2(Y-bhY*q) + pow_2(Z) + pow_2(sma*1e-3))
                      -Mbh/(1+q)*q/sqrt(pow_2(X+bhX) + pow_2(Y+bhY)   + pow_2(Z) + pow_2(sma*1e-3)); //!!! softening
        } else 
            result += -Mbh/sqrt(pow_2(X) + pow_2(Y) + pow_2(Z));
    }
    return result;
};

void CPotentialComposite::Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv) const
{
    force[0]=force[1]=force[2]=0;
    if(forceDeriv) {
        forceDeriv[0]=forceDeriv[1]=forceDeriv[2]=forceDeriv[3]=forceDeriv[4]=forceDeriv[5]=0;
    }
    if(Mbh!=0) {
        if(sma==0) {
            double X=xyz[0];
            double Y=xyz[1];
            double Z=xyz[2];
            double r2=pow_2(X) + pow_2(Y) + pow_2(Z) + 1e-100;
            double Mrminus3= Mbh/pow(r2, 1.5);
            force[0] = -X*Mrminus3;
            force[1] = -Y*Mrminus3;
            force[2] = -Z*Mrminus3;
            if(forceDeriv!=NULL)
            {
                forceDeriv[0] = Mrminus3*(3*X*X/r2-1);
                forceDeriv[1] = Mrminus3*(3*Y*Y/r2-1);
                forceDeriv[2] = Mrminus3*(3*Z*Z/r2-1);
                forceDeriv[3] = Mrminus3*3*X*Y/r2;
                forceDeriv[4] = Mrminus3*3*Y*Z/r2;
                forceDeriv[5] = Mrminus3*3*Z*X/r2;
            }
        } else {
            double omegabh=sqrt(Mbh/(sma*sma*sma));
            double bhX=0, bhY=0;
            kepler_orbit(omegabh*t+phase, ecc, &bhX, &bhY);
            bhX*=sma/(1+q);
            bhY*=sma/(1+q);
            double X=xyz[0]-bhX*q;
            double Y=xyz[1]-bhY*q;
            double Z=xyz[2];
            double r2=pow_2(X) + pow_2(Y) + pow_2(Z) + pow_2(sma*1e-3); //!!! softening
            double Mrminus3= Mbh/((1+q)*r2*sqrt(r2));
            force[0] = -X*Mrminus3;
            force[1] = -Y*Mrminus3;
            force[2] = -Z*Mrminus3;
            if(forceDeriv!=NULL)
            {
                forceDeriv[0] = Mrminus3*(3*X*X/r2-1);
                forceDeriv[1] = Mrminus3*(3*Y*Y/r2-1);
                forceDeriv[2] = Mrminus3*(3*Z*Z/r2-1);
                forceDeriv[3] = Mrminus3*3*X*Y/r2;
                forceDeriv[4] = Mrminus3*3*Y*Z/r2;
                forceDeriv[5] = Mrminus3*3*Z*X/r2;
            }
            X=xyz[0]+bhX;
            Y=xyz[1]+bhY;
            Z=xyz[2];
            r2=pow_2(X) + pow_2(Y) + pow_2(Z) + pow_2(sma*1e-3); //!!! softening
            Mrminus3= Mbh*q/((1+q)*r2*sqrt(r2));
            force[0]+= -X*Mrminus3;
            force[1]+= -Y*Mrminus3;
            force[2]+= -Z*Mrminus3;
            if(forceDeriv!=NULL)
            {
                forceDeriv[0] += Mrminus3*(3*X*X/r2-1);
                forceDeriv[1] += Mrminus3*(3*Y*Y/r2-1);
                forceDeriv[2] += Mrminus3*(3*Z*Z/r2-1);
                forceDeriv[3] += Mrminus3*3*X*Y/r2;
                forceDeriv[4] += Mrminus3*3*Y*Z/r2;
                forceDeriv[5] += Mrminus3*3*Z*X/r2;
            }
        }
    }
    double tmpForce[N_DIM], tmpForceDeriv[N_DIM*(N_DIM+1)/2];
    for(size_t index=0; index<components.size(); index++) {
        components[index]->Force(xyz, t, tmpForce, forceDeriv!=NULL ? tmpForceDeriv : NULL);
        for(size_t d=0; d<N_DIM; d++) force[d]+=tmpForce[d];
        if(forceDeriv!=NULL)
            for(size_t d=0; d<N_DIM*(N_DIM+1)/2; d++) forceDeriv[d]+=tmpForceDeriv[d];
    }
};

//----------------------------------------------------------------------------//
// Logarithmic potential

double CPotentialLog::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m2=pow_2(Rc) + pow_2(X) + pow_2(Y)/pow_2(q) + pow_2(Z)/pow_2(p);
    return halfsigma2/M_PI * (0.5*(1+1/pow_2(q)+1/pow_2(p))/m2 - (pow_2(X) + pow_2(Y/pow_2(q)) + pow_2(Z/pow_2(p)))/pow_2(m2));
}

double CPotentialLog::Phi(double X, double Y, double Z, double /*t*/) const
{
    return halfsigma2*log(pow_2(Rc) + pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
}

void CPotentialLog::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2];
    double m2=(pow_2(X) + pow_2(Y)/pow_2(q) + pow_2(Z)/pow_2(p) + pow_2(Rc));
    force[0] = halfsigma2*-2*X/m2;
    force[1] = halfsigma2*-2*Y/m2/pow_2(q);
    force[2] = halfsigma2*-2*Z/m2/pow_2(p);
    if(forceDeriv!=NULL)
    {   // variational equation
        forceDeriv[0] = halfsigma2*(-2/m2         +4*pow_2(X/m2));
        forceDeriv[1] = halfsigma2*(-2/m2/pow_2(q)+4*pow_2(Y/(m2*q*q)));
        forceDeriv[2] = halfsigma2*(-2/m2/pow_2(p)+4*pow_2(Z/(m2*p*p)));
        forceDeriv[3] = halfsigma2*X*Y * 4/pow_2(q*m2);
        forceDeriv[4] = halfsigma2*Y*Z * 4/pow_2(q*p*m2);
        forceDeriv[5] = halfsigma2*Z*X * 4/pow_2(p*m2);
    }
}

//----------------------------------------------------------------------------//
// simple harmonic oscillator potential

double CPotentialHarmonic::Rho(double /*X*/, double /*Y*/, double /*Z*/, double /*t*/) const
{
    return mass*(1+1/q/q+1/p/p);
}

double CPotentialHarmonic::Phi(double X, double Y, double Z, double /*t*/) const
{
    return mass/2*(X*X + Y*Y/q/q + Z*Z/p/p);
}

void CPotentialHarmonic::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    force[0] = -v[0]*mass;
    force[1] = -v[1]*mass/q/q;
    force[2] = -v[2]*mass/p/p;
    if(forceDeriv!=NULL)
    {
        forceDeriv[0] = -mass;
        forceDeriv[1] = -mass/pow_2(q);
        forceDeriv[2] = -mass/pow_2(p);
    }
}

//----------------------------------------------------------------------------//
// arbitrary spherical potential

CPotentialSpherical::CPotentialSpherical(const CMassModel& _model) :
  CPotential(), model(new CMassModel(_model)) {};

CPotentialSpherical::CPotentialSpherical(const vectord& radii, const vectord& mass) :
  CPotential(), model(new CMassModel(radii, mass)) {};

template<typename NumT>
CPotentialSpherical::CPotentialSpherical(const CPointMassSet<NumT> &points, double smoothfactor, const vectord *_gridradii) :
  CPotential()
{
    vectord gridradii, gridmass;
    if(_gridradii!=NULL) 
        gridradii.assign(_gridradii->begin(), _gridradii->end());
    size_t nbody=points.size();
    std::vector<NumT> nbodyrad(nbody), nbodymass(nbody);
    for(size_t i=0; i<nbody; i++)
    {
        nbodyrad[i]=sqrt(pow_2(points[i].first.Pos[0]) + pow_2(points[i].first.Pos[1]) + pow_2(points[i].first.Pos[2]) );
        nbodymass[i]=points[i].second;
    }
    if(!fitMassModel(nbodyrad, nbodymass, &gridradii, &gridmass, smoothfactor) || gridradii.size()==0) {
        gridradii.clear();  // model creation will fail but at least an instance will be initialized
        gridmass.clear();
    }
    model=new CMassModel(gridradii, gridmass);
}

template
CPotentialSpherical::CPotentialSpherical(const CPointMassSet<double> &points, double smoothfactor, const vectord *_gridradii);
template
CPotentialSpherical::CPotentialSpherical(const CPointMassSet<float> &points, double smoothfactor, const vectord *_gridradii);

CPotentialSpherical::~CPotentialSpherical()
{
    delete model;
}

CPotential* CPotentialSpherical::clone() const
{ 
    CPotentialSpherical* newpot=new CPotentialSpherical(*this); 
    newpot->model=new CMassModel(*model);
    return newpot;
}

double CPotentialSpherical::Rho(double X, double Y, double Z, double /*t*/) const
{
    return model->dens(sqrt(X*X+Y*Y+Z*Z));
}

double CPotentialSpherical::Phi(double X, double Y, double Z, double /*t*/) const
{
    return model->pot(sqrt(X*X+Y*Y+Z*Z));
}

double CPotentialSpherical::Mass(const double r) const
{
    return model->mass(r);
}

void CPotentialSpherical::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2];
    double r=sqrt(X*X+Y*Y+Z*Z);
    double M=model->mass(r);
    double Mrminus3=M/(r*r*r);
    force[0] = -X*Mrminus3;
    force[1] = -Y*Mrminus3;
    force[2] = -Z*Mrminus3;
    if(forceDeriv!=NULL)
    {
        double Mderrminus4=model->dens(r)/(4*M_PI*pow(r,6)), r2=r*r;
        forceDeriv[0] = Mrminus3*(3*X*X/r2-1)-Mderrminus4*X*X;
        forceDeriv[1] = Mrminus3*(3*Y*Y/r2-1)-Mderrminus4*Y*Y;
        forceDeriv[2] = Mrminus3*(3*Z*Z/r2-1)-Mderrminus4*Z*Z;
        forceDeriv[3] = Mrminus3*3*X*Y/r2-Mderrminus4*X*Y;
        forceDeriv[4] = Mrminus3*3*Y*Z/r2-Mderrminus4*Y*Z;
        forceDeriv[5] = Mrminus3*3*Z*X/r2-Mderrminus4*Z*X;
    }
}

//----------------------------------------------------------------------------//
// Miyamoto-Nagai potential

double CPotentialMiyamotoNagai::Rho(double X, double Y, double Z, double /*t*/) const
{
    double zb=sqrt(Z*Z+B*B);
    double azb2=pow_2(A+zb);
    double R2=X*X+Y*Y;
    return mass * B*B * (A*R2 + (A+3*zb)*azb2) / (4*M_PI * pow(R2+azb2, 2.5) * zb*zb*zb);
}

double CPotentialMiyamotoNagai::Phi(double X, double Y, double Z, double /*t*/) const
{
    return - mass/sqrt(X*X + Y*Y + pow_2(A+sqrt(Z*Z+B*B)));
}

void CPotentialMiyamotoNagai::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2];
    double zb=sqrt(Z*Z+B*B);
    double azb2=pow_2(A+zb);
    double R2=X*X+Y*Y;
    double denom3=mass*pow(R2+azb2, -1.5);

    force[0] = -X*denom3;
    force[1] = -Y*denom3;
    force[2] = -Z*denom3*(A/zb+1);
    if(forceDeriv!=NULL)
    {
        double denom5=mass*pow(R2+azb2, -2.5);
        double azb=sqrt(azb2);
        forceDeriv[0] = 3*X*X*denom5-denom3;
        forceDeriv[1] = 3*Y*Y*denom5-denom3;
        forceDeriv[2] = ((2*azb2-R2)*Z*Z*zb*denom5 - B*B*azb*denom3)/(zb*zb*zb);
        forceDeriv[3] = 3*X*Y*denom5;
        forceDeriv[4] = Y*Z * 3*azb*denom5/zb;
        forceDeriv[5] = Z*X * 3*azb*denom5/zb;
    }
}

//----------------------------------------------------------------------------//
// Ferrers n=2 potential

CPotentialFerrers::CPotentialFerrers(double _mass, double _R, double _q, double _p): 
CPotential(), a(_R), b(_R*_q), c(_R*_p), mass(_mass),  //n(_n), 
//rho0( mass*gsl_sf_gamma(n+2.5)/pow(M_PI, 1.5)/gsl_sf_gamma(n+1)/q/p/pow(R, 3) )
rho0( mass*105./(32*M_PI*a*b*c) )
{
    if(a<=0) my_error(FUNCNAME, "Error, scale radius should be positive!");
    computeW(0, W0);
};

double CPotentialFerrers::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m2 = pow_2(X/a) + pow_2(Y/b) + pow_2(Z/c);
    return m2>1 ? 0 : rho0*pow_2(1-m2);
}

// helper routine to find lambda: param=(x^2,y^2,z^2,a^2,b^2,c^2), 
// x^2/(lambda+a^2) + y^2/(lambda+b^2) + z^2/(lambda+c^2) = 1

double findLambda(double lambda, void* param)
{
    double* p=(double*)param;
    return p[0]/(lambda+p[3]) + p[1]/(lambda+p[4]) + p[2]/(lambda+p[5]) - 1;
}

double CPotentialFerrers::Phi(double X, double Y, double Z, double /*t*/) const
{
    double m2 = pow_2(X/a) + pow_2(Y/b) + pow_2(Z/c);
    double r2 = X*X+Y*Y+Z*Z;
    if(r2>a*a*4*4) {   // use spherical-harmonic expansion up to l=2 beyond 4 scale radii
        double Moverr=rho0*32*M_PI*a*b*c/105/sqrt(r2);
        double val=1 + ( (2*c*c-a*a-b*b)/36 * (2*Z*Z-X*X-Y*Y) + (b*b-a*a)/12 * (Y*Y-X*X) )/(r2*r2);
        return -Moverr*val;
    }
    double Wcurr[20];  // temp.coefs for lambda>0 if needed
    const double *W;   // coefs used in computation (either pre-computed or temp.)
    if(m2>1) {
        gsl_function F;
        double param[6] = { X*X, Y*Y, Z*Z, a*a, b*b, c*c };
        F.function=&findLambda;
        F.params=param;
        double lmax=1; 
        while(findLambda(lmax, param)>0) lmax*=2;
        double lambda=findRoot(&F, 0, lmax);
        computeW(lambda, Wcurr);
        W=Wcurr;
    } else 
        W=W0;  // use pre-computed coefs inside the model
    X*=X; Y*=Y; Z*=Z;
    return -rho0*M_PI/3*a*b*c* ( W[0] - 6*W[10]*X*Y*Z
        + X*( X*(3*W[7]-X*W[17]) + 3*Y*(2*W[4]-Y*W[11]-X*W[14]) - 3*W[1] )
        + Y*( Y*(3*W[8]-Y*W[18]) + 3*Z*(2*W[5]-Z*W[12]-Y*W[15]) - 3*W[2] )
        + Z*( Z*(3*W[9]-Z*W[19]) + 3*X*(2*W[6]-X*W[13]-Z*W[16]) - 3*W[3] ) );
}

void CPotentialFerrers::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const 
{
    double X2=v[0]*v[0], Y2=v[1]*v[1], Z2=v[2]*v[2];  // note that these are squared coords
    double m2 = X2/(a*a) + Y2/(b*b) + Z2/(c*c);
    double r2 = X2+Y2+Z2;
    if(r2>a*a*4*4) {   // use spherical-harmonic expansion up to l=2 beyond 4 scale radii, to speed up (and increase the accuracy)
        double Moverr3=rho0*32*M_PI*a*b*c/(105*r2*sqrt(r2));
        double C20=(2*c*c-a*a-b*b)/36, C22=(b*b-a*a)/12;
        double add=5*(C20 * (2*Z2-X2-Y2) + C22 * (Y2-X2) )/(r2*r2);
        force[0] = -v[0]*Moverr3*(1+add+2*(C20+C22)/r2);
        force[1] = -v[1]*Moverr3*(1+add+2*(C20-C22)/r2);
        force[2] = -v[2]*Moverr3*(1+add-4*C20/r2);
        if(forceDeriv!=NULL) {
            forceDeriv[0] = -Moverr3*( 1 - 3*X2/r2 + add*(1-7*X2/r2) + 2*(C20+C22)/r2*(1-10*X2/r2) );
            forceDeriv[1] = -Moverr3*( 1 - 3*Y2/r2 + add*(1-7*Y2/r2) + 2*(C20-C22)/r2*(1-10*Y2/r2) );
            forceDeriv[2] = -Moverr3*( 1 - 3*Z2/r2 + add*(1-7*Z2/r2) - 4*C20/r2*(1-10*Z2/r2) );
            forceDeriv[3] = -Moverr3*v[0]*v[1]/r2*( - 3 - add*7 - 20*C20/r2 );
            forceDeriv[4] = -Moverr3*v[1]*v[2]/r2*( - 3 - add*7 + 10*(C20+C22)/r2 );
            forceDeriv[5] = -Moverr3*v[2]*v[0]/r2*( - 3 - add*7 + 10*(C20-C22)/r2 );
        }
        return;
    }
    double Wcurr[20];  // temp.coefs for lambda>0 if needed
    const double *W;   // coefs used in computation (either pre-computed or temp.)
    if(m2>1) {
        gsl_function F;
        double param[6] = { X2, Y2, Z2, a*a, b*b, c*c };
        F.function=&findLambda;
        F.params=param;
        double lmax=1; 
        while(findLambda(lmax, param)>0) lmax*=2;
        double lambda=findRoot(&F, 0, lmax);
        computeW(lambda, Wcurr);
        W=Wcurr;
    } else 
        W=W0;  // use pre-computed coefs inside the model
    double mult=-2*M_PI*rho0*a*b*c;
    force[0] = v[0]*mult*(W[1] +
        X2*(X2*W[17]+2*Y2*W[14]-2*W[7]) +
        Y2*(Y2*W[11]+2*Z2*W[10]-2*W[4]) +
        Z2*(Z2*W[16]+2*X2*W[13]-2*W[6]) );
    force[1] = v[1]*mult*(W[2] +
        X2*(X2*W[14]+2*Y2*W[11]-2*W[4]) +
        Y2*(Y2*W[18]+2*Z2*W[15]-2*W[8]) +
        Z2*(Z2*W[12]+2*X2*W[10]-2*W[5]) );
    force[2] = v[2]*mult*(W[3] +
        X2*(X2*W[13]+2*Y2*W[10]-2*W[6]) +
        Y2*(Y2*W[15]+2*Z2*W[12]-2*W[5]) +
        Z2*(Z2*W[19]+2*X2*W[16]-2*W[9]) );
    if(forceDeriv!=NULL) {
        forceDeriv[0] = mult*(W[1] + 
            X2*(5*X2*W[17]+6*Y2*W[14]-6*W[7]) +
            Y2*(  Y2*W[11]+2*Z2*W[10]-2*W[4]) +
            Z2*(  Z2*W[16]+6*X2*W[13]-2*W[6]) );
        forceDeriv[1] = mult*(W[2] +
            X2*(  X2*W[14]+6*Y2*W[11]-2*W[4]) +
            Y2*(5*Y2*W[18]+6*Z2*W[15]-6*W[8]) +
            Z2*(  Z2*W[12]+2*X2*W[10]-2*W[5]) );
        forceDeriv[2] = mult*(W[3] +
            X2*(  X2*W[13]+2*Y2*W[10]-2*W[6]) +
            Y2*(  Y2*W[15]+6*Z2*W[12]-2*W[5]) +
            Z2*(5*Z2*W[19]+6*X2*W[16]-6*W[9]) );
        forceDeriv[3] = mult*4*v[0]*v[1]*(X2*W[14]+Y2*W[11]+Z2*W[10]-W[4]);
        forceDeriv[4] = mult*4*v[1]*v[2]*(X2*W[10]+Y2*W[15]+Z2*W[12]-W[5]);
        forceDeriv[5] = mult*4*v[2]*v[0]*(X2*W[13]+Y2*W[10]+Z2*W[16]-W[6]);
    }
}

void CPotentialFerrers::computeW(double lambda, double W[20]) const
{
    double ab2=(a*a-b*b),ab=sqrt(ab2);
    double ac2=(a*a-c*c),ac=sqrt(ac2);
    double bc2=(b*b-c*c);
    double denom=sqrt((lambda+a*a)*(lambda+b*b)*(lambda+c*c));
    double argphi=asin(sqrt(ac2/(a*a+lambda)));
    double E=gsl_sf_ellint_E(argphi, ab/ac, GSL_PREC_SINGLE);
    double F=gsl_sf_ellint_F(argphi, ab/ac, GSL_PREC_SINGLE);
    W[0] = 2*F/ac;   // W_000
    W[1] = 2*(F-E)/(ab2*ac);  // W_100
    W[3] = 2/bc2*(b*b+lambda)/denom - 2*E/bc2/ac;  // W_001
    W[2] = 2/denom - W[1] - W[3]; // W_010
    W[4] = (W[2] - W[1])/ab2;     // W_110
    W[5] = (W[3] - W[2])/bc2;     // W_011
    W[6] =-(W[1] - W[3])/ac2;     // W_101
    W[7] = (2/denom/(a*a+lambda) - W[4] - W[6])/3;  // W_200
    W[8] = (2/denom/(b*b+lambda) - W[5] - W[4])/3;  // W_020
    W[9] = (2/denom/(c*c+lambda) - W[6] - W[5])/3;  // W_002
    W[10]=-(W[4] - W[5])/ac2;     // W_111
    W[11]= (W[8] - W[4])/ab2;     // W_120
    W[12]= (W[9] - W[5])/bc2;     // W_012
    W[13]=-(W[7] - W[6])/ac2;     // W_201
    W[14]= (W[4] - W[7])/ab2;     // W_210  !! inverse sign w.r.t. Pfenniger'1984 paper, but matches his fortran code
    W[15]= (W[5] - W[8])/bc2;     // W_021  !! --"--
    W[16]=-(W[6] - W[9])/ac2;     // W_102  !! --"--
    W[17]= (2/denom/pow_2(a*a+lambda) - W[14] - W[13])/5;  // W_300
    W[18]= (2/denom/pow_2(b*b+lambda) - W[15] - W[11])/5;  // W_030
    W[19]= (2/denom/pow_2(c*c+lambda) - W[16] - W[12])/5;  // W_003
}

//----------------------------------------------------------------------------//
// Dehnen potential
CPotentialDehnen::CPotentialDehnen(double _mass, double _scalerad, double _q, double _p, double _Gamma): 
    CPotential(), mass(_mass), scalerad(_scalerad), q(_q), p(_p),
    Gamma(std::max<double>(0, std::min<double>(_Gamma, 2.0))) {}

double CPotentialDehnen::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p));
    return mass*scalerad*(3-Gamma)/(4*M_PI*p*q) * pow(m, -Gamma) * pow(scalerad+m, Gamma-4);
}

/// \cond INTERNAL_DOCS
struct CPotentialParamD{
    double X, Y, Z;
    double Gamma;
    double a2;  // squared scale radius in i-th coordinate
    double C1, C2, C3, C4, C5;  // coefficients in computation as in Merritt&Fridman 1996
};
double intDehnenPhi(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    double q= ((CPotentialParamD*)params)->C1;
    double p= ((CPotentialParamD*)params)->C2;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double m = s * sqrt( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)));
    double numerator = (Gamma==2)? (log((1+m)*s/m) - 1/(1+m) - log(s)) :
        (1 - (3-Gamma)*pow(m/(m+1), 2-Gamma) + (2-Gamma)*pow(m/(m+1), 3-Gamma))/(2-Gamma);
    return numerator / sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s)));
}
double intDehnenForce(double s, void* params)
{
    double a2=((CPotentialParamD*)params)->a2;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double s2=s*s;
    double m = s * sqrt( 
        pow_2(((CPotentialParamD*)params)->X) / (a2 + ((CPotentialParamD*)params)->C1*s2) +
        pow_2(((CPotentialParamD*)params)->Y) / (a2 + ((CPotentialParamD*)params)->C2*s2) +
        pow_2(((CPotentialParamD*)params)->Z) / (a2 + ((CPotentialParamD*)params)->C3*s2) );
    return s2 * pow(m, -Gamma) * pow(1+m, Gamma-4) / sqrt( 
        (a2 + ((CPotentialParamD*)params)->C1*s2) * 
        (a2 + ((CPotentialParamD*)params)->C2*s2) * 
        (a2 + ((CPotentialParamD*)params)->C3*s2) );
}
double intDehnenForceDer(double s, void* params)
{
    double a2=((CPotentialParamD*)params)->a2;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double s2=s*s;
    double m = s * sqrt( 
        pow_2(((CPotentialParamD*)params)->X) / (a2 + ((CPotentialParamD*)params)->C1*s2) +
        pow_2(((CPotentialParamD*)params)->Y) / (a2 + ((CPotentialParamD*)params)->C2*s2) +
        pow_2(((CPotentialParamD*)params)->Z) / (a2 + ((CPotentialParamD*)params)->C3*s2) );
    return s2 * s2 * (Gamma + 4*m) * pow(m, -Gamma-2) * pow(1+m, Gamma-5) / sqrt( 
        (a2 + ((CPotentialParamD*)params)->C1*s2) * 
        (a2 + ((CPotentialParamD*)params)->C2*s2) * 
        (a2 + ((CPotentialParamD*)params)->C3*s2) );
}
double intDehnenForceDerMixed(double s, void* params)
{
    double a2=((CPotentialParamD*)params)->a2;
    double Gamma= ((CPotentialParamD*)params)->Gamma;
    double s2=s*s;
    double m = s * sqrt( 
        pow_2(((CPotentialParamD*)params)->X) / (a2 + ((CPotentialParamD*)params)->C1*s2) +
        pow_2(((CPotentialParamD*)params)->Y) / (a2 + ((CPotentialParamD*)params)->C2*s2) +
        pow_2(((CPotentialParamD*)params)->Z) / (a2 + ((CPotentialParamD*)params)->C3*s2) );
    return s2 * s2 * (Gamma + 4*m) * pow(m, -Gamma-2) * pow(1+m, Gamma-5) /
        sqrt(a2 + ((CPotentialParamD*)params)->C5*s2) * pow(a2 + ((CPotentialParamD*)params)->C4*s2, -1.5); 
}
/// \endcond

// accurate adaptive integration of potential and forces (muuuuch slower, used only for a high-accuracy test run)
#define DEHNENPOTENTIAL_ACCURATE_QUADRATURE 0

double CPotentialDehnen::Phi(double X, double Y, double Z, double /*t*/) const
{
    if(q==1 && p==1) {  // analytical expression for spherical potential
        double r=sqrt(X*X+Y*Y+Z*Z);
        return mass/scalerad * (Gamma==2 ? log(r/(r+scalerad)) : (1-pow(r/(r+scalerad), 2-Gamma))/(Gamma-2) );
    }
    double result, error;
    CPotentialParamD PP;
    PP.X=X/scalerad; PP.Y=Y/scalerad; PP.Z=Z/scalerad;
    PP.C1=q; PP.C2=p; PP.Gamma=Gamma;
    gsl_function F;
    F.function=&intDehnenPhi;
    F.params=&PP;
    size_t neval;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    if(fabs(error/result)>1e-3) {
        gsl_integration_workspace * ws = gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &result, &error); 
        gsl_integration_workspace_free(ws);
    }
#endif
    return -result*mass/scalerad;
}

void CPotentialDehnen::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2];
    if(q==1 && p==1) {  // analytical expressions for spherical potential
        double r=sqrt(X*X+Y*Y+Z*Z);
        double val=-mass*pow(r, -Gamma)*pow(r+scalerad, Gamma-3);
        force[0] = val*X;
        force[1] = val*Y;
        force[2] = val*Z;
        if(forceDeriv!=NULL) {
            double val2=val*(scalerad*(1-Gamma)-2*r)/(r*r*(r+scalerad));
            forceDeriv[0] = val2*X*X + val*(1-pow_2(X/r));
            forceDeriv[1] = val2*Y*Y + val*(1-pow_2(Y/r));
            forceDeriv[2] = val2*Z*Z + val*(1-pow_2(Z/r));
            forceDeriv[3] = (val2-val/(r*r))*X*Y;
            forceDeriv[4] = (val2-val/(r*r))*Y*Z;
            forceDeriv[5] = (val2-val/(r*r))*Z*X;
        }
        return;
    }
    // force calculation requires integration
    double result, resultd, error; 
    size_t neval;
    gsl_function F;
    CPotentialParamD PP;
    F.params=&PP;
    F.function=&intDehnenForce;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.Gamma=Gamma;
    double Rs2=pow_2(scalerad);
    PP.a2=Rs2;
    PP.C1=0;
    PP.C2=(q*q-1)*Rs2;
    PP.C3=(p*p-1)*Rs2;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    gsl_integration_workspace * ws = NULL;
    if(fabs(error/result)>0.001) {
        if(!ws) ws=gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &result, &error);
    }
#endif
    force[0] = -(3-Gamma)*mass*X * result;
    if(forceDeriv!=NULL) {
        F.function=&intDehnenForceDer;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultd, &error, &neval);
        forceDeriv[0] = (3-Gamma)*mass * (pow_2(X/scalerad) * resultd - result);
        F.function=&intDehnenForceDerMixed;
        PP.C4=PP.C2;
        PP.C5=PP.C3;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultd, &error, &neval);
        forceDeriv[3] = (3-Gamma)*mass*X*Y/scalerad * resultd;
        F.function=&intDehnenForce;
    }

    PP.a2=Rs2*q*q;
    PP.C1=(1-q*q)*Rs2;
    PP.C2=0;
    PP.C3=(p*p-q*q)*Rs2;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    if(fabs(error/result)>0.001) {
        if(!ws) ws=gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &result, &error);
    }
#endif
    force[1] = -(3-Gamma)*mass*Y * result;
    if(forceDeriv!=NULL) {
        F.function=&intDehnenForceDer;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultd, &error, &neval);
        forceDeriv[1] = (3-Gamma)*mass * (pow_2(Y/(scalerad*q)) * resultd - result);
        F.function=&intDehnenForceDerMixed;
        PP.C4=PP.C3;
        PP.C5=PP.C1;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultd, &error, &neval);
        forceDeriv[4] = (3-Gamma)*mass*Y*Z/scalerad/q * resultd;
        F.function=&intDehnenForce;
    }

    PP.a2=Rs2*p*p;
    PP.C1=(1-p*p)*Rs2;
    PP.C2=(q*q-p*p)*Rs2;
    PP.C3=0;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    if(fabs(error/result)>0.001) {
        if(!ws) ws=gsl_integration_workspace_alloc (1000);
        gsl_integration_qag (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, 3,ws, &result, &error);
    }
#endif
    force[2] = -(3-Gamma)*mass*Z * result;
    if(forceDeriv!=NULL) {
        F.function=&intDehnenForceDer;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultd, &error, &neval);
        forceDeriv[2] = (3-Gamma)*mass * (pow_2(Z/(scalerad*p)) * resultd - result);
        F.function=&intDehnenForceDerMixed;
        PP.C4=PP.C1;
        PP.C5=PP.C2;
        gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &resultd, &error, &neval);
        forceDeriv[5] = (3-Gamma)*mass*Z*X/scalerad/p * resultd;
    }
#if DEHNENPOTENTIAL_ACCURATE_QUADRATURE
    gsl_integration_workspace_free(ws);
#endif
}

//----------------------------------------------------------------------------//
// CPotentialSH -- parent class for all potentials using angular expansion in spherical harmonics

void CPotentialSH::assignlmrange()
{
    lmax = (mysymmetry & ST_SPHSYM)    ==ST_SPHSYM     ? 0 : static_cast<int>(Ncoefs_angular);  // if spherical model, use only l=0,m=0 term
    lstep= (mysymmetry & ST_REFLECTION)==ST_REFLECTION ? 2 : 1;  // if reflection symmetry, use only even l
    mmax = (mysymmetry & ST_ZROTSYM)   ==ST_ZROTSYM    ? 0 : 1;  // if axisymmetric model, use only m=0 terms, otherwise all terms up to l (1 is the multiplying factor)
    mmin = (mysymmetry & ST_PLANESYM)  ==ST_PLANESYM   ? 0 :-1;  // if triaxial symmetry, do not use sine terms which correspond to m<0
    mstep= (mysymmetry & ST_PLANESYM)  ==ST_PLANESYM   ? 2 : 1;  // if triaxial symmetry, use only even m
}

/// \cond INTERNAL_DOCS
// helper functions for integration over angles (common for BSE and Spline potentials), used in computing coefficients from analytic density model
struct CPotentialParamSH{
    const CDensity* P;
    int sh_n, sh_l, sh_m;
    double costheta, sintheta, r, Alpha;
    double theta_max, phi_max;    // Pi/2, Pi or even 2*Pi, depending on potential symmetry
};
double intSH_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamSH*)params)->costheta;
    double sintheta= ((CPotentialParamSH*)params)->sintheta;
    double r= ((CPotentialParamSH*)params)->r;
    const CDensity* P=((CPotentialParamSH*)params)->P;
    int sh_l=((CPotentialParamSH*)params)->sh_l;
    int sh_m=((CPotentialParamSH*)params)->sh_m;
    return P->Rho(r*sintheta*cos(phi), r*sintheta*sin(phi), r*costheta)
        * gsl_sf_legendre_sphPlm(sh_l, abs(sh_m), costheta) * (sh_m>=0 ? cos(sh_m*phi) : sin(-sh_m*phi));
}
double intSH_theta(double theta, void* params)
{
    double result, error;
    size_t neval;
    ((CPotentialParamSH*)params)->costheta=cos(theta);
    ((CPotentialParamSH*)params)->sintheta=sin(theta);
    gsl_function F;
    F.function=&intSH_phi;
    F.params=params;
    if( (((CPotentialParamSH*)params)->P->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC)
    {   // don't integrate in phi
        result = ((CPotentialParamSH*)params)->sh_m != 0 ? 0 :
            ((CPotentialParamSH*)params)->phi_max * 
            gsl_sf_legendre_sphPlm(((CPotentialParamSH*)params)->sh_l, 0, ((CPotentialParamSH*)params)->costheta)* 
            ((CPotentialParamSH*)params)->P->Rho(
            ((CPotentialParamSH*)params)->r*((CPotentialParamSH*)params)->sintheta, 0, 
            ((CPotentialParamSH*)params)->r*((CPotentialParamSH*)params)->costheta);
    } else
        gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->phi_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    return result * sin(theta) / (2*M_SQRTPI);   // factor \sqrt{4\pi} coming from the definition of spherical function Y_l^m
}
/// \endcond

double CPotentialSH::Rho(double X, double Y, double Z, double /*t*/) const
{
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(pow_2(X)+pow_2(Y)+pow_2(Z)));
    double phi=atan2(Y, X);
    double r=sqrt(X*X+Y*Y+Z*Z);
    double result=0;
    // arrays where angular expansion coefficients will be accumulated by calling computeSHcoefs() for derived classes
    double coefsF[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];      // F(theta,phi)
    double coefsdFdr[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];   // dF(theta,phi)/dr
    double coefsd2Fdr2[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR]; // d2F(theta,phi)/dr2
    computeSHCoefs(r, coefsF, coefsdFdr, coefsd2Fdr2);         // implemented in derived classes
    double legendre_array[MAX_NCOEFS_ANGULAR];
    for(int m=0; m<=mmax*lmax; m+=mstep)
    {
        gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array);
        int lmin = lstep==2 ? (m+1)/2*2 : m;   // if lstep is even and m is odd, start from next even number greater than m
        double cosmphi = (m==0 ? 1 : cos(m*phi)*M_SQRT2) * 2*M_SQRTPI;   // factor \sqrt{4\pi} from the definition of spherical function Y_l^m absorbed into this term
        for(int l=lmin; l<=lmax; l+=lstep)
        {
            result += (coefsd2Fdr2[l*(l+1)+m] + 2/r*coefsdFdr[l*(l+1)+m] - l*(l+1)/(r*r)*coefsF[l*(l+1)+m]) 
                * legendre_array[l-m] * cosmphi;
        }
        if(mmin<0 && m>0)  // use sin terms as well
        {
            double sinmphi = sin(m*phi)*M_SQRT2 * 2*M_SQRTPI;
            for(int l=lmin; l<=lmax; l+=mstep)
                result += (coefsd2Fdr2[l*(l+1)-m] + 2/r*coefsdFdr[l*(l+1)-m] - l*(l+1)/(r*r)*coefsF[l*(l+1)-m]) 
                    * legendre_array[l-m] * sinmphi;
        }
    }
    result *= 1/(4*M_PI);
    if(result<0 || gsl_isnan(result)) result=0;
    return result;
}

double CPotentialSH::Phi(double X, double Y, double Z, double /*t*/) const
{
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(pow_2(X)+pow_2(Y)+pow_2(Z)));
    double phi=atan2(Y, X);
    double r=sqrt(X*X+Y*Y+Z*Z);
    double res=0;
    // arrays where angular expansion coefficients will be accumulated by calling computeSHcoefs() for derived classes
    double coefsF[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];      // F(theta,phi)
    computeSHCoefs(r, coefsF, NULL, NULL);  // implemented in derived classes
    double legendre_array[MAX_NCOEFS_ANGULAR];
    for(int m=0; m<=mmax*lmax; m+=mstep)
    {
        gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array);
        double cosmphi = (m==0 ? 1 : cos(m*phi)*M_SQRT2) * 2*M_SQRTPI;   // factor \sqrt{4\pi} from the definition of spherical function Y_l^m absorbed into this term
        int lmin = lstep==2 ? (m+1)/2*2 : m;   // if lstep is even and m is odd, start from next even number greater than m
        for(int l=lmin; l<=lmax; l+=lstep)
            res +=  coefsF[l*(l+1)+m] * legendre_array[l-m] * cosmphi;
        if(mmin<0 && m>0)
        {
            double sinmphi = sin(m*phi)*M_SQRT2 * 2*M_SQRTPI;
            for(int l=lmin; l<=lmax; l+=lstep)
                res +=  coefsF[l*(l+1)-m] * legendre_array[l-m] * sinmphi;
        }
    }
    return res;
}

void CPotentialSH::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2]; 
    double dFdR=0, dFdTheta=0, dFdPhi=0; 
    double d2FdR2=0, d2FdRdTheta=0, d2FdRdPhi=0, d2FdTheta2=0, d2FdThetadPhi=0, d2FdPhi2=0;
    double r2 = pow_2(X) + pow_2(Y) + pow_2(Z);
    if(r2==0) {
        force[0]=force[1]=force[2]=0;
        if(forceDeriv!=NULL) forceDeriv[0]=forceDeriv[1]=forceDeriv[2]=forceDeriv[3]=forceDeriv[4]=forceDeriv[5]=0;
        return;
    }
    double r  = sqrt(r2);
    double rxy2=pow_2(X)+pow_2(Y);
    double rxy =sqrt(rxy2);
    double theta=atan2(rxy, Z); 
    double phi  =atan2(Y, X);
    double costheta=(Z==0)?0:cos(theta);
    double sintheta=sin(theta);
    double cosphi=(X==0)?0:cos(phi);
    double sinphi=(Y==0)?0:sin(phi);
    // arrays where angular expansion coefficients will be accumulated by calling computeSHcoefs() for derived classes
    double coefsF[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];      // F(theta,phi)
    double coefsdFdr[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR];   // dF(theta,phi)/dr
    double coefsd2Fdr2[MAX_NCOEFS_ANGULAR*MAX_NCOEFS_ANGULAR]; // d2F(theta,phi)/dr2
    computeSHCoefs(r, coefsF, coefsdFdr, forceDeriv!=NULL ? coefsd2Fdr2 : NULL);  // implemented in derived classes
    double legendre_array[MAX_NCOEFS_ANGULAR];
    double legendre_deriv_array[MAX_NCOEFS_ANGULAR];
    for(int m=0; m<=mmax*lmax; m+=mstep)
    {
#ifdef NO_OPTIMIZATION
        gsl_sf_legendre_sphPlm_deriv_array(lmax, m, costheta, legendre_array, legendre_deriv_array);
#else
        // compute unnormalized polynomials and then normalize manually, which is faster than computing normalized ones.
        // This is not suitable for large l,m (when overflow may occur), but in our application we aren't going to have such large values.
        gsl_sf_legendre_Plm_deriv_array(lmax, m, costheta, legendre_array, legendre_deriv_array);
        double prefact=sqrt(1.0/4/M_PI/gsl_sf_fact(2*m));
        for(int l=m; l<=lmax; l++)
        {
            double prefactl=sqrt(2*l+1.0)*prefact;
            legendre_array[l-m] *= prefactl;
            legendre_deriv_array[l-m] *= prefactl;
            prefact *= sqrt((l+1.0-m)/(l+1.0+m));
        }
#endif
        double cosmphi = (m==0 ? 1 : cos(m*phi)*M_SQRT2) * 2*M_SQRTPI;   // factor \sqrt{4\pi} from the definition of spherical function Y_l^m absorbed into this term
        double sinmphi = (sin(m*phi)*M_SQRT2) * 2*M_SQRTPI;
        int lmin = lstep==2 ? (m+1)/2*2 : m;   // if lstep is even and m is odd, start from next even number greater than m
        for(int l=lmin; l<=lmax; l+=lstep)
        {
            int indx=l*(l+1)+m;
            dFdR +=  coefsdFdr[indx] * legendre_array[l-m] * cosmphi;
            dFdTheta += coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * cosmphi;
            dFdPhi   += coefsF[indx] * legendre_array[l-m] * (-m)*sinmphi;
            if(forceDeriv!=NULL)
            {
                d2FdR2 +=  coefsd2Fdr2[indx] * legendre_array[l-m] * cosmphi;
                d2FdRdTheta+=coefsdFdr[indx] * legendre_deriv_array[l-m] * (-sintheta) * cosmphi;
                d2FdRdPhi += coefsdFdr[indx] * legendre_array[l-m] * (-m)*sinmphi;
                d2FdTheta2   += coefsF[indx] * (costheta * legendre_deriv_array[l-m] - (l*(l+1)-pow_2(m/sintheta)) * legendre_array[l-m]) * cosmphi;
                d2FdThetadPhi+= coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * (-m)*sinmphi;
                d2FdPhi2     += coefsF[indx] * legendre_array[l-m] * cosmphi * -m*m;
            }
            if(mmin<0 && m>0)
            {
                indx=l*(l+1)-m;
                dFdR +=  coefsdFdr[indx] * legendre_array[l-m] * sinmphi;
                dFdTheta += coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * sinmphi;
                dFdPhi   += coefsF[indx] * legendre_array[l-m] * m*cosmphi;
                if(forceDeriv!=NULL)
                {
                    d2FdR2 +=  coefsd2Fdr2[indx] * legendre_array[l-m] * sinmphi;
                    d2FdRdTheta+=coefsdFdr[indx] * legendre_deriv_array[l-m] * (-sintheta) * sinmphi;
                    d2FdRdPhi += coefsdFdr[indx] * legendre_array[l-m] * m*cosmphi;
                    d2FdTheta2   += coefsF[indx] * (costheta * legendre_deriv_array[l-m] - (l*(l+1)-pow_2(m/sintheta)) * legendre_array[l-m]) * sinmphi;
                    d2FdThetadPhi+= coefsF[indx] * legendre_deriv_array[l-m] * (-sintheta) * m*cosmphi;
                    d2FdPhi2     += coefsF[indx] * legendre_array[l-m] * sinmphi * -m*m;
                }
            }
        }
    }
    if((mysymmetry & ST_TRIAXIAL) == ST_TRIAXIAL) {
        if((X==0)||(Y==0)) dFdPhi=0; // to avoid singularity
        if(Z==0) dFdTheta=0;
    }
    if(rxy==0) rxy=1e-10;  // avoid singularity in denom
    double J[3][3]   = { {sintheta*cosphi, sintheta*sinphi, costheta}, {costheta*cosphi/r, costheta*sinphi/r, -sintheta/r}, {-sinphi/rxy, cosphi/rxy, 0} };
    force[0] = -( dFdR * J[0][0] + dFdTheta * J[1][0] + dFdPhi * J[2][0]);
    force[1] = -( dFdR * J[0][1] + dFdTheta * J[1][1] + dFdPhi * J[2][1]);
    force[2] = -( dFdR * J[0][2] + dFdTheta * J[1][2]);
    if(forceDeriv!=NULL)
    {
        double X2=X*X, Y2=Y*Y, Z2=Z*Z;
        double d2F[3][3] = { { d2FdR2, d2FdRdTheta, d2FdRdPhi }, {d2FdRdTheta, d2FdTheta2, d2FdThetadPhi}, {d2FdRdPhi, d2FdThetadPhi, d2FdPhi2} };
        double m=1/(r*r2);
        double d2R[3][3] = { {(Y2+Z2)*m, -X*Y*m, -X*Z*m}, {-X*Y*m, (X2+Z2)*m, -Y*Z*m}, {-X*Z*m, -Y*Z*m, (X2+Y2)*m} };
        m=1/(rxy*rxy2*r2*r2);
        double d2Theta[3][3] = { 
            {-Z*(2*X2*X2+Y2*X2-Y2*Y2-Z2*Y2)*m, -X*Y*Z*(3*X2+3*Y2+Z2)*m, X*(X2+Y2-Z2)*rxy2*m},
            {-X*Y*Z*(3*X2+3*Y2+Z2)*m, -Z*(2*Y2*Y2+Y2*X2-X2*X2-Z2*X2)*m, Y*(X2+Y2-Z2)*rxy2*m},
            {X*(X2+Y2-Z2)*rxy2*m, Y*(X2+Y2-Z2)*rxy2*m, 2*Z*rxy2*rxy2*m} };
        m=1/(rxy2*rxy2);
        double d2Phi[3][3] = { {2*X*Y*m, (Y2-X2)*m, 0}, {(Y2-X2)*m, -2*X*Y*m, 0}, {0, 0, 0} };
        // convert to Cartesian coordinates
        double k0 = (J[0][0]*d2F[0][0]+J[1][0]*d2F[0][1]+J[2][0]*d2F[0][2]);
        double k1 = (J[0][0]*d2F[0][1]+J[1][0]*d2F[1][1]+J[2][0]*d2F[1][2]);
        double k2 = (J[0][0]*d2F[0][2]+J[1][0]*d2F[1][2]+J[2][0]*d2F[2][2]);
        double d2Fdx2 = (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[0][0] + dFdTheta*d2Theta[0][0] + dFdPhi*d2Phi[0][0]);
        double d2Fdxdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[0][1] + dFdTheta*d2Theta[0][1] + dFdPhi*d2Phi[0][1]);
        //double d2Fdxdz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[0][2] + dFdTheta*d2Theta[0][2] + dFdPhi*d2Phi[0][2]);
        k0 = (J[0][1]*d2F[0][0]+J[1][1]*d2F[0][1]+J[2][1]*d2F[0][2]);
        k1 = (J[0][1]*d2F[0][1]+J[1][1]*d2F[1][1]+J[2][1]*d2F[1][2]);
        k2 = (J[0][1]*d2F[0][2]+J[1][1]*d2F[1][2]+J[2][1]*d2F[2][2]);
        //double d2Fdydx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[1][0] + dFdTheta*d2Theta[1][0] + dFdPhi*d2Phi[1][0]);
        double d2Fdy2 = (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[1][1] + dFdTheta*d2Theta[1][1] + dFdPhi*d2Phi[1][1]);
        double d2Fdydz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[1][2] + dFdTheta*d2Theta[1][2] + dFdPhi*d2Phi[1][2]);
        k0 = (J[0][2]*d2F[0][0]+J[1][2]*d2F[0][1]+J[2][2]*d2F[0][2]);
        k1 = (J[0][2]*d2F[0][1]+J[1][2]*d2F[1][1]+J[2][2]*d2F[1][2]);
        k2 = (J[0][2]*d2F[0][2]+J[1][2]*d2F[1][2]+J[2][2]*d2F[2][2]);
        double d2Fdzdx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[2][0] + dFdTheta*d2Theta[2][0] + dFdPhi*d2Phi[2][0]);
        //double d2Fdzdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[2][1] + dFdTheta*d2Theta[2][1] + dFdPhi*d2Phi[2][1]);
        double d2Fdz2 = (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[2][2] + dFdTheta*d2Theta[2][2] + dFdPhi*d2Phi[2][2]);

        forceDeriv[0]=-d2Fdx2;
        forceDeriv[1]=-d2Fdy2;
        forceDeriv[2]=-d2Fdz2;
        forceDeriv[3]=-d2Fdxdy;
        forceDeriv[4]=-d2Fdydz;
        forceDeriv[5]=-d2Fdzdx;
    }
}

//----------------------------------------------------------------------------//
// Basis-set expansion for arbitrary potential (using Zhao(1995) basis set)

template<typename NumT> 
CPotentialBSE::CPotentialBSE(
    double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym, bool checkMonotonic):
  CPotentialSH(_Ncoefs_angular), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _Ncoefs_radial)),
  Alpha(_Alpha)
{
    mysymmetry=_sym;
    assignlmrange();
    if(points.size()==0) { initDefault(); return; }
    prepareCoefsDiscrete(points);
    if(checkMonotonic) checkSymmetry();  // not called from BSE-Schwarzschild model
}
template 
CPotentialBSE::CPotentialBSE(double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, const smile::CPointMassSet<double>&points, smile::CDensity::SYMMETRYTYPE _sym, bool checkMonotonic);
template 
CPotentialBSE::CPotentialBSE(double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, const smile::CPointMassSet<float> &points, smile::CDensity::SYMMETRYTYPE _sym, bool checkMonotonic);

CPotentialBSE::CPotentialBSE(double _Alpha, const std::vector< vectord > &coefs):
  CPotentialSH((assert(coefs.size()>0), static_cast<size_t>(sqrt(coefs[0].size()*1.0)-1))), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, static_cast<size_t>(coefs.size()-1))),
  Alpha(std::max<double>(0.5,_Alpha))  // here Alpha!=0 - no autodetect
{
    if(_Alpha<0.5) { 
        my_error(FUNCNAME, "Error, invalid parameter Alpha="+convertToString(_Alpha)); 
        initDefault(); 
        return; 
    }
    assert(coefs[0].size()==pow_2(Ncoefs_angular+1));
    SHcoefs = coefs;  // do not check array size -- assumed to be done beforehand (in load routine)...
    checkSymmetry();
}

CPotentialBSE::CPotentialBSE(double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CDensity* density):    // init potential from analytic mass model
  CPotentialSH(_Ncoefs_angular), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _Ncoefs_radial)),
  Alpha(_Alpha)
{
    mysymmetry=density->symmetry();
    assignlmrange();
    prepareCoefsAnalytic(density);
    checkSymmetry();
}

void CPotentialBSE::initDefault()
{
    Ncoefs_radial=0; 
    Ncoefs_angular=0;
    assignlmrange();
    Alpha=1;
    SHcoefs.resize(1);
    SHcoefs[0].assign(1,1);
}

void CPotentialBSE::checkSymmetry()
{ 
    SYMMETRYTYPE sym=ST_SPHERICAL;  // too optimistic:))
    const double MINCOEF=1e-8;
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if(fabs(SHcoefs[n][l*(l+1)+m])>MINCOEF) 
                {   // nonzero coef.: check if that breaks any symmetry
                    if(l%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_REFLECTION);
                    if(m<0 || m%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_PLANESYM);
                    if(m!=0) sym = (SYMMETRYTYPE)(sym & ~ST_ZROTSYM);
                    if(l>0) sym = (SYMMETRYTYPE)(sym & ~ST_SPHSYM);
                }
    }
    // now set all coefs excluded by the inferred symmetry  to zero
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if( (l>0 && (sym & ST_SPHSYM)) ||
                    (m!=0 && (sym & ST_ZROTSYM)) ||
                    ((m<0 || m%2==1) && (sym & ST_PLANESYM)) ||
                    (l%2==1 && (sym & ST_REFLECTION)) ) 
                        SHcoefs[n][l*(l+1)+m] = 0;
    }
    mysymmetry = sym; 
    assignlmrange();
    bool densityNonzero = checkDensityNonzero();
    bool massMonotonic  = checkMassMonotonic();
    if(!massMonotonic || !densityNonzero) 
        my_message(FUNCNAME, "Warning, " + 
        std::string(!massMonotonic ? "mass does not monotonically increase with radius" : "") +
        std::string(!massMonotonic && !densityNonzero ? " and " : "") + 
        std::string(!densityNonzero ? "density drops to zero at a finite radius" : "") + "!");
}

/// \cond INTERNAL_DOCS
double intBSE_xi(double xi, void* params)   // integrate over scaled radial variable
{
    if(xi>=1.0 || xi<=-1.0) return 0;
    double Alpha=((CPotentialParamSH*)params)->Alpha;
    double ralpha=(1+xi)/(1-xi);
    double r=pow(ralpha, Alpha);
    int n=((CPotentialParamSH*)params)->sh_n;
    int l=((CPotentialParamSH*)params)->sh_l;
    double phi_nl = gsl_sf_gegenpoly_n(n, (2*l+1)*Alpha+0.5, xi) * pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
    ((CPotentialParamSH*)params)->r=r;
    gsl_function F;
    F.function=&intSH_theta;
    F.params=params;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    return result * phi_nl * 8*M_PI*Alpha *r*r*r/(1-xi*xi);
}
/// \endcond

void CPotentialBSE::prepareCoefsAnalytic(const CDensity* density)
{
    if(Alpha<0.5)
    {  // find best-suited value depending on inner density slope
        double Gamma=density->getGamma();
        Alpha = (Gamma<=1? 1.0+Gamma :
        (Gamma<=1.5? 2.0 :
        (Gamma<1.75? 1./(2-Gamma) : 4.0)));
        if(density->PotentialType()==CDensity::PT_PLUMMER) Alpha=0.5;  // the default value for plummer potential only; however even in this case alpha=1 might perform better..
    }
    SHcoefs.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++)
        SHcoefs[n].assign(pow_2(Ncoefs_angular+1), 0);
    CPotentialParamSH PP;
    PP.P=density;
    PP.Alpha=Alpha;
    PP.theta_max= (mysymmetry & ST_REFLECTION)==ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= (mysymmetry & ST_PLANESYM)==ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = ((mysymmetry & ST_PLANESYM)==ST_PLANESYM ? 4 : 1) * 
        ((mysymmetry & ST_REFLECTION)==ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc (1000);
    double interval[2]={-1.0, 1.0};
    gsl_function F;
    F.function=&intBSE_xi;
    F.params=&PP;
    for(size_t n=0; n<=Ncoefs_radial; n++)
        for(int l=0; l<=lmax; l+=lstep)
        {
            PP.sh_n=static_cast<int>(n);  PP.sh_l=l;
            double w=(2*l+1)*Alpha+0.5;
            double Knl = (4*pow_2(n+w)-1)/8/pow_2(Alpha);
            double Inl = Knl * 4*M_PI*Alpha * 
                exp( gsl_sf_lngamma(n+2*w) - 2*gsl_sf_lngamma(w) - gsl_sf_lnfact(PP.sh_n) - 4*w*log(2.0)) / (n+w);
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                PP.sh_m=m;
                double result, error;
                gsl_integration_qagp(&F, interval, 2, 0, EPSREL_POTENTIAL_INT, 1000, ws, &result, &error);
                SHcoefs[n][l*(l+1)+m] = result * multfactor * (m==0 ? 1 : M_SQRT2) / Inl;
            }
        }
    gsl_integration_workspace_free (ws);
}

template<typename NumT> 
void CPotentialBSE::prepareCoefsDiscrete(const CPointMassSet<NumT> &points)
{
    SHcoefs.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++)
        SHcoefs[n].assign(pow_2(1+Ncoefs_angular), 0);
    size_t npoints=points.size();
    if(Alpha<0.5)  // autodetect.. the heuristic below is not terribly good, so if you know the inner density slope beforehand, it's better to set alpha=1/(2-gamma) manually
    {
        vectorpd radmass(npoints);
        for(size_t i=0; i<npoints; i++) 
            radmass[i] = paird(sqrt(pow_2(points[i].first.Pos[0])+pow_2(points[i].first.Pos[1])+pow_2(points[i].first.Pos[2])), points[i].second);
        std::sort(radmass.begin(), radmass.end());
        size_t n_est=std::min<size_t>(std::max<size_t>(npoints/50,1000),npoints-1); 
        size_t n_min=n_est/100;
        double inm=0, r_n_est=radmass[n_est].first;
        while(n_min<n_est && radmass[n_min].first<r_n_est*0.03) n_min++;  // neglect some first entries
        gsl_matrix* X=gsl_matrix_alloc(n_est-n_min, 3);  // matrix of coefficients
        gsl_vector* Y=gsl_vector_alloc(n_est-n_min);     // vector of r.h.s. values (mass(r))
        gsl_vector* W=gsl_vector_alloc(n_est-n_min);     // weights 
        for(size_t i=0; i<n_est; i++)
        {
            inm+=radmass[i].second;
            if(i>=n_min){
            gsl_matrix_set(X, i-n_min, 0, 1);  // constant term
            gsl_matrix_set(X, i-n_min, 1, log(radmass[i].first+SPLINE_MIN_RADIUS));  // to avoid log zero values
            gsl_matrix_set(X, i-n_min, 2, radmass[i].first);
            gsl_vector_set(Y, i-n_min, log(inm));
            gsl_vector_set(W, i-n_min, 1/*radmass[i].first/r_n_est*(1-radmass[i].first/r_n_est)*/);  // simple weighting function disfavouring lower and higher radii
            }
        }
        if(n_est-n_min>100)  { // meaningful fit
            // approximate m(r) by power law function:  m(r) = C * r^(3-Gamma) * (1 + k*r), three parameters: C, Gamma, k
            gsl_vector* fit=gsl_vector_alloc(3);
            gsl_matrix* cov=gsl_matrix_alloc(3,3);
            double chisq;
            gsl_multifit_linear_workspace* ws=gsl_multifit_linear_alloc(n_est-n_min, 3);
            gsl_multifit_wlinear(X, W, Y, fit, cov, &chisq, ws);
            gsl_multifit_linear_free(ws);
            double Gamma = 3-gsl_vector_get(fit, 1);   // mass ~ r^(3-gamma)
            Alpha = (Gamma<=0.5? 1.0 :
              (Gamma<=1.5? 0.5+Gamma :
              (Gamma<1.75? 1./(2-Gamma) : 4.0)));
            gsl_vector_free(fit);
            gsl_matrix_free(cov);
#ifdef DEBUGPRINT
            my_message(FUNCNAME, "Gamma_est="+convertToString(Gamma)+", alpha="+convertToString(Alpha));
#endif
        }  else Alpha=1.;  // default value
        gsl_matrix_free(X);
        gsl_vector_free(Y);
        gsl_vector_free(W);
    }
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR-1];
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    double Inl[MAX_NCOEFS_RADIAL][MAX_NCOEFS_ANGULAR];
    for(int l=0; l<=lmax; l+=lstep)
    {   // pre-compute coefficients
        double w=(2*l+1)*Alpha+0.5;
        for(unsigned int n=0; n<=Ncoefs_radial; n++)
            Inl[n][l] = 4*M_PI*Alpha * 
              exp( gsl_sf_lngamma(n+2*w) - 2*gsl_sf_lngamma(w) - gsl_sf_lnfact(n) - 4*w*log(2.0)) /
              (n+w) * (4*(n+w)*(n+w)-1)/(8*Alpha*Alpha);
    }
    for(size_t i=0; i<npoints; i++)
    {
        double massi=points[i].second;
        double r=sqrt(pow_2(points[i].first.Pos[0])+pow_2(points[i].first.Pos[1])+pow_2(points[i].first.Pos[2]));
        double costheta= (r==0) ? 0 : (points[i].first.Pos[2]/r);
        double phi=atan2(points[i].first.Pos[1], points[i].first.Pos[0]);
        double ralpha=pow(r, 1/Alpha);
        double xi=(ralpha-1)/(ralpha+1);
        for(int m=0; m<=lmax; m+=mstep)
            gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array[m]);

        for(int l=0; l<=lmax; l+=lstep)
        {
            double w=(2*l+1)*Alpha+0.5;
            double phil=pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
            gsl_sf_gegenpoly_array(static_cast<int>(Ncoefs_radial), w, xi, gegenpoly_array);
            for(size_t n=0; n<=Ncoefs_radial; n++)
            {
                double mult= massi * gegenpoly_array[n] * phil * 2*M_SQRTPI / Inl[n][l];
                for(int m=0; m<=l*mmax; m+=mstep)
                    SHcoefs[n][l*(l+1)+m] += mult * legendre_array[m][l-m] * cos(m*phi) * (m==0 ? 1 : M_SQRT2);
                if(mmin)
                    for(int m=mmin*l; m<0; m+=mstep)
                        SHcoefs[n][l*(l+1)+m] += mult * legendre_array[-m][l+m] * sin(-m*phi) * M_SQRT2;
            }
        }
    }
}

double CPotentialBSE::Mass(const double r) const
{
    if(r<=0) return 0;
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    gsl_sf_gegenpoly_array(static_cast<int>(Ncoefs_radial), Alpha+0.5, xi, gegenpoly_array);
    double multr = pow(1+ralpha, -Alpha);
    double multdr= -ralpha/((ralpha+1)*r);
    if(gsl_isnan(multdr)) multdr=0;  // safety measure to avoid NaN
    double result=0;
    for(int n=0; n<=static_cast<int>(Ncoefs_radial); n++)
    {
        double dGdr=(n>0 ? (-n*xi*gegenpoly_array[n] + (n+2*Alpha)*gegenpoly_array[n-1])/(2*Alpha*r) : 0);
        result += SHcoefs[n][0] * multr * (multdr * gegenpoly_array[n] + dGdr);
    }
    return -result * r*r;   // d Phi(r)/d r = - G M(r) /r^2
}

void CPotentialBSE::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    if(coefsF)      for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsF     [k] = 0;
    if(coefsdFdr)   for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsdFdr  [k] = 0;
    if(coefsd2Fdr2) for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsd2Fdr2[k] = 0;
    for(int l=0; l<=lmax; l+=lstep)
    {
        double w=(2*l+1)*Alpha+0.5;
        gsl_sf_gegenpoly_array(static_cast<int>(Ncoefs_radial), w, xi, gegenpoly_array);
        double multr = -pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
        double multdr= (l-(l+1)*ralpha)/((ralpha+1)*r);
        if(gsl_isnan(multdr)) multdr=0;  // safety measure to avoid NaN
        for(int n=0; n<=(int)Ncoefs_radial; n++)
        {
            double multdFdr=0, multd2Fdr2=0, dGdr=0;
            if(coefsdFdr!=NULL) {
                dGdr=(n>0 ? (-n*xi*gegenpoly_array[n] + (n+2*w-1)*gegenpoly_array[n-1])/(2*Alpha*r) : 0);
                multdFdr= multdr * gegenpoly_array[n] + dGdr;
                if(coefsd2Fdr2!=NULL)
                    multd2Fdr2 = ( (l+1)*(l+2)*pow_2(ralpha) + 
                                   ((1-2*l*(l+1)) - (2*n+1)*(2*l+1)/Alpha - n*(n+1)/pow_2(Alpha))*ralpha + 
                                   l*(l-1) 
                                 )/pow_2((1+ralpha)*r)*gegenpoly_array[n] - dGdr*2/r;
            }
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int indx=l*(l+1)+m;
                double coef = SHcoefs[n][indx] * multr;
                if(coefsF)      coefsF     [indx] += coef * gegenpoly_array[n];
                if(coefsdFdr)   coefsdFdr  [indx] += coef * multdFdr;
                if(coefsd2Fdr2) coefsd2Fdr2[indx] += coef * multd2Fdr2;
            }
        }
    }
}

//----------------------------------------------------------------------------//
// Basis-set expansion for a compact system (using spherical Bessel functions)

template<typename NumT> 
CPotentialBSECompact::CPotentialBSECompact(
    double _Rmax, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym /*, bool checkMonotonic*/):
  CPotentialSH(_Ncoefs_angular), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _Ncoefs_radial)),
  Rmax(_Rmax)
{
    mysymmetry=_sym;
    assignlmrange();
    prepareAlpha();
    if(points.size()==0) { initDefault(); return; }
    prepareCoefsDiscrete(points);
    //if(checkMonotonic) checkSymmetry();  // not called from BSE-Schwarzschild model
}
template 
CPotentialBSECompact::CPotentialBSECompact(double _Rmax, size_t _Ncoefs_radial, size_t _Ncoefs_angular, const smile::CPointMassSet<double>&points, smile::CDensity::SYMMETRYTYPE _sym/*, bool checkMonotonic*/);
template 
CPotentialBSECompact::CPotentialBSECompact(double _Rmax, size_t _Ncoefs_radial, size_t _Ncoefs_angular, const smile::CPointMassSet<float> &points, smile::CDensity::SYMMETRYTYPE _sym/*, bool checkMonotonic*/);

CPotentialBSECompact::CPotentialBSECompact(double _Rmax, const std::vector< vectord > &coefs):
  CPotentialSH((assert(coefs.size()>0), static_cast<size_t>(sqrt(coefs[0].size()*1.0)-1))), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, static_cast<size_t>(coefs.size()-1))),
  Rmax(_Rmax)
{
    assert(coefs[0].size()==pow_2(Ncoefs_angular+1));
    SHcoefs = coefs;  // do not check array size -- assumed to be done beforehand (in load routine)...
    prepareAlpha();
    checkSymmetry();
}

CPotentialBSECompact::CPotentialBSECompact(double _Rmax, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CDensity* density):    // init potential from analytic mass model
  CPotentialSH(_Ncoefs_angular), 
  Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _Ncoefs_radial)),
  Rmax(_Rmax)
{
    mysymmetry=density->symmetry();
    assignlmrange();
    prepareAlpha();
    prepareCoefsAnalytic(density);
    checkSymmetry();
}

void CPotentialBSECompact::initDefault()
{
    Ncoefs_radial=0; 
    Ncoefs_angular=0;
    assignlmrange();
    Rmax=1;
    SHcoefs.resize(1);
    SHcoefs[0].assign(1,1);
}

void CPotentialBSECompact::checkSymmetry()
{ 
    SYMMETRYTYPE sym=ST_SPHERICAL;  // too optimistic:))
    const double MINCOEF=1e-8;
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if(fabs(SHcoefs[n][l*(l+1)+m])>MINCOEF) 
                {   // nonzero coef.: check if that breaks any symmetry
                    if(l%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_REFLECTION);
                    if(m<0 || m%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_PLANESYM);
                    if(m!=0) sym = (SYMMETRYTYPE)(sym & ~ST_ZROTSYM);
                    if(l>0) sym = (SYMMETRYTYPE)(sym & ~ST_SPHSYM);
                }
    }
    // now set all coefs excluded by the inferred symmetry  to zero
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if( (l>0 && (sym & ST_SPHSYM)) ||
                    (m!=0 && (sym & ST_ZROTSYM)) ||
                    ((m<0 || m%2==1) && (sym & ST_PLANESYM)) ||
                    (l%2==1 && (sym & ST_REFLECTION)) ) 
                        SHcoefs[n][l*(l+1)+m] = 0;
    }
    mysymmetry = sym; 
    assignlmrange();
    bool densityNonzero = checkDensityNonzero();
    bool massMonotonic  = checkMassMonotonic();
    if(!massMonotonic || !densityNonzero) 
        my_message(FUNCNAME, "Warning, " + 
        std::string(!massMonotonic ? "mass does not monotonically increase with radius" : "") +
        std::string(!massMonotonic && !densityNonzero ? " and " : "") + 
        std::string(!densityNonzero ? "density drops to zero at a finite radius" : "") + "!");
}

void CPotentialBSECompact::prepareAlpha()
{
    alpha.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++) {
        alpha[n].resize(Ncoefs_angular+1);
        for(size_t l=0; l<=Ncoefs_angular; l++) {
            double zero = (l==0) ? (1+2*n)*M_PI/2 : gsl_sf_bessel_zero_Jnu(l-0.5, n+1);
            alpha[n][l] = zero/Rmax;
        }
    }
}

/// \cond INTERNAL_DOCS
double intBSECompact_r(double r, void* params)   // integrate over radial variable
{
    double scale=((CPotentialParamSH*)params)->Alpha;
    int l=((CPotentialParamSH*)params)->sh_l;
    double phi_nl = -sqrt(2/M_PI*scale) * gsl_sf_bessel_jl(l, scale*r);
    ((CPotentialParamSH*)params)->r=r;
    gsl_function F;
    F.function=&intSH_theta;
    F.params=params;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    return result * phi_nl * 4*M_PI*r*r;
}
/// \endcond

void CPotentialBSECompact::prepareCoefsAnalytic(const CDensity* density)
{
    SHcoefs.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++)
        SHcoefs[n].assign(pow_2(Ncoefs_angular+1), 0);
    CPotentialParamSH PP;
    PP.P=density;
    PP.theta_max= mysymmetry & ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= mysymmetry & ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = (mysymmetry & ST_PLANESYM ? 4 : 1) * (mysymmetry & ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    gsl_function F;
    F.function=&intBSECompact_r;
    F.params=&PP;
    for(size_t n=0; n<=Ncoefs_radial; n++)
        for(int l=0; l<=lmax; l+=lstep)
        {
            PP.Alpha=alpha[n][l];
            PP.sh_l=l;
            double Inl = -pow(alpha[n][l]*Rmax, 3.0)*pow_2(gsl_sf_bessel_jl(l, alpha[n][l]*Rmax))/M_PI;
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                PP.sh_m=m;
                double result, error;
                size_t neval;
                gsl_integration_qng(&F, 0, Rmax, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
                SHcoefs[n][l*(l+1)+m] = result * multfactor * (m==0 ? 1 : M_SQRT2) / Inl;
            }
        }
}

template<typename NumT> 
void CPotentialBSECompact::prepareCoefsDiscrete(const CPointMassSet<NumT> &points)
{
    SHcoefs.resize(Ncoefs_radial+1);
    for(size_t n=0; n<=Ncoefs_radial; n++)
        SHcoefs[n].assign(pow_2(1+Ncoefs_angular), 0);
    size_t npoints=points.size();
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR-1];
    double Inl[MAX_NCOEFS_RADIAL][MAX_NCOEFS_ANGULAR];
    for(int l=0; l<=lmax; l+=lstep)
    {   // pre-compute normalization coefficients
        for(unsigned int n=0; n<=Ncoefs_radial; n++)
            Inl[n][l] = -pow(alpha[n][l]*Rmax, 3.0)*pow_2(gsl_sf_bessel_jl(l, alpha[n][l]*Rmax))/M_PI;
    }
    for(size_t i=0; i<npoints; i++)
    {
        double massi=points[i].second;
        double r=sqrt(pow_2(points[i].first.Pos[0])+pow_2(points[i].first.Pos[1])+pow_2(points[i].first.Pos[2]));
        if(r>Rmax) 
            continue;   // don't consider points beyond the compact support
        double costheta= (r==0) ? 0 : (points[i].first.Pos[2]/r);
        double phi=atan2(points[i].first.Pos[1], points[i].first.Pos[0]);
        for(int m=0; m<=lmax; m+=mstep)
            gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array[m]);

        for(int l=0; l<=lmax; l+=lstep)
        {
            for(size_t n=0; n<=Ncoefs_radial; n++)
            {
                double mult= massi * -sqrt(2/M_PI*alpha[n][l]) * gsl_sf_bessel_jl(l, alpha[n][l]*r) * 2*M_SQRTPI / Inl[n][l];
                for(int m=0; m<=l*mmax; m+=mstep)
                    SHcoefs[n][l*(l+1)+m] += mult * legendre_array[m][l-m] * cos(m*phi) * (m==0 ? 1 : M_SQRT2);
                if(mmin)
                    for(int m=mmin*l; m<0; m+=mstep)
                        SHcoefs[n][l*(l+1)+m] += mult * legendre_array[-m][l+m] * sin(-m*phi) * M_SQRT2;
            }
        }
    }
}

double CPotentialBSECompact::Mass(const double r) const
{
    if(r<=0) return 0;
    double rmod=std::min<double>(r, Rmax);
    double result=0;
    for(int n=0; n<=static_cast<int>(Ncoefs_radial); n++)
    {
        double scale = alpha[n][0];
        result += SHcoefs[n][0] * sqrt(scale*2/M_PI)*scale * gsl_sf_bessel_j1(rmod*scale);
    }
    return result * rmod*rmod;   // d Phi(r)/d r = G M(r) /r^2
}

void CPotentialBSECompact::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    if(coefsF)      for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsF     [k] = 0;
    if(coefsdFdr)   for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsdFdr  [k] = 0;
    if(coefsd2Fdr2) for(size_t k=0; k<pow_2(Ncoefs_angular+1); k++) coefsd2Fdr2[k] = 0;
    double Rfact=1.0;
    if(r>Rmax) {   // outside the finite support use multipole expansion for Laplace equation
        Rfact=Rmax/r;
    }
    for(int l=0; l<=lmax; l+=lstep)
    {
        double RfactPow = pow(Rfact, l+1);
        for(int n=0; n<=(int)Ncoefs_radial; n++)
        {
            double scale = alpha[n][l];
            double rscaled = std::min<double>(r, Rmax)*scale;
            double mult  = -sqrt(scale*2/M_PI);
            double jl = gsl_sf_bessel_jl(l, rscaled);
            double multF = jl * RfactPow;
            double multdFdr=0, multd2Fdr2=0;
            if(coefsdFdr!=NULL) {
                if(Rfact<1) {  // use multipole expansion for empty space
                    multdFdr = -(l+1)*multF/r;
                    if(coefsd2Fdr2!=NULL)
                        multd2Fdr2 = -(l+2)*multdFdr/r;
                } else {
                    if(l==0) {
                        double j1 = gsl_sf_bessel_j1(rscaled);
                        multdFdr = -scale*j1;
                        if(coefsd2Fdr2!=NULL)
                            multd2Fdr2 = -rscaled*( rscaled*jl - 2*j1 )/pow_2(r);
                    } else {
                        double jlminus1 = gsl_sf_bessel_jl(l-1, rscaled);
                        multdFdr = scale*jlminus1 - (l+1)/r*jl;
                        if(coefsd2Fdr2!=NULL)
                            multd2Fdr2 = -( 2*rscaled*jlminus1 + (pow_2(rscaled)-l*l-3*l-2)*jl )/pow_2(r);
                    }
                }
            }
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int indx=l*(l+1)+m;
                double coef = SHcoefs[n][indx] * mult;
                if(coefsF)      coefsF     [indx] += coef * multF;
                if(coefsdFdr)   coefsdFdr  [indx] += coef * multdFdr;
                if(coefsd2Fdr2) coefsd2Fdr2[indx] += coef * multd2Fdr2;
            }
        }
    }
}

//----------------------------------------------------------------------------//
// Spherical-harmonic expansion of arbitrary potential, radial part is spline interpolated on a grid

template<typename NumT>
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym, double smoothfactor, const vectord *radii):  // init coefs from point mass set
    CPotentialSH(_Ncoefs_angular),
    Ncoefs_radial(std::max<size_t>(5,_Ncoefs_radial))
{
    mysymmetry=_sym;
    assignlmrange();
    if(points.size() <= Ncoefs_radial*10) { 
        if(points.size()>0)
            my_error(FUNCNAME, "Number of points is too small in Spline potential initialization");
        initDefault(); 
        return; 
    }
    prepareCoefsDiscrete(points, smoothfactor, radii);
}
template
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSet<double> &points, SYMMETRYTYPE _sym, double smoothfactor, const vectord *radii);
template
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CPointMassSet<float> &points, SYMMETRYTYPE _sym, double smoothfactor, const vectord *radii);

CPotentialSpline::CPotentialSpline(
    const vectord &_gridradii, const std::vector< vectord > &_coefs):
    CPotentialSH((assert(_coefs.size()>0), static_cast<size_t>(sqrt(_coefs[0].size()*1.0)-1))), 
    Ncoefs_radial(std::min<size_t>(MAX_NCOEFS_RADIAL-1, _coefs.size()-1))  // init from existing coefs
{
    assert(_coefs[0].size()==pow_2(Ncoefs_angular+1));
    initSpline(_gridradii, _coefs);
}

CPotentialSpline::CPotentialSpline(size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
    const CDensity* density, const vectord *radii):    // init potential from analytic mass model
    CPotentialSH(_Ncoefs_angular), 
    Ncoefs_radial(std::max<size_t>(5,_Ncoefs_radial))
{
    mysymmetry=density->symmetry();
    assignlmrange();
    prepareCoefsAnalytic(density, radii);
}

// not a constructor, but a method to compute coefs at given radii
template<typename NumT> 
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_angular, 
    const CPointMassSet<NumT> &points, const vectord *radii, std::vector< vectord > *coefsArray):
    CPotentialSH(_Ncoefs_angular),
    Ncoefs_radial(radii->size()-1)
{
    computeCoefsFromPoints(points, radii, NULL, coefsArray);
}
template 
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_angular, const CPointMassSet<double> &points, const vectord *radii, std::vector< vectord > *coefsArray);
template 
CPotentialSpline::CPotentialSpline(size_t _Ncoefs_angular, const CPointMassSet<float> &points, const vectord *radii, std::vector< vectord > *coefsArray);

void CPotentialSpline::initDefault()
{
    mysymmetry=ST_SPHERICAL;
    assignlmrange();
    Ncoefs_radial=5; Ncoefs_angular=0;
    vectord nodes(Ncoefs_radial+1);
    std::vector< vectord > values(Ncoefs_radial+1);
    nodes[0]=0; 
    nodes[1]=0.01;
    nodes[2]=0.1;
    nodes[3]=1.0;
    nodes[4]=10.;
    nodes[5]=100;
    for(size_t i=0; i<=5; i++) //default initialization for Dehnen gamma=1 model
        values[i].push_back(-1.0/(nodes[i]+1.0));
    initSpline(nodes, values);
}

CPotentialSpline::~CPotentialSpline()
{
    for(size_t i=0; i<splines.size(); i++)
        if(splines[i]!=NULL)  gsl_spline_free(splines[i]);
}

CPotential* CPotentialSpline::clone() const 
{   // need to duplicate spline arrays explicitly; everything else is copied automatically
    CPotentialSpline* newpot = new CPotentialSpline(*this); 
    for(size_t i=0; i<splines.size(); i++)
        newpot->splines[i]=copyspline(splines[i]);  // a special copy function that takes into account clamped splines properly
    return newpot;
}

/// \cond INTERNAL_DOCS
double intSpline_r(double r, void* params)
{
    if(r==0) return 0;
    int n=((CPotentialParamSH*)params)->sh_n;  // power index for s^n, either l+2 or 1-l
    double result, error;
    size_t neval;
    ((CPotentialParamSH*)params)->r=r;
    gsl_function F;
    F.function=&intSH_theta;
    F.params=params;
    gsl_integration_qng(&F, 0, ((CPotentialParamSH*)params)->theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    if(((CPotentialParamSH*)params)->sh_l>2 && fabs(result)<EPSABS_POTENTIAL_INT)  // should add some workaround if a coefficient is very small due to unnoticed symmetry?
        result=0;  // to avoid huge drop in performance for very small coef values
    return result * pow(r, n*1.0);
}
/// \endcond

void CPotentialSpline::prepareCoefsAnalytic(const CDensity* density, const vectord *srcradii)
{
    std::vector< vectord > coefsArray(Ncoefs_radial+1);  // SHE coefficients to pass to initspline routine
    vectord radii(Ncoefs_radial+1);  // true radii to pass to initspline routine
    for(size_t i=0; i<=Ncoefs_radial; i++)
        coefsArray[i].assign(pow_2(1+Ncoefs_angular), 0);
    bool initUserRadii = (srcradii!=NULL && srcradii->size()==Ncoefs_radial+1 && srcradii->front()==0);
    if(srcradii!=NULL && !initUserRadii)  // something went wrong with manually supplied radii, initializing by default
        my_error(FUNCNAME, 
            "Warning, invalid call to constructor of Spline potential or SHGrid Schwarzschild model");
    if(initUserRadii)
        radii= *srcradii;
    else
    {
        // find inner/outermost radius
        double totalMass=density->totalMass();
        double epsout = 0.1/sqrt(pow_2(Ncoefs_radial)+0.01*pow(Ncoefs_radial*1.0,4.0));      // how far should be the outer node (leave out this fraction of mass)
        double epsin = 5.0/pow(Ncoefs_radial*1.0,3.0);                                       // how close can we get to zero, in terms of innermost grid node
        double rout = totalMass<0 ? 10 : density->getRadiusByMass(totalMass*(1-epsout));     // totalMass<0 should not really happen! means that mass is infinite
        double rin = std::min<double>(epsin, density->getRadiusByMass(totalMass*epsin*0.1)); // somewhat arbitrary choice for min/max radii, but probably reasonable
        createNonuniformGrid(Ncoefs_radial+1, rin, rout, true, &radii); 
#ifdef DEBUGPRINT
        my_message(FUNCNAME, 
            "Mtotal="+convertToString(totalMass)+", rin="+convertToString(radii[1])+", rout="+convertToString(rout));
#endif
    }
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    CPotentialParamSH PP;
    PP.P=density;
    PP.theta_max= mysymmetry & ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= mysymmetry & ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = (mysymmetry & ST_PLANESYM ? 4 : 1) * (mysymmetry & ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    gsl_function F;
    F.function=&intSpline_r;
    F.params=&PP;
    vectord coefsInner, coefsOuter;
    radii.front()=SPLINE_MIN_RADIUS*radii[1];  // to prevent log divergence for gamma=2 potentials
    for(int l=0; l<=lmax; l+=lstep)
    {
        PP.sh_l=l;
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            PP.sh_m=m;
            // first precompute inner and outer density integrals at each radial grid point, summing contributions from each interval of radial grid
            coefsInner.assign(Ncoefs_radial+1, 0);
            coefsOuter.assign(Ncoefs_radial+1, 0);
            // loop over inner intervals
            double result, error;
            PP.sh_n = l+2;
            for(size_t c=0; c<Ncoefs_radial; c++)
            {   // the reason we give absolute error threshold is that for flat-density-core profiles the high-l,m coefs at small radii are extremely tiny and their exact calculation is impractically slow
                gsl_integration_qags(&F, radii[c], radii[c+1], EPSABS_POTENTIAL_INT, EPSREL_POTENTIAL_INT, 1000, w, &result, &error);
                coefsInner[c+1] = result + coefsInner[c];
            }
            // loop over outer intervals, starting from infinity backwards
            PP.sh_n = 1-l;
            gsl_integration_qagiu(&F, radii.back(), EPSABS_POTENTIAL_INT, EPSREL_POTENTIAL_INT, 1000, w, &result, &error);
            coefsOuter.back() = result;
            for(size_t c=Ncoefs_radial; c>static_cast<size_t>(l==0 ? 0 : 1); c--)
            {
                gsl_integration_qags(&F, radii[c-1], radii[c], EPSREL_POTENTIAL_INT*fabs(coefsOuter[c]), EPSREL_POTENTIAL_INT, 1000, w, &result, &error);
                coefsOuter[c-1] = result + coefsOuter[c];
            }
            // now compute the coefs of potential expansion themselves
            for(size_t c=0; c<=Ncoefs_radial; c++)
            {
                coefsArray[c][l*(l+1) + m] = ((c>0 ? coefsInner[c]*pow(radii[c], -l-1.0) : 0) + coefsOuter[c]*pow(radii[c], l*1.0)) *
                    multfactor* -4*M_PI/(2*l+1) * (m==0 ? 1 : M_SQRT2);
            }
#ifdef DEBUGPRINT
            my_message(FUNCNAME, "l="+convertToString(l)+",m="+convertToString(m));
#endif
        }
    }
    radii.front()=0;
    gsl_integration_workspace_free (w);
    initSpline(radii, coefsArray);
}

/// \cond INTERNAL_DOCS
// auxiliary structure to sort particles in radius
struct CParticle{
    float r;   // float not double, to save memory
    size_t index;
    CParticle() : r(0), index(0) {}
    CParticle(const std::pair< CPosVelPoint<double>, double> &src, size_t srcindex) :
        r(static_cast<float>(sqrt(pow_2(src.first.Pos[0])+pow_2(src.first.Pos[1])+pow_2(src.first.Pos[2])))),
        index(srcindex) {};
    inline bool operator<(CParticle p) const
    { return (r<p.r); }
};
/// \endcond

template<typename NumT>
void CPotentialSpline::computeCoefsFromPoints(const CPointMassSet<NumT> &srcpoints, 
    const vectord* srcradii, vectord* outradii, std::vector< vectord > *outcoefs)
{
    assert(outcoefs!=NULL);
    bool initUserRadii = srcradii!=NULL;   // if true, then it was called in the context of SHgrid Schwarzschild model
    std::vector< vectord > *CoefsInner=NULL, *CoefsOuter=NULL;  // coefficients of expansion at each point
    try{
        if(initUserRadii) 
        {
            outcoefs->resize(srcradii->size());
            for(size_t i=0; i<outcoefs->size(); i++)
                outcoefs->at(i).assign(pow_2(Ncoefs_angular+1), 0);
        }
        double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR-1];
        size_t npoints=srcpoints.size();
        if(npoints==0) return;
        std::vector<CParticle> points;
        points.reserve(npoints);
        size_t hadNegativeMass=0;
        for(size_t i=0; i<npoints; i++)
        {
            if(srcpoints[i].second>0)  // don't consider zero-mass points
                points.push_back(CParticle(srcpoints[i], i));
            if(srcpoints[i].second<0) hadNegativeMass++;
        }
        if(hadNegativeMass>0)   // don't fail entirely, but display an error message
            my_error(FUNCNAME,
                convertToString(hadNegativeMass)+" particle(s) with negative mass are discarded");
        std::sort(points.begin(), points.end());
        npoints=points.size();   // may be smaller than before if we throw out zero masses
        if(npoints<=Ncoefs_radial*10)
        {
            if(!initUserRadii) 
                my_error(FUNCNAME, 
                    "Number of points "+convertToString(npoints)+" is too small");
            return;
        }
        // having sorted particles in radius, may now initialize coefs
        if(!initUserRadii)  // will need to output coefs at each particle's radius
        {
            outradii->resize(npoints);
            for(size_t i=0; i<npoints; i++)
                outradii->at(i)=points[i].r;
            outcoefs->resize(pow_2(Ncoefs_angular+1));   // note that array indexing is swapped w.r.t. coefsArray - to save memory by not initializing unnecessary coefs
            CoefsOuter = outcoefs;   // re-use memory allocated for target coefs array  for the intermediate calculations
        }
        else 
            CoefsOuter = new std::vector< vectord> (pow_2(Ncoefs_angular+1));
        CoefsInner = new std::vector< vectord> (pow_2(Ncoefs_angular+1)); 
        for(int l=0; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
                CoefsOuter->at(l*(l+1)+m).assign(npoints, 0);  // reserve memory only for those coefs that will be used
        for(int l=0; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
                CoefsInner->at(l*(l+1)+m).assign(npoints, 0);  // yes do it separately from the above, to allow contiguous block of memory to be freed after deleting CoefsInner
 
        double addMbh=0;       // count particles at r=0; in principle, there shouldn't be any, but in case there are, add their mass to the central black hole.
        size_t addNpoints=0;   // note: this assignment of r=0 particles to the black hole will not be reflected on the computed and stored coefficients
        // initialize SH expansion coefs at each point's location
        for(size_t i=0; i<npoints; i++)
        {
            const CPosVelPoint<double> &point = srcpoints[ points[i].index ].first;
            double mass=srcpoints[ points[i].index ].second;
            double costheta = points[i].r>0 ?   // don't use points[i].r because it's float while Pos[] is double -- may end up having cos>1...
                point.Pos[2]/sqrt(pow_2(point.Pos[0])+pow_2(point.Pos[1])+pow_2(point.Pos[2])) : 0;
            if(points[i].r==0) {
                addNpoints++;
                addMbh+=mass;
                mass=0;
            }
            double phi = atan2(point.Pos[1], point.Pos[0]); 
            for(int m=0; m<=lmax; m+=mstep)
                gsl_sf_legendre_sphPlm_array(lmax, m, costheta, legendre_array[m]);
            for(int l=0; l<=lmax; l+=lstep)
                for(int m=l*mmin; m<=l*mmax; m+=mstep)
                {
                    int coefind=l*(l+1)+m;
                    int absm=abs(m);  // negative m correspond to sine, positive - to cosine
                    double mult= -sqrt(4*M_PI)/(2*l+1) * (m==0 ? 1 : M_SQRT2) * mass *
                        legendre_array[absm][l-absm] * (m>=0 ? cos(m*phi) : sin(-m*phi));
                    CoefsOuter->at(coefind)[i] = points[i].r>0 ? mult * pow(points[i].r, -(1+l)) : 0;
                    CoefsInner->at(coefind)[i] = mult * pow(points[i].r, l);
                }
        }
        // sum inner coefs interior and outer coefs exterior to each point's location
        for(int l=0; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int coefind=l*(l+1)+m;
                for(size_t i=1; i<npoints; i++)
                    CoefsInner->at(coefind)[i] += CoefsInner->at(coefind)[i-1];
                for(size_t i=npoints-1; i>0; i--)
                    CoefsOuter->at(coefind)[i-1] += CoefsOuter->at(coefind)[i];
            }
        // initialize potential expansion coefs by multiplying inner and outer coefs by r^(-1-l) and r^l, correspondingly
        if(initUserRadii)   // if using manually supplied srcradii, output coefficients only at these radii
        {
            size_t indPoint=0;
            for(size_t indGrid=0; indGrid<srcradii->size(); indGrid++)
            {   // find indPoint so that points[indPoint-1].r <= srcRadii[indGrid] < points[indPoint].r
                double radGrid=srcradii->at(indGrid);
                while(indPoint<npoints && points[indPoint].r<=radGrid) indPoint++;
                for(int l=0; l<=lmax; l+=lstep)
                    for(int m=l*mmin; m<=l*mmax; m+=mstep)
                    {
                        int coefind=l*(l+1)+m;
                        outcoefs->at(indGrid)[coefind] = 
                            (indPoint>0 ? CoefsInner->at(coefind)[indPoint-1] * pow(radGrid, -(1+l)) : 0) + 
                            (indPoint<npoints ? CoefsOuter->at(coefind)[indPoint] * pow(radGrid, l) : 0);
                    }
            }
        } else {   // compute coefs at every particle radius
            for(size_t i=0; i<npoints; i++)
            {
                for(int l=0; l<=lmax; l+=lstep)
                    for(int m=l*mmin; m<=l*mmax; m+=mstep)
                    {
                        int coefind=l*(l+1)+m;
                        outcoefs->at(coefind)[i] = 
                            (i>0 ? CoefsInner->at(coefind)[i-1] * pow(points[i].r, -(1+l)) : 0) + 
                            (i<npoints-1 ? CoefsOuter->at(coefind)[i+1] * pow(points[i].r, l) : 0);
                    }
            }
        }
        if(CoefsOuter!=NULL && initUserRadii) delete CoefsOuter;   // if !initUserRadii, this array actually holds the output coefs so it should not be deleted
        if(CoefsInner!=NULL) delete CoefsInner;
    }
    catch(...)   // there could well be an out-of-memory exception, but for the sake of generality we catch everything including an asteroid hitting earth
    {
        outcoefs->clear();  // marks an error
        if(CoefsInner!=NULL && initUserRadii) delete CoefsInner;   // in any case we need to free temporary arrays
        if(CoefsOuter!=NULL) delete CoefsOuter;
    }
}

/// \cond INTERNAL_DOCS
// auxiliary functions to find outer slope and 1st order correction to inner power-law slope for potential
struct CGammaParam{
    double r1,r2,r3,K,K2,K3;
};
double findGammaOut(double y, void* params)
{
    CGammaParam* par=static_cast<CGammaParam*>(params);
    return( pow(par->r2, 3-y) - pow(par->r1, 3-y))/( pow(par->r3, 3-y) - pow(par->r2, 3-y)) - par->K;
}
double findBcorrIn(double B, void* params)
{
    CGammaParam* par=static_cast<CGammaParam*>(params);
    return (par->K2 - log( (1-B*par->r2)/(1-B*par->r1) ))*log(par->r3/par->r1) - (par->K3 - log( (1-B*par->r3)/(1-B*par->r1) ))*log(par->r2/par->r1);
}
/// \endcond

template<typename NumT>
void CPotentialSpline::prepareCoefsDiscrete(const CPointMassSet<NumT> &srcpoints, double smoothfactor, const vectord *userradii)
{
  try{

    std::vector<double> pointRadii;      // radii of each point and coefficients of expansion at each point
    std::vector< vectord > pointCoefs;   // note that array indexing is swapped w.r.t. coefsArray - to save memory by not initializing unnecessary coefs
    computeCoefsFromPoints(srcpoints, NULL, &pointRadii, &pointCoefs);
    if(pointCoefs.size()==0) {
        my_error(FUNCNAME, "Not enough memory in computing SH coefs");
        initDefault();
        return;
    }
    size_t npoints=pointRadii.size();
    vectord radii(Ncoefs_radial+1);                         // radii of grid nodes to pass to initspline routine
    std::vector< vectord > coefsArray(Ncoefs_radial+1);     // SHE coefficients to pass to initspline routine
    for(size_t i=0; i<=Ncoefs_radial; i++)
        coefsArray[i].assign(pow_2(Ncoefs_angular+1), 0);

    // choose the radial grid parameters: innermost cell contains minBinMass and outermost radial node should encompass cutoffMass
    // spline definition region extends up to outerRadius which is ~several times larger than outermost radial node, however coefficients at that radius are not used in the potential computation later
    const size_t minBinPoints=10;
    bool initUserRadii = (userradii!=NULL && userradii->size()==Ncoefs_radial+1 && userradii->front()==0 && userradii->at(1)>= pointRadii[minBinPoints]);
    if(userradii!=NULL && !initUserRadii)  // something went wrong with manually supplied radii, initializing by default
        my_error(FUNCNAME, "Warning, invalid radial grid (" +
            (userradii->size()>0 ? convertToString(userradii->at(1))+":"+convertToString(userradii->back()) : "empty") +
            ") provided to constructor of SHGrid Schwarzschild model");
    if(initUserRadii)
        radii= *userradii;
    else
    {
        size_t npointsMargin    = static_cast<size_t>(sqrt(npoints*0.1));
        size_t npointsInnerGrid = std::max<size_t>(minBinPoints, npointsMargin);    // number of points within 1st grid radius
        size_t npointsOuterGrid = npoints - std::max<size_t>(minBinPoints, npointsMargin);   // number of points within outermost grid radius
        double innerBinRadius = pointRadii[npointsInnerGrid];
        double outerBinRadius = pointRadii[npointsOuterGrid];
        createNonuniformGrid(Ncoefs_radial+1, innerBinRadius, outerBinRadius, true, &radii);
    }
    // find index of the inner- and outermost points which are used in b-spline fitting
    size_t npointsInnerSpline = 0;
    while(pointRadii[npointsInnerSpline]<radii[1]) npointsInnerSpline++;
    npointsInnerSpline = std::min<size_t>(npointsInnerSpline-2, std::max<size_t>(minBinPoints, npointsInnerSpline/2));
    // index of last point used in b-spline fitting (it is beyond outer grid radius, since b-spline definition region is larger than the range of radii for which the spline approximation will eventually be constructed)
    double outerRadiusSpline = pow_2(radii[Ncoefs_radial])/radii[Ncoefs_radial-1];  // roughly equally logarithmically spaced from the last two points
    size_t npointsOuterSpline = npoints-1;
    while(pointRadii[npointsOuterSpline]>outerRadiusSpline) npointsOuterSpline--;
    //!!!FIXME: what happens if the outermost pointRadius is < outerBinRadius ?

    size_t numPointsUsed = npointsOuterSpline-npointsInnerSpline;  // outer and inner points are ignored
    size_t numBSplineKnots = Ncoefs_radial+2;  // including zero and outermost point; only interior nodes are actually used for computing best-fit coefs (except l=0 coef, for which r=0 is also used)
    vectord scaledPointRadii(numPointsUsed), scaledPointCoefs(numPointsUsed);  // transformed x- and y- values of original data points which will be approximated by a spline regression
    vectord scaledKnotRadii(numBSplineKnots), scaledSplineValues;              // transformed x- and y- values of regression spline knots

    // open block so that temp variable "appr" will be destroyed upon closing this block
    {
        // first construct spline for zero-order term (radial dependence)
        potcenter=pointCoefs[0][0];  // value of potential at origin (times 1/2\sqrt{\pi} ?)
        for(size_t p=1; p<=Ncoefs_radial; p++)
            scaledKnotRadii[p] = log(radii[p]);
        scaledKnotRadii[Ncoefs_radial+1] = log(outerRadiusSpline);
        scaledKnotRadii[0] = log(pointRadii[npointsInnerSpline]);
        for(size_t i=0; i<numPointsUsed; i++)
        {
            scaledPointRadii[i] = log(pointRadii[i+npointsInnerSpline]);
            scaledPointCoefs[i] = log(1/(1/potcenter - 1/pointCoefs[0][i+npointsInnerSpline]));
        }
        CSplineApprox appr(scaledPointRadii, scaledKnotRadii);
        if(appr.status()==CSplineApprox::AS_SINGULAR)
            my_message(FUNCNAME, 
                "Warning, in Spline potential initialization: singular matrix for least-square fitting; fallback to a slow algorithm");
        if(appr.status()==CSplineApprox::AS_NOMEM)
        {
            my_error(FUNCNAME, 
                "Not enough memory in performing spline fitting to potential coefficients");
            // in principle we could just pick up coefficients at radii corresponding to grid points, 
            // instead of creating a smoother curve, but this doesn't make much sense anyway - 
            // it's better to reduce the number of input points
            initDefault();
            return;
        }
        appr.loadyvalues(scaledPointCoefs);
        CSplineApprox::APPROXSTATUS status = appr.fitData(0, &scaledSplineValues);
        if(status != CSplineApprox::AS_OK && status != CSplineApprox::AS_SINGULAR)
            scaledSplineValues.assign(Ncoefs_radial+1, 0);   // in case of error fallback to zero values
        // now store fitted values in coefsArray to pass to initspline routine
        coefsArray[0][0] = potcenter;
        for(size_t c=1; c<=Ncoefs_radial; c++)
            coefsArray[c][0] = -1./(exp(-scaledSplineValues[c])-1/potcenter);
    }
    if(lmax>0) {  // construct splines for all l>0 spherical-harmonic terms separately
        // first estimate the asymptotic power-law slope of coefficients at r=0 and r=infinity
        double gammaInner = 2-log((coefsArray[1][0]-potcenter)/(coefsArray[2][0]-potcenter))/log(radii[1]/radii[2]);
        if(gammaInner<0) gammaInner=0; 
        if(gammaInner>2) gammaInner=2;
        // this was the estimate of density slope. Now we need to convert it to the estimate of power-law slope of l>0 coefs
        gammaInner = 2.0-gammaInner;   // the same recipe as used later in initSpline
        double gammaOuter = -1.0;      // don't freak out, assume default value
#ifdef DEBUGPRINT
        my_message(FUNCNAME, 
            "Estimated slope: inner="+convertToString(gammaInner)+", outer="+convertToString(gammaOuter));
#endif
        // init x-coordinates from scaling transformation
        for(size_t p=0; p<=Ncoefs_radial; p++)
            scaledKnotRadii[p] = log(1+radii[p]);
        scaledKnotRadii[Ncoefs_radial+1] = log(1+outerRadiusSpline);
        for(size_t i=0; i<numPointsUsed; i++)
            scaledPointRadii[i] = log(1+pointRadii[i+npointsInnerSpline]);
        CSplineApprox appr(scaledPointRadii, scaledKnotRadii);
        if(appr.status()==CSplineApprox::AS_SINGULAR)
            my_message(FUNCNAME, 
                "Warning, in Spline potential initialization: singular matrix for least-square fitting; fallback to a slow algorithm without smoothing");
        // loop over l,m
        for(int l=lstep; l<=lmax; l+=lstep)
        {
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int coefind=l*(l+1) + m;
                // init matrix of values to fit
                for(size_t i=0; i<numPointsUsed; i++)
                    scaledPointCoefs[i] = pointCoefs[coefind][i+npointsInnerSpline]/pointCoefs[0][i+npointsInnerSpline];// *
                    //pow(std::min<double>(1.0, pointRadii[i+npointsInnerSpline]/radii[2]), gammaInner);    // a trick to downplay wild fluctuations at r->0
                appr.loadyvalues(scaledPointCoefs);
                double edf=0;  // equivalent number of free parameters in the fit; if it is ~2, fit is oversmoothed to death (i.e. to a linear regression, which means we should ignore it)
                CSplineApprox::APPROXSTATUS status = appr.fitDataOversmooth(smoothfactor, &scaledSplineValues, NULL, &edf);
                if((status != CSplineApprox::AS_OK && status != CSplineApprox::AS_SINGULAR) || edf<3.0)
                    scaledSplineValues.assign(Ncoefs_radial+1, 0);   // in case of error or an oversmoothed fit fallback to zero values
#ifdef DEBUGPRINT
                my_message(FUNCNAME, 
                    "l="+convertToString(l)+", m="+convertToString(m)+" - EDF="+convertToString(edf));
#endif
                // now store fitted values in coefsArray to pass to initspline routine
                coefsArray[0][coefind] = 0;  // unused
                for(size_t c=1; c<=Ncoefs_radial; c++)
                    coefsArray[c][coefind] = scaledSplineValues[c] * coefsArray[c][0];  // scale back (multiply by l=0,m=0 coefficient)
                // correction to avoid fluctuation at first and last grid radius
                if( coefsArray[1][coefind] * coefsArray[2][coefind] < 0 || coefsArray[1][coefind]/coefsArray[2][coefind] > pow(radii[1]/radii[2], gammaInner))
                    coefsArray[1][coefind] = coefsArray[2][coefind] * pow(radii[1]/radii[2], gammaInner);   // make the smooth curve drop to zero at least as fast as gammaInner'th power of radius
                if( coefsArray[Ncoefs_radial][coefind] * coefsArray[Ncoefs_radial-1][coefind] < 0 || 
                    coefsArray[Ncoefs_radial][coefind] / coefsArray[Ncoefs_radial-1][coefind] > pow(radii[Ncoefs_radial]/radii[Ncoefs_radial-1], gammaOuter))
                    coefsArray[Ncoefs_radial][coefind] = coefsArray[Ncoefs_radial-1][coefind] * pow(radii[Ncoefs_radial]/radii[Ncoefs_radial-1], gammaOuter);
            }
        }
    }
    initSpline(radii, coefsArray);

  }  //< try
  catch(const std::bad_alloc&) {
    my_error(FUNCNAME, "Not enough memory in initializing Spline potential");
    initDefault();
  }
}

void CPotentialSpline::checkSymmetry(const std::vector< vectord > &coefsArray)
{ 
    SYMMETRYTYPE sym=ST_SPHERICAL;  // too optimistic:))
    const double MINCOEF=1e-8;   // if ALL coefs of a certain subset of indices are below this value, assume some symmetry
    for(size_t n=0; n<=Ncoefs_radial; n++)
    {
        for(int l=0; l<=(int)Ncoefs_angular; l++)
            for(int m=-l; m<=l; m++)
                if(fabs(coefsArray[n][l*(l+1)+m])>MINCOEF) 
                {   // nonzero coef.: check if that breaks any symmetry
                    if(l%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_REFLECTION);
                    if(m<0 || m%2==1)  sym = (SYMMETRYTYPE)(sym & ~ST_PLANESYM);
                    if(m!=0) sym = (SYMMETRYTYPE)(sym & ~ST_ZROTSYM);
                    if(l>0) sym = (SYMMETRYTYPE)(sym & ~ST_SPHSYM);
                }
    }
    mysymmetry = sym; 
    assignlmrange();
}

void CPotentialSpline::initSpline(const vectord &_radii, const std::vector< vectord > &_coefsArray)
{
    if(_radii[0]!=0) 
        my_error(FUNCNAME, "Warning, in Spline potential: radii[0]!=0 (assumed to be zero)");
    if(_radii.size()!=Ncoefs_radial+1 || _coefsArray.size()!=Ncoefs_radial+1) {
        my_error(FUNCNAME, "Warning, in Spline potential: Ncoefs_radial+1 != array size");  // likely to get segmentation fault shortly soon
        initDefault();
        return;
    }
    potcenter=_coefsArray[0][0];    // here assumed that gridradii[0]=0, i.e. innermost node is at origin
    // safety measure: if zero-order coefs are the same as potcenter for r>0, skip these elements
    vectord newRadii;
    std::vector< vectord > newCoefsArray;
    size_t nskip=0;
    while(nskip+1<_coefsArray.size() && _coefsArray[nskip+1][0]==potcenter) nskip++;
    if(nskip>0) {  // skip some elements
        newRadii=_radii;
        newRadii.erase(newRadii.begin()+1, newRadii.begin()+nskip+1);
        Ncoefs_radial=newRadii.size()-1;
        newCoefsArray=_coefsArray;
        newCoefsArray.erase(newCoefsArray.begin()+1, newCoefsArray.begin()+nskip+1);
        if(newRadii.size()<5) {
            my_error(FUNCNAME, "Warning, in Spline potential: too few radial points");
            initDefault();
            return;
        }
    }
    const vectord &radii = nskip==0 ? _radii : newRadii;
    const std::vector< vectord > &coefsArray = nskip==0 ? _coefsArray : newCoefsArray;
    checkSymmetry(coefsArray);   // assign nontrivial symmetry class if some of coefs are equal or close to zero
    gridradii=radii;    // copy real radii
    minr=gridradii[1];
    maxr=gridradii.back();
    std::vector<double> spnodes(Ncoefs_radial);  // scaled radii
    std::vector<double> spvalues(Ncoefs_radial+2);
    splines.assign(pow_2(Ncoefs_angular+1), (gsl_spline*)NULL);
    slopein. assign(pow_2(Ncoefs_angular+1), 1.);
    slopeout.assign(pow_2(Ncoefs_angular+1), -1.);

    // estimate outermost slope  (needed for accurate extrapolation beyond last grid point)
    gsl_function F;
    CGammaParam PP;
    PP.r1=radii[Ncoefs_radial];
    PP.r2=radii[Ncoefs_radial-1];
    PP.r3=radii[Ncoefs_radial-2];
    PP.K = (coefsArray[Ncoefs_radial][0]*PP.r1 - coefsArray[Ncoefs_radial-1][0]*PP.r2)/(coefsArray[Ncoefs_radial-1][0]*PP.r2 - coefsArray[Ncoefs_radial-2][0]*PP.r3);
    F.function=&findGammaOut;
    F.params=&PP;
    if(gsl_finite(PP.K)) {
        gammaout = findRoot(&F, 3.01, 10., 4.0);
        coefout = (1 - coefsArray[Ncoefs_radial-1][0]*PP.r2/(coefsArray[Ncoefs_radial][0]*PP.r1)) / (pow(PP.r2/PP.r1, 3-gammaout) - 1);
        if(coefout<0) coefout=0;
    } else {
        gammaout=10.0;
        coefout=0;
    }

    // estimate innermost slope with 1st order correction for non-zero radii 
    F.function=&findBcorrIn;
    PP.r1=radii[1];
    PP.r2=radii[2];
    PP.r3=radii[3];
    PP.K2=log((coefsArray[2][0]-potcenter)/(coefsArray[1][0]-potcenter));
    PP.K3=log((coefsArray[3][0]-potcenter)/(coefsArray[1][0]-potcenter));
    double B = findRoot(&F, 0, 0.9/radii[3], 0);
    gammain = 2. - (log((coefsArray[2][0]-potcenter)/(coefsArray[1][0]-potcenter)) - log((1-B*radii[2])/(1-B*radii[1])))/log(radii[2]/radii[1]);
    double gammainuncorr = 2. - log((coefsArray[2][0]-potcenter)/(coefsArray[1][0]-potcenter)) /log(radii[2]/radii[1]);
    if(gammain>=1) gammain=gammainuncorr;
    if(gammain<0) gammain=0; 
    if(gammain>2) gammain=2;
    coefin = (1-coefsArray[1][0]/potcenter)/pow(radii[1],2-gammain);
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "gammain="+convertToString(gammain)+
        " ("+convertToString(gammainuncorr)+");  gammaout="+convertToString(gammaout));
#endif
    potmax=coefsArray.back()[0];
    potminr=coefsArray[1][0];
    // first init l=0 spline which has radial scaling "log(r)" and nontrivial transformation 1/(1/phi-1/phi0)
    for(size_t i=0; i<Ncoefs_radial; i++)
    {
        spnodes[i] = log(gridradii[i+1]);
        spvalues[i]= log(1/ (1/potcenter - 1/coefsArray[i+1][0]));
    }
    spvalues[Ncoefs_radial] = -(2-gammain)*potcenter/coefsArray[1][0];   // derivative at leftmost node
    spvalues[Ncoefs_radial+1] = - (1+coefout*(3-gammaout))/(1 - potmax/potcenter);  // derivative at rightmost node
    gsl_spline *spline = gsl_spline_alloc (gsl_interp_ccspline, Ncoefs_radial);
    gsl_spline_init (spline, &(spnodes.front()), &(spvalues.front()), Ncoefs_radial);
    splines[0]=spline;
    coef0(maxr, NULL, NULL, &der2out);
    // next init all higher-order splines which have radial scaling log(a+r) and value scaled to l=0,m=0 coefficient
    const double ascale=1.0;
    for(size_t i=0; i<Ncoefs_radial; i++)
        spnodes[i]=log(ascale+gridradii[i+1]);
    double C00val, C00der;
    coef0(minr, &C00val, &C00der, NULL);
    for(int l=lstep; l<=lmax; l+=lstep)
    {
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            int coefind=l*(l+1)+m;
            for(size_t i=0; i<Ncoefs_radial; i++)
                spvalues[i] = coefsArray[i+1][coefind]/coefsArray[i+1][0];
            slopein[coefind] = log(coefsArray[2][coefind]/coefsArray[1][coefind]) / log(gridradii[2]/gridradii[1]);   // estimate power-law slope of Clm(r) at r->0
            if(gsl_isnan(slopein[coefind])) slopein[coefind]=1.0;  // default
            slopein[coefind] = std::max<double>(slopein[coefind], std::min<double>(l, 2-gammain));  // the asymptotic power-law behaviour of the coefficient expected for power-law density profile
            spvalues[Ncoefs_radial] = spvalues[0] * (1+ascale/minr) * (slopein[coefind] - minr*C00der/C00val);   // derivative at innermost node
            slopeout[coefind] = log(coefsArray[Ncoefs_radial][coefind]/coefsArray[Ncoefs_radial-1][coefind]) / log(gridradii[Ncoefs_radial]/gridradii[Ncoefs_radial-1]) + 1;   // estimate slope of Clm(r)/C00(r) at r->infinity (+1 is added because C00(r) ~ 1/r at large r)
            if(gsl_isnan(slopeout[coefind])) slopeout[coefind]=-1.0;  // default
            slopeout[coefind] = std::min<double>(slopeout[coefind], std::max<double>(-l, 3-gammaout));
            spvalues[Ncoefs_radial+1] = spvalues[Ncoefs_radial-1] * (1+ascale/maxr) * slopeout[coefind];   // derivative at outermost node
            spline = gsl_spline_alloc (gsl_interp_ccspline, Ncoefs_radial);
            gsl_spline_init (spline, &(spnodes.front()), &(spvalues.front()), Ncoefs_radial);
            splines[coefind]=spline;
#ifdef DEBUGPRINT
            my_message(FUNCNAME, "l="+convertToString(l)+", m="+convertToString(m)+
                " - inner="+convertToString(slopein[coefind])+", outer="+convertToString(slopeout[coefind]));
#endif
        }
    }
    bool densityNonzero = checkDensityNonzero();
    bool massMonotonic  = checkMassMonotonic();
    if(!massMonotonic || !densityNonzero) 
        my_message(FUNCNAME, "Warning, " + 
        std::string(!massMonotonic ? "mass does not monotonically increase with radius" : "") +
        std::string(!massMonotonic && !densityNonzero ? " and " : "") + 
        std::string(!densityNonzero ? "density drops to zero at a finite radius" : "") + "!");
}

void CPotentialSpline::getCoefs(vectord *radii, std::vector< vectord > *coefsArray, bool useNodes) const
{
    if(radii==NULL || coefsArray==NULL) return;
    if(useNodes)
    {
        radii->resize(Ncoefs_radial+1);
        for(size_t i=0; i<=Ncoefs_radial; i++)
            (*radii)[i] = gridradii[i];
    }
    size_t numrad=radii->size();
    coefsArray->resize(numrad);
    for(size_t i=0; i<numrad; i++)
    {
        double rad=(*radii)[i];
        double xi=log(1+rad);
        double Coef00;
        coef0(rad, &Coef00, NULL, NULL);
        (*coefsArray)[i].assign(pow_2(Ncoefs_angular+1), 0);
        (*coefsArray)[i][0] = Coef00;
        for(int l=lstep; l<=lmax; l+=lstep)
            for(int m=l*mmin; m<=l*mmax; m+=mstep)
            {
                int coefind=l*(l+1)+m;
                coeflm(coefind, rad, xi, &((*coefsArray)[i][l*(l+1)+m]), NULL, NULL, Coef00);
            }
    }
}

void CPotentialSpline::coeflm(size_t lm, double r, double xi, double *val, double *der, double *der2, double c0val, double c0der, double c0der2) const  // works only for l2>0
{
    const double ascale=1.0;
    double cval=0, cder=0, cder2=0;   // value and derivatives of \tilde Clm = Clm(r)/C00(r)
    if(r < maxr)
    {
        if(r > minr)  // normal interpolation
        {
#ifdef NO_OPTIMIZATION
            cval= gsl_spline_eval(splines[lm], xi, NULL);
            if(der!=NULL){ cder = gsl_spline_eval_deriv(splines[lm], xi, NULL)/(r+ascale);    // to save time, evaluate these only if needed
            if(der2!=NULL) cder2=(gsl_spline_eval_deriv2(splines[lm], xi, NULL)/(r+ascale) - cder)/(r+ascale); } // obviously, if der2!=NULL then assuming that der!=NULL
#else
            // use optimized routines to compute value and derivatives at once
            if(der==NULL) {
                cspline_eval_all(splines[lm], xi, &cval, NULL, NULL);
            } else if(der2==NULL) {
                cspline_eval_all(splines[lm], xi, &cval, &cder, NULL);
                cder /= r+ascale;
            } else {
                cspline_eval_all(splines[lm], xi, &cval, &cder, &cder2);
                cder /= r+ascale;
                cder2 = (cder2/(r+ascale) - cder)/(r+ascale);
            }
#endif
        }
        else  // power-law asymptotics at r<minr
        {
            cval = splines[lm]->y[0] * potminr;
            if(val!=NULL)  *val = cval * pow(r/minr, slopein[lm]);
            if(der!=NULL){ *der = (*val) * slopein[lm]/r;
            if(der2!=NULL) *der2= (*der) * (slopein[lm]-1)/r; }
            return;   // for r<minr, Clm is not scaled by C00
        }
    }
    else  // power-law asymptotics for r>maxr
    {
        cval = splines[lm]->y[splines[lm]->interp->size-1] * pow(r/maxr, slopeout[lm]);
        cder = cval * slopeout[lm]/r;
        cder2= cder * (slopeout[lm]-1)/r;
        double ximax=splines[lm]->x[splines[lm]->interp->size-1];
        double der2left = (gsl_spline_eval_deriv2(splines[lm], ximax, NULL) - gsl_spline_eval_deriv(splines[lm], ximax, NULL))/pow_2(r+ascale);
        double der2right = splines[lm]->y[splines[lm]->interp->size-1]*slopeout[lm]*(slopeout[lm]-1)/pow_2(maxr);
        double acorr = (der2left-der2right)*0.5;
        double slopecorr = slopeout[lm]-4;
        double powcorr = pow(r/maxr, slopecorr);
        cval += acorr*powcorr*pow_2(r-maxr);
        cder += acorr*powcorr*(r-maxr)*(2 + slopecorr*(r-maxr)/r);
        cder2+= acorr*powcorr*(2 + 4*slopecorr*(r-maxr)/r + pow_2(1-maxr/r)*slopecorr*(slopecorr-1));
    }
    // scale by C00
    if(val!=NULL)  *val = cval*c0val;
    if(der!=NULL)  *der = cder*c0val + cval*c0der;
    if(der2!=NULL) *der2= cder2*c0val + 2*cder*c0der + cval*c0der2;
}

void CPotentialSpline::coef0(double r, double *val, double *der, double *der2) const  // works only for l=0
{
    if(r<=maxr)
    {
        double logr=log(r);
        double sval, sder, sder2;
        if(r<minr)
        {
            double ratio = 1-coefin*pow(r, 2-gammain);  // C00(r)/C00(0)
            sval = log(-potcenter/coefin) - (2-gammain)*logr + log(ratio);
            sder = -(2-gammain)/ratio;
            sder2= -pow_2(sder)*(1-ratio);
        }
        else
        {
#ifdef NO_OPTIMIZATION
            sval = gsl_spline_eval(splines[0], logr, NULL);
            sder = gsl_spline_eval_deriv(splines[0], logr, NULL);
            sder2= gsl_spline_eval_deriv2(splines[0], logr, NULL);
#else
            cspline_eval_all(splines[0], logr, &sval, &sder, &sder2);
#endif
        }
        double sexp = (r>0? exp(-sval) : 0);
        double vval = 1./(sexp-1/potcenter);
        if(val!=NULL)  *val = -vval;
        if(der!=NULL)  *der = -vval*vval*sexp/r * sder;  // this would not work for r=0 anyway...
        if(der2!=NULL) *der2= -pow_2(vval/r)*sexp * (sder2 - sder + sder*sder*(2*vval*sexp-1) );
    }
    else
    {
        double r_over_maxr_g=pow(r/maxr, 3-gammaout);
        double der2right = -2*potmax*maxr/pow(r,3) * (1 - coefout*(r_over_maxr_g*(gammaout-1)*(gammaout-2)/2 - 1));
        double slopecorr = -gammaout-4;
        double acorr = 0*pow(r/maxr, slopecorr) * (der2out-der2right)*0.5;   // apparently unused, but why?
        if(val!=NULL)  *val = -(-potmax*maxr/r * (1 - coefout*(r_over_maxr_g-1))  + acorr*pow_2(r-maxr) );
        if(der!=NULL)  *der = -( potmax*maxr/r/r * (1 - coefout*(r_over_maxr_g*(gammaout-2) - 1))  + acorr*(r-maxr)*(2 + slopecorr*(r-maxr)/r) );
        if(der2!=NULL) *der2= -(der2right  + acorr*(2 + 4*slopecorr*(r-maxr)/r + pow_2(1-maxr/r)*slopecorr*(slopecorr-1)) );
    }
}

double CPotentialSpline::Mass(const double r) const
{
    if(r<=0) return 0;
    double der;
    coef0(r, NULL, &der, NULL);
    return der * r*r;   // d Phi(r)/d r = G M(r) /r^2
}

void CPotentialSpline::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    const double ascale=1.0;
    double xi = log(r+ascale);
    double val00, der00, der200;
    coef0(r, &val00, &der00, &der200);  // compute value and two derivatives of l=0,m=0 spline
    if(coefsF)    coefsF[0]    = val00;
    if(coefsdFdr) coefsdFdr[0] = der00;
    if(coefsd2Fdr2) coefsd2Fdr2[0] = der200;
    for(int l=lstep; l<=lmax; l+=lstep)
    {
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            int coefind=l*(l+1)+m;
            coeflm(coefind, r, xi, 
                coefsF!=NULL ? coefsF+coefind : NULL, 
                coefsdFdr!=NULL ? coefsdFdr+coefind : NULL, 
                coefsd2Fdr2!=NULL ? coefsd2Fdr2+coefind : NULL, 
                val00, der00, der200);
        }
    }
}

#ifdef HAVE_INTERP2D
//-------- CylSpline section --------//
// faster evaluation of hypergeometric function by spline approximation
#define HYPERGEOM_APPROX
#ifdef HYPERGEOM_APPROX
gsl_spline* init_spline_hypergeom(double m)
{
    const double ymin=log(1e-6), ymax=log(100.0), ystepinit=1;
    const size_t numpt=static_cast<size_t>((ymax-ymin)/ystepinit);
    vectord yval(numpt), fval(numpt);
    for(size_t i=0; i<numpt; i++) {
        double y=ymin+ystepinit*i;
        double x=pow_2(exp(y)+1);
        double val=x*(gsl_sf_hyperg_2F1(1+m/2, 0.5+m/2, 1.5+m, 1/x)-1);
        yval[i]=y;
        fval[i]=val;
    }
    size_t ind=1;
    while(ind<yval.size()-1) {
        // check second derivative
        double der2 = 
            ((fval[ind+1]-fval[ind])/(yval[ind+1]-yval[ind]) - 
            ( fval[ind]-fval[ind-1])/(yval[ind]-yval[ind-1])) / 
            ((yval[ind+1]-yval[ind-1]) * fval[ind]);
        if( yval[ind]-yval[ind-1]>0.1 && 
            der2*(yval[ind+1]-yval[ind])*pow_2(yval[ind]-yval[ind-1])*(yval[ind+1]-yval[ind-1])>1e-3) {  // add a point in between
            double yadd=(yval[ind]+yval[ind-1])/2;
            double x=pow_2(exp(yadd)+1);
            double val=x*(gsl_sf_hyperg_2F1(1+m/2, 0.5+m/2, 1.5+m, 1/x)-1);
            yval.insert(yval.begin()+ind, yadd);
            fval.insert(fval.begin()+ind, val); 
            //std::cerr << "m="<<m<<", y="<<yval[ind]<<", val="<<val<<", der2="<<der2overf<<"\n";
        }
        else ind++;
    }
    //for(size_t ind=0; ind<yval.size(); ind++) { 
        //std::cerr << yval[ind] << "\t" << fval[ind] << "\n";
    //}
    //std::cerr << yval.size() << " [points for m="<<m<<"]\n";
    gsl_spline* spl=gsl_spline_alloc(gsl_interp_cspline, yval.size());
    gsl_spline_init(spl, &(yval.front()), &(fval.front()), yval.size());
    return spl;
}
double my_hypergeom(double x, const gsl_spline* spl)
{
    /*if(x>=1) {  // shouldn't occur
        my_message("Invalid argument for hypergeometric function!");
        return 0;
    }*/
    return 1+x*gsl_spline_evalx(spl, log(1/sqrt(x)-1), NULL);
}
#endif

CPotentialDirect::CPotentialDirect(const CDensity* _density, size_t _mmax, ACCURACYMODE _accuracymode) : 
  density(_density), mysymmetry(_density->symmetry()), accuracymode(_accuracymode), points(NULL)
{
#ifdef HYPERGEOM_APPROX
    spl_hyperg.assign(_mmax+1, NULL);
    for(size_t m=0; m<=_mmax; m++)
        spl_hyperg[m]=init_spline_hypergeom(m-0.5);
#endif
    if((mysymmetry & ST_AXISYMMETRIC)==ST_AXISYMMETRIC)
        return;  // no further action necessary
    // otherwise prepare interpolating splines in (R,z) for Fourier expansion of density in angle phi
    int mmax=_mmax;
    splines.assign(mmax*2+1, static_cast<interp2d_spline*>(NULL));
    double Rmin=0.001, Rmax=1000;  // arbirary values, but if they are inappropriate, it only will slowdown the computation 
    // but not deteriorate its accuracy, because the interpolation is not used outside the grid
    double delta=0.05;  // relative difference between grid nodes = log(x[n+1]/x[n]) 
    size_t numNodes = static_cast<size_t>(log(Rmax/Rmin)/delta);
    vectord grid;
    createNonuniformGrid(numNodes, Rmin, Rmax, true, &grid);
    vectord gridz(2*grid.size()-1);
    for(size_t i=0; i<grid.size(); i++) {
        gridz[grid.size()-1-i] =-grid[i];
        gridz[grid.size()-1+i] = grid[i];
    }
    vectord values(grid.size()*gridz.size());
    bool zsymmetry = (density->symmetry()&ST_PLANESYM)==ST_PLANESYM;      // whether densities at z and -z are different
    int mmin = (density->symmetry() & ST_PLANESYM)==ST_PLANESYM ? 0 :-1;  // if triaxial symmetry, do not use sine terms which correspond to m<0
    int mstep= (density->symmetry() & ST_PLANESYM)==ST_PLANESYM ? 2 : 1;  // if triaxial symmetry, use only even m
    for(int m=mmax*mmin; m<=mmax; m+=mstep) {
        splines[mmax+m] = interp2d_spline_alloc(interp2d_bicubic, grid.size(), gridz.size());
        for(size_t iR=0; iR<grid.size(); iR++)
            for(size_t iz=0; iz<grid.size(); iz++) {
                double val=computeRho_m(grid[iR], grid[iz], m);
                if(gsl_isnan(val) || gsl_isinf(val)) {
                    if(iR==0 && iz==0)  // may have a singularity at origin, substitute the infinite density with something reasonable
                        val=std::max<double>(computeRho_m(grid[1], grid[0], m), computeRho_m(grid[0], grid[1], m));
                    else val=0;
                }
                values[INDEX_2D(iR, grid.size()-1+iz, grid.size(), gridz.size())] = val;
                if(!zsymmetry && iz>0) {
                    val=computeRho_m(grid[iR], -grid[iz], m);
                    if(gsl_isnan(val) || gsl_isinf(val)) val=0;  // don't let rubbish in
                }
                values[INDEX_2D(iR, grid.size()-1-iz, grid.size(), gridz.size())] = val;
            }
        interp2d_spline_init(splines[mmax+m], &(grid.front()), &(gridz.front()), &(values.front()), grid.size(), gridz.size());
    }
}

CPotentialDirect::CPotentialDirect(const CPointMassSet<double> *_points, size_t _mmax, SYMMETRYTYPE sym) :
  density(NULL), mysymmetry(sym), accuracymode(AM_FAST), points(_points) 
{
#ifdef HYPERGEOM_APPROX
    spl_hyperg.assign(_mmax+1, NULL);
    for(size_t m=0; m<=_mmax; m++)
        spl_hyperg[m]=init_spline_hypergeom(m-0.5);
#endif
};

CPotentialDirect::~CPotentialDirect()
{
    for(size_t m=0; m<splines.size(); m++)
        if(splines[m]!=NULL)
            interp2d_spline_free(splines[m]);
    for(size_t m=0; m<spl_hyperg.size(); m++)
        if(spl_hyperg[m]!=NULL)
            gsl_spline_free(spl_hyperg[m]);
}

double CPotentialDirect::totalMass() const
{
    if(density!=NULL) return density->totalMass();
    if(points!=NULL) {
        double mass=0;
        for(CPointMassSet<double>::Type::const_iterator pt=points->data.begin(); pt!=points->data.end(); pt++) 
            mass+=pt->second;
        return mass;
    }
    return 0;  // shouldn't reach here
}

double CPotentialDirect::Mass(const double r) const
{
    if(density!=NULL) return density->Mass(r);
    if(points!=NULL) {
        double mass=0;
        for(CPointMassSet<double>::Type::const_iterator pt=points->data.begin(); pt!=points->data.end(); pt++) 
            if(pow_2(pt->first.Pos[0])+pow_2(pt->first.Pos[1])+pow_2(pt->first.Pos[2]) <= pow_2(r))
                mass+=pt->second;
        return mass;
    }
    return 0;  // shouldn't reach here
}

double CPotentialDirect::Rho(double X, double Y, double Z, double) const
{
    if(density==NULL) return 0;  // not applicable in discrete point set mode
    if(splines.size()==0)  // no interpolation
        return density->Rho(X,Y,Z); 
    else {
        double R=sqrt(X*X+Y*Y);
        double phi=atan2(Y, X);
        double val=0;
        int mmax=splines.size()/2;
        for(int m=-mmax; m<=mmax; m++)
            val+=Rho_m(R, Z, m) * (m>=0 ? cos(m*phi) : sin(-m*phi));
        return std::max<double>(val, 0);
    }
}

double CPotentialDirect::Rho_m(double R, double z, int m) const
{
    if(splines.size()==0) {
        if(m==0 && density!=NULL) return density->Rho(R,0,z);
        else return 0;
    }
    size_t mmax=splines.size()/2;
    if(splines[mmax+m]==NULL) return 0;
    if( R>splines[mmax+m]->interp_object.xmax || (fabs(z)+R)<splines[mmax+m]->xarr[1]*2 || 
        z<splines[mmax+m]->interp_object.ymin || z>splines[mmax+m]->interp_object.ymax)
        return computeRho_m(R, z, m);  // outside interpolating grid -- compute directly by integration
    else 
        return interp2d_spline_eval(splines[mmax+m], R, z, NULL, NULL);
}

/// \cond INTERNAL_DOCS
struct CPotentialDirectParam {
    const CDensity* density;
    const CPotentialDirect* potential;
    double R, z;   // point at which the potential is evaluated
    int m;         // azimuthal harmonic (m>=0 correspond to cos(m*phi), m<0 -- to sin(|m|*phi)
    double R1, z1; // point at which the density is taken
    bool zfirst;   //
    double gammamult;
    gsl_integration_workspace* ws;
    gsl_spline* spl_hypergeom;
};
/// \endcond
double integrandPotentialDirectDensity(double phi, void* param)
{
    double mult=((CPotentialDirectParam*)param)->m==0 ? 1 : ((CPotentialDirectParam*)param)->m>0 ? 
        2*cos(((CPotentialDirectParam*)param)->m*phi) : 2*sin(-((CPotentialDirectParam*)param)->m*phi);
    return mult*((CPotentialDirectParam*)param)->density->Rho( 
        ((CPotentialDirectParam*)param)->R*cos(phi), ((CPotentialDirectParam*)param)->R*sin(phi), ((CPotentialDirectParam*)param)->z);
}

double CPotentialDirect::computeRho_m(double R, double z, int m) const
{
    gsl_function F;
    CPotentialDirectParam params;
    F.function=&integrandPotentialDirectDensity;
    F.params=&params;
    params.density=density;
    params.R=R;
    params.z=z;
    params.m=m;
    double result, error;
    size_t neval;
    double phimax=(density->symmetry() & ST_PLANESYM)==ST_PLANESYM ? M_PI_2 : 2*M_PI;
    gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result/phimax;
}

#define Q_HYPERG 1
/// Legendre function Q(m,x) expressed through Gauss hypergeometric function 
double LegendreQ(double m, double x, const gsl_spline* spl)
{
#if Q_HYPERG
#ifdef HYPERGEOM_APPROX
    double val=my_hypergeom(1/(x*x), spl);
#else
    double val= gsl_sf_hyperg_2F1(1+m/2, 0.5+m/2, 1.5+m, 1/(x*x));// : (1+(0.345+0.25*m)/pow_2(x));
#endif
    if(gsl_isnan(val)) val=1;  // not quite correct, but allows to continue computation
    val*=pow(2*x,-m-1);
    return val;
#else
    double mu=sqrt(2.0/(1+x));
    double m1,m0=mu*gsl_sf_ellint_Kcomp(mu, GSL_PREC_SINGLE);
    if(m==-0.5) return m0;
    if(m>=0.5) m1=x*m0-(1+x)*mu*gsl_sf_ellint_Ecomp(mu, GSL_PREC_SINGLE);
    if(m==0.5) return m1;
    // otherwise use recurrence relations from Cohl&Tohline1999
    double mcurr=2.0;
    do{
        double m2=4*(mcurr-1)/(2*mcurr-1)*x*m1 - (2*mcurr-3)/(2*mcurr-1)*m0;
        if(mcurr==m+0.5)
            return m2;
        m0=m1; m1=m2; mcurr+=1;
    } while(mcurr<100);  // avoid infinite loop
    return 0;  // shouldn't happen?
#endif
}
/// $\int_0^\infinity J_m(a*x)*J_m(b*x)*exp(-|c|*x) dx / (\sqrt(\pi)*\Gamma(m+1/2)/\Gamma(m+1))$
double IntBessel(double m, double a, double b, double c, const gsl_spline* spl, double eps2=0)
{
    if(fabs(a)<1e-10 || fabs(b)<1e-10)
        return m==0 ? 1/(M_PI*sqrt(a*a+b*b+c*c+eps2)) : 0;
    else
        return LegendreQ(m-0.5, (a*a+b*b+c*c+eps2)/(2*a*b), spl)/(M_PI*sqrt(a*b));
}

double integrandPotentialDirectVar2(double var, void* params)
{
    double R1,z1;
    if(((CPotentialDirectParam*)params)->zfirst) {
        R1=var/(1-var);
        z1=((CPotentialDirectParam*)params)->z1;
    } else {
        z1=var/(1-var);
        R1=((CPotentialDirectParam*)params)->R1;
    } 
    double val = ((CPotentialDirectParam*)params)->potential->Rho_m(R1, z1, ((CPotentialDirectParam*)params)->m) * 
        -2*M_PI*R1 * IntBessel( abs(((CPotentialDirectParam*)params)->m), 
        ((CPotentialDirectParam*)params)->R, R1, ((CPotentialDirectParam*)params)->z-z1, 
        ((CPotentialDirectParam*)params)->spl_hypergeom );
    z1*=-1;
    val += ((CPotentialDirectParam*)params)->potential->Rho_m(R1, z1, ((CPotentialDirectParam*)params)->m) * 
        -2*M_PI*R1 * IntBessel( abs(((CPotentialDirectParam*)params)->m), 
        ((CPotentialDirectParam*)params)->R, R1, ((CPotentialDirectParam*)params)->z-z1, 
        ((CPotentialDirectParam*)params)->spl_hypergeom );
    return val/pow_2(1-var);
}

double integrandPotentialDirectVar1(double var, void* params)
{
    gsl_function F;
    F.function=&integrandPotentialDirectVar2;
    F.params=params;
    if(((CPotentialDirectParam*)params)->zfirst) 
        ((CPotentialDirectParam*)params)->z1=var/(1-var);
    else
        ((CPotentialDirectParam*)params)->R1=var/(1-var);
    double result, errr;
    size_t neval;
    if(((CPotentialDirectParam*)params)->ws==NULL)
        gsl_integration_qng(&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &errr, &neval);
    else 
        gsl_integration_qags(&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, ((CPotentialDirectParam*)params)->ws, &result, &errr);
    return result/pow_2(1-var);
}

double CPotentialDirect::Phi_m(double R, double Z, int m) const
{
    CPotentialDirectParam params;
#ifdef HYPERGEOM_APPROX
    params.spl_hypergeom=spl_hyperg[abs(m)];
#endif
    if(density==NULL) {  // invoked in the discrete point set mode
        if(points==NULL) return 0;  // shouldn't normally happen
        double val=0;
        for(CPointMassSet<double>::Type::const_iterator pt=points->data.begin(); pt!=points->data.end(); pt++) {
            double R1=sqrt(pow_2(pt->first.Pos[0])+pow_2(pt->first.Pos[1]));
            double phi1=atan2(pt->first.Pos[1], pt->first.Pos[0]);
            double Z1=pt->first.Pos[2];
            double eps2=0;
            double val1=IntBessel(abs(m), R, R1, Z-Z1, params.spl_hypergeom,eps2);
            if((mysymmetry & ST_PLANESYM)==ST_PLANESYM)   // add symmetric contribution from -Z
                val1 = (val1 + IntBessel(abs(m), R, R1, Z+Z1, params.spl_hypergeom))/2;
            if(!gsl_isnan(val1) && gsl_finite(val1))
                val += pt->second * val1 /* R1*/ * (m==0 ? 1 : m>0 ? 2*cos(m*phi1) : 2*sin(-m*phi1) );
        }
#if Q_HYPERG
        val*=M_SQRTPI*gsl_sf_gamma(abs(m)+0.5)/gsl_sf_gamma(abs(m)+1);
#endif
        return -val;
    }
    // otherwise invoked in the smooth density profile mode
    int mmax=splines.size()/2;
    if(splines.size()>0 && splines[mmax+m]==NULL) return 0;  // using splines for m-components of density but it is identically zero at this m
    gsl_function F;
    F.function=&integrandPotentialDirectVar1;
    F.params=&params;
    params.potential=this;
    params.R=R;
    params.z=Z;
    params.zfirst=true;
    params.m=m;
    if(accuracymode!=AM_FAST && abs(m)<=4)  // more precise computation only for low-m terms
        params.ws=gsl_integration_workspace_alloc(1000);
    else
        params.ws=NULL;
    double result, error;
    size_t neval;
    if(accuracymode!=AM_SLOW)
        gsl_integration_qng(&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
    else {
        gsl_integration_workspace* ws=gsl_integration_workspace_alloc(1000);
        gsl_integration_qags(&F, 0, 1, 0, EPSREL_POTENTIAL_INT, 1000, ws, &result, &error);
        gsl_integration_workspace_free(ws);
    }
    if(params.ws!=NULL)
        gsl_integration_workspace_free(params.ws);
#if Q_HYPERG
    result*=M_SQRTPI*gsl_sf_gamma(abs(params.m)+0.5)/gsl_sf_gamma(abs(params.m)+1);
#endif
    return result;
};

double CPotentialDirect::Phi(double X, double Y, double Z, double /*t*/) const
{
    double R=sqrt(X*X+Y*Y);
    double phi=atan2(Y, X);
    double val=0;
    int mmax=splines.size()/2;
    for(int m=-mmax; m<=mmax; m++)
        val+=Phi_m(R, Z, m) * (m>=0 ? cos(m*phi) : sin(-m*phi));
    return val;
}

//----------------------------------------------------------------------------//
// Cylindrical spline potential 

CPotentialCylSpline::CPotentialCylSpline(size_t Ncoefs_R, size_t Ncoefs_z, size_t Ncoefs_phi, 
    const CDensity* density, double radius_min, double radius_max, double z_min, double z_max)
{
    const CPotential* pot_tmp=new CPotentialDirect(density, Ncoefs_phi, CPotentialDirect::AM_MEDIUM);
    initPot(Ncoefs_R, Ncoefs_z, Ncoefs_phi, pot_tmp, radius_min, radius_max, z_min, z_max);
    delete pot_tmp;
}

CPotentialCylSpline::CPotentialCylSpline(size_t Ncoefs_R, size_t Ncoefs_z, size_t Ncoefs_phi, 
    const CPotential* potential, double radius_min, double radius_max, double z_min, double z_max)
{
    initPot(Ncoefs_R, Ncoefs_z, Ncoefs_phi, potential, radius_min, radius_max, z_min, z_max);
}

CPotentialCylSpline::CPotentialCylSpline(size_t Ncoefs_R, size_t Ncoefs_z, size_t Ncoefs_phi, 
    const CPointMassSet<double> &points, SYMMETRYTYPE _sym, double radius_min, double radius_max, double z_min, double z_max)
{
    mysymmetry=_sym;
    if(Ncoefs_phi==0)
        mysymmetry=(SYMMETRYTYPE)(mysymmetry | ST_ZROTSYM);
    const CPotential* pot_tmp=new CPotentialDirect(&points, Ncoefs_phi, mysymmetry);
    initPot(Ncoefs_R, Ncoefs_z, Ncoefs_phi, pot_tmp, radius_min, radius_max, z_min, z_max);
    delete pot_tmp;
}

CPotentialCylSpline::CPotentialCylSpline(const vectord &gridR, const vectord& gridz, const std::vector< vectord > &coefs)
{
    if(coefs.size()==0 || coefs.size()%2!=1 || coefs[coefs.size()/2].size()!=gridR.size()*gridz.size()) {
        my_error(FUNCNAME, "Invalid parameters in the constructor");
        initDefault();
    } else {
        grid_R=gridR;
        grid_z=gridz;
        initSplines(coefs);
        // check symmetry
        int mmax=static_cast<int>(splines.size()/2);
        mysymmetry=mmax==0 ? ST_AXISYMMETRIC : ST_TRIAXIAL;
        for(int m=-mmax; m<=mmax; m++)
            if(splines[m+mmax]!=NULL) {
                if(m<0 || (m>0 && m%2==1))
                    mysymmetry = ST_NONE;//(SYMMETRYTYPE)(mysymmetry & ~ST_PLANESYM & ~ST_ZROTSYM);
            }
    }
}

CPotentialCylSpline::~CPotentialCylSpline()
{
    for(size_t k=0; k<splines.size(); k++)
        if(splines[k]!=NULL)
            interp2d_spline_free(splines[k]);
}

double integrandPotAvgPhi(double phi, void* param)
{
    int m=((CPotentialParam*)param)->m;
    return ((CPotentialParam*)param)->P->Phi( 
        ((CPotentialParam*)param)->R*cos(phi), ((CPotentialParam*)param)->R*sin(phi), ((CPotentialParam*)param)->Z) *
    (m==0 ? 1 : m>0 ? 2*cos(m*phi) : 2*sin(-m*phi));
}

double CPotentialCylSpline::computePhi_m(double R, double z, int m, const CPotential* potential) const
{
    if(potential->PotentialType()==PT_DIRECT) {
        return ((const CPotentialDirect*)potential)->Phi_m(R, z, m);
    }
    else {  // compute azimuthal Fourier harmonic coefficient for the given m
        if(R==0 && m!=0) return 0;
        gsl_function F;
        CPotentialParam PP;
        PP.P=potential;
        PP.R=R;
        PP.Z=z;
        PP.m=m;
        F.function=integrandPotAvgPhi;
        F.params=&PP;
        double result, error;
        size_t neval;
        double phimax=(potential->symmetry() & ST_PLANESYM)==ST_PLANESYM ? M_PI_2 : 2*M_PI;
        gsl_integration_qng(&F, 0, phimax, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
        return result/phimax;
    }
}

void CPotentialCylSpline::initPot(size_t _Ncoefs_R, size_t _Ncoefs_z, size_t _Ncoefs_phi, 
    const CPotential* potential, double radius_min, double radius_max, double z_min, double z_max)
{
    mysymmetry=potential->symmetry();
    bool zsymmetry= (mysymmetry & ST_PLANESYM)==ST_PLANESYM;  // whether we need to compute potential at z<0 independently from z>0
    int mmax = (mysymmetry & ST_AXISYMMETRIC) == ST_AXISYMMETRIC ? 0 : _Ncoefs_phi;
    int mmin = (mysymmetry & ST_PLANESYM)==ST_PLANESYM ? 0 :-1;  // if triaxial symmetry, do not use sine terms which correspond to m<0
    int mstep= (mysymmetry & ST_PLANESYM)==ST_PLANESYM ? 2 : 1;  // if triaxial symmetry, use only even m
    if(radius_max==0 || radius_min==0) {
        double totalmass=potential->totalMass();
        if(radius_max==0)
            radius_max=potential->getRadiusByMass(totalmass*(1-1.0/(_Ncoefs_R*_Ncoefs_z)));
        if(radius_max<=0) radius_max=100; // arbitrary!
        if(radius_min==0) 
            radius_min=std::min<double>(radius_max/_Ncoefs_R, potential->getRadiusByMass(totalmass/(_Ncoefs_R*_Ncoefs_z)));
        if(radius_min<=0) radius_min=radius_max/_Ncoefs_R;
    }
    vectord splineRad;
    createNonuniformGrid(_Ncoefs_R, radius_min, radius_max, true, &splineRad);
    grid_R=splineRad;
    if(z_max==0) z_max=radius_max;
    if(z_min==0) z_min=radius_min;
    z_min=std::min<double>(z_min, z_max/_Ncoefs_z);
    createNonuniformGrid(_Ncoefs_z, z_min, z_max, true, &splineRad);
    grid_z.assign(2*_Ncoefs_z-1,0);
    for(size_t i=0; i<_Ncoefs_z; i++) {
        grid_z[_Ncoefs_z-1-i] =-splineRad[i];
        grid_z[_Ncoefs_z-1+i] = splineRad[i];
    }
    size_t Ncoefs_R=grid_R.size();
    size_t Ncoefs_z=grid_z.size();
    std::vector<vectord> coefs(2*mmax+1);
    for(int m=mmax*mmin; m<=mmax; m+=mstep) {
        coefs[mmax+m].assign(Ncoefs_R*Ncoefs_z,0);
    }
    bool correct=true;
    // for some unknown reason, switching on the OpenMP parallelization makes the results of computation of m=4 coef 
    // irreproducible (adds a negligible random error). Assumed to be unimportant...
#pragma omp parallel for
    for(int iR=0; iR<static_cast<int>(Ncoefs_R); iR++) {
        for(size_t iz=0; iz<=Ncoefs_z/2; iz++) {
            for(int m=mmax*mmin; m<=mmax; m+=mstep) {
                double val=computePhi_m(grid_R[iR], grid_z[Ncoefs_z/2+iz], m, potential);
                if(!gsl_finite(val)) {
#ifdef DEBUGPRINT
                    my_message(FUNCNAME, "Error in computing potential at R="+convertToString(grid_R[iR])+
                        ", z="+convertToString(grid_z[iz+Ncoefs_z/2])+", m="+convertToString(m));
#endif
                    val=0;
                    correct=false;
                }
                coefs[mmax+m][(Ncoefs_z/2+iz)*Ncoefs_R+iR] = val;
                if(!zsymmetry && iz>0)   // no symmetry about x-y plane
                    val=computePhi_m(grid_R[iR], grid_z[Ncoefs_z/2-iz], m, potential);  // compute potential at -z independently
                coefs[mmax+m][(Ncoefs_z/2-iz)*Ncoefs_R+iR] = val;
            }
        }
#ifdef DEBUGPRINT
        my_message(FUNCNAME, "R="+convertToString(grid_R[iR]));
#endif
    }
    if(!correct) 
        my_message(FUNCNAME, "Errors in computing the potential are replaced by zeros");
    initSplines(coefs);
}

void CPotentialCylSpline::initSplines(const std::vector< vectord > &coefs)
{
    size_t Ncoefs_R=grid_R.size();
    size_t Ncoefs_z=grid_z.size();
    int mmax=coefs.size()/2;
    assert(coefs[mmax].size()==Ncoefs_R*Ncoefs_z);  // check that at least m=0 coefficients are present
    // compute multipole coefficients for extrapolating the potential and forces beyond the grid,
    // by fitting them to the potential at the grid boundary
    C00=C20=C22=C40=0;
    bool fitm2=mmax>=2 && coefs[mmax+2].size()==Ncoefs_R*Ncoefs_z;  // whether to fit m=2
    size_t npointsboundary=2*(Ncoefs_R-1)+Ncoefs_z;
    gsl_matrix* X0=gsl_matrix_alloc(npointsboundary, 3);  // matrix of coefficients  for m=0
    gsl_vector* Y0=gsl_vector_alloc(npointsboundary);     // vector of r.h.s. values for m=0
    gsl_vector* W0=gsl_vector_alloc(npointsboundary);     // vector of weights
    vectord X2(npointsboundary);     // vector of coefficients  for m=2
    vectord Y2(npointsboundary);     // vector of r.h.s. values for m=2
    for(size_t i=0; i<npointsboundary; i++) {
        size_t iR=i<2*Ncoefs_R ? i/2 : Ncoefs_R-1;
        size_t iz=i<2*Ncoefs_R ? (i%2)*(Ncoefs_z-1) : i-2*Ncoefs_R+1;
        double R=grid_R[iR];
        double z=grid_z[iz];
        double oneoverr=1/sqrt(R*R+z*z);
        gsl_vector_set(Y0, i, coefs[mmax][iz*Ncoefs_R+iR]);
        gsl_matrix_set(X0, i, 0, oneoverr);
        gsl_matrix_set(X0, i, 1, pow(oneoverr,5.0)*(2*z*z-R*R));
        gsl_matrix_set(X0, i, 2, pow(oneoverr,9.0)*(8*pow(z,4.0)-24*z*z*R*R+3*pow(R,4.0)) );
        // weight proportionally to the value of potential itself (so that we minimize sum of squares of relative differences)
        gsl_vector_set(W0, i, 1.0/pow_2(coefs[mmax][iz*Ncoefs_R+iR]));
        if(fitm2) {
            X2[i]=R*R*pow(oneoverr,5.0);
            Y2[i]=coefs[mmax+2][iz*Ncoefs_R+iR];
        }
    }
    // fit m=0 by three parameters
    gsl_vector* fit=gsl_vector_alloc(3);
    gsl_matrix* cov=gsl_matrix_alloc(3,3);
    double chisq;
    gsl_multifit_linear_workspace* ws=gsl_multifit_linear_alloc(npointsboundary, 3);
    if(gsl_multifit_wlinear(X0, W0, Y0, fit, cov, &chisq, ws) == GSL_SUCCESS) {
        C00=gsl_vector_get(fit, 0);  // C00 ~= -Mtotal
        C20=gsl_vector_get(fit, 1);
        C40=gsl_vector_get(fit, 2);
    }
    gsl_multifit_linear_free(ws);
    gsl_vector_free(fit);
    gsl_matrix_free(cov);
    // fit m=2 if necessary
    if(fitm2) {
        double dummy1, dummy2;
        if(gsl_fit_mul(&(X2.front()), 1, &(Y2.front()), 1, npointsboundary, &C22, &dummy1, &dummy2) != GSL_SUCCESS)
            C22=0;
    }
    // assign Rscale so that it approximately equals -Mtotal/Phi(r=0)
    Rscale=C00/coefs[mmax][(Ncoefs_z/2)*Ncoefs_R];
    if(Rscale<=0) Rscale=std::min<double>(grid_R.back(), grid_z.back())*0.5; // shouldn't occur?
#ifdef DEBUGPRINT
    my_message(FUNCNAME,  "Rscale="+convertToString(Rscale)+", C00="+convertToString(C00)+", C20="+convertToString(C20)+", C22="+convertToString(C22)+", C40="+convertToString(C40));
#endif

    const size_t N_ADD=4;
    vectord grid_Rscaled(Ncoefs_R+N_ADD);
    vectord grid_zscaled(Ncoefs_z);
    for(size_t i=0; i<Ncoefs_R; i++) {
        grid_Rscaled[i+N_ADD] = log(1+grid_R[i]/Rscale);
        if(i>=1 && i<=N_ADD)   // add three points symmetric about 0, to ensure that spline derivative is (close to) zero at R=0
            grid_Rscaled[N_ADD-i] = -grid_Rscaled[N_ADD+i];
    }
    for(size_t i=0; i<Ncoefs_z; i++) {
        grid_zscaled[i] = log(1+fabs(grid_z[i])/Rscale)*(grid_z[i]>=0?1:-1);
    }
    splines.assign(coefs.size(), NULL);
    vectord values((Ncoefs_R+N_ADD)*Ncoefs_z);
    for(size_t m=0; m<coefs.size(); m++) {
        if(coefs[m].size()!=Ncoefs_R*Ncoefs_z) 
            continue;
        bool allzero=true;
        for(size_t iR=0; iR<Ncoefs_R; iR++) {
            for(size_t iz=0; iz<Ncoefs_z; iz++) {
                double scaling=sqrt(pow_2(Rscale)+pow_2(grid_R[iR])+pow_2(grid_z[iz]));
                double val=coefs[m][iz*Ncoefs_R+iR] * scaling;
                values[INDEX_2D(iR+N_ADD, iz, Ncoefs_R+N_ADD, Ncoefs_z)] = val;
                if(iR>=1&&iR<=N_ADD)
                    values[INDEX_2D(N_ADD-iR, iz, Ncoefs_R+N_ADD, Ncoefs_z)] = val;
                if(val!=0) allzero=false;
            }
        }
        if(allzero)
            continue;
        splines[m]=interp2d_spline_alloc(interp2d_bicubic, Ncoefs_R+N_ADD, Ncoefs_z);
        if(splines[m]==NULL) { 
            my_error(FUNCNAME, "Error allocating memory for splines");
            initDefault();
            return;
        }
        if(interp2d_spline_init(splines[m], &(grid_Rscaled.front()), &(grid_zscaled.front()), 
            &(values.front()), Ncoefs_R+N_ADD, Ncoefs_z) != GSL_SUCCESS)  {
            my_error(FUNCNAME, "Can't initialize splines");
            initDefault();
            return;
        }
    }
}

void CPotentialCylSpline::initDefault()
{
    Rscale=1.0;
    C00=C20=C22=0;
    mysymmetry=ST_SPHERICAL;
    grid_R.resize(2); grid_R[0]=0; grid_R[1]=1;
    grid_z.resize(3); grid_z[0]=-1; grid_z[1]=0; grid_z[2]=1;
    splines.assign(1, static_cast<interp2d_spline*>(NULL));
}

CPotential* CPotentialCylSpline::clone() const 
{   // need to duplicate spline arrays explicitly; everything else is copied automatically
    CPotentialCylSpline* newpot = new CPotentialCylSpline(*this); 
    for(size_t i=0; i<splines.size(); i++)
        if(splines[i]!=NULL)
        {
            interp2d_spline *spline = interp2d_spline_alloc(interp2d_bicubic, 
                splines[i]->interp_object.xsize, splines[i]->interp_object.ysize);
            interp2d_spline_init(spline, splines[i]->xarr, splines[i]->yarr, splines[i]->zarr, 
                splines[i]->interp_object.xsize, splines[i]->interp_object.ysize);
            newpot->splines[i]=spline;
        }
        else newpot->splines[i]=NULL;
    return newpot;
}

void CPotentialCylSpline::getCoefs(vectord *gridR, vectord* gridz, std::vector< vectord > *coefs) const
{
    *gridR = grid_R;
    *gridz = grid_z;
    coefs->resize(splines.size());
    for(size_t m=0; m<splines.size(); m++)
        coefs->at(m).assign(splines[m]!=NULL ? gridz->size()*gridR->size() : 0, 0);
    for(size_t iz=0; iz<gridz->size(); iz++)
        for(size_t iR=0; iR<gridR->size(); iR++) {
            double Rscaled=log(1+gridR->at(iR)/Rscale);
            double zscaled=log(1+fabs(gridz->at(iz))/Rscale)*(gridz->at(iz)>=0?1:-1);
            for(size_t m=0; m<splines.size(); m++)
                if(splines[m]!=NULL) {
                    double scaling=sqrt(pow_2(Rscale)+pow_2(grid_R[iR])+pow_2(grid_z[iz]));
                    coefs->at(m)[iz*gridR->size()+iR] = interp2d_spline_eval(splines[m], Rscaled, zscaled, NULL, NULL)/scaling;
                }
        }
}

double CPotentialCylSpline::Mass(const double r) const
{
    if(r<=0) return 0;
    gsl_function F;
    CDensityParam params;
    F.function=&getDensityRcyl;
    F.params=&params;
    params.P=this;
    params.r=r;  // upper limit of the integration
    params.mumble=grid_z.back();  // upper limit on integration in Z
    double R=std::min<double>(r, grid_R.back());
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, R/(1+R), 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result;
}
    
#ifndef NO_OPTIMIZATION
/// \cond INTERNAL_DOCS
struct bicubic_state_t
{
    double * zx;
    double * zy;
    double * zxy;
};
/// \endcond

/// optimized routine for computing 2d spline value, first and second derivatives at once (to save time)
void interp2d_spline_eval_all(const interp2d_spline* interp, const double x, const double y, 
                              double *z, double *z_x, double *z_y, double *z_xx, double *z_xy, double *z_yy)
{
    bicubic_state_t *state = (bicubic_state_t *) interp->interp_object.state;
    const double* xarr = interp->xarr;
    const double* yarr = interp->yarr;
    const double* zarr = interp->zarr;
    const size_t xsize = interp->interp_object.xsize;
    const size_t ysize = interp->interp_object.ysize;
    double v;
    // First compute the indices into the data arrays where we are interpolating
    const size_t xi = gsl_interp_bsearch(xarr, x, 0, xsize - 1);
    const size_t yi = gsl_interp_bsearch(yarr, y, 0, ysize - 1);
    // Find the minimum and maximum values on the grid cell in each dimension
    const double xmin = xarr[xi];
    const double xmax = xarr[xi + 1];
    const double ymin = yarr[yi];
    const double ymax = yarr[yi + 1];
    const double zminmin = zarr[INDEX_2D(xi, yi, xsize, ysize)];
    const double zminmax = zarr[INDEX_2D(xi, yi + 1, xsize, ysize)];
    const double zmaxmin = zarr[INDEX_2D(xi + 1, yi, xsize, ysize)];
    const double zmaxmax = zarr[INDEX_2D(xi + 1, yi + 1, xsize, ysize)];
    // Get the width and height of the grid cell
    const double dx = xmax - xmin;
    const double dy = ymax - ymin;
    // t and u are the positions within the grid cell at which we are computing
    // the interpolation, in units of grid cell size
    const double t = (x - xmin)/dx;
    const double u = (y - ymin)/dy;
    const double dt = 1./dx; // partial t / partial x
    const double du = 1./dy; // partial u / partial y
    const double zxminmin = state->zx[INDEX_2D(xi, yi, xsize, ysize)]/dt;
    const double zxminmax = state->zx[INDEX_2D(xi, yi + 1, xsize, ysize)]/dt;
    const double zxmaxmin = state->zx[INDEX_2D(xi + 1, yi, xsize, ysize)]/dt;
    const double zxmaxmax = state->zx[INDEX_2D(xi + 1, yi + 1, xsize, ysize)]/dt;
    const double zyminmin = state->zy[INDEX_2D(xi, yi, xsize, ysize)]/du;
    const double zyminmax = state->zy[INDEX_2D(xi, yi + 1, xsize, ysize)]/du;
    const double zymaxmin = state->zy[INDEX_2D(xi + 1, yi, xsize, ysize)]/du;
    const double zymaxmax = state->zy[INDEX_2D(xi + 1, yi + 1, xsize, ysize)]/du;
    const double zxyminmin = state->zxy[INDEX_2D(xi, yi, xsize, ysize)]/(dt*du);
    const double zxyminmax = state->zxy[INDEX_2D(xi, yi + 1, xsize, ysize)]/(dt*du);
    const double zxymaxmin = state->zxy[INDEX_2D(xi + 1, yi, xsize, ysize)]/(dt*du);
    const double zxymaxmax = state->zxy[INDEX_2D(xi + 1, yi + 1, xsize, ysize)]/(dt*du);
    const double t0 = 1;
    const double t1 = t;
    const double t2 = t*t;
    const double t3 = t*t2;
    const double u0 = 1;
    const double u1 = u;
    const double u2 = u*u;
    const double u3 = u*u2;
    const double t0u0 = t0*u0, t0u1=t0*u1, t0u2=t0*u2, t0u3=t0*u3;
    const double t1u0 = t1*u0, t1u1=t1*u1, t1u2=t1*u2, t1u3=t1*u3;
    const double t2u0 = t2*u0, t2u1=t2*u1, t2u2=t2*u2, t2u3=t2*u3;
    const double t3u0 = t3*u0, t3u1=t3*u1, t3u2=t3*u2, t3u3=t3*u3;

    double zval =0;
    double zderx=0;
    double zdery=0;
    double zd_xx=0;
    double zd_xy=0;
    double zd_yy=0;
    v = zminmin;
    zval += v*t0u0;
    v = zyminmin;
    zval += v*t0u1;
    zdery+= v*t0u0;
    v = -3*zminmin + 3*zminmax - 2*zyminmin - zyminmax;
    zval += v*t0u2;
    zdery+= 2*v*t0u1;
    zd_yy+= 2*v*t0u0;
    v = 2*zminmin - 2*zminmax + zyminmin + zyminmax;
    zval += v*t0u3;
    zdery+= 3*v*t0u2;
    zd_yy+= 6*v*t0u1;
    v = zxminmin;
    zval += v*t1u0;
    zderx+= v*t0u0;
    v = zxyminmin;
    zval += v*t1u1;
    zderx+= v*t0u1;
    zdery+= v*t1u0;
    zd_xy+= v*t0u0;
    v = -3*zxminmin + 3*zxminmax - 2*zxyminmin - zxyminmax;
    zval += v*t1u2;
    zderx+= v*t0u2;
    zdery+= 2*v*t1u1;
    zd_xy+= 2*v*t0u1;
    zd_yy+= 2*v*t1u0;
    v = 2*zxminmin - 2*zxminmax + zxyminmin + zxyminmax;
    zval += v*t1u3;
    zderx+= v*t0u3;
    zdery+= 3*v*t1u2;
    zd_xy+= 3*v*t0u2;
    zd_yy+= 6*v*t1u1;
    v = -3*zminmin + 3*zmaxmin - 2*zxminmin - zxmaxmin;
    zval += v*t2u0;
    zderx+= 2*v*t1u0;
    zd_xx+= 2*v*t0u0;
    v = -3*zyminmin + 3*zymaxmin - 2*zxyminmin - zxymaxmin;
    zval += v*t2u1;
    zderx+= 2*v*t1u1;
    zdery+= v*t2u0;
    zd_xx+= 2*v*t0u1;
    zd_xy+= 2*v*t1u0;
    v = 9*zminmin - 9*zmaxmin + 9*zmaxmax - 9*zminmax + 6*zxminmin + 3*zxmaxmin - 3*zxmaxmax - 6*zxminmax 
      + 6*zyminmin - 6*zymaxmin - 3*zymaxmax + 3*zyminmax + 4*zxyminmin + 2*zxymaxmin + zxymaxmax + 2*zxyminmax;
    zval += v*t2u2;
    zderx+= 2*v*t1u2;
    zdery+= 2*v*t2u1;
    zd_xx+= 2*v*t0u2;
    zd_xy+= 4*v*t1u1;
    zd_yy+= 2*v*t2u0;
    v = -6*zminmin + 6*zmaxmin - 6*zmaxmax + 6*zminmax - 4*zxminmin - 2*zxmaxmin + 2*zxmaxmax + 4*zxminmax 
      - 3*zyminmin + 3*zymaxmin + 3*zymaxmax - 3*zyminmax - 2*zxyminmin - zxymaxmin - zxymaxmax - 2*zxyminmax;
    zval += v*t2u3;
    zderx+= 2*v*t1u3;
    zdery+= 3*v*t2u2;
    zd_xx+= 2*v*t0u3;
    zd_xy+= 6*v*t1u2;
    zd_yy+= 6*v*t2u1;
    v = 2*zminmin - 2*zmaxmin + zxminmin + zxmaxmin;
    zval += v*t3u0;
    zderx+= 3*v*t2u0;
    zd_xx+= 6*v*t1u0;
    v = 2*zyminmin - 2*zymaxmin + zxyminmin + zxymaxmin;
    zval += v*t3u1;
    zderx+= 3*v*t2u1;
    zdery+= v*t3u0;
    zd_xx+= 6*v*t1u1;
    zd_xy+= 3*v*t2u0;
    v = -6*zminmin + 6*zmaxmin - 6*zmaxmax + 6*zminmax - 3*zxminmin - 3*zxmaxmin + 3*zxmaxmax + 3*zxminmax 
      - 4*zyminmin + 4*zymaxmin + 2*zymaxmax - 2*zyminmax - 2*zxyminmin - 2*zxymaxmin - zxymaxmax - zxyminmax;
    zval += v*t3u2;
    zderx+= 3*v*t2u2;
    zdery+= 2*v*t3u1;
    zd_xx+= 6*v*t1u2;
    zd_xy+= 6*v*t2u1;
    zd_yy+= 2*v*t3u0;
    v = 4*zminmin - 4*zmaxmin + 4*zmaxmax - 4*zminmax + 2*zxminmin + 2*zxmaxmin - 2*zxmaxmax - 2*zxminmax 
      + 2*zyminmin - 2*zymaxmin - 2*zymaxmax + 2*zyminmax + zxyminmin + zxymaxmin + zxymaxmax + zxyminmax;
    zval += v*t3u3;
    zderx+= 3*v*t2u3;
    zdery+= 3*v*t3u2;
    zd_xx+= 6*v*t1u3;
    zd_xy+= 9*v*t2u2;
    zd_yy+= 6*v*t3u1;

    if(z   !=NULL) *z=zval;
    if(z_x !=NULL) *z_x=zderx*dt;
    if(z_y !=NULL) *z_y=zdery*du;
    if(z_xx!=NULL) *z_xx=zd_xx*dt*dt;
    if(z_xy!=NULL) *z_xy=zd_xy*dt*du;
    if(z_yy!=NULL) *z_yy=zd_yy*du*du;
}
#endif

double CPotentialCylSpline::Phi(double X, double Y, double Z, double /*t*/) const
{
    double R=sqrt(X*X+Y*Y);
    if(R>=grid_R.back() || Z<=grid_z.front() || Z>=grid_z.back()) 
    { // fallback mechanism for extrapolation beyond grid definition region
        double r2=X*X+Y*Y+Z*Z;
        return (C00 + ((2*Z*Z-R*R)*C20 + (X*X-Y*Y)*C22 + (35*pow_2(Z*Z/r2)-30*Z*Z/r2+3)*C40)/pow_2(r2))/sqrt(r2);
    }
    double S=1/sqrt(pow_2(Rscale)+pow_2(R)+pow_2(Z));  // scaling
    double phi=atan2(Y,X);
    double Rscaled=log(1+R/Rscale);
    double zscaled=log(1+fabs(Z)/Rscale)*(Z>=0?1:-1);
    double val=0;
    int mmax=splines.size()/2;
    for(int m=-mmax; m<=mmax; m++)
        if(splines[m+mmax]!=NULL) {
            val += interp2d_spline_eval(splines[m+mmax], Rscaled, zscaled, NULL, NULL) * (m>=0 ? cos(m*phi) : sin(-m*phi));
        }
    return val*S;
}

double CPotentialCylSpline::Rho(double X, double Y, double Z, double /*t*/) const
{
    double R=sqrt(X*X+Y*Y);
    if(R>=grid_R.back() || fabs(Z)>=grid_z.back())
        return 0;  // no extrapolation beyong grid 
    double phi=atan2(Y,X);
    double Rscaled=log(1+R/Rscale);
    double zscaled=log(1+fabs(Z)/Rscale)*(Z>=0?1:-1);
    double dRscaleddR=1/(Rscale+R), d2RscaleddR2=-pow_2(dRscaleddR);
    double dzscaleddz=1/(Rscale+fabs(Z)), d2zscaleddz2=-pow_2(dzscaleddz)*(Z>=0?1:-1);
    double Phi_tot=0, dPhidRscaled=0, dPhidzscaled=0, d2PhidRscaled2=0, d2Phidzscaled2=0, d2Phidphi2=0;
    int mmax=splines.size()/2;
    for(int m=-mmax; m<=mmax; m++)
        if(splines[m+mmax]!=NULL) { 
            double cosmphi=m>=0 ? cos(m*phi) : sin(-m*phi);
#ifdef NO_OPTIMIZATION
            double dPhi_m_dRscaled=interp2d_spline_eval_deriv_x(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
            double dPhi_m_dzscaled=interp2d_spline_eval_deriv_y(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
            double d2Phi_m_dRscaled2=interp2d_spline_eval_deriv_xx(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
            double d2Phi_m_dzscaled2=interp2d_spline_eval_deriv_yy(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
            double Phi_m=interp2d_spline_eval(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
#else
            double Phi_m, dPhi_m_dRscaled, dPhi_m_dzscaled, d2Phi_m_dRscaled2, d2Phi_m_dzscaled2;
            interp2d_spline_eval_all(splines[m+mmax], Rscaled, zscaled, 
                &Phi_m, &dPhi_m_dRscaled, &dPhi_m_dzscaled, &d2Phi_m_dRscaled2, NULL, &d2Phi_m_dzscaled2);
#endif
            dPhidRscaled+=dPhi_m_dRscaled * cosmphi;
            dPhidzscaled+=dPhi_m_dzscaled * cosmphi;
            d2PhidRscaled2+=d2Phi_m_dRscaled2 * cosmphi;
            d2Phidzscaled2+=d2Phi_m_dzscaled2 * cosmphi;
            Phi_tot+=Phi_m*cosmphi;
            d2Phidphi2+=Phi_m * -m*m*cosmphi;
        }
    double S=1/sqrt(pow_2(Rscale)+pow_2(R)+pow_2(Z));  // scaling
    double dSdr_over_r=-S*S*S;
    double d2Sdr2=(pow_2(Rscale)-2*(R*R+Z*Z))*dSdr_over_r*S*S;
    double ddd=R*dPhidRscaled*dRscaleddR + Z*dPhidzscaled*dzscaleddz + Phi_tot;
    double LaplacePhi = (R>0 ? dPhidRscaled*(dRscaleddR/R+d2RscaleddR2) + d2Phidphi2/R/R : 0)
        + d2PhidRscaled2*pow_2(dRscaleddR) + dPhidzscaled*d2zscaleddz2 + d2Phidzscaled2*pow_2(dzscaleddz);
    double result = S*LaplacePhi + 2*dSdr_over_r*ddd + d2Sdr2*Phi_tot;
    return std::max<double>(0, result)/(4*M_PI);
}

void CPotentialCylSpline::Force(const double v[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=v[0], Y=v[1], Z=v[2];
    double R2=X*X+Y*Y, R=sqrt(R2);
    if(R>=grid_R.back() || fabs(Z)>=grid_z.back()) { // fallback mechanism for extrapolation beyond grid definition region
        double X2=X*X, Y2=Y*Y, Z2=Z*Z, R2=X2+Y2, r2=R2+Z2;
        double oneoverr3=pow(r2,-1.5), oneoverr4=1/(r2*r2), oneoverr8=pow_2(oneoverr4);
        double add =5*(C20 * (2*Z2-R2) + C22 * (X2-Y2) ) * oneoverr4;
        double add4=C40*15*(21*Z2*Z2 - 14*Z2*r2 + r2*r2) * oneoverr8;
        double add4z=C40*5*(63*Z2*Z2 - 70*Z2*r2 + 15*r2*r2) * oneoverr8;
        force[0] = X*oneoverr3*(C00+add+2*(C20-C22)/r2 + add4);
        force[1] = Y*oneoverr3*(C00+add+2*(C20+C22)/r2 + add4);
        force[2] = Z*oneoverr3*(C00+add-4*C20/r2 + add4z);
        if(forceDeriv!=NULL) {
            double X2r2=X2/r2, Y2r2=Y2/r2, Z2r2=Z2/r2;
            double add7=C40*60*(1-7*Z2r2)*oneoverr4 - add4*11;
            forceDeriv[0] = oneoverr3*( C00*(1 - 3*X2r2) + add*(1-7*X2r2) + 2*(C20-C22)/r2*(1-10*X2r2) 
                                      + add4 + add7*X2r2 );
            forceDeriv[1] = oneoverr3*( C00*(1 - 3*Y2r2) + add*(1-7*Y2r2) + 2*(C20+C22)/r2*(1-10*Y2r2)
                                      + add4 + add7*Y2r2 );
            forceDeriv[2] = oneoverr3*( C00*(1 - 3*Z2r2) + add*(1-7*Z2r2) - 4*C20/r2*(1-10*Z2r2) 
                                      + add4z*(1-11*Z2r2) + C40*80*Z2r2*(7*Z2r2-5)*oneoverr4);
            double addmix = - 3*C00 - add*7;
            forceDeriv[3] = oneoverr3*X*Y/r2*(addmix - 20*C20/r2 + add7);
            addmix += -11*add4z + C40*100*(3-7*Z2r2)*oneoverr4;
            forceDeriv[4] = oneoverr3*Y*Z/r2*(addmix + 10*(C20-C22)/r2);
            forceDeriv[5] = oneoverr3*Z*X/r2*(addmix + 10*(C20+C22)/r2);
        }
        return;
    }
    double phi=atan2(Y,X);
    double Rscaled=log(1+R/Rscale);
    double zscaled=log(1+fabs(Z)/Rscale)*(Z>=0?1:-1);
    double Phi_tot=0, dPhidRscaled=0, dPhidzscaled=0, dPhidphi=0;
    double d2PhidRscaled2=0, d2PhidRscaleddzscaled=0, d2Phidzscaled2=0, d2Phidphi2=0, d2PhidRscaleddphi=0, d2Phidzscaleddphi=0;
    int mmax=splines.size()/2;
    for(int m=-mmax; m<=mmax; m++)
        if(splines[m+mmax]!=NULL) {
            double cosmphi=m>=0?cos(m*phi):sin(-m*phi);
            double sinmphi=m>=0?sin(m*phi):cos(-m*phi);
#ifdef NO_OPTIMIZATION
            double Phi_m=interp2d_spline_eval(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
            double dPhi_m_dRscaled=interp2d_spline_eval_deriv_x(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
            double dPhi_m_dzscaled=interp2d_spline_eval_deriv_y(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
#else
            double Phi_m, dPhi_m_dRscaled, dPhi_m_dzscaled, d2Phi_m_dRscaled2, d2Phi_m_dRscaleddzscaled, d2Phi_m_dzscaled2;
            interp2d_spline_eval_all(splines[m+mmax], Rscaled, zscaled, &Phi_m, 
                &dPhi_m_dRscaled, &dPhi_m_dzscaled, &d2Phi_m_dRscaled2, &d2Phi_m_dRscaleddzscaled, &d2Phi_m_dzscaled2);
#endif
            dPhidRscaled+=dPhi_m_dRscaled*cosmphi;
            dPhidzscaled+=dPhi_m_dzscaled*cosmphi;
            Phi_tot+=Phi_m*cosmphi;
            dPhidphi+=Phi_m * -m*sinmphi;
            if(forceDeriv!=NULL)
            {
#ifdef NO_OPTIMIZATION
                double d2Phi_m_dRscaled2=interp2d_spline_eval_deriv_xx(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
                double d2Phi_m_dRscaleddzscaled=interp2d_spline_eval_deriv_xy(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
                double d2Phi_m_dzscaled2=interp2d_spline_eval_deriv_yy(splines[m+mmax], Rscaled, zscaled, NULL, NULL);
#endif
                d2PhidRscaled2+=d2Phi_m_dRscaled2 * cosmphi;
                d2Phidzscaled2+=d2Phi_m_dzscaled2 * cosmphi;
                d2Phidphi2+=Phi_m * -m*m*cosmphi;
                d2PhidRscaleddzscaled+=d2Phi_m_dRscaleddzscaled * cosmphi;
                d2PhidRscaleddphi+=dPhi_m_dRscaled* -m*sinmphi;
                d2Phidzscaleddphi+=dPhi_m_dzscaled* -m*sinmphi;
            }
        }
    double r2=R*R+Z*Z;
    double S=1/sqrt(pow_2(Rscale)+r2);  // scaling
    double dSdr_over_r=-S*S*S;
    double dRscaleddR=1/(Rscale+R);
    double dzscaleddz=1/(Rscale+fabs(Z));
    dPhidphi*=S;
    double dPhidR=dPhidRscaled*dRscaleddR*S + Phi_tot*R*dSdr_over_r;
    if(R>0) {
        force[0] = -(dPhidR*X-dPhidphi*Y/R)/R;
        force[1] = -(dPhidR*Y+dPhidphi*X/R)/R;
    } else { 
        force[0]=force[1]=0;
    }
    if(Z==0 && (mysymmetry & ST_PLANESYM)==ST_PLANESYM) { // symmetric about z -> -z
        dPhidzscaled=0;
        d2Phidzscaleddphi=0;
        d2PhidRscaleddzscaled=0;
    }
    double dPhidz=dPhidzscaled*dzscaleddz*S + Phi_tot*Z*dSdr_over_r;
    force[2] = -dPhidz;
    if(forceDeriv!=NULL && R>0)
    {
        d2Phidphi2*=S;
        double d2RscaleddR2=-pow_2(dRscaleddR);
        double d2zscaleddz2=-pow_2(dzscaleddz)*(Z>=0?1:-1);
        double d2Sdr2=(pow_2(Rscale)-2*r2)*dSdr_over_r*S*S;
        double d2PhidR2 = (R*R*d2Sdr2+Z*Z*dSdr_over_r)/r2*Phi_tot + 2*dSdr_over_r*R*dRscaleddR*dPhidRscaled +
            S*(d2PhidRscaled2*pow_2(dRscaleddR)+dPhidRscaled*d2RscaleddR2);
        double d2Phidz2 = (Z*Z*d2Sdr2+R*R*dSdr_over_r)/r2*Phi_tot + 2*dSdr_over_r*Z*dzscaleddz*dPhidzscaled +
            S*(d2Phidzscaled2*pow_2(dzscaleddz)+dPhidzscaled*d2zscaleddz2);
        double d2PhidRdz = (d2Sdr2-dSdr_over_r)*R*Z/r2*Phi_tot + d2PhidRscaleddzscaled*S*dRscaleddR*dzscaleddz +
            dSdr_over_r*(Z*dRscaleddR*dPhidRscaled+R*dzscaleddz*dPhidzscaled);
        double d2PhidRdphi = dSdr_over_r*R*dPhidphi/S + S*dRscaleddR*d2PhidRscaleddphi;
        double d2Phidzdphi = dSdr_over_r*Z*dPhidphi/S + S*dzscaleddz*d2Phidzscaleddphi;
        double Rinv4 = 1/pow_2(R2);
        double X2=X*X, Y2=Y*Y, XY=X*Y, X2mY2=X2-Y2;
        forceDeriv[0] = -(dPhidR*Y2*R + dPhidphi* 2*XY + d2PhidR2*X2*R2 + d2PhidRdphi*-2*XY*R + d2Phidphi2*Y2) *Rinv4;
        forceDeriv[1] = -(dPhidR*X2*R + dPhidphi*-2*XY + d2PhidR2*Y2*R2 + d2PhidRdphi* 2*XY*R + d2Phidphi2*X2) *Rinv4;
        forceDeriv[3] = -(dPhidR*-XY*R + dPhidphi*-X2mY2 + d2PhidR2*XY*R2 + d2PhidRdphi*X2mY2*R + d2Phidphi2*-XY) *Rinv4;
        forceDeriv[2] = -d2Phidz2;
        forceDeriv[4] = -(d2PhidRdz*Y*R + d2Phidzdphi*X)/R2;
        forceDeriv[5] = -(d2PhidRdz*X*R - d2Phidzdphi*Y)/R2;
    }
}
#endif  // HAVE_INTERP2D

//----------------------------------------------------------------------------//
// Scale-free potential 
CPotentialScaleFree::CPotentialScaleFree(double mass, double scalerad, double _q, double _p, double _Gamma):
  CPotential(), norm((3-_Gamma)*mass/(4*M_PI*scalerad*scalerad*scalerad)), q(_q), p(_p), Gamma(_Gamma) 
{};

double CPotentialScaleFree::Rho(double X, double Y, double Z, double /*t*/) const
{
    return norm*pow(pow_2(X) + pow_2(Y/q) + pow_2(Z/p), -Gamma/2);
}

/// \cond INTERNAL_DOCS
struct CPotentialParamSF {
    double theta, phi;
    double X, Y, Z;
    double p, q, Gamma;
    int sh_l, sh_m;
    double costheta;
    double phi_max;
};

double intPhiSF_angular(double s, void* params)
{
    double theta= ((CPotentialParamSF*)params)->theta;
    double phi  = ((CPotentialParamSF*)params)->phi;
    double p= ((CPotentialParamSF*)params)->p;
    double q= ((CPotentialParamSF*)params)->q;
    double Gamma= ((CPotentialParamSF*)params)->Gamma;
    return 2*pow(pow_2(s)*( pow_2(sin(theta))*(pow_2(cos(phi)) + 
        pow_2(sin(phi))/(1-(1-pow_2(q))*pow_2(s))) + pow_2(cos(theta))/(1-(1-pow_2(p))*pow_2(s)) ), 1-Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s)));
}

double intForceSFx(double s, void* params)
{
    double X= ((CPotentialParamSF*)params)->X;
    double Y= ((CPotentialParamSF*)params)->Y;
    double Z= ((CPotentialParamSF*)params)->Z;
    double p= ((CPotentialParamSF*)params)->p;
    double q= ((CPotentialParamSF*)params)->q;
    double Gamma= ((CPotentialParamSF*)params)->Gamma;
    return pow(s, 2-Gamma) * pow( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)), -Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s)));
}
double intForceSFy(double s, void* params)
{
    double X= ((CPotentialParamSF*)params)->X;
    double Y= ((CPotentialParamSF*)params)->Y;
    double Z= ((CPotentialParamSF*)params)->Z;
    double p= ((CPotentialParamSF*)params)->p;
    double q= ((CPotentialParamSF*)params)->q;
    double Gamma= ((CPotentialParamSF*)params)->Gamma;
    return pow(s, 2-Gamma) * pow( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)), -Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s))) / (1-(1-pow_2(q))*pow_2(s));
}
double intForceSFz(double s, void* params)
{
    double X= ((CPotentialParamSF*)params)->X;
    double Y= ((CPotentialParamSF*)params)->Y;
    double Z= ((CPotentialParamSF*)params)->Z;
    double p= ((CPotentialParamSF*)params)->p;
    double q= ((CPotentialParamSF*)params)->q;
    double Gamma= ((CPotentialParamSF*)params)->Gamma;
    return pow(s, 2-Gamma) * pow( pow_2(X) + pow_2(Y)/(1-(1-pow_2(q))*pow_2(s)) + pow_2(Z)/(1-(1-pow_2(p))*pow_2(s)), -Gamma/2) /
        sqrt( (1-(1-pow_2(q))*pow_2(s)) * (1-(1-pow_2(p))*pow_2(s))) / (1-(1-pow_2(p))*pow_2(s));
}
/// \endcond 

double CPotentialScaleFree::Phi(double X, double Y, double Z, double /*t*/) const
{
    double result, error;
    size_t neval;
    CPotentialParamSF PP;
    PP.theta=atan2(sqrt(pow_2(X)+pow_2(Y)), Z); 
    PP.phi=atan2(Y, X);
    PP.p=p; PP.q=q; PP.Gamma=Gamma;
    gsl_function F;
    F.function=&intPhiSF_angular;
    F.params=&PP;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    return result * norm*2*M_PI/(2-Gamma)*q*p * pow(pow_2(X)+pow_2(Y)+pow_2(Z),1-Gamma/2);
}

void CPotentialScaleFree::Force(const double xyz[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double X=xyz[0];
    double Y=xyz[1];
    double Z=xyz[2];
    double mult=-4*M_PI*q*p*norm;
    double result, error; 
    size_t neval;
    CPotentialParamSF PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.p=p; PP.q=q; PP.Gamma=Gamma;
    gsl_function F;
    F.params=&PP;
    F.function=&intForceSFx;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    force[0] = mult * X * result;
    F.function=&intForceSFy;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    force[1] = mult * Y * result;
    F.function=&intForceSFz;
    gsl_integration_qng (&F, 0, 1, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval); 
    force[2] = mult * Z * result;
    if(forceDeriv) {    // variational equation - not implemented in the scale-free model; may use Scale-free SH for that purpose
        forceDeriv[0]=forceDeriv[1]=forceDeriv[2]=forceDeriv[3]=forceDeriv[4]=forceDeriv[5]=0;
    }
}

//----------------------------------------------------------------------------//
// scale-free potential - spherical harmonic expansion
CPotentialScaleFreeSH::CPotentialScaleFreeSH(double mass, double scalerad, 
    double q, double p, double _Gamma, size_t _Ncoefs_angular): 
CPotentialSH(_Ncoefs_angular), Gamma(_Gamma)
{
    mysymmetry = (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL);
    assignlmrange();
    prepareCoefsAnalytic((3-_Gamma)*mass/(4*M_PI*scalerad*scalerad*scalerad), q, p);
}

CPotentialScaleFreeSH::CPotentialScaleFreeSH(double _Gamma, const vectord &_coefs): 
CPotentialSH(static_cast<size_t>(sqrt(_coefs.size()*1.0)-1)), Gamma(_Gamma)
{
    assert(_coefs.size()==pow_2(Ncoefs_angular+1));
    SHcoefs=_coefs;  // no array size check performed!
    mysymmetry=ST_TRIAXIAL;  // no check performed if it is the case
}

/// \cond INTERNAL_DOCS
double intSF_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamSF*)params)->costheta;
    size_t sh_m=((CPotentialParamSF*)params)->sh_m;
    double q=((CPotentialParamSF*)params)->q;
    double p=((CPotentialParamSF*)params)->p;
    double Gamma=((CPotentialParamSF*)params)->Gamma;
    return pow((1-pow_2(costheta))*(1+(1/pow_2(q)-1)*pow_2(sin(phi))) + pow_2(costheta/p), -Gamma/2) * cos(sh_m*phi);
}
double intSF_theta(double theta, void* params)
{
    double result, error;
    double costheta=cos(theta);
    ((CPotentialParamSF*)params)->costheta=costheta;
    int sh_l=((CPotentialParamSF*)params)->sh_l;
    int sh_m=((CPotentialParamSF*)params)->sh_m;
    gsl_function F;
    F.function=&intSF_phi;
    F.params=params;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    gsl_integration_qag(&F, 0, ((CPotentialParamSF*)params)->phi_max, 0, EPSREL_POTENTIAL_INT, 1000, 3, w, &result, &error);
    gsl_integration_workspace_free (w);
    return result * sin(theta) * gsl_sf_legendre_sphPlm(sh_l, sh_m, costheta) / (2*M_SQRTPI);
}
/// \endcond

void CPotentialScaleFreeSH::prepareCoefsAnalytic(double norm, double q, double p)
{   // this function assumes that potential in question has triaxial symmetry, unlike the more general function for BSE potential
    SHcoefs.assign(pow_2(Ncoefs_angular+1), 0);
    CPotentialParamSF PP;
    PP.q=q; PP.p=p;
    PP.Gamma=std::max<double>(Gamma,1e-5);  // HACK to avoid singularity in gamma=0 coefficient computation
    //gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    gsl_function F;
    F.function=&intSF_theta;
    F.params=&PP;
    double theta_max= mysymmetry & ST_REFLECTION ? M_PI_2 : M_PI;  // if symmetries exist, no need to integrate over whole space
    PP.phi_max= mysymmetry & ST_PLANESYM ? M_PI_2 : 2*M_PI;
    int multfactor = (mysymmetry & ST_PLANESYM ? 4 : 1) * (mysymmetry & ST_REFLECTION ? 2 : 1);  // compensates integration of only half- or 1/8-space
    for(int l=0; l<=lmax; l+=lstep)
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            double result, error;
            size_t neval;
            PP.sh_l=l; PP.sh_m=m;
            gsl_integration_qng(&F, 0, theta_max, 0, EPSREL_POTENTIAL_INT, &result, &error, &neval);
            SHcoefs[l*(l+1)+m] = norm*multfactor*result * 4*M_PI/(l+3-Gamma)/(2-l-Gamma) * (m==0 ? 1 : M_SQRT2);
        }
    //gsl_integration_workspace_free (w);
}

void CPotentialScaleFreeSH::computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const
{
    double multr=pow(r, 2-Gamma);
    for(int l=0; l<=lmax; l+=lstep)
    {
        for(int m=l*mmin; m<=l*mmax; m+=mstep)
        {
            int indx=l*(l+1)+m;
            double coef = SHcoefs[indx] * multr;
            if(coefsF)      coefsF     [indx] = coef;
            if(coefsdFdr)   coefsdFdr  [indx] = coef * (2-Gamma)/r;
            if(coefsd2Fdr2) coefsd2Fdr2[indx] = coef * (2-Gamma)*(1-Gamma)/(r*r);
        }
    }
}

//----------------------------------------------------------------------------//
// Frozen-N-body potential (adapted from Barnes&Hut treecode from 'hackforce')
CPotentialNB::CPotentialNB() :
    CPotential(), eps(0), tol(0)
{ 
    maketree();
}

template<typename NumT> 
CPotentialNB::CPotentialNB(double _eps, double _tol, const CPointMassSet<NumT> &points):
    CPotential(), eps(_eps), tol(_tol)
{
    bodytab.resize(points.size());
    for(size_t i=0; i<points.size(); i++)
    {
        bodytab[i].type=BODY;
        bodytab[i].mass=points[i].second;
        bodytab[i].pos[0]=points[i].first.Pos[0];
        bodytab[i].pos[1]=points[i].first.Pos[1];
        bodytab[i].pos[2]=points[i].first.Pos[2];
    }
    maketree();  // should be called anyway
}
// explicit instantiations of constructor for two variants of template argument
template 
CPotentialNB::CPotentialNB(double _eps, double _tol, const CPointMassSet<double> &points);
template 
CPotentialNB::CPotentialNB(double _eps, double _tol, const CPointMassSet<float > &points);

double CPotentialNB::Phi(double X, double Y, double Z, double /*t*/) const
{
    double phi=0;
    vec3 point={X,Y,Z};
    walktree(troot, pow_2(rsize), point, CALCPHI, &phi);
    return -phi;
}

double CPotentialNB::Rho(double X, double Y, double Z, double /*t*/) const
{
    vectorpd vec;
    vec3 point={X,Y,Z};
    // first determine which cell does this point belong to, and its size (which will be used as an estimate of neighbour search radius)
    double searchrad=rsize;
    nodeptr *qptr=(nodeptr*)&troot;
    int xp[N_DIM];
    if(!intcoord(xp, point)) return 0;   // if point is outside root cell, density is zero
    int l = IMAX >> 1;                /* start with top bit       */
    while (*qptr != NULL && l>0) {  /* loop descending tree     */
        if ((*qptr)->type == BODY) {/*   reached a "leaf"?      */
            l=0;
        }
        else{
            qptr = &(((cellptr)(*qptr))->subp[subindex(xp, l)]);    /*   move down one level    */
            l = l >> 1;                /*   and test next bit      */
            searchrad/=2;
        }
    }
    searchrad *= 4;  // to have somewhat more than one neighbour!
    // now look for neighbours inside searchrad
    size_t numneighbour=0;
    do{ // recursively scan tree and collect the list of particles closer to our position than the search radius
        vec.clear();
        walktreedens(troot, rsize*rsize, searchrad*searchrad, point, xp, IMAX >> 1, &vec);
        numneighbour=vec.size();
        searchrad*=2;
    } while(numneighbour<TREECODE_DENSITY_KTH_NEIGHBOUR);  // if the number of particles was too low, repeat the search
    std::sort(vec.begin(), vec.end(), comparepair<double, double>);
    size_t kth=std::min<size_t>(vec.size(), TREECODE_DENSITY_KTH_NEIGHBOUR);
    if(kth<3) return 0; // failure to get density
    double mass=0;
    double smoothrad=vec[kth-1].first;
    for(size_t k=0; k<vec.size(); k++)
    {
        double radrel=vec[k].first/smoothrad;  // smoothing radius is the distance to Kth neighbour
        double weight=0; 
#if TREECODE_DENSITY_SPLINE
        if(radrel<0.5) weight=8-48*radrel*radrel+48*radrel*radrel*radrel;  // use spline weighting function with compact support
        else if(radrel<1) weight=16*(1-radrel)*(1-radrel)*(1-radrel);
#else
        if(radrel<1) weight=15.0/8*(1-radrel*radrel);  // use simpler Ferrer F1 kernel
#endif
        mass += vec[k].second*weight;
        if(weight==0) k=vec.size(); // all other neighbours are farther than spline support
    }
    mass /= (M_PI*pow(smoothrad, 3.0));  // for spline
    return mass;
}

void CPotentialNB::Force(const double xyz[N_DIM], const double /*t*/, double* force, double* forceDeriv) const
{
    double data[N_DIM+1]={0,0,0,rsize*rsize};  // init min distance (squared)
    walktree(troot, pow_2(rsize), xyz, CALCACC, data);  // recursively scan tree from the root and compute accelerations and distance to the nearest particle
    force[0]=data[0]; 
    force[1]=data[1]; 
    force[2]=data[2];
    if(forceDeriv!=NULL)
        forceDeriv[0]=sqrt(data[N_DIM]);  // this holds the minimum distance (unsoftened) to the nearest particle
}

double CPotentialNB::Mass(const double r) const
{
    double m=0, r2=r*r;
    for(size_t i=0; i<bodytab.size(); i++)
        if(pow_2(bodytab[i].pos[0])+pow_2(bodytab[i].pos[1])+pow_2(bodytab[i].pos[2]) < r2)
            m+=bodytab[i].mass;
    return m;
}

void CPotentialNB::maketree()    // initialize tree from set of bodies
{
    int celltabsize = (int) (100+bodytab.size()*0.52);  // should be enough, but if it was not, repeat the tree construction with larger allocated number of cells
    bool result=true; 
    do{
        troot=NULL;
        for(unsigned int d=0; d<N_DIM; d++) rmin[d]=-2.0;
        rsize=4.0;  // init box size
        celltab.clear();
        celltab.reserve(celltabsize); 
        result=true; 
        size_t b=0;
        for(; b<bodytab.size() && result; b++)
        {
            result = expandbox(b);   /* expand root to fit */
            result&= appendtree(b);  /* insert into tree   */
        }
        if(!result)
        {
            if(celltab.size()>=celltab.capacity() && b>bodytab.size()/2)  // error due to insufficient allocation of celltab
            {
                celltabsize = (int) (celltab.capacity() *1.05* bodytab.size()/b); // increase it and repeat
                my_error(FUNCNAME, 
                    "Warning, restarting tree initialization because of insufficient initial allocation of celltab");
            }
            else
            {
                troot=NULL;
                celltab.clear();
                my_error(FUNCNAME, 
                    "Error (unknown) in tree construction, perhaps pathological mass distribution");
                return;  // unknown error...
            }
        }
    } while(!result);

    centerofmass(troot, rmin, rsize,0);                /* find c-of-m coordinates  */

    double avgmass=troot->mass/bodytab.size();   // average particle mass (to assign variable softening length depending on mass)
    if(eps<0) {  // need to initialize each particle's softening length according to local density
#if TREECODE_EPS_FROM_DENSITY
        // this is more exact method, based on local density, but much more costly
        for(size_t i=0; i<bodytab.size(); i++)
        {
            double dens=Rho(bodytab[i].pos[0], bodytab[i].pos[1], bodytab[i].pos[2]);
            bodytab[i].eps = fabs(eps)*sqrt(bodytab[i].mass/avgmass) * pow(dens/avgmass, -1.0/3);
        }
#else
        // this is a more approximate method, based on just cell sizes, with iterative averaging of epsilons for nearby nodes in the tree
        // it is a lot faster (at least unless the density computation is made substantially more efficient) and gives slightly larger spread in epsilon
        // compared to the density-based method; however, as the effect of spatially adaptive softening depends only weakly on the softening parameter,
        // at least for the (recommended) Epanechnikov kernel, this approximation should be sufficient for practical purposes.
        for(int iter=0; iter<10; iter++)   // iterations to reduce 'binning' effect when eps ~ size of the smallest cell containing this one particle
            assigneps(troot, 0, rsize);    // this computes the approximation to local softening length for each point (stored is eps_i=|eps|*n^{-1/3})
        for(size_t i=0; i<bodytab.size(); i++)   // now adjust these values for variation with particle mass
            bodytab[i].eps = sqrt(bodytab[i].mass/avgmass) * bodytab[i].eps;
#endif
    } else {  // no spatial variation of softening length, but retain variation with particle mass
        for(size_t b=0; b<bodytab.size(); b++)
            bodytab[b].eps = eps*sqrt(bodytab[b].mass/avgmass);
    }
    propagateeps(troot);   // initialize cells' epsilon by averaging from underlying nodes

}

bool CPotentialNB::expandbox(const size_t b)  // expand bounding box if needed to accomodate for a new body
{
    int xtmp[N_DIM];
    while (! intcoord(xtmp, bodytab[b].pos)) {  /* expand box (rarely)      */
       vec3 rmid;
       for(unsigned int d=0; d<N_DIM; d++)
            rmid[d]=rmin[d]+0.5*rsize;          /*   find box midpoint      */
        for(unsigned int d=0; d<N_DIM; d++)     /*   loop over dimensions   */
            if (bodytab[b].pos[d] < rmid[d])    /*     is p left of mid?    */
                rmin[d] -= rsize;               /*       extend to left     */
        rsize = 2.0 * rsize;                    /*   double length of box   */
        if (troot != NULL) {                    /*   repot existing tree?   */
            cell c;                             /*     create new root cell */
            int xmid[N_DIM]={0,0,0};
            if(!intcoord(xmid, rmid)) return false;   /*     locate old root cell */
            int s = subindex(xmid, IMAX >> 1);  /*     find old tree index  */
            c.subp[s] = troot;                  /*     graft old on new     */
            if(celltab.size()>=celltab.capacity()) return false;  /* cannot reallocate vector, otherwise all pointers will be invalidated */
            celltab.push_back(c);
            troot = (nodeptr) (&(celltab.back()));             /*     plant new tree       */
        }
    }
    return true;
}

bool CPotentialNB::appendtree(const size_t b)            // append a body to the tree, creating new cell if necessary
{
    int xp[N_DIM]={0,0,0}, xq[N_DIM]={0,0,0};
    nodeptr *qptr;

    intcoord(xp, bodytab[b].pos);   /* form integer coords      */
    int l = IMAX >> 1;              /* start with top bit       */
    qptr = &troot;                  /* start with tree root     */
    while (*qptr != NULL) {         /* loop descending tree     */
        if(l==0) {                  /*   dont run out of bits   */
            (*qptr)->mass += bodytab[b].mass;  /* if two particles coincide, just sum up their mass */
            return true;
        }
        if ((*qptr)->type == BODY) {/*   reached a "leaf"?      */
            cell c;                 /*     alloc a new cell     */
            if(!intcoord(xq, (*qptr)->pos)) return false;   /*     get integer coords   */
            c.subp[subindex(xq, l)] = *qptr;    /*     put body in cell     */
            if(celltab.size()>=celltab.capacity()) return false;  /* cannot reallocate vector, otherwise all pointers will be invalidated */
            celltab.push_back(c);
            *qptr = (nodeptr) (&(celltab.back()));        /*     link cell in tree    */
        }
        qptr = &(((cellptr)(*qptr))->subp[subindex(xp, l)]);    /*   move down one level    */
        l = l >> 1;                /*   and test next bit      */
    }
    *qptr = (nodeptr) (&(bodytab[b]));            /* found place, store p     */
    return true;
}

bool CPotentialNB::intcoord(    // compute integer coords, return false if point is out of bounds for current root cell
            int xp[N_DIM],   /* integerized coordinate vec3 [0,IMAX) */
            const vec3 rp)   const   /* double coordinate vec3 (system coords) */
{
    bool inb = true;                    /* use to check bounds      */
    for (unsigned int k = 0; k < N_DIM; k++) {        /* loop over dimensions     */
        double xsc = (rp[k] - rmin[k]) / rsize; /*   scale to range [0,1)   */
        if (0.0 <= xsc && xsc < 1.0)            /*   within unit interval?  */
            xp[k] = (int) floor(IMAX * xsc);    /*     then integerize      */
        else                                    /*   out of range           */
            inb = false;                        /*     then remember that   */
    }
    return inb;
}

int CPotentialNB::subindex(   // determine the subcell that a given position belongs to
           const int x[N_DIM],        /* integerized coordinates of particle */
           const int l)     const      /* current level of tree */
{
    int i = 0;                                  /* sum index in i           */
    for (unsigned int k = 0; k < N_DIM; k++)             /* check each dimension     */
        if (x[k] & l)                           /*   if beyond midpoint     */
            i += NSUB >> (k + 1);               /*     skip over subcells   */
    return (i);
}

void CPotentialNB::centerofmass(const nodeptr q, const vec3 corner, const double cellsize, int l)      // recursive center-of-mass finder, also initializes rmax and quadrupole moment
{
    if(q==NULL) return;
    if(q->type == CELL) {                      /* is this a cell?          */
        q->mass = 0.0;                          /*   init total mass        */
        for(unsigned int d=0; d<N_DIM; d++) q->pos[d]=0; /*   and c. of m.           */
        for(unsigned int i=0; i< NSUB; i++) {            /*   loop over subcells     */
            vec3 subcellcorner;
            nodeptr r = ((cellptr)q)->subp[i];
            if(r != NULL) {                    /*     does subcell exist?  */
                if(r->type==BODY) 
                    ((cellptr)q)->numbody++; 
                else
                {
                    double halfsize=cellsize/2;
                    for(unsigned int d=0; d<N_DIM; d++)
                        subcellcorner[d]=corner[d] + halfsize*((i & (NSUB >> (d+1))) > 0);
                    centerofmass(r, subcellcorner, halfsize, l+1);  /*       find subcell cm    */
                    ((cellptr)q)->numbody+=((cellptr)r)->numbody;   /*  count total number of bodies */
                }
                q->mass += r->mass;             /*       sum total mass     */
                for(unsigned int d=0; d<N_DIM; d++)      /*   sum c.o.m. coordinates */
                    q->pos[d] += r->mass * r->pos[d];
            }
        }
        if(q->mass>0)
            for(unsigned int d=0; d<N_DIM; d++)
                q->pos[d] /= q->mass;
        // now we have center of mass coordinates, may compute quadrupole moment and other stuff
#if TREECODE_QUADRUPOLE
        for(unsigned int d1=0; d1<N_DIM; d1++)
            for(unsigned int d2=0; d2<N_DIM; d2++)
                ((cellptr)q)->quad[d1][d2]=0;
        for(unsigned int i = 0; i < NSUB; i++) {        /*   loop over subnodes     */
            nodeptr r = ((cellptr)q)->subp[i];
            if (r != NULL) {                    /*     does subnode exist?  */
                vec3 dr={r->pos[0]-q->pos[0], r->pos[1]-q->pos[1], r->pos[2]-q->pos[2]};  /* displacement vect. */
                for(unsigned int d1=0; d1<N_DIM; d1++)
                    for(unsigned int d2=0; d2<N_DIM; d2++)
                    {
                        ((cellptr)q)->quad[d1][d2] += dr[d1]*dr[d2] * r->mass;
                        if(r->type==CELL)       /*       if subnode is cell */
                            ((cellptr)q)->quad[d1][d2] += ((cellptr)r)->quad[d1][d2];  /*         use its moment   */
                    }
            }
        }
#endif
#if TREECODE_DEHNEN_MAC
        double rmax1=0, rmax2=0;
        // loop over sub-nodes
        for (unsigned int i = 0; i < NSUB; i++) {        /*   loop over subnodes     */
            nodeptr r = ((cellptr)q)->subp[i];
            if (r != NULL) {            /*     does subnode exist?  */
                vec3 dr={r->pos[0]-q->pos[0], r->pos[1]-q->pos[1], r->pos[2]-q->pos[2]};  /* displacement vect. */
                double drmag=sqrt(pow_2(dr[0])+pow_2(dr[1])+pow_2(dr[2]));
                rmax1 = std::max<double>(rmax1, (r->type==BODY ? 0 : ((cellptr)r)->rmax) + drmag );
            }
        }
        // find distance from this cell's c.o.m. to its most distant corner
        for(unsigned int c=0; c<NSUB; c++)  // loop over corners
        {
            double dist2corner=0;
            for(unsigned int d=0; d<N_DIM; d++)
                dist2corner += pow_2(corner[d] + ((c & (NSUB >> (d+1))) > 0)*cellsize - q->pos[d]);
            rmax2 = std::max<double>(rmax2, sqrt(dist2corner));
        }
        // assign cell's rmax to lower of these two values (Dehnen 2000)
        ((cellptr)q)->rmax=std::min<double>(rmax1, rmax2);
#if TREECODE_QUADRUPOLE
        // check that quadrupole terms are not too large, otherwise increase rmax (a modification to Dehnen's criterion to combat worst-case configurations)
        //double quadfact = (((cellptr)q)->quad[0][0]+((cellptr)q)->quad[1][1]+((cellptr)q)->quad[2][2])/(q->mass*pow_2(((cellptr)q)->rmax));
        //if(quadfact>0.3)  ((cellptr)q)->rmax *= sqrt(quadfact/0.3);
#endif
#else
        ((cellptr)q)->rmax=cellsize;   // use Barnes&Hut cell opening criterion based on geometric size of the cell only
#endif
    }
}

/** recursive assign of softening length: 
    compute the number density of points in a given cell and average it with the parent cell,
    repeat process for the entire tree is scanned several times to perform back and forth 
    averaging over parent/daughter nodes.
    eps_i = |eps| * n^{-1/3}, where eps is the global softening parameter and n is 
    the average number density in the vicinity of the node.
    epsparent is the softening length of the parent cell (zero on the first iteration).
*/
void CPotentialNB::assigneps(const nodeptr q, const double epsparent, const double cellsize)
{
    if(q==NULL) return;
    double epsavg=0; int numsub=0;
    // if on the first pass, compute epsilon from the number density of particles in this cell, otherwise don't re-assign it
    double epscur = q->eps>0 ? q->eps : fabs(eps)*cellsize/(q->type==BODY ? 1.0 : pow(((cellptr)q)->numbody*1.0, 1.0/3) );
    if(epsparent>0)
        epscur = sqrt(epsparent*epscur);     // geometric average with the parent cell's eps (which is an averaged value for all daughter nodes)
    q->eps=epscur;   // assign some preliminary guess
    if(q->type==BODY) 
        return;  // no descendants
    for(unsigned int i=0; i< NSUB; i++) {   // this node is a cell -- loop over daughter nodes
        nodeptr r = ((cellptr)q)->subp[i];
        if (r != NULL) {
            assigneps(r, ((cellptr)q)->eps, cellsize/2);  // descend one level deeper into the tree
            epsavg += pow(r->eps, -1./2);                 // add averaged value (weighted with some power, index chosen empirically!)
            numsub++;
        }
    }
    q->eps = pow(epsavg/numsub, -2./1);  // compute averaged value of non-empty daughter nodes' mean number density; 
    // note that for cells it is no longer the mean density within the cell, but rather the mean density averaged over its non-empty nodes
    // otherwise we will end up having the root cell with a very low average density, which will then be spread down into its daughter cells via iterative averaging.
}

void CPotentialNB::propagateeps(const nodeptr q)   // recursively compute averaged epsilons for tree cells from its child nodes
{
    if(q==NULL) return;
    double epsavg=0;
    for(unsigned int i=0; i< NSUB; i++) {    // assign eps to tree cells by averaging the values in sub-cells/leafs
        nodeptr r = ((cellptr)q)->subp[i];
        if (r != NULL) {
            if(r->type==CELL) propagateeps(r);
            epsavg += r->mass * pow(r->eps, -2.);  // add averaged value, weighted with the node mass and an empirically chosen power
        }
    }
    q->eps = pow(epsavg/q->mass, -0.5);   // the softening for a tree cell is the average over its non-empty subcells
}

void CPotentialNB::walktree(const nodeptr p,   // pointer to interacting node or cell 
           const double dsq,                   // cell size squared
           const vec3 pos0,                    // position to calculate quantity in question
           const WALKOP operation,             // operation to perform during tree-walk
           double data[])  const               // data accumulated according to the operation
{
    if(p==NULL) return;
    vec3 dr={p->pos[0]-pos0[0], p->pos[1]-pos0[1], p->pos[2]-pos0[2]};
    double distsq=pow_2(dr[0])+pow_2(dr[1])+pow_2(dr[2]);
    // test whether cell should be opened
    if(p->type==CELL && tol*tol*distsq < pow_2(((cellptr)p)->rmax) ) {
        for(unsigned int c=0; c<NSUB; c++)
            if(((cellptr)p)->subp[c] != NULL)  /* walk into sub-cells */
                walktree(((cellptr)p)->subp[c], dsq/4.0, pos0, operation, data);
    }
    else if(distsq>0) 
    {   // perform requested operation on leaf or unopened cell
        // assign softened distance, based either on Plummer softening or on F1 compact kernel; 
        // we need to compute the gravitation field (including quadrupole) of the softened particle,
        // so we compute additional quantities g1-g3 (notation is taken from Springel et al.2001)
#if TREECODE_EPS_COMPACT_KERNEL
        double eps2=p->eps*p->eps, eps3=eps2*p->eps;    /* use Epanechnikov kernel with compact support */
        double dist_sq_over_eps = p->eps==0?2 : distsq/eps2;
        double dist_inv_eps = dist_sq_over_eps>=1 ? 1/sqrt(distsq) : (15./8 - 5./4*dist_sq_over_eps + 3./8*pow_2(dist_sq_over_eps) )/p->eps;
        double g1 = dist_sq_over_eps>=1 ? dist_inv_eps/distsq : (2.5-1.5*dist_sq_over_eps)/eps3;
        double g2 = dist_sq_over_eps>=1 ? -3*g1/distsq : -3/(eps2*eps3);
        double g3 = dist_sq_over_eps>=1 ? -5*g2/distsq : 0;
#else
        double dist_sq_eps = distsq + pow_2(p->eps);    /* use Plummer softening */
        double dist_inv_eps = 1/sqrt(dist_sq_eps);
        double g1 = dist_inv_eps/dist_sq_eps;
        double g2 = -3*g1/dist_sq_eps;
        double g3 = -5*g2/dist_sq_eps;
#endif
#if TREECODE_QUADRUPOLE
        vec3 quaddr;
        double drquaddr=0, quadtrace=0;
        if(p->type == CELL){        /* if cell, add quad. term  */
            quadtrace=( ((cellptr)p)->quad[0][0]+((cellptr)p)->quad[1][1]+((cellptr)p)->quad[2][2] );
            for(unsigned int d1=0; d1<N_DIM; d1++)
            {
                quaddr[d1] = 
                    ((cellptr)p)->quad[d1][0] * dr[0] + 
                    ((cellptr)p)->quad[d1][1] * dr[1] + 
                    ((cellptr)p)->quad[d1][2] * dr[2];
                drquaddr += quaddr[d1]*dr[d1];
            }
        }
#endif
        switch(operation){
            case CALCPHI: 
                data[0] += p->mass*dist_inv_eps; 
#if TREECODE_QUADRUPOLE
                data[0] -= 0.5*(g2*drquaddr + g1*quadtrace);
#endif
                break;
            case CALCACC: 
                for(unsigned int d=0; d<N_DIM; d++)
                    data[d] += p->mass*dr[d] * g1;
#if TREECODE_QUADRUPOLE
                if(p->type == CELL) {        /* if cell, add quad. term  */
                    for(unsigned int d=0; d<N_DIM; d++)
                        data[d] += g2*quaddr[d] + 0.5*(g3*drquaddr+g2*quadtrace)*dr[d];
                }
#endif
                if(distsq<data[N_DIM]) data[N_DIM]=distsq;   // update minimum distance
                //data[N_DIM]=std::min<double>(data[N_DIM], 1/pow_2(dist_inv_eps));
                break;
        }
    }
}

void CPotentialNB::walktreedens(const nodeptr p,   // pointer to interacting node or cell 
           const double dsq,      // cell size squared
           const double searchradsq, // search radius squared
           const vec3 pos0,       // position to calculate quantity in question
           const int xp[N_DIM],   // integerized coordinates of the point in question
           const int lev,         // current grid level
           vectorpd* data)  const // array to store neighbours
{
    if(p==NULL) return;
    vec3 dr={p->pos[0]-pos0[0], p->pos[1]-pos0[1], p->pos[2]-pos0[2]};
    double distsq=pow_2(dr[0])+pow_2(dr[1])+pow_2(dr[2]);
    if(p->type==CELL) { 
        if(distsq < dsq*3 + searchradsq) {  // worst case - CoM may lie on the other side of diagonal
            for(unsigned int c=0; c<NSUB; c++)
                if(((cellptr)p)->subp[c] != NULL)  /* walk into sub-cells */
                    walktreedens(((cellptr)p)->subp[c], dsq/4, searchradsq, pos0, xp, lev>>1, data);
        }
        else {  // point is far from our CoM, but lies within one of child cells, which still needs to be opened
            int subcell=subindex(xp, lev);
            if(((cellptr)p)->subp[subcell] != NULL)  /* walk into this sub-cell */
                walktreedens(((cellptr)p)->subp[subcell], dsq/4, searchradsq, pos0, xp, lev>>1, data);
        }
    }
    else if (distsq < searchradsq)   // this is a leaf within search radius
        data->push_back(paird(sqrt(distsq),p->mass));
}

//----------------------------------------------------------------------------//

// simple density models, without potential

double CDensityPlummer::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m2 = pow_2(X) + pow_2(Y/q) + pow_2(Z/p) + pow_2(scalerad);
    return mass*pow_2(scalerad)*3/(4*M_PI*p*q) * pow(m2, -2.5);
}

double CDensityIsochrone::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m2 = pow_2(X) + pow_2(Y/q) + pow_2(Z/p);
    double a=sqrt(m2+scalerad*scalerad);
    return mass*(3*(scalerad+a)*a*a-m2*(scalerad+3*a))/(4*M_PI*p*q*pow(a*(scalerad+a),3));
}

double CDensityPerfectEllipsoid::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m2 = pow_2(X) + pow_2(Y/q) + pow_2(Z/p) + pow_2(scalerad);
    return mass*scalerad/(M_PI*M_PI*p*q*pow_2(m2));
}

//----------------------------------------------------------------------------//
double CDensityExpDisk::Rho(double X, double Y, double Z, double /*t*/) const
{
    double R=sqrt(X*X+Y*Y);
    double val=mass*exp(-R/scaleR)/(2*M_PI*scaleR*scaleR);
    if(edtype==ED_EXP) 
        val/=exp(fabs(Z)/scalez)*2*scalez;
    else if(edtype==ED_SECH2)
        val/=0.5*scalez*pow_2(exp(Z/scalez)+exp(-Z/scalez));
    return val;
}

//----------------------------------------------------------------------------//
/// \cond INTERNAL_DOCS
// some more machinery for NFW potential with a gradual cutoff in radius
double findNFWrcutoff(double r, void* params)
{
    return 4*M_PI*(r!=1 ? r*(r*log(r)-r+1)/pow_2(r-1) : 0.5) - *((double*)params);
}
/// \endcond
CDensityNFW::CDensityNFW(double _mass, double _scalerad, double _q, double _p, double c):
  mass(_mass), scalerad(_scalerad), q(_q), p(_p)
{
    if(c<1.0) c=1.0;
    double totalMassNondim=(4*M_PI*(log(1+c)-c/(c+1)));
    /*  "c" is the standard NFW concentration, defined as the ratio of virial radius to scale radius, 
        and assuming that the density drops instantly to zero beyond that radius.
        We introduce another radius "rcutoff" such that (1) density drops more gradually (as r^-4) 
        beyond rcutoff, and (2) total mass is the same as for truncated NFW model with given concentration.
        Now the task is to find rcutoff. */
    gsl_function F;
    F.function=&findNFWrcutoff;
    F.params=&totalMassNondim;
    rcutoff = findRoot(&F, 0.25, c);
    norm = mass*pow(scalerad, -3.0)/(totalMassNondim*p*q);
}

double CDensityNFW::Rho(double X, double Y, double Z, double /*t*/) const
{
    double m = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p))/scalerad;
    return norm / (m * pow_2(1.0 + m) * (1 + m/rcutoff));
}

//----------------------------------------------------------------------------//
/// \cond INTERNAL_DOCS
// find the numerical scaling factor for radius
double find_sersic_b(double x, void* param)
{
    double n=((double*)param)[0];
    return gsl_sf_gamma(2*n)-2*gsl_sf_gamma_inc(2*n, x);
}
// integrand for deprojected density
double int_sersic_dens(double x, void* param)
{
    double n=((double*)param)[0];
    double c=((double*)param)[1];
    double lncosh=gsl_sf_lncosh(x);
    return exp(lncosh*(1/n-1)-c*exp(lncosh/n));
}
/// \endcond

CDensitySersic::CDensitySersic(double _mass, double _Reff, double _q, double _p, double _n):
  mass(_mass), Reff(_Reff), q(_q), p(_p), n(_n), dens(NULL)
{
    // first find the scaling factor b from the equation Gamma(2*n)=2*Gamma(2*n,b)
    gsl_function F;
    F.function=&find_sersic_b;
    double par[2]={n,0};
    F.params=&par;
    b=findRoot(&F, 2*n-1, 2*n, -1);
    if(b<0) 
        b=1.9992*n-0.3271; // approximation value, in case that numerical root couldn't be obtained (unlikely)
    // next compute the deprojected density at a grid of radial scaling points
    double xscale=0.2,xoffset=-5.0 -(n<2?(2-n)*2.5:0);
    size_t npoints=static_cast<size_t>((5.0-xoffset)/xscale);
    vectord xval(npoints), yval(npoints);
    gsl_integration_workspace *ws=gsl_integration_workspace_alloc(1000);
    double result, error;
    F.function=&int_sersic_dens;
    double mult=mass*pow(Reff, -3.0)*pow(b, 1+2*n)/(2*M_PI*M_PI*n*n*gsl_sf_gamma(2*n)*q*p);
    double prevresult=0;
    for(size_t i=0; i<npoints; i++) {
        double scale=xscale*i+xoffset;  // [exp(-5)..exp(5)]
        double r_over_Reff_pow1n=exp(scale)/b;
        par[1]=b*r_over_Reff_pow1n;
        gsl_integration_qagiu(&F, 0, 0, EPSREL_DENSITY_INT, 1000, ws, &result, &error);
        if(i>0 && result<prevresult*1e-3) {
            npoints=i;
            break;
        } else 
            prevresult=result;
        xval[i]=log(b*r_over_Reff_pow1n);
        yval[i]=log(mult*result*pow(r_over_Reff_pow1n, 1-n));
    }
    dens = gsl_spline_alloc(gsl_interp_cspline, npoints);
    gsl_spline_init (dens, &(xval.front()), &(yval.front()), npoints);
}

CDensitySersic::~CDensitySersic()
{
    if(dens!=NULL) gsl_spline_free(dens);
}

double CDensitySersic::Rho(double X, double Y, double Z, double /*t*/) const
{
    if(dens==NULL) return 0;
    double r_over_Reff = sqrt(pow_2(X) + pow_2(Y/q) + pow_2(Z/p))/Reff;
    double xval=log(r_over_Reff)/n+log(b);
    double yval=gsl_spline_evalx(dens, xval, NULL);
    return exp(yval);
}
//----------------------------------------------------------------------------//
// even more machinery for Ellipsoidal mass model with varying axis ratios and user-specified cumulative mass profile
CDensityEllipsoidal::CDensityEllipsoidal(const vectord &_radii, const vectord &_inmass, const vectord &_axesradii, const vectord &_qvals, const vectord &_pvals)
{
    qvals=pvals=dens=NULL;
    if(_radii.size()<=2 || _radii.size()!=_inmass.size() || _axesradii.size()!=_qvals.size() || _axesradii.size()!=_pvals.size())
    {
        my_error(FUNCNAME, "Error, input arrays are not equal in length");
        return;
    }
    CMassModel* massModel = new CMassModel(_radii, _inmass);
    if(massModel->getStatus()!=CMassModel::MS_OK && massModel->getStatus()!=CMassModel::MS_DF_NEGATIVE)
    { 
        my_error(FUNCNAME, "Error in creating spherical mass model!");
        return; 
    }
    // get values of density at the grid in radius and construct our own spline (for efficiency)
    vectord radii, densval;
    for(size_t i=0; i<_radii.size(); i++)
        if(_radii[i]>0) {
            radii.push_back(log(_radii[i]));
            densval.push_back(log(massModel->dens(_radii[i])));
        }
    dens = gsl_spline_alloc(gsl_interp_cspline, radii.size());
    gsl_spline_init (dens, &(radii.front()), &(densval.front()), densval.size());
    delete massModel;  // not needed anymore - we have our own spline
    // check if all values of q, p are equal
    if(_axesradii.size()==0) {
        q=1; p=1; return;
    }
    q=_qvals[0]; p=_pvals[0];
    for(size_t i=1; i<_axesradii.size() && p>0; i++)
        if(_qvals[i]!=q || _pvals[i]!=p) p=0;
    if(p>0) return;  // no need to initialize splines -- constant axis ratios
    // initialize radius-dependent axis ratio
    vectord axesradii, axesq, axesp;
    for(size_t i=0; i<_axesradii.size(); i++)
        if(_axesradii[i]>0) {
            axesradii.push_back(log(_axesradii[i]));
            axesq.push_back(_qvals[i]);
            axesp.push_back(_pvals[i]);
        }
    qvals = gsl_spline_alloc(gsl_interp_linear, axesradii.size());
    gsl_spline_init (qvals, &(axesradii.front()), &(axesq.front()), axesradii.size());
    pvals = gsl_spline_alloc(gsl_interp_linear, axesradii.size());
    gsl_spline_init (pvals, &(axesradii.front()), &(axesp.front()), axesradii.size());
}

CDensityEllipsoidal::~CDensityEllipsoidal()
{
    if(dens !=NULL) gsl_spline_free(dens);
    if(qvals!=NULL) gsl_spline_free(qvals);
    if(pvals!=NULL) gsl_spline_free(pvals);
}

/*CDensityEllipsoidal::Mass(const double r) const
{
    return massModel->mass(r);  // spherically averaged approximation
}*/

double CDensityEllipsoidal::Rho(double X, double Y, double Z, double /*t*/) const
{
    if((X==0&&Y==0&&Z==0) || dens==NULL) return 0;
    double qs, ps;
    if(p>0) { qs=q; ps=p; }  // simplification for constant axis ratio
    else {  // find from spline interpolation
        double rslog=log(X*X+Y*Y+Z*Z)/2;  // used to find q and p
        if(rslog<=qvals->x[0]) { qs = qvals->y[0]; ps = pvals->y[0]; } else
        if(rslog>=qvals->x[qvals->size-1]) { qs = qvals->y[qvals->size-1]; ps = pvals->y[qvals->size-1]; } else
        {
            qs = gsl_spline_eval(qvals, rslog, 0);
            ps = gsl_spline_eval(pvals, rslog, 0);
        }
    }
    double qp=pow(qs*ps, -1.0/3);
    qs*=qp; ps*=qp;   // scale axis ratio so that their product is 1
    double logr = log(pow_2(X/qp)+pow_2(Y/qs)+pow_2(Z/ps))/2.0;
    return exp(gsl_spline_evalx(dens, logr, NULL));
}

//----------------------------------------------------------------------------//
// Multi-gaussian expansion of density profile
CDensityMGE::CDensityMGE(const smile::CDensityMGE::vectorg &_data) : data(_data)
{
    // check if all values of q, p are equal
    if(data.empty()) { q=p=1; return; }
    q=data[0].q; p=data[0].p;
    for(size_t i=1; i<data.size(); i++) {
        if(data[i].q!=q) q=0;
        if(data[i].p!=p) p=0;
    }
}

double CDensityMGE::Rho(double X, double Y, double Z, double /*t*/) const
{
    double result=0;
    for(size_t i=0; i<data.size(); i++)
    {
        double r2=pow_2(X) + pow_2(Y/data[i].q) + pow_2(Z/data[i].p);
        result += data[i].mass/(pow(2*M_PI*pow_2(data[i].width),1.5)*data[i].p*data[i].q)
            * exp(-r2/(2*pow_2(data[i].width)));
    }
    return result;
}

//----------------------------------------------------------------------------//
// 'class factory'  (makes correspondence between enum potential and symmetry types and string names)
/// lists all 'true' potentials, i.e. those providing a complete density-potential(-force) pair
typedef std::map<CDensity::POTENTIALTYPE, const char*> PotentialNameMapType;

/// lists all analytic density profiles 
/// (including those that don't have corresponding potential, but excluding general-purpose expansions)
typedef std::map<CDensity::POTENTIALTYPE, const char*> DensityNameMapType;

/// lists available symmetry types
typedef std::map<CDensity::SYMMETRYTYPE,  const char*> SymmetryNameMapType;

PotentialNameMapType PotentialNames;
DensityNameMapType DensityNames;
SymmetryNameMapType SymmetryNames;
bool mapinitialized = false;

/// create a correspondence between names and enum identifiers for potential, density and symmetry types
void initPotentialAndSymmetryNameMap()
{
    PotentialNames.clear();
    PotentialNames[CDensity::PT_LOG] = CPotentialLog::myName();
    PotentialNames[CDensity::PT_HARMONIC] = CPotentialHarmonic::myName();
    PotentialNames[CDensity::PT_DEHNEN] = CPotentialDehnen::myName();
    PotentialNames[CDensity::PT_SCALEFREE] = CPotentialScaleFree::myName();
    PotentialNames[CDensity::PT_SCALEFREESH] = CPotentialScaleFreeSH::myName();
    PotentialNames[CDensity::PT_BSE] = CPotentialBSE::myName();
    PotentialNames[CDensity::PT_BSECOMPACT] = CPotentialBSECompact::myName();
    PotentialNames[CDensity::PT_SPLINE] = CPotentialSpline::myName();
#ifdef HAVE_INTERP2D
    PotentialNames[CDensity::PT_CYLSPLINE] = CPotentialCylSpline::myName();
#endif
    PotentialNames[CDensity::PT_SPHERICAL] = CPotentialSpherical::myName();
    PotentialNames[CDensity::PT_MIYAMOTONAGAI] = CPotentialMiyamotoNagai::myName();
    PotentialNames[CDensity::PT_FERRERS] = CPotentialFerrers::myName();
    PotentialNames[CDensity::PT_NB] = CPotentialNB::myName();

    // list of density models available for BSE and Spline approximation
    DensityNames.clear();
    DensityNames[CDensity::PT_NB] = CPotentialNB::myName();   // denotes not the actual tree-code potential, but rather the fact that the density model comes from discrete points in Nbody file
    DensityNames[CDensity::PT_ELLIPSOIDAL] = CDensityEllipsoidal::myName();
    DensityNames[CDensity::PT_MGE] = CDensityMGE::myName();
    DensityNames[CDensity::PT_COEFS] = "Coefs";  // denotes that potential expansion coefs are loaded from a text file rather than computed from a density model
    DensityNames[CDensity::PT_DEHNEN] = CPotentialDehnen::myName();
    DensityNames[CDensity::PT_MIYAMOTONAGAI] = CPotentialMiyamotoNagai::myName();
    DensityNames[CDensity::PT_PLUMMER] = CDensityPlummer::myName();
    DensityNames[CDensity::PT_PERFECTELLIPSOID] = CDensityPerfectEllipsoid::myName();
    DensityNames[CDensity::PT_ISOCHRONE] = CDensityIsochrone::myName();
    DensityNames[CDensity::PT_EXPDISK] = CDensityExpDisk::myName();
    DensityNames[CDensity::PT_NFW] = CDensityNFW::myName();
    DensityNames[CDensity::PT_SERSIC] = CDensitySersic::myName();
    DensityNames[CDensity::PT_FERRERS] = CPotentialFerrers::myName();

    SymmetryNames[CDensity::ST_NONE]="None";
    SymmetryNames[CDensity::ST_REFLECTION]="Reflection";
    SymmetryNames[CDensity::ST_TRIAXIAL]="Triaxial";
    SymmetryNames[CDensity::ST_AXISYMMETRIC]="Axisymmetric";
    SymmetryNames[CDensity::ST_SPHERICAL]="Spherical";

    mapinitialized=true;
}
   
const char* getPotentialNameByType(CDensity::POTENTIALTYPE type)
{
    if(!mapinitialized) initPotentialAndSymmetryNameMap();
    PotentialNameMapType::const_iterator iter=PotentialNames.find(type);
    if(iter!=PotentialNames.end()) return iter->second;
    return "";
}

const char* getDensityNameByType(CDensity::POTENTIALTYPE type)
{
    if(!mapinitialized) initPotentialAndSymmetryNameMap();
    DensityNameMapType::const_iterator iter=DensityNames.find(type);
    if(iter!=DensityNames.end()) return iter->second;
    return "";
}

const char* getSymmetryNameByType(CDensity::SYMMETRYTYPE type)
{
    if(!mapinitialized) initPotentialAndSymmetryNameMap();
    SymmetryNameMapType::const_iterator iter=SymmetryNames.find(type);
    if(iter!=SymmetryNames.end()) return iter->second;
    return "";
}
    
CDensity::POTENTIALTYPE getPotentialTypeByName(const std::string& PotentialName)
{
    if(!mapinitialized) initPotentialAndSymmetryNameMap();
    for(PotentialNameMapType::const_iterator iter=PotentialNames.begin(); iter!=PotentialNames.end(); iter++)
        if(strings_equal(PotentialName, iter->second)) return iter->first;
    return CDensity::PT_UNKNOWN;
}

CDensity::POTENTIALTYPE getDensityTypeByName(const std::string& DensityName)
{
    if(!mapinitialized) initPotentialAndSymmetryNameMap();
    for(DensityNameMapType::const_iterator iter=DensityNames.begin(); iter!=DensityNames.end(); iter++)
        if(strings_equal(DensityName, iter->second)) return iter->first;
    return CDensity::PT_UNKNOWN;
}

CDensity::SYMMETRYTYPE getSymmetryTypeByName(const std::string& SymmetryName)
{
    if(!mapinitialized) initPotentialAndSymmetryNameMap();
    if(SymmetryName.empty()) return CDensity::ST_DEFAULT;
    // compare only the first letter (should abandon this simplification 
    // if more than one symmetry types are defined that could start with the same letter)
    for(SymmetryNameMapType::const_iterator iter=SymmetryNames.begin(); iter!=SymmetryNames.end(); iter++)
        if(tolower(SymmetryName[0]) == tolower(iter->second[0])) return iter->first;
    return CDensity::ST_DEFAULT;
}

const char* coefFileExtension(CDensity::POTENTIALTYPE pottype)
{
    switch(pottype) {
        case CDensity::PT_BSE: return ".coef_bse";
        case CDensity::PT_BSECOMPACT: return ".coef_bsec";
        case CDensity::PT_SPLINE: return ".coef_spl";
        case CDensity::PT_CYLSPLINE: return ".coef_cyl";
        case CDensity::PT_SCALEFREESH: return ".coef_sf";
        case CDensity::PT_SPHERICAL: return ".mass";
        default: return "";
    }
}

CDensity::POTENTIALTYPE coefFileType(const std::string& fileName)
{
    if(ends_with_str(fileName, ".coef_bse"))
        return CDensity::PT_BSE;
    else if(ends_with_str(fileName, ".coef_bsec"))
        return CDensity::PT_BSECOMPACT;
    else if(ends_with_str(fileName, ".coef_spl"))
        return CDensity::PT_SPLINE;
    else if(ends_with_str(fileName, ".coef_cyl"))
        return CDensity::PT_CYLSPLINE;
    else if(ends_with_str(fileName, ".coef_sf"))
        return CDensity::PT_SCALEFREESH;
    else if(ends_with_str(fileName, ".mass") || ends_with_str(fileName, ".tab"))
        return CDensity::PT_SPHERICAL;
    else
        return CDensity::PT_UNKNOWN;
}

}; // namespace
